/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2001 Gary Henderson (gary@daniver.demon.co.uk) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (Ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
 * C4 C code (c) Copyright 2001 Gary Henderson (gary@daniver.demon.co.uk).
 *
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */
#include "snes9x.h"
#include "memmap.h"
#include "spc700.h"
#include "apu_orig.h"
#include "soundux.h"
#include "snesapu.h"

/* For note-triggered SPC dump support */
#include "snapshot.h"

//#define APU_DEBUG
#ifdef APU_DEBUG
static const char apu_log_name[] = "apu.log";
static FILE *apu_fs = NULL;
extern FILE *trace;
#endif

int spc_is_dumping = 0;
uint8 spc_is_dumping_temp;
uint8 spc_dump_dsp[0x100];

uint8 APUROM [64] =
{
  0xCD,0xEF,0xBD,0xE8,0x00,0xC6,0x1D,0xD0,0xFC,0x8F,0xAA,0xF4,0x8F,0xBB,0xF5,0x78,
  0xCC,0xF4,0xD0,0xFB,0x2F,0x19,0xEB,0xF4,0xD0,0xFC,0x7E,0xF4,0xD0,0x0B,0xE4,0xF5,
  0xCB,0xF4,0xD7,0x00,0xFC,0xD0,0xF3,0xAB,0x01,0x10,0xEF,0x7E,0xF4,0x10,0xEB,0xBA,
  0xF6,0xDA,0x00,0xBA,0xF4,0xC4,0xF4,0xDD,0x5D,0xD0,0xDB,0x1F,0x00,0x00,0xC0,0xFF
};
/*
	;APU ROM code

ffc0:
	mov x, #$ef
	mov sp, x
	mov a, #$00
ffc5:
	mov [x], a
	dec x
	bne $ffc5
	mov $f4, #$aa
	mov $f5, #$bb
ffcf:
	cmp $f4, #$cc
	bne $ffcf
	bra $ffef
ffd6:
	mov y, $f4
	bne $ffd6
ffda
	cmp y, $f4
	bne $ffe9
	mov a, $f5
	mov $f4, y
	mov [$00]+y, a
	inc y
	bne $ffda
	inc $01
ffe9:
	bpl $ffda
	cmp y, $f4
	bpl $ffda
ffef:
	movw ya, $f6
	movw $00, ya
	movw ya, $f4
	mov $f4, a
	mov a, y
	mov x, a
	bne $ffd6
	jmp [$0000+x]
	di
	stop
*/

bool8
S9xInitAPU ()
{
  IAPU.RAM = (uint8 *) malloc (0x10000);

  if (!IAPU.RAM || !InitAPUDSP ())
    {
      S9xDeinitAPU ();
      return (FALSE);
    }
  return (TRUE);
}

void
S9xDeinitAPU ()
{
  if (IAPU.RAM)
    {
      free ((char *) IAPU.RAM);
      IAPU.RAM = NULL;
    }
  DeinitAPUDSP ();
}

void
S9xResetAPU ()
{
  int i;

  Settings.APUEnabled = Settings.NextAPUEnabled;
  if (SNESGameFixes.APU_Disable)
    Settings.APUEnabled = FALSE;

  ZeroMemory (spc_dump_dsp, 0x100);

  //ZeroMemory(IAPU.RAM, 0x100);
  memset (IAPU.RAM, Settings.APURAMInitialValue, 0x10000);

  ZeroMemory (APU.OutPorts, 4);

  IAPU.TwoCycles = IAPU.OneCycle * 2;

  APU.Cycles = CPU.Cycles;
  APU.Cycles_err = 0;
  APU.t_ticks_err = 0;

  for(i = 0; i < 3; i++)
    {
      APU.t_ticks[i] = 0;
      APU.t_stage1_ticks[i] = 0;
      APU.t_current_line[i] = 0;
    }

  APU.t64kCycles = 0;
  APU.t64kCycles_err = 0;

  APURegisters.PC       = 0xffc0;
  APURegisters.YA.B.A   = 0x00;
  APURegisters.X        = 0x00;
  APURegisters.YA.B.Y   = 0x00;
  APURegisters.S        = 0xef;
  APURegisters.P        = 0x02;

  S9xAPUUnpackStatus ();

  IAPU.APUExecuting = Settings.APUEnabled;

  EXT.NextAPUTimerPos = APU.Cycles;
  EXT.APUTimerCounter_err = 0;

  S9xSetAPUTimerCounter ();

  EXT.t64Cnt = 0;

  EXT.NextAPUTimerPos +=
    ((EXT.APUTimerCounter_err += EXT.APUTimerCounter) >> FIXED_POINT_SHIFT);
  EXT.APUTimerCounter_err &= FIXED_POINT_REMAINDER;

  IAPU.RAM[0xf0] = 0x0A;//timers_enabled & ram_writable

  APU.ShowROM = TRUE;
  IAPU.RAM[0xf1] = 0x80;

  for (i = 0; i < 3; i++)
    {
      APU.TimerEnabled[i] = FALSE;
      APU.TimerValueWritten[i] = 0;
      APU.TimerTarget[i] = 0;
      APU.Timer[i] = 0;
    }

  //soundux.cpp SoundData
  S9xResetSound ();

  //SNESAPU
  ResetAPUDSP ();

  spc_is_dumping_temp = 0;

}

void S9xSetAPUCycle(int32 onecycle)
{
  IAPU.OneCycle = onecycle;
  IAPU.TwoCycles = onecycle * 2;
}

void
S9xSetAPUControl (uint8 byte)
{
  //if (byte & 0x40)
  //printf ("*** Special SPC700 timing enabled\n");
  if ((byte & 1) != 0 && !APU.TimerEnabled[0])
    {
      APU.Timer[0] = 0;
      IAPU.RAM[0xfd] = 0;
      if ((APU.TimerTarget[0] = IAPU.RAM[0xfa]) == 0)
	APU.TimerTarget[0] = 0x100;
    }
  if ((byte & 2) != 0 && !APU.TimerEnabled[1])
    {
      APU.Timer[1] = 0;
      IAPU.RAM[0xfe] = 0;
      if ((APU.TimerTarget[1] = IAPU.RAM[0xfb]) == 0)
	APU.TimerTarget[1] = 0x100;
    }
  if ((byte & 4) != 0 && !APU.TimerEnabled[2])
    {
      APU.Timer[2] = 0;
      IAPU.RAM[0xff] = 0;
      if ((APU.TimerTarget[2] = IAPU.RAM[0xfc]) == 0)
	APU.TimerTarget[2] = 0x100;
    }
  APU.TimerEnabled[0] = byte & 1;
  APU.TimerEnabled[1] = (byte & 2) >> 1;
  APU.TimerEnabled[2] = (byte & 4) >> 2;

  if (byte & 0x10)
    {
      sapuport_w (0, 0x00);
      sapuport_w (1, 0x00);
    }

  if (byte & 0x20)
    {
      sapuport_w (2, 0x00);
      sapuport_w (3, 0x00);
    }

  if (byte & 0x80)
    APU.ShowROM = TRUE;
  else
    APU.ShowROM = FALSE;
}

void
S9xSetAPUTimerCounter ()
{
  EXT.APUTimerCounter =
    (long) ((((uint64)SNES_CYCLES_PER_SECOND << FIXED_POINT_SHIFT) * Timings.H_Max_Master)
	    / (SNES_CYCLES_PER_SCANLINE * RATE_64K));
  return;
}

uint8
S9xGetAPUDSP (uint8 reg)
{
  uint8 byte;

  switch (reg)
    {
    case APU_ENDX:
      byte = APU.DSP[reg];
      break;

    default:
      switch (reg & 0x0f)
	{
	case APU_OUTX:
	  byte = APU.DSP[reg];
	  break;

	case APU_ENVX:
	  if (Settings.SoundEnvelopeHeightReading)
	    byte = APU.DSP[reg] & 0x7f;
	  else
	    byte = 0x00;
	  break;

	default:
	  byte = APU.DSP[reg];
	  break;
	}
      break;
    }
#ifdef APU_DEBUG
  if (Options.toggle && !apu_fs)
    apu_fs = fopen (apu_log_name, "a");
  else if (!Options.toggle && apu_fs)
    {
      fclose (apu_fs);
      apu_fs = NULL;
    }
  if (apu_fs)
    fprintf (apu_fs,
	     "Address %02X r Byte %02X\n",
	     (unsigned int) IAPU.RAM[0xf2], (unsigned int) byte);
#endif
  return byte;
}

void
S9xSetAPUDSP (uint8 byte, uint8 reg)
{
#ifdef APU_DEBUG
  if (Options.toggle && !apu_fs)
    apu_fs = fopen (apu_log_name, "a");
  else if (!Options.toggle && apu_fs)
    {
      fclose (apu_fs);
      apu_fs = NULL;
    }
  if (apu_fs)
    fprintf (apu_fs,
	     "Address %02X w Byte %02X\n",
	     (unsigned int) IAPU.RAM[0xf2], (unsigned int) byte);
#endif
  APUDSPIn (reg, byte);
}

#if 1
uint8
sapuport_r (int n)
{
  return IAPU.RAM[0xf4 + n];
}

void
sapuport_w (int n, uint8 val)
{
  IAPU.RAM[0xf4 + n] = val;
}

uint8
sport_r (int n)
{
  return APU.OutPorts[n];
}

void
sport_w (int n, uint8 val)
{
  APU.OutPorts[n] = val;
}
#endif

#if 0
uint8 APU_InPorts[0x10000][4];
uint8 APU_OutPorts[0x10000][4];
uint16 APU_In_in;
uint16 APU_In_out;
uint16 APU_Out_in;
uint16 APU_Out_out;

uint8
sapuport_r (int n)
{
  uint8 r = APU_InPorts[APU_In_in][n];
  if(APU_In_in != APU_In_out)
    APU_In_in++;
  return r;
}

void
sapuport_w (int n, uint8 val)
{
  if(APU_InPorts[APU_In_out][n] == val)
    return;
  APU_InPorts[APU_In_out][n] = val;
  uint16 next = APU_In_out + 1;
  APU_InPorts[next][0] = APU_InPorts[APU_In_out][0];
  APU_InPorts[next][1] = APU_InPorts[APU_In_out][1];
  APU_InPorts[next][2] = APU_InPorts[APU_In_out][2];
  APU_InPorts[next][3] = APU_InPorts[APU_In_out][3];
  APU_In_out = next;
}

uint8
sport_r (int n)
{
  uint8 r = APU_OutPorts[APU_Out_in][n];
  if(APU_Out_in != APU_Out_out)
    APU_Out_in++;
  return r;
}

void
sport_w (int n, uint8 val)
{
  if(APU_OutPorts[APU_Out_out][n] == val)
    return;
  APU_OutPorts[APU_Out_out][n] = val;
  uint16 next = APU_Out_out + 1;
  APU_OutPorts[next][0] = APU_OutPorts[APU_Out_out][0];
  APU_OutPorts[next][1] = APU_OutPorts[APU_Out_out][1];
  APU_OutPorts[next][2] = APU_OutPorts[APU_Out_out][2];
  APU_OutPorts[next][3] = APU_OutPorts[APU_Out_out][3];
  APU_Out_out = next;
}
#endif

void S9xDumpSPCSnapshot (void)
{
  spc_is_dumping = 1;
}

