#include "snes9x.h"
#include "memmap.h"
#include "apu_orig.h"
#include "snesapu.h"
#include "apumem.h"

static inline int calc_t64kCycles(int add)
{
  int a = add * SNES_CYCLES_PER_SCANLINE;
  int b;
  if(Settings.TurboMode && Settings.SoundConstantTempo)
    b = (Settings.TurboSkipFrames == 0 ? 1 : Settings.TurboSkipFrames) * Timings.H_Max_Master;
  else
    b = Timings.H_Max_Master;

  APU.t64kCycles += (a + APU.t64kCycles_err) / b;
  APU.t64kCycles_err = (a + APU.t64kCycles_err) % b;

  int r = APU.t64kCycles / FREQ_64K;
  APU.t64kCycles %= FREQ_64K;

  return r;
}

static inline void APU_EXECUTE1()
{
  //0% speed -- locks S-SMP
  if((IAPU.RAM[0xf0] & 0xC0) == 0x80)
    {
      APU.Cycles = CPU.Cycles;
      return;
    }

#ifdef DEBUGGER
  if (APU.Flags & TRACE_FLAG)
    S9xTraceAPU ();
#endif

  uint8 op = S9xAPUGetByte (APURegisters.PC);
  smp_add_clocks(S9xAPUCycleLengths[op]);
  (*S9xApuOpcodes[op]) ();
}

static inline void APU_EXECUTE()
{
  if (IAPU.APUExecuting)
    {
      while (APU.Cycles < CPU.Cycles)
	APU_EXECUTE1();
    }
  else
    APU.Cycles = CPU.Cycles;
}

const static int t_freq[3] = {192, 192, 24};

static inline void t_tick(int t, int step)
{
  APU.t_ticks[t] += step;
  if(APU.t_ticks[t] < t_freq[t])
    return;
  APU.t_ticks[t] -= t_freq[t];
  APU.t_stage1_ticks[t] ^= 1;
  t_sync_stage1(t);
}

void t_sync_stage1(int t)
{
  bool8 new_line = APU.t_stage1_ticks[t];

  //timers_enabled & timers_disabled
  if(!(IAPU.RAM[0xf0] & 0x08) || (IAPU.RAM[0xf0] & 0x01))
    new_line = FALSE;

  bool8 old_line = APU.t_current_line[t];
  APU.t_current_line[t] = new_line;
  if(old_line != 1 || new_line != 0) return;  //only pulse on 1->0 transition
  if(!APU.TimerEnabled[t]) return;

  //8-bit counter increment
  APU.Timer[t] = (APU.Timer[t] + 1) & 0xff;

  if (APU.Timer[t] != (APU.TimerTarget[t] & 0xff))
    return;

  APU.Timer[t] = 0;
  IAPU.RAM[0xfd + t] = (IAPU.RAM[0xfd + t] + 1) & 0xf;
}

void smp_add_clocks(int32 clocks)
{
  int a = clocks * IAPU.OneCycle * SNES_CYCLES_PER_SCANLINE;
  int b;
  if(Settings.TurboMode && Settings.SoundConstantTempo)
    b = SMP_ONE_CYCLE * (Settings.TurboSkipFrames == 0 ? 1 : Settings.TurboSkipFrames) * Timings.H_Max_Master;
  else
    b = SMP_ONE_CYCLE * Timings.H_Max_Master;
    
  int t_ticks = (a + APU.t_ticks_err) / b;
  APU.t_ticks_err = (a + APU.t_ticks_err) % b;

  int step = (1 << ((IAPU.RAM[0xf0] >> 6) & 3)) + (2 << ((IAPU.RAM[0xf0] >> 4) & 3));

  int c;
  for(c = 0; c < t_ticks; c++)
    {
      t_tick(0, step);
      t_tick(1, step);
      t_tick(2, step);
    }

  int speed;
  switch((IAPU.RAM[0xf0] & 0xC0))
    {
    case 0x00: speed = 1;
    case 0x40: speed = 2;
    case 0xC0: speed = 10;
    default: speed = 1;
    }

  int addc = IAPU.OneCycle * clocks * speed;
  int64 x = (int64)SNES_CYCLES_PER_SECOND * addc;
  int64 y = (int64)IAPU.MasterClock;

  APU.Cycles += (x + APU.Cycles_err) / y;
  APU.Cycles_err = (x + APU.Cycles_err) % y;

  int tc = calc_t64kCycles(addc);
  for(c = 0; c < tc; c++, EXT.t64Cnt++)
    if(!(EXT.t64Cnt & 1))
      smpdsp_run();
}

void
S9xAPUMainLoop ()
{
  if(APU.scpu_sync)
    return;
  APU_EXECUTE ();
}

