/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#ifndef _apumemory_h_
#define _apumemory_h_

static inline uint8 APU_GETBYTE(uint16 addr)
{
  uint8 r;
  switch(addr)
    {
    case 0xf0:    //TEST -- write-only register
      r = 0x00;
      break;

    case 0xf1:    //CONTROL -- write-only register
      r = 0x00;
      break;

    case 0xf2:    //DSPADDR
      r = IAPU.RAM [addr];
      break;

    case 0xf3:    //DSPDATA
      //0x80-0xff are read-only mirrors of 0x00-0x7f
      r = S9xGetAPUDSP (IAPU.RAM[0xf2] & 0x7f);
      break;

    case 0xf4:    //CPUIO0
    case 0xf5:    //CPUIO1
    case 0xf6:    //CPUIO2
    case 0xf7:    //CPUIO3
      //if(!CPU.ssmp_sync && !CPU.InDMAorHDMA && !Settings.SPCPlaying)
      //scpu_sync();
      r = sapuport_r(addr - 0xf4);
      break;

    case 0xf8:    //RAM0
    case 0xf9:    //RAM1
      r = IAPU.RAM [addr];
      break;

    case 0xfa:    //T0TARGET
    case 0xfb:    //T1TARGET
    case 0xfc:    //T2TARGET -- write-only registers
      r = 0x00;
      break;

    case 0xfd:    //T0OUT -- 4-bit counter value
    case 0xfe:    //T1OUT -- 4-bit counter value
    case 0xff:    //T2OUT -- 4-bit counter value
      r = IAPU.RAM [addr] & 15;
      IAPU.RAM [addr] = 0;
      break;

    default:
      if(addr >= 0xffc0 && APU.ShowROM)
	r = APUROM[addr & 0x3f];
      else if(IAPU.RAM [0xf0] & 0x04)//ram_disabled
	r = 0x5a;//0xff on mini-SNES
      else
	r = IAPU.RAM [addr];
      break;
    }
  return r;
}

static inline uint8
S9xAPUGetByte (uint16 Address)
{
  return APU_GETBYTE (Address);
}

static inline uint8
S9xAPUGetByteDP (uint8 Address)
{
  return APU_GETBYTE ((APUCheckDirectPage() ? 0x100 : 0) + Address);
}

static inline void
APU_SETBYTE (uint8 data, uint16 addr)
{
  switch(addr)
    {
    case 0xf0:  //TEST
      if(APUCheckDirectPage())  //writes only valid when P flag is clear
	break;
      IAPU.RAM [addr] = data;
      t_sync_stage1(0);
      t_sync_stage1(1);
      t_sync_stage1(2);
      break;

    case 0xf1:  //CONTROL
      S9xSetAPUControl (data);
      IAPU.RAM [addr] = data;
      break;

    case 0xf2:  //DSPADDR
      IAPU.RAM [addr] = data;
      break;

    case 0xf3: //DSPDATA
      //0x80-0xff is a read-only mirror of 0x00-0x7f
      if(!(IAPU.RAM[0xf2] & 0x80))
	S9xSetAPUDSP (data, IAPU.RAM[0xf2] & 0x7f);
      IAPU.RAM [addr] = data;
      break;

    case 0xf4:    //CPUIO0
    case 0xf5:    //CPUIO1
    case 0xf6:    //CPUIO2
    case 0xf7:    //CPUIO3
      if(!CPU.ssmp_sync && !CPU.InDMAorHDMA && !Settings.SPCPlaying)
	scpu_sync();//"HIOUDEN" "ZAN3 SFC"
      sport_w(addr - 0xf4, data);
      break;

    case 0xf8:    //RAM0
    case 0xf9:    //RAM1
      IAPU.RAM [addr] = data;
      break;

    case 0xfa:    //T0TARGET
    case 0xfb:    //T1TARGET
    case 0xfc:    //T2TARGET
      IAPU.RAM [addr] = data;
      APU.TimerTarget [addr - 0xfa] = data == 0 ? 0x100 : data;
      break;

    case 0xfd:    //T0OUT
    case 0xfe:    //T1OUT
    case 0xff:    //T2OUT -- read-only registers
      break;

    default:
      //writes to $ffc0-$ffff always go to apuram, even if the iplrom is enabled
      if((IAPU.RAM [0xf0] & 0x02) && !(IAPU.RAM [0xf0] & 0x04))
	IAPU.RAM [addr] = data;
      break;
    }
}

static inline void
S9xAPUSetByte (uint8 byte, uint16 Address)
{
  APU_SETBYTE(byte, Address);
}

static inline void
S9xAPUSetByteDP (uint8 byte, uint8 Address)
{
  APU_SETBYTE(byte, (APUCheckDirectPage() ? 0x100 : 0) + Address);
}

#endif
