/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#include "snes9x.h"
#include "memmap.h"
#include "cpuops.h"
#include "ppu.h"
#include "cpuexec.h"
#include "debug.h"
#include "snapshot.h"
#include "gfx.h"
#include "apu_orig.h"
#include "dma.h"
#include "fxemu.h"
#include "sa1.h"
#include "spc7110.h"

#ifdef DEBUGGER
#include "missing.h"
#endif

void S9xMainLoop (void)
{
  for (;;) {
#ifdef DEBUG_MAXCOUNT
    CPU.GlobalLoopCount++;
    if(Settings.MaxCount && CPU.GlobalLoopCount == Settings.MaxCount) {
      fprintf(stderr, "Max loop count reached: %ld \n", Settings.MaxCount);
      S9xExit();
    }
#endif
    if (CPU.Flags) {
      if (CPU.Flags & NMI_FLAG) {
	if (Timings.NMITriggerPos <= CPU.Cycles) {
	    CPU.Flags &= ~NMI_FLAG;
	    Timings.NMITriggerPos = 0xffff;
	    if (CPU.WaitingForInterrupt) {
	      CPU.WaitingForInterrupt = FALSE;
	      CPU.PC++;
	      Registers.PCw++;
	    }
	    S9xOpcode_NMI ();
	  }
	}
#ifdef DEBUGGER
	if ((CPU.Flags & BREAK_FLAG) &&
	    !(CPU.Flags & SINGLE_STEP_FLAG)) {
	  for (int Break = 0; Break != 6; Break++) {
	    if (S9xBreakpoint[Break].Enabled &&
		S9xBreakpoint[Break].Bank == Registers.PB &&
		S9xBreakpoint[Break].Address == CPU.PC - CPU.PCBase) {
	      if (S9xBreakpoint[Break].Enabled == 2)
		S9xBreakpoint[Break].Enabled = TRUE;
	      else
		CPU.Flags |= DEBUG_MODE_FLAG;
	    }
	  }
	}
#endif

	if (CPU.Flags & IRQ_FLAG) {
	  if (CPU.IRQPending == 0) {
	    if (CPU.WaitingForInterrupt) {
	      CPU.WaitingForInterrupt = FALSE;
	      CPU.PC++;
	      Registers.PCw++;
	    }
	    if (CPU.IRQActive) {
	      if (!CheckFlag (IRQ) && !Settings.DisableIRQ)
		S9xOpcode_IRQ ();
	    }
	    else
	      CPU.Flags &= ~IRQ_FLAG;
	  }
	  else if(--CPU.IRQPending == 0 && CheckFlag (IRQ)) {
#if 0
	    // "Mystic Quest"
	    if (CPU.WaitingForInterrupt) {
	      CPU.WaitingForInterrupt = FALSE;
	      CPU.PC++;
	      Registers.PCw++;
	    }
#endif
	    CPU.IRQPending = 1;
	  }
	}

#ifdef DEBUGGER
	if (CPU.Flags & DEBUG_MODE_FLAG)
	  break;
#endif
	if (CPU.Flags & SCAN_KEYS_FLAG)
	  break;
#ifdef DEBUGGER
	if (CPU.Flags & TRACE_FLAG)
	  S9xTrace ();

	if (CPU.Flags & SINGLE_STEP_FLAG) {
	  CPU.Flags &= ~SINGLE_STEP_FLAG;
	  CPU.Flags |= DEBUG_MODE_FLAG;
	}
#endif
      } // if (CPU.Flags)
#ifdef CPU_SHUTDOWN
#ifdef OLD_PCBASE
      CPU.PBPCAtOpcodeStart = (ICPU.ShiftedPB + CPU.PC - CPU.PCBase);
#else
      CPU.PBPCAtOpcodeStart = Registers.PBPC;
#endif
#endif
      //CPU.Cycles += CPU.MemSpeed;

      CPU.PC++;
      //Registers.PCw++;
      (*ICPU.S9xOpcodes [S9xGetByte(ICPU.ShiftedPB + Registers.PCw++)].S9xOpcode) ();
      if (SA1.Executing)
	S9xSA1MainLoop ();

    if (CPU.Cycles >= CPU.NextEvent)
	S9xDoHEventProcessing ();

      //apu execute by aputimer pulse
      if (CPU.Cycles >= EXT.NextAPUTimerPos)
	S9xAPUMainLoop ();

    } // for (;;)
		
    // unneeded
    //Registers.PC = CPU.PC - CPU.PCBase;
    //S9xPackStatus ();
    //APURegisters.PC = IAPU.PC - IAPU.RAM;
    //S9xAPUPackStatus ();

    if (CPU.Flags & SCAN_KEYS_FLAG) {
#ifdef DEBUGGER
      if (!(CPU.Flags & FRAME_ADVANCE_FLAG))
#endif
	S9xSyncSpeed ();
      CPU.Flags &= ~SCAN_KEYS_FLAG;
    }
}

