/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#if defined(__unix) || defined(__linux) || defined(__sun) || defined(__DJGPP) || defined(__MINGW32__)
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "snapshot.h"
#include "memmap.h"
#include "dma.h"
#include "snes9x.h"
#include "65c816.h"
#include "ppu.h"
#include "cpuexec.h"
#include "display.h"
#include "controls.h"
#include "apu_orig.h"
#include "soundux.h"
#include "snesapu.h"
#include "fxinst.h"
#include "fxemu.h"
#include "sa1.h"
#include "srtc.h"
#include "sdd1.h"
#include "spc7110.h"
#include "bsx.h"
#include "movie.h"
#include "language.h"

//#undef TransferBytes
//#undef IndirectAddress

typedef struct {
    int offset;
    int size;
    int type;
} FreezeData;

enum
{
	INT_V,
	uint8_ARRAY_V,
	uint16_ARRAY_V,
	uint32_ARRAY_V,
	uint8_INDIR_ARRAY_V,
	uint16_INDIR_ARRAY_V,
	uint32_INDIR_ARRAY_V,
	POINTER_V
};

#define Offset(field,structure) \
((int) (((char *) (&(((structure)NULL)->field))) - ((char *) NULL)))

#define Offset2(field1,field2,structure) \
((int) (((char *) (&(((structure)NULL)->field1))) - ((char *) NULL)) | \
((int) (((char *) (&(((structure)NULL)->field2))) - ((char *) NULL)) << 16))

#define COUNT(ARRAY) (sizeof (ARRAY) / sizeof (ARRAY[0]))

struct SnapshotMovieInfo
{
	uint32	MovieInputDataSize;
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SnapshotMovieInfo *)

static FreezeData SnapMovie [] = {
    {OFFSET (MovieInputDataSize), 4, INT_V},
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SCPUState *)

static FreezeData SnapCPU [] = {
    {OFFSET (Flags), 4, INT_V},
    {OFFSET (BranchSkip), 1, INT_V},
    {OFFSET (NMIActive), 1, INT_V},
    {OFFSET (IRQActive), 1, INT_V},
    {OFFSET (WaitingForInterrupt), 1, INT_V},
    {OFFSET (WhichEvent), 1, INT_V},
    {OFFSET (Cycles), 4, INT_V},
    {OFFSET (NextEvent), 4, INT_V},
    {OFFSET (V_Counter), 4, INT_V},
    {OFFSET (MemSpeed), 4, INT_V},
    {OFFSET (MemSpeedx2), 4, INT_V},
    {OFFSET (FastROMSpeed), 4, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SEXTState *)

static FreezeData SnapEXT [] = {
    {OFFSET (Clock), 4, INT_V},
    {OFFSET (t64Cnt), 4, INT_V},
    {OFFSET (NMICycleCount), 4, INT_V},
    {OFFSET (IRQCycleCount), 4, INT_V},
    {OFFSET (SA1NMICycleCount), 4, INT_V},
    {OFFSET (SA1IRQCycleCount), 4, INT_V},
    {OFFSET (NextAPUTimerPos), 4, INT_V},
    {OFFSET (APUTimerCounter), 4, INT_V},
    {OFFSET (APUTimerCounter_err), 4, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SRegisters *)

static FreezeData SnapRegisters [] = {
    {OFFSET (PB),  1, INT_V},
    {OFFSET (DB),  1, INT_V},
    {OFFSET (P.W), 2, INT_V},
    {OFFSET (A.W), 2, INT_V},
    {OFFSET (D.W), 2, INT_V},
    {OFFSET (S.W), 2, INT_V},
    {OFFSET (X.W), 2, INT_V},
    {OFFSET (Y.W), 2, INT_V},
    {OFFSET (PC),  2, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SPPU *)

static FreezeData SnapPPU [] = {
    {OFFSET (BGMode), 1, INT_V},
    {OFFSET (BG3Priority), 1, INT_V},
    {OFFSET (Brightness), 1, INT_V},
    {OFFSET (VMA.High), 1, INT_V},
    {OFFSET (VMA.Increment), 1, INT_V},
    {OFFSET (VMA.Address), 2, INT_V},
    {OFFSET (VMA.Mask1), 2, INT_V},
    {OFFSET (VMA.FullGraphicCount), 2, INT_V},
    {OFFSET (VMA.Shift), 2, INT_V},
    {OFFSET (BG[0].SCBase), 2, INT_V},
    {OFFSET (BG[0].VOffset), 2, INT_V},
    {OFFSET (BG[0].HOffset), 2, INT_V},
    {OFFSET (BG[0].BGSize), 1, INT_V},
    {OFFSET (BG[0].NameBase), 2, INT_V},
    {OFFSET (BG[0].SCSize), 2, INT_V},
	
    {OFFSET (BG[1].SCBase), 2, INT_V},
    {OFFSET (BG[1].VOffset), 2, INT_V},
    {OFFSET (BG[1].HOffset), 2, INT_V},
    {OFFSET (BG[1].BGSize), 1, INT_V},
    {OFFSET (BG[1].NameBase), 2, INT_V},
    {OFFSET (BG[1].SCSize), 2, INT_V},
	
    {OFFSET (BG[2].SCBase), 2, INT_V},
    {OFFSET (BG[2].VOffset), 2, INT_V},
    {OFFSET (BG[2].HOffset), 2, INT_V},
    {OFFSET (BG[2].BGSize), 1, INT_V},
    {OFFSET (BG[2].NameBase), 2, INT_V},
    {OFFSET (BG[2].SCSize), 2, INT_V},
	
    {OFFSET (BG[3].SCBase), 2, INT_V},
    {OFFSET (BG[3].VOffset), 2, INT_V},
    {OFFSET (BG[3].HOffset), 2, INT_V},
    {OFFSET (BG[3].BGSize), 1, INT_V},
    {OFFSET (BG[3].NameBase), 2, INT_V},
    {OFFSET (BG[3].SCSize), 2, INT_V},
	
    {OFFSET (CGFLIP), 1, INT_V},
    {OFFSET (CGDATA), 256, uint16_ARRAY_V},
    {OFFSET (FirstSprite), 1, INT_V},
#define O(N) \
    {OFFSET (OBJ[N].HPos), 2, INT_V}, \
    {OFFSET (OBJ[N].VPos), 2, INT_V}, \
    {OFFSET (OBJ[N].Name), 2, INT_V}, \
    {OFFSET (OBJ[N].VFlip), 1, INT_V}, \
    {OFFSET (OBJ[N].HFlip), 1, INT_V}, \
    {OFFSET (OBJ[N].Priority), 1, INT_V}, \
    {OFFSET (OBJ[N].Palette), 1, INT_V}, \
    {OFFSET (OBJ[N].Size), 1, INT_V}
	
    O(  0), O(  1), O(  2), O(  3), O(  4), O(  5), O(  6), O(  7),
    O(  8), O(  9), O( 10), O( 11), O( 12), O( 13), O( 14), O( 15),
    O( 16), O( 17), O( 18), O( 19), O( 20), O( 21), O( 22), O( 23),
    O( 24), O( 25), O( 26), O( 27), O( 28), O( 29), O( 30), O( 31),
    O( 32), O( 33), O( 34), O( 35), O( 36), O( 37), O( 38), O( 39),
    O( 40), O( 41), O( 42), O( 43), O( 44), O( 45), O( 46), O( 47),
    O( 48), O( 49), O( 50), O( 51), O( 52), O( 53), O( 54), O( 55),
    O( 56), O( 57), O( 58), O( 59), O( 60), O( 61), O( 62), O( 63),
    O( 64), O( 65), O( 66), O( 67), O( 68), O( 69), O( 70), O( 71),
    O( 72), O( 73), O( 74), O( 75), O( 76), O( 77), O( 78), O( 79),
    O( 80), O( 81), O( 82), O( 83), O( 84), O( 85), O( 86), O( 87),
    O( 88), O( 89), O( 90), O( 91), O( 92), O( 93), O( 94), O( 95),
    O( 96), O( 97), O( 98), O( 99), O(100), O(101), O(102), O(103),
    O(104), O(105), O(106), O(107), O(108), O(109), O(110), O(111),
    O(112), O(113), O(114), O(115), O(116), O(117), O(118), O(119),
    O(120), O(121), O(122), O(123), O(124), O(125), O(126), O(127),
#undef O
    {OFFSET (OAMPriorityRotation), 1, INT_V},
    {OFFSET (OAMAddr), 2, INT_V},
    {OFFSET (OAMFlip), 1, INT_V},
    {OFFSET (OAMTileAddress), 2, INT_V},
    {OFFSET (IRQVBeamPos), 2, INT_V},
    {OFFSET (IRQHBeamPos), 2, INT_V},
    {OFFSET (VBeamPosLatched), 2, INT_V},
    {OFFSET (HBeamPosLatched), 2, INT_V},
    {OFFSET (HBeamFlip), 1, INT_V},
    {OFFSET (VBeamFlip), 1, INT_V},
    {OFFSET (HVBeamCounterLatched), 1, INT_V},
    {OFFSET (MatrixA), 2, INT_V},
    {OFFSET (MatrixB), 2, INT_V},
    {OFFSET (MatrixC), 2, INT_V},
    {OFFSET (MatrixD), 2, INT_V},
    {OFFSET (CentreX), 2, INT_V},
    {OFFSET (CentreY), 2, INT_V},
    {OFFSET (Joypad1ButtonReadPos), 1, INT_V},
    {OFFSET (Joypad2ButtonReadPos), 1, INT_V},
    {OFFSET (Joypad3ButtonReadPos), 1, INT_V},
    {OFFSET (CGADD), 1, INT_V},
    {OFFSET (FixedColourRed), 1, INT_V},
    {OFFSET (FixedColourGreen), 1, INT_V},
    {OFFSET (FixedColourBlue), 1, INT_V},
    {OFFSET (SavedOAMAddr), 2, INT_V},
    {OFFSET (ScreenHeight), 2, INT_V},
    {OFFSET (WRAM), 4, INT_V},
    {OFFSET (ForcedBlanking), 1, INT_V},
    {OFFSET (OBJNameSelect), 2, INT_V},
    {OFFSET (OBJSizeSelect), 1, INT_V},
    {OFFSET (OBJNameBase), 2, INT_V},
    {OFFSET (OAMReadFlip), 1, INT_V},
    {OFFSET (VTimerEnabled), 1, INT_V},
    {OFFSET (HTimerEnabled), 1, INT_V},
    {OFFSET (HTimerPosition), 2, INT_V},
    {OFFSET (Mosaic), 1, INT_V},
    {OFFSET (Mode7HFlip), 1, INT_V},
    {OFFSET (Mode7VFlip), 1, INT_V},
    {OFFSET (Mode7Repeat), 1, INT_V},
    {OFFSET (Window1Left), 1, INT_V},
    {OFFSET (Window1Right), 1, INT_V},
    {OFFSET (Window2Left), 1, INT_V},
    {OFFSET (Window2Right), 1, INT_V},
#define O(N) \
    {OFFSET (ClipWindowOverlapLogic[N]), 1, INT_V}, \
    {OFFSET (ClipWindow1Enable[N]), 1, INT_V}, \
    {OFFSET (ClipWindow2Enable[N]), 1, INT_V}, \
    {OFFSET (ClipWindow1Inside[N]), 1, INT_V}, \
    {OFFSET (ClipWindow2Inside[N]), 1, INT_V}
	
    O(0), O(1), O(2), O(3), O(4), O(5),
	
#undef O
	
    {OFFSET (CGFLIPRead), 1, INT_V},
    {OFFSET (Need16x8Mulitply), 1, INT_V},
    {OFFSET (BGMosaic), 4, uint8_ARRAY_V},
    {OFFSET (OAMData), 512 + 32, uint8_ARRAY_V},
    {OFFSET (Need16x8Mulitply), 1, INT_V},
    {OFFSET (MouseSpeed), 2, uint8_ARRAY_V}
};


#if 1

#undef OFFSET
#define OFFSET(f) Offset(f,struct SDMA *)

static FreezeData SnapDMA [] = {
#define O(N) \
    {OFFSET (TransferDirection) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (AAddressFixed) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (AAddressDecrement) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (TransferMode) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (ABank) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (AAddress) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (Address) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (BAddress) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (TransferBytes) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (HDMAIndirectAddressing) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (IndirectAddress) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (IndirectBank) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (Repeat) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (LineCount) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (FirstLine) + N * sizeof (struct SDMA), 1, INT_V}

    O(0), O(1), O(2), O(3), O(4), O(5), O(6), O(7)
#undef O
};

#else

#undef OFFSET
#define OFFSET(f) Offset(f,struct SDMA *)

static FreezeData SnapDMA [] = {
#define O(N) \
    {OFFSET (ReverseTransfer) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (HDMAIndirectAddressing) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (UnusedBit43x0) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (AAddressFixed) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (AAddressDecrement) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (TransferMode) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (BAddress) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (AAddress) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (ABank) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (DMACount_Or_HDMAIndirectAddress) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (IndirectBank) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (Address) + N * sizeof (struct SDMA), 2, INT_V}, \
    {OFFSET (Repeat) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (LineCount) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (UnknownByte) + N * sizeof (struct SDMA), 1, INT_V}, \
    {OFFSET (DoTransfer) + N * sizeof (struct SDMA), 1, INT_V} 

    O(0), O(1), O(2), O(3), O(4), O(5), O(6), O(7)
#undef O
};

#endif

#undef OFFSET
#define OFFSET(f) Offset(f,struct SAPU *)

static FreezeData SnapAPU [] = {
    {OFFSET (Cycles), 4, INT_V},
    {OFFSET (ShowROM), 1, INT_V},
    {OFFSET (Flags), 1, INT_V},
    {OFFSET (KeyedChannels), 1, INT_V},
    {OFFSET (OutPorts), 4, uint8_ARRAY_V},
    {OFFSET (DSP), 0x80, uint8_ARRAY_V},
    {OFFSET (ExtraRAM), 64, uint8_ARRAY_V},
    {OFFSET (Timer), 3, uint16_ARRAY_V},
    {OFFSET (TimerTarget), 3, uint16_ARRAY_V},
    {OFFSET (TimerEnabled), 3, uint8_ARRAY_V},
    {OFFSET (TimerValueWritten), 3, uint8_ARRAY_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SAPURegisters *)

static FreezeData SnapAPURegisters [] = {
    {OFFSET (P), 1, INT_V},
    {OFFSET (YA.W), 2, INT_V},
    {OFFSET (X), 1, INT_V},
    {OFFSET (S), 1, INT_V},
    {OFFSET (PC), 2, INT_V},
};

#undef OFFSET
#define OFFSET(f) Offset(f,SSoundData *)

static FreezeData SnapSoundData [] = {
    {OFFSET (master_volume_left), 2, INT_V},
    {OFFSET (master_volume_right), 2, INT_V},
    {OFFSET (echo_volume_left), 2, INT_V},
    {OFFSET (echo_volume_right), 2, INT_V},
    {OFFSET (echo_enable), 4, INT_V},
    {OFFSET (echo_feedback), 4, INT_V},
    {OFFSET (echo_ptr), 4, INT_V},
    {OFFSET (echo_buffer_size), 4, INT_V},
    {OFFSET (echo_write_enabled), 4, INT_V},
    {OFFSET (echo_channel_enable), 4, INT_V},
    {OFFSET (pitch_mod), 4, INT_V},
    {OFFSET (dummy), 3, uint32_ARRAY_V},
#define O(N) \
    {OFFSET (channels [N].state), 4, INT_V}, \
    {OFFSET (channels [N].type), 4, INT_V}, \
    {OFFSET (channels [N].volume_left), 2, INT_V}, \
    {OFFSET (channels [N].volume_right), 2, INT_V}, \
    {OFFSET (channels [N].hertz), 4, INT_V}, \
    {OFFSET (channels [N].count), 4, INT_V}, \
    {OFFSET (channels [N].loop), 1, INT_V}, \
    {OFFSET (channels [N].envx), 4, INT_V}, \
    {OFFSET (channels [N].left_vol_level), 2, INT_V}, \
    {OFFSET (channels [N].right_vol_level), 2, INT_V}, \
    {OFFSET (channels [N].envx_target), 2, INT_V}, \
    {OFFSET (channels [N].env_error), 4, INT_V}, \
    {OFFSET (channels [N].erate), 4, INT_V}, \
    {OFFSET (channels [N].direction), 4, INT_V}, \
    {OFFSET (channels [N].attack_rate), 4, INT_V}, \
    {OFFSET (channels [N].decay_rate), 4, INT_V}, \
    {OFFSET (channels [N].sustain_rate), 4, INT_V}, \
    {OFFSET (channels [N].release_rate), 4, INT_V}, \
    {OFFSET (channels [N].sustain_level), 4, INT_V}, \
    {OFFSET (channels [N].sample), 2, INT_V}, \
    {OFFSET (channels [N].decoded), 16, uint16_ARRAY_V}, \
    {OFFSET (channels [N].previous16), 2, uint16_ARRAY_V}, \
    {OFFSET (channels [N].sample_number), 2, INT_V}, \
    {OFFSET (channels [N].last_block), 1, INT_V}, \
    {OFFSET (channels [N].needs_decode), 1, INT_V}, \
    {OFFSET (channels [N].block_pointer), 4, INT_V}, \
    {OFFSET (channels [N].sample_pointer), 4, INT_V}, \
    {OFFSET (channels [N].mode), 4, INT_V}
	
    O(0), O(1), O(2), O(3), O(4), O(5), O(6), O(7)
#undef O
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SControlSnapshot *)

static FreezeData SnapControls [] = {
    {OFFSET (ver), 1, INT_V},
    {OFFSET (port1_read_idx), 2, uint8_ARRAY_V},
    {OFFSET (dummy1), 4, uint8_ARRAY_V},
    {OFFSET (port2_read_idx), 2, uint8_ARRAY_V},
    {OFFSET (dummy2), 4, uint8_ARRAY_V},
    {OFFSET (mouse_speed), 2, uint8_ARRAY_V},
    {OFFSET (justifier_select), 1, INT_V},
    {OFFSET (dummy3), 8, uint8_ARRAY_V},
    {OFFSET (pad_read), 1, INT_V},
    {OFFSET (pad_read_last), 1, INT_V},
    {OFFSET (internal), 60, uint8_ARRAY_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct STimings *)

static FreezeData SnapTimings [] = {
    {OFFSET (H_Max_Master), 4, INT_V},
    {OFFSET (H_Max), 4, INT_V},
    {OFFSET (V_Max_Master), 4, INT_V},
    {OFFSET (V_Max), 4, INT_V},
    {OFFSET (HBlankStart), 4, INT_V},
    {OFFSET (HBlankEnd), 4, INT_V},
    {OFFSET (HDMAInit), 4, INT_V},
    {OFFSET (HDMAStart), 4, INT_V},
    {OFFSET (NMITriggerPos), 4, INT_V},
    {OFFSET (WRAMRefreshPos), 4, INT_V},
    {OFFSET (RenderPos), 4, INT_V},
    {OFFSET (InterlaceField), 1, INT_V},
    {OFFSET (DMACPUSync), 4, INT_V},
    {OFFSET (NMIDMADelay), 4, INT_V},
    {OFFSET (IRQPendCount), 4, INT_V},
    {OFFSET (APUSpeedup), 4, INT_V}
};

#ifndef ZSNES_FX

#undef OFFSET
#define OFFSET(f) Offset(f,struct FxRegs_s *)

#undef OFFSET2
#define OFFSET2(f1,f2) Offset2(f1,f2,struct FxRegs_s *)

static FreezeData SnapFX [] = {
    {OFFSET (avReg), 16, uint32_ARRAY_V},
    {OFFSET (vColorReg), 4, INT_V},
    {OFFSET (vPlotOptionReg), 4, INT_V},
    {OFFSET (vStatusReg), 4, INT_V},
    {OFFSET (vPrgBankReg), 4, INT_V},
    {OFFSET (vRomBankReg), 4, INT_V},
    {OFFSET (vRamBankReg), 4, INT_V},
    {OFFSET (vCacheBaseReg), 4, INT_V},
    {OFFSET (vCacheFlags), 4, INT_V},
    {OFFSET (vLastRamAdr), 4, INT_V},
    {OFFSET2 (pvDreg, avRegAddr), 4, POINTER_V},
    {OFFSET2 (pvSreg, avRegAddr), 4, POINTER_V},
    {OFFSET (vRomBuffer), 1, INT_V},
    {OFFSET (vPipe), 1, INT_V},
    {OFFSET (vPipeAdr), 4, INT_V},
    {OFFSET (vSign), 4, INT_V},
    {OFFSET (vZero), 4, INT_V},
    {OFFSET (vCarry), 4, INT_V},
    {OFFSET (vOverflow), 4, INT_V},
    {OFFSET (vErrorCode), 4, INT_V},
    {OFFSET (vIllegalAddress), 4, INT_V},
    {OFFSET (bBreakPoint), 1, INT_V},
    {OFFSET (vBreakPoint), 4, INT_V},
    {OFFSET (vStepPoint), 4, INT_V},
    {OFFSET (nRamBanks), 4, INT_V},
    {OFFSET (nRomBanks), 4, INT_V},
    {OFFSET (vMode), 4, INT_V},
    {OFFSET (vPrevMode), 4, INT_V},
    {OFFSET2 (pvScreenBase, pvRam), 4, POINTER_V},
#define O(N) \
	{OFFSET2 (apvScreen[N], pvRam), 4, POINTER_V}
	O(  0), O(  1), O(  2), O(  3), O(  4), O(  5), O(  6), O(  7),
	O(  8), O(  9), O( 10), O( 11), O( 12), O( 13), O( 14), O( 15),
	O( 16), O( 17), O( 18), O( 19), O( 20), O( 21), O( 22), O( 23),
	O( 24), O( 25), O( 26), O( 27), O( 28), O( 29), O( 30), O( 31),
#undef O
    {OFFSET (x), 32, uint32_ARRAY_V},
    {OFFSET (vScreenHeight), 4, INT_V},
    {OFFSET (vScreenRealHeight), 4, INT_V},
    {OFFSET (vPrevScreenHeight), 4, INT_V},
    {OFFSET (vScreenSize), 4, INT_V},
    {OFFSET2 (pvRamBank, apvRamBank), 4, POINTER_V},
    {OFFSET2 (pvRomBank, apvRomBank), 4, POINTER_V},
    {OFFSET2 (pvPrgBank, apvRomBank), 4, POINTER_V},
#define O(N) \
	{OFFSET2 (apvRamBank[N], pvRam), 4, POINTER_V}
	O(0), O(1), O(2), O(3),
#undef O
    {OFFSET (bCacheActive), 1, INT_V},
    {OFFSET2 (pvCache, pvRegisters), 4, POINTER_V},
    {OFFSET (avCacheBackup), 512, uint8_ARRAY_V},
    {OFFSET (vCounter), 4, INT_V},
    {OFFSET (vInstCount), 4, INT_V},
    {OFFSET (vSCBRDirty), 4, INT_V}
};

#undef OFFSET2

#endif

#undef OFFSET
#define OFFSET(f) Offset(f,struct SSA1Registers *)

static FreezeData SnapSA1Registers [] = {
    {OFFSET (PB),  1, INT_V},
    {OFFSET (DB),  1, INT_V},
    {OFFSET (P.W), 2, INT_V},
    {OFFSET (A.W), 2, INT_V},
    {OFFSET (D.W), 2, INT_V},
    {OFFSET (S.W), 2, INT_V},
    {OFFSET (X.W), 2, INT_V},
    {OFFSET (Y.W), 2, INT_V},
    {OFFSET (PC),  2, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SSA1 *)

static FreezeData SnapSA1 [] = {
    {OFFSET (Flags), 4, INT_V},
    {OFFSET (NMIActive), 1, INT_V},
    {OFFSET (IRQActive), 1, INT_V},
    {OFFSET (WaitingForInterrupt), 1, INT_V},
    {OFFSET (op1), 2, INT_V},
    {OFFSET (op2), 2, INT_V},
    {OFFSET (arithmetic_op), 4, INT_V},
    {OFFSET (sum), 8, INT_V},
    {OFFSET (overflow), 1, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SDSP1 *)

static FreezeData SnapDSP1 [] = {
    {OFFSET (waiting4command), 1, INT_V},
    {OFFSET (first_parameter), 1, INT_V},
    {OFFSET (command), 1, INT_V},
    {OFFSET (in_count), 4, INT_V},
    {OFFSET (in_index), 4, INT_V},
    {OFFSET (out_count), 4, INT_V},
    {OFFSET (out_index), 4, INT_V},
    {OFFSET (parameters), 512, uint8_ARRAY_V},
    {OFFSET (output), 512, uint8_ARRAY_V},
    {OFFSET (CentreX), 2, INT_V},
    {OFFSET (CentreY), 2, INT_V},
    {OFFSET (VOffset), 2, INT_V},
    {OFFSET (VPlane_C), 2, INT_V},
    {OFFSET (VPlane_E), 2, INT_V},
    {OFFSET (SinAas), 2, INT_V},
    {OFFSET (CosAas), 2, INT_V},
    {OFFSET (SinAzs), 2, INT_V},
    {OFFSET (CosAzs), 2, INT_V},
    {OFFSET (SinAZS), 2, INT_V},
    {OFFSET (CosAZS), 2, INT_V},
    {OFFSET (SecAZS_C1), 2, INT_V},
    {OFFSET (SecAZS_E1), 2, INT_V},
    {OFFSET (SecAZS_C2), 2, INT_V},
    {OFFSET (SecAZS_E2), 2, INT_V},
    {OFFSET (Nx), 2, INT_V},
    {OFFSET (Ny), 2, INT_V},
    {OFFSET (Nz), 2, INT_V},
    {OFFSET (Gx), 2, INT_V},
    {OFFSET (Gy), 2, INT_V},
    {OFFSET (Gz), 2, INT_V},
    {OFFSET (C_Les), 2, INT_V},
    {OFFSET (E_Les), 2, INT_V},
    {OFFSET (G_Les), 2, INT_V},
#define O(N) \
    {OFFSET (matrixA[N]), 3, uint16_ARRAY_V},\
    {OFFSET (matrixB[N]), 3, uint16_ARRAY_V},\
    {OFFSET (matrixC[N]), 3, uint16_ARRAY_V}
    O(0), O(1), O(2),
#undef O
    {OFFSET (Op00Multiplicand), 2, INT_V},
    {OFFSET (Op00Multiplier), 2, INT_V},
    {OFFSET (Op00Result), 2, INT_V},
    {OFFSET (Op20Multiplicand), 2, INT_V},
    {OFFSET (Op20Multiplier), 2, INT_V},
    {OFFSET (Op20Result), 2, INT_V},
    {OFFSET (Op10Coefficient), 2, INT_V},
    {OFFSET (Op10Exponent), 2, INT_V},
    {OFFSET (Op10CoefficientR), 2, INT_V},
    {OFFSET (Op10ExponentR), 2, INT_V},
    {OFFSET (Op04Angle), 2, INT_V},
    {OFFSET (Op04Radius), 2, INT_V},
    {OFFSET (Op04Sin), 2, INT_V},
    {OFFSET (Op04Cos), 2, INT_V},
    {OFFSET (Op0CA), 2, INT_V},
    {OFFSET (Op0CX1), 2, INT_V},
    {OFFSET (Op0CY1), 2, INT_V},
    {OFFSET (Op0CX2), 2, INT_V},
    {OFFSET (Op0CY2), 2, INT_V},
    {OFFSET (Op02FX), 2, INT_V},
    {OFFSET (Op02FY), 2, INT_V},
    {OFFSET (Op02FZ), 2, INT_V},
    {OFFSET (Op02LFE), 2, INT_V},
    {OFFSET (Op02LES), 2, INT_V},
    {OFFSET (Op02AAS), 2, INT_V},
    {OFFSET (Op02AZS), 2, INT_V},
    {OFFSET (Op02VOF), 2, INT_V},
    {OFFSET (Op02VVA), 2, INT_V},
    {OFFSET (Op02CX), 2, INT_V},
    {OFFSET (Op02CY), 2, INT_V},
    {OFFSET (Op0AVS), 2, INT_V},
    {OFFSET (Op0AA), 2, INT_V},
    {OFFSET (Op0AB), 2, INT_V},
    {OFFSET (Op0AC), 2, INT_V},
    {OFFSET (Op0AD), 2, INT_V},
    {OFFSET (Op06X), 2, INT_V},
    {OFFSET (Op06Y), 2, INT_V},
    {OFFSET (Op06Z), 2, INT_V},
    {OFFSET (Op06H), 2, INT_V},
    {OFFSET (Op06V), 2, INT_V},
    {OFFSET (Op06M), 2, INT_V},
    {OFFSET (Op01m), 2, INT_V},
    {OFFSET (Op01Zr), 2, INT_V},
    {OFFSET (Op01Xr), 2, INT_V},
    {OFFSET (Op01Yr), 2, INT_V},
    {OFFSET (Op11m), 2, INT_V},
    {OFFSET (Op11Zr), 2, INT_V},
    {OFFSET (Op11Xr), 2, INT_V},
    {OFFSET (Op11Yr), 2, INT_V},
    {OFFSET (Op21m), 2, INT_V},
    {OFFSET (Op21Zr), 2, INT_V},
    {OFFSET (Op21Xr), 2, INT_V},
    {OFFSET (Op21Yr), 2, INT_V},
    {OFFSET (Op0DX), 2, INT_V},
    {OFFSET (Op0DY), 2, INT_V},
    {OFFSET (Op0DZ), 2, INT_V},
    {OFFSET (Op0DF), 2, INT_V},
    {OFFSET (Op0DL), 2, INT_V},
    {OFFSET (Op0DU), 2, INT_V},
    {OFFSET (Op1DX), 2, INT_V},
    {OFFSET (Op1DY), 2, INT_V},
    {OFFSET (Op1DZ), 2, INT_V},
    {OFFSET (Op1DF), 2, INT_V},
    {OFFSET (Op1DL), 2, INT_V},
    {OFFSET (Op1DU), 2, INT_V},
    {OFFSET (Op2DX), 2, INT_V},
    {OFFSET (Op2DY), 2, INT_V},
    {OFFSET (Op2DZ), 2, INT_V},
    {OFFSET (Op2DF), 2, INT_V},
    {OFFSET (Op2DL), 2, INT_V},
    {OFFSET (Op2DU), 2, INT_V},
    {OFFSET (Op03F), 2, INT_V},
    {OFFSET (Op03L), 2, INT_V},
    {OFFSET (Op03U), 2, INT_V},
    {OFFSET (Op03X), 2, INT_V},
    {OFFSET (Op03Y), 2, INT_V},
    {OFFSET (Op03Z), 2, INT_V},
    {OFFSET (Op13F), 2, INT_V},
    {OFFSET (Op13L), 2, INT_V},
    {OFFSET (Op13U), 2, INT_V},
    {OFFSET (Op13X), 2, INT_V},
    {OFFSET (Op13Y), 2, INT_V},
    {OFFSET (Op13Z), 2, INT_V},
    {OFFSET (Op23F), 2, INT_V},
    {OFFSET (Op23L), 2, INT_V},
    {OFFSET (Op23U), 2, INT_V},
    {OFFSET (Op23X), 2, INT_V},
    {OFFSET (Op23Y), 2, INT_V},
    {OFFSET (Op23Z), 2, INT_V},
    {OFFSET (Op14Zr), 2, INT_V},
    {OFFSET (Op14Xr), 2, INT_V},
    {OFFSET (Op14Yr), 2, INT_V},
    {OFFSET (Op14U), 2, INT_V},
    {OFFSET (Op14F), 2, INT_V},
    {OFFSET (Op14L), 2, INT_V},
    {OFFSET (Op14Zrr), 2, INT_V},
    {OFFSET (Op14Xrr), 2, INT_V},
    {OFFSET (Op14Yrr), 2, INT_V},
    {OFFSET (Op0EH), 2, INT_V},
    {OFFSET (Op0EV), 2, INT_V},
    {OFFSET (Op0EX), 2, INT_V},
    {OFFSET (Op0EY), 2, INT_V},
    {OFFSET (Op0BX), 2, INT_V},
    {OFFSET (Op0BY), 2, INT_V},
    {OFFSET (Op0BZ), 2, INT_V},
    {OFFSET (Op0BS), 2, INT_V},
    {OFFSET (Op1BX), 2, INT_V},
    {OFFSET (Op1BY), 2, INT_V},
    {OFFSET (Op1BZ), 2, INT_V},
    {OFFSET (Op1BS), 2, INT_V},
    {OFFSET (Op2BX), 2, INT_V},
    {OFFSET (Op2BY), 2, INT_V},
    {OFFSET (Op2BZ), 2, INT_V},
    {OFFSET (Op2BS), 2, INT_V},
    {OFFSET (Op28X), 2, INT_V},
    {OFFSET (Op28Y), 2, INT_V},
    {OFFSET (Op28Z), 2, INT_V},
    {OFFSET (Op28R), 2, INT_V},
    {OFFSET (Op1CX), 2, INT_V},
    {OFFSET (Op1CY), 2, INT_V},
    {OFFSET (Op1CZ), 2, INT_V},
    {OFFSET (Op1CXBR), 2, INT_V},
    {OFFSET (Op1CYBR), 2, INT_V},
    {OFFSET (Op1CZBR), 2, INT_V},
    {OFFSET (Op1CXAR), 2, INT_V},
    {OFFSET (Op1CYAR), 2, INT_V},
    {OFFSET (Op1CZAR), 2, INT_V},
    {OFFSET (Op1CX1), 2, INT_V},
    {OFFSET (Op1CY1), 2, INT_V},
    {OFFSET (Op1CZ1), 2, INT_V},
    {OFFSET (Op1CX2), 2, INT_V},
    {OFFSET (Op1CY2), 2, INT_V},
    {OFFSET (Op1CZ2), 2, INT_V},
    {OFFSET (Op0FRamsize), 2, INT_V},
    {OFFSET (Op0FPass), 2, INT_V},
    {OFFSET (Op2FUnknown), 2, INT_V},
    {OFFSET (Op2FSize), 2, INT_V},
    {OFFSET (Op08X), 2, INT_V},
    {OFFSET (Op08Y), 2, INT_V},
    {OFFSET (Op08Z), 2, INT_V},
    {OFFSET (Op08Ll), 2, INT_V},
    {OFFSET (Op08Lh), 2, INT_V},
    {OFFSET (Op18X), 2, INT_V},
    {OFFSET (Op18Y), 2, INT_V},
    {OFFSET (Op18Z), 2, INT_V},
    {OFFSET (Op18R), 2, INT_V},
    {OFFSET (Op18D), 2, INT_V},
    {OFFSET (Op38X), 2, INT_V},
    {OFFSET (Op38Y), 2, INT_V},
    {OFFSET (Op38Z), 2, INT_V},
    {OFFSET (Op38R), 2, INT_V},
    {OFFSET (Op38D), 2, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SDSP2 *)

static FreezeData SnapDSP2 [] = {
    {OFFSET (waiting4command), 1, INT_V},
    {OFFSET (command), 1, INT_V},
    {OFFSET (in_count), 4, INT_V},
    {OFFSET (in_index), 4, INT_V},
    {OFFSET (out_count), 4, INT_V},
    {OFFSET (out_index), 4, INT_V},
    {OFFSET (parameters), 512, uint8_ARRAY_V},
    {OFFSET (output), 512, uint8_ARRAY_V},
    {OFFSET (Op05HasLen), 1, INT_V},
    {OFFSET (Op05Len), 4, INT_V},
    {OFFSET (Op05Transparent), 1, INT_V},
    {OFFSET (Op06HasLen), 1, INT_V},
    {OFFSET (Op06Len), 4, INT_V},
    {OFFSET (Op09Word1), 2, INT_V},
    {OFFSET (Op09Word2), 2, INT_V},
    {OFFSET (Op0DHasLen), 1, INT_V},
    {OFFSET (Op0DOutLen), 4, INT_V},
    {OFFSET (Op0DInLen), 4, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SDSP4 *)

static FreezeData SnapDSP4 [] = {
    {OFFSET (waiting4command), 1, INT_V},
    {OFFSET (half_command), 1, INT_V},
    {OFFSET (command), 2, INT_V},
    {OFFSET (in_count), 4, INT_V},
    {OFFSET (in_index), 4, INT_V},
    {OFFSET (out_count), 4, INT_V},
    {OFFSET (out_index), 4, INT_V},
    {OFFSET (parameters), 512, uint8_ARRAY_V},
    {OFFSET (output), 512, uint8_ARRAY_V},
    {OFFSET (byte), 1, INT_V},
    {OFFSET (address), 2, INT_V},
    {OFFSET (Logic), 1, INT_V},
    {OFFSET (lcv), 2, INT_V},
    {OFFSET (distance), 2, INT_V},
    {OFFSET (raster), 2, INT_V},
    {OFFSET (segments), 2, INT_V},
    {OFFSET (world_x), 4, INT_V},
    {OFFSET (world_y), 4, INT_V},
    {OFFSET (world_dx), 4, INT_V},
    {OFFSET (world_dy), 4, INT_V},
    {OFFSET (world_ddx), 2, INT_V},
    {OFFSET (world_ddy), 2, INT_V},
    {OFFSET (world_xenv), 4, INT_V},
    {OFFSET (world_yofs), 2, INT_V},
    {OFFSET (view_x1), 2, INT_V},
    {OFFSET (view_y1), 2, INT_V},
    {OFFSET (view_x2), 2, INT_V},
    {OFFSET (view_y2), 2, INT_V},
    {OFFSET (view_dx), 2, INT_V},
    {OFFSET (view_dy), 2, INT_V},
    {OFFSET (view_xofs1), 2, INT_V},
    {OFFSET (view_yofs1), 2, INT_V},
    {OFFSET (view_xofs2), 2, INT_V},
    {OFFSET (view_yofs2), 2, INT_V},
    {OFFSET (view_yofsenv), 2, INT_V},
    {OFFSET (view_turnoff_x), 2, INT_V},
    {OFFSET (view_turnoff_dx), 2, INT_V},
    {OFFSET (viewport_cx), 2, INT_V},
    {OFFSET (viewport_cy), 2, INT_V},
    {OFFSET (viewport_left), 2, INT_V},
    {OFFSET (viewport_right), 2, INT_V},
    {OFFSET (viewport_top), 2, INT_V},
    {OFFSET (viewport_bottom), 2, INT_V},
    {OFFSET (sprite_x), 2, INT_V},
    {OFFSET (sprite_y), 2, INT_V},
    {OFFSET (sprite_attr), 2, INT_V},
    {OFFSET (sprite_size), 1, INT_V},
    {OFFSET (sprite_clipy), 2, INT_V},
    {OFFSET (sprite_count), 2, INT_V},
#define O(N) \
    {OFFSET (poly_clipLf[N]), 2, uint16_ARRAY_V},	\
    {OFFSET (poly_clipRt[N]), 2, uint16_ARRAY_V},	\
    {OFFSET (poly_ptr[N]), 2, uint16_ARRAY_V},\
    {OFFSET (poly_raster[N]), 2, uint16_ARRAY_V},\
    {OFFSET (poly_top[N]), 2, uint16_ARRAY_V},\
    {OFFSET (poly_bottom[N]), 2, uint16_ARRAY_V},\
    {OFFSET (poly_cx[N]), 2, uint16_ARRAY_V}
    O(0), O(1),
#undef O
    {OFFSET (poly_start), 2, uint16_ARRAY_V},
    {OFFSET (poly_plane), 2, uint16_ARRAY_V},
    {OFFSET (OAM_attr), 16, uint16_ARRAY_V},
    {OFFSET (OAM_index), 2, INT_V},
    {OFFSET (OAM_bits), 2, INT_V},
    {OFFSET (OAM_RowMax), 2, INT_V},
    {OFFSET (OAM_Row), 32, uint16_ARRAY_V}
};


#undef OFFSET
#define OFFSET(f) Offset(f,struct SST010 *)

static FreezeData SnapST010 [] = {
    {OFFSET (input_params), 16, uint8_ARRAY_V},
    {OFFSET (output_params), 16, uint8_ARRAY_V},
    {OFFSET (op_reg), 1, INT_V},
    {OFFSET (execute), 1, INT_V},
    {OFFSET (control_enable), 1, INT_V}
};


#undef OFFSET
#define OFFSET(f) Offset(f,struct SOBC1 *)

static FreezeData SnapOBC1 [] = {
    {OFFSET (address), 2, INT_V},
    {OFFSET (basePtr), 2, INT_V},
    {OFFSET (shift), 2, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SSPC7110Snapshot *)

static FreezeData SnapSPC7110Snap [] = {
    {OFFSET (r4801), 1, INT_V},
    {OFFSET (r4802), 1, INT_V},
    {OFFSET (r4803), 1, INT_V},
    {OFFSET (r4804), 1, INT_V},
    {OFFSET (r4805), 1, INT_V},
    {OFFSET (r4806), 1, INT_V},
    {OFFSET (r4807), 1, INT_V},
    {OFFSET (r4808), 1, INT_V},
    {OFFSET (r4809), 1, INT_V},
    {OFFSET (r480a), 1, INT_V},
    {OFFSET (r480b), 1, INT_V},
    {OFFSET (r480c), 1, INT_V},
    {OFFSET (r4811), 1, INT_V},
    {OFFSET (r4812), 1, INT_V},
    {OFFSET (r4813), 1, INT_V},
    {OFFSET (r4814), 1, INT_V},
    {OFFSET (r4815), 1, INT_V},
    {OFFSET (r4816), 1, INT_V},
    {OFFSET (r4817), 1, INT_V},
    {OFFSET (r4818), 1, INT_V},
    {OFFSET (r481x), 1, INT_V},
    {OFFSET (r4814_latch), 1, INT_V},
    {OFFSET (r4815_latch), 1, INT_V},
    {OFFSET (r4820), 1, INT_V},
    {OFFSET (r4821), 1, INT_V},
    {OFFSET (r4822), 1, INT_V},
    {OFFSET (r4823), 1, INT_V},
    {OFFSET (r4824), 1, INT_V},
    {OFFSET (r4825), 1, INT_V},
    {OFFSET (r4826), 1, INT_V},
    {OFFSET (r4827), 1, INT_V},
    {OFFSET (r4828), 1, INT_V},
    {OFFSET (r4829), 1, INT_V},
    {OFFSET (r482a), 1, INT_V},
    {OFFSET (r482b), 1, INT_V},
    {OFFSET (r482c), 1, INT_V},
    {OFFSET (r482d), 1, INT_V},
    {OFFSET (r482e), 1, INT_V},
    {OFFSET (r482f), 1, INT_V},
    {OFFSET (r4830), 1, INT_V},
    {OFFSET (r4831), 1, INT_V},
    {OFFSET (r4832), 1, INT_V},
    {OFFSET (r4833), 1, INT_V},
    {OFFSET (r4834), 1, INT_V},
    {OFFSET (dx_offset), 4, INT_V},
    {OFFSET (ex_offset), 4, INT_V},
    {OFFSET (fx_offset), 4, INT_V},
    {OFFSET (r4840), 1, INT_V},
    {OFFSET (r4841), 1, INT_V},
    {OFFSET (r4842), 1, INT_V},
    {OFFSET (rtc_state), 4, INT_V},
    {OFFSET (rtc_mode), 4, INT_V},
    {OFFSET (rtc_index), 4, INT_V},
    {OFFSET (decomp_mode), 4, INT_V},
    {OFFSET (decomp_offset), 4, INT_V},
    {OFFSET (decomp_buffer), SPC7110_DECOMP_BUFFER_SIZE, uint8_ARRAY_V},
    {OFFSET (decomp_buffer_rdoffset), 4, INT_V},
    {OFFSET (decomp_buffer_wroffset), 4, INT_V},
    {OFFSET (decomp_buffer_length), 4, INT_V},
#define O(N) \
    {OFFSET (context[N].index), 1, INT_V}, \
    {OFFSET (context[N].invert), 1, INT_V}
    O(  0), O(  1), O(  2), O(  3), O(  4), O(  5), O(  6), O(  7),
    O(  8), O(  9), O( 10), O( 11), O( 12), O( 13), O( 14), O( 15),
    O( 16), O( 17), O( 18), O( 19), O( 20), O( 21), O( 22), O( 23),
    O( 24), O( 25), O( 26), O( 27), O( 28), O( 29), O( 30), O( 31)
#undef O
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SSRTCSnapshot *)

static FreezeData SnapSRTCSnap [] = {
    {OFFSET (rtc_mode), 4, INT_V},
    {OFFSET (rtc_index), 4, INT_V}
};

#undef OFFSET
#define OFFSET(f) Offset(f,struct SBSX *)

static FreezeData SnapBSX [] = {
    {OFFSET (dirty), 1, INT_V},
    {OFFSET (dirty2), 1, INT_V},
    {OFFSET (bootup), 1, INT_V},
    {OFFSET (flash_enable), 1, INT_V},
    {OFFSET (write_enable), 1, INT_V},
    {OFFSET (read_enable), 1, INT_V},
    {OFFSET (flash_command), 4, INT_V},
    {OFFSET (old_write), 4, INT_V},
    {OFFSET (new_write), 4, INT_V},
    {OFFSET (out_index), 1, INT_V},
    {OFFSET (output), 32, uint8_ARRAY_V},
    {OFFSET (PPU), 32, uint8_ARRAY_V},
    {OFFSET (MMC), 16, uint8_ARRAY_V},
    {OFFSET (prevMMC), 16, uint8_ARRAY_V},
    {OFFSET (test2192), 32, uint8_ARRAY_V}
};

#pragma pack(push,1)
typedef struct _NewPPU
{
	uint16	OAMWriteRegister;//new
	short	VTimerPosition;//new
	uint16	GunHLatch;//new
	uint16	GunVLatch;//new
	short	M7HOFS;//new
	short	M7VOFS;//new
	uint8	M7byte;//new
	uint8	HDMA;//new
	uint8	HDMAEnded;//new
	uint8	OpenBus1;
	uint8	OpenBus2;
} NewPPU, *NewPPUp;
#pragma pack(pop)

static char ROMFilename [_MAX_PATH];
//static char SnapshotFilename [_MAX_PATH];

static int UnfreezeBlock (STREAM, const char *, uint8 *, int);
static int UnfreezeBlockCopy (STREAM, const char *, uint8 **, int);
static int UnfreezeStruct (STREAM, const char *, void *, FreezeData *, int);
static int UnfreezeStructCopy (STREAM, const char *, uint8 **, FreezeData *, int);
static void UnfreezeStructFromCopy (void *, FreezeData *, int, uint8 *);
static void FreezeBlock (STREAM, const char *, uint8 *, int);
static void FreezeStruct (STREAM, const char *, void *, FreezeData *, int);

bool8 Snapshot (const char *filename)
{
    return (S9xFreezeGame (filename));
}

bool8 S9xFreezeGame (const char *filename)
{
    STREAM stream = NULL;
	
    if (S9xOpenSnapshotFile (filename, FALSE, &stream))
    {
		S9xFreezeToStream (stream);
		S9xCloseSnapshotFile (stream);

		if(S9xMovieActive())
		{
			sprintf(String, "Movie snapshot %s", S9xBasename (filename));
			S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, String);
		}
		else
		{
			sprintf(String, "Saved %s", S9xBasename (filename));
			S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, String);
		}

		return (TRUE);
    }
    return (FALSE);
}

bool8 S9xLoadSnapshot (const char *filename)
{
    return (S9xUnfreezeGame (filename));
}

bool8 S9xUnfreezeGame (const char *filename)
{
    STREAM snapshot = NULL;
    if (S9xOpenSnapshotFile (filename, TRUE, &snapshot))
    {
		int result;
		if ((result = S9xUnfreezeFromStream (snapshot)) != SUCCESS)
		{
			switch (result)
			{
			case WRONG_FORMAT:
				S9xMessage (S9X_ERROR, S9X_WRONG_FORMAT, 
					"File not in Snes9x freeze format");
				break;
			case WRONG_VERSION:
				S9xMessage (S9X_ERROR, S9X_WRONG_VERSION,
					"Incompatable Snes9x freeze file format version");
				break;
			case WRONG_MOVIE_SNAPSHOT:
				S9xMessage (S9X_ERROR, S9X_WRONG_MOVIE_SNAPSHOT, MOVIE_ERR_SNAPSHOT_WRONG_MOVIE);
				break;
			case NOT_A_MOVIE_SNAPSHOT:
				S9xMessage (S9X_ERROR, S9X_NOT_A_MOVIE_SNAPSHOT, MOVIE_ERR_SNAPSHOT_NOT_MOVIE);
				break;
			default:
			case FILE_NOT_FOUND:
				sprintf (String, "ROM image \"%s\" for freeze file not found",
					ROMFilename);
				S9xMessage (S9X_ERROR, S9X_ROM_NOT_FOUND, String);
				break;
			}
			S9xCloseSnapshotFile (snapshot);
			return (FALSE);
		}

		if(!S9xMovieActive())
		{
			sprintf(String, "Loaded %s", S9xBasename (filename));
			S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, String);
		}

		S9xCloseSnapshotFile (snapshot);

		return (TRUE);
    }
    return (FALSE);
}

void S9xFreezeToStream (STREAM stream)
{
    char buffer [1024];
    int i;
	
    S9xSetSoundMute (TRUE);
#ifdef ZSNES_FX
    if (Settings.SuperFX)
		S9xSuperFXPreSaveState ();
#endif

    for (i = 0; i < 8; i++)
    {
		SoundData.channels [i].previous16 [0] = (int16) SoundData.channels [i].previous [0];
		SoundData.channels [i].previous16 [1] = (int16) SoundData.channels [i].previous [1];
    }
    sprintf (buffer, "%s:%04d\n", SNAPSHOT_MAGIC, SNAPSHOT_VERSION);
    WRITE_STREAM (buffer, strlen (buffer), stream);
    sprintf (buffer, "NAM:%06d:%s%c", strlen (Memory.ROMFilename) + 1,
		Memory.ROMFilename, 0);
    WRITE_STREAM (buffer, strlen (buffer) + 1, stream);

	// Store Registers
    //Registers.PCw = CPU.PC - CPU.PCBase;
    S9xPackStatus ();

    FreezeStruct (stream, "CPU", &CPU, SnapCPU, COUNT (SnapCPU));
    FreezeStruct (stream, "REG", &Registers, SnapRegisters, COUNT (SnapRegisters));

    struct SControlSnapshot	ctl_snap;
    S9xControlPreSaveState(&ctl_snap);

    //old snapshot
    PPU.Joypad1ButtonReadPos = ctl_snap.port1_read_idx[0];
    PPU.Joypad2ButtonReadPos = ctl_snap.port2_read_idx[0];
    PPU.Joypad3ButtonReadPos = ctl_snap.port2_read_idx[1];
    PPU.MouseSpeed[0] = ctl_snap.mouse_speed[0];
    PPU.MouseSpeed[1] = ctl_snap.mouse_speed[1];

    FreezeStruct (stream, "PPU", &PPU, SnapPPU, COUNT (SnapPPU));

    FreezeStruct (stream, "DMA", DMA, SnapDMA, COUNT (SnapDMA));

	// RAM and VRAM
    FreezeBlock (stream, "VRA", Memory.VRAM, 0x10000);
    FreezeBlock (stream, "RAM", Memory.RAM, 0x20000);
    FreezeBlock (stream, "SRA", Memory.SRAM, 0x20000);

    //seiken2 it's needed saving this value for old snapshot
    for(i = 0; i < 8; i++)
      Memory.FillRAM[0x430F + (i << 4)] = DMA[i].DoTransfer;

    //Save new valuables
    NewPPUp npp = (NewPPUp)Memory.FillRAM;
    npp->OAMWriteRegister = PPU.OAMWriteRegister;
    npp->VTimerPosition = PPU.VTimerPosition;
    npp->GunHLatch = PPU.GunHLatch;
    npp->GunVLatch = PPU.GunVLatch;
    npp->M7HOFS = PPU.M7HOFS;
    npp->M7VOFS = PPU.M7VOFS;
    npp->M7byte = PPU.M7byte;
    npp->HDMA = PPU.HDMA;
    npp->HDMAEnded = PPU.HDMAEnded;
    npp->OpenBus1 = PPU.OpenBus1;
    npp->OpenBus2 = PPU.OpenBus2;

    WRITE_DWORD(&Memory.FillRAM[0x20], APU.Cycles_err);
    WRITE_DWORD(&Memory.FillRAM[0x24], APU.t_ticks_err);

    for(i = 0; i < 3; i++)
      {
	WRITE_DWORD(&Memory.FillRAM[0x28 + i * 4], APU.t_ticks[i]);
	Memory.FillRAM[0x28 + 12 + i] = APU.t_stage1_ticks[i];
	Memory.FillRAM[0x28 + 12 + 3 + i] = APU.t_current_line[i];
      }

    WRITE_DWORD(&Memory.FillRAM[0x3A], APU.t64kCycles);
    WRITE_DWORD(&Memory.FillRAM[0x3E], APU.t64kCycles_err);

    memcpy(&Memory.FillRAM[0x100], Loop, FIRBUF * sizeof(int16));
    Memory.FillRAM[0x100 + FIRBUF * sizeof(int16)] = firCur;

    FreezeBlock (stream, "FIL", Memory.FillRAM, 0x8000);
    if (Settings.APUEnabled)
    {
		// SNESAPU
		StoreAPUDSP();

		// Store APURegisters
		S9xAPUPackStatus ();

		//direct access iplrom
		memcpy(APU.ExtraRAM, &IAPU.RAM[0xffc0], 64);

		// APU
		FreezeStruct (stream, "APU", &APU, SnapAPU, COUNT (SnapAPU));
		FreezeStruct (stream, "ARE", &APURegisters, SnapAPURegisters,
			COUNT (SnapAPURegisters));
		FreezeBlock (stream, "ARA", IAPU.RAM, 0x10000);
		FreezeStruct (stream, "SOU", &SoundData, SnapSoundData,
			COUNT (SnapSoundData));
    }


#ifndef ZSNES_FX
	if (Settings.SuperFX)
	{
		GSU.avRegAddr = (uint8 *) &GSU.avReg;
		FreezeStruct(stream, "SFX", &GSU, SnapFX, COUNT(SnapFX));
	}
#endif

    if (Settings.SA1)
    {
      //SA1Registers.PCw = SA1.PC - SA1.PCBase;
		S9xSA1PackStatus ();
		FreezeStruct (stream, "SA1", &SA1, SnapSA1, COUNT (SnapSA1));
		FreezeStruct (stream, "SAR", &SA1Registers, SnapSA1Registers, 
			COUNT (SnapSA1Registers));
    }
	
	if (Settings.DSP == 1)
		FreezeStruct(stream, "DP1", &DSP1, SnapDSP1, COUNT(SnapDSP1));

	if (Settings.DSP == 2)
		FreezeStruct(stream, "DP2", &DSP2, SnapDSP2, COUNT(SnapDSP2));

	if (Settings.DSP == 4)
		FreezeStruct(stream, "DP4", &DSP4, SnapDSP4, COUNT(SnapDSP4));

	if (Settings.C4)
#ifndef ZSNES_C4
		FreezeBlock (stream, "CX4", Memory.C4RAM, 8192);
#else
		FreezeBlock (stream, "CX4", C4Ram, 8192);
#endif

	if (Settings.SETA == ST_010)
		FreezeStruct(stream, "ST0", &ST010, SnapST010, COUNT(SnapST010));

	if (Settings.OBC1)
	{
		FreezeStruct(stream, "OBC", &OBC1, SnapOBC1, COUNT(SnapOBC1));
		FreezeBlock (stream, "OBM", Memory.OBC1RAM, 8192);
	}

	if (Settings.SPC7110)
	{
		S9xSPC7110PreSaveState();
		FreezeStruct(stream, "S71", &s7snap, SnapSPC7110Snap, COUNT(SnapSPC7110Snap));
	}

	if (Settings.SRTC)
	{
		S9xSRTCPreSaveState();
		FreezeStruct(stream, "SRT", &srtcsnap, SnapSRTCSnap, COUNT(SnapSRTCSnap));
	}

	if (Settings.SRTC || Settings.SPC7110RTC)
		FreezeBlock (stream, "CLK", RTCData.reg, 20);

	if (Settings.BS)
		FreezeStruct(stream, "BSX", &BSX, SnapBSX, COUNT(SnapBSX));

	if (S9xMovieActive ())
	{
		uint8* movie_freeze_buf;
		uint32 movie_freeze_size;

		S9xMovieFreeze(&movie_freeze_buf, &movie_freeze_size);
		if(movie_freeze_buf)
		{
			struct SnapshotMovieInfo mi;
			mi.MovieInputDataSize = movie_freeze_size;
			FreezeStruct (stream, "MOV", &mi, SnapMovie, COUNT (SnapMovie));
		    FreezeBlock (stream, "MID", movie_freeze_buf, movie_freeze_size);
			delete [] movie_freeze_buf;
		}
	}

    if (Settings.BS)
    {
		FreezeBlock (stream, "PSR", Memory.BSRAM, 0x80000);
	}

    EXT.IRQCycleCount = CPU.IRQPending;
    FreezeStruct (stream, "EXT", &EXT, SnapEXT, COUNT (SnapEXT));
    FreezeStruct(stream, "CTL", &ctl_snap, SnapControls, COUNT(SnapControls));
    FreezeStruct (stream, "TIM", &Timings, SnapTimings, COUNT (SnapTimings));

	S9xSetSoundMute (FALSE);
#ifdef ZSNES_FX
	if (Settings.SuperFX)
		S9xSuperFXPostSaveState ();
#endif
}

int S9xUnfreezeFromStream (STREAM stream)
{
    char buffer [_MAX_PATH + 1];
    char rom_filename [_MAX_PATH + 1];
    int result;
	
    int version;
    int len = strlen (SNAPSHOT_MAGIC) + 1 + 4 + 1;
    if (READ_STREAM (buffer, len, stream) != len)
		return (WRONG_FORMAT);
    if (strncmp (buffer, SNAPSHOT_MAGIC, strlen (SNAPSHOT_MAGIC)) != 0)
		return (WRONG_FORMAT);
    if ((version = atoi (&buffer [strlen (SNAPSHOT_MAGIC) + 1])) > SNAPSHOT_VERSION)
		return (WRONG_VERSION);
	
    if ((result = UnfreezeBlock (stream, "NAM", (uint8 *) rom_filename, _MAX_PATH)) != SUCCESS)
		return (result);
	
    if (strcasecmp (rom_filename, Memory.ROMFilename) != 0 &&
		strcasecmp (S9xBasename (rom_filename), S9xBasename (Memory.ROMFilename)) != 0)
    {
		S9xMessage (S9X_WARNING, S9X_FREEZE_ROM_NAME,
			"Current loaded ROM image doesn't match that required by freeze-game file.");
    }
	
// ## begin load ##
	uint8* local_cpu = NULL;
	uint8* local_registers = NULL;
	uint8* local_ppu = NULL;
	uint8* local_dma = NULL;
	uint8* local_vram = NULL;
	uint8* local_ram = NULL;
	uint8* local_sram = NULL;
	uint8* local_fillram = NULL;
	uint8* local_apu = NULL;
	uint8* local_apu_registers = NULL;
	uint8* local_apu_ram = NULL;
	uint8* local_apu_sounddata = NULL;
	uint8	*local_superfx       = NULL;
	uint8* local_sa1 = NULL;
	uint8* local_sa1_registers = NULL;
	uint8	*local_dsp1          = NULL;
	uint8	*local_dsp2          = NULL;
	uint8	*local_dsp4          = NULL;
	uint8	*local_cx4_data      = NULL;
	uint8	*local_st010         = NULL;
	uint8	*local_obc1          = NULL;
	uint8	*local_obc1_data     = NULL;
	uint8* local_spc7110       = NULL;
	uint8* local_srtc          = NULL;
	uint8* local_rtc_data      = NULL;
	uint8	*local_bsx_data      = NULL;
	uint8* local_movie_data = NULL;
	uint8* local_bsram = NULL;
	uint8* local_ext = NULL;
	uint8	*local_control_data  = NULL;
	uint8* local_timing_data = NULL;

	do
	{
		if ((result = UnfreezeStructCopy (stream, "CPU", &local_cpu, SnapCPU, COUNT (SnapCPU))) != SUCCESS)
			break;
		if ((result = UnfreezeStructCopy (stream, "REG", &local_registers, SnapRegisters, COUNT (SnapRegisters))) != SUCCESS)
			break;
		if ((result = UnfreezeStructCopy (stream, "PPU", &local_ppu, SnapPPU, COUNT (SnapPPU))) != SUCCESS)
			break;
		if ((result = UnfreezeStructCopy (stream, "DMA", &local_dma, SnapDMA, COUNT (SnapDMA))) != SUCCESS)
			break;
		if ((result = UnfreezeBlockCopy (stream, "VRA", &local_vram, 0x10000)) != SUCCESS)
			break;
		if ((result = UnfreezeBlockCopy (stream, "RAM", &local_ram, 0x20000)) != SUCCESS)
			break;
		if ((result = UnfreezeBlockCopy (stream, "SRA", &local_sram, 0x20000)) != SUCCESS)
			break;
		if ((result = UnfreezeBlockCopy (stream, "FIL", &local_fillram, 0x8000)) != SUCCESS)
			break;
		if (UnfreezeStructCopy (stream, "APU", &local_apu, SnapAPU, COUNT (SnapAPU)) == SUCCESS)
		{
			if ((result = UnfreezeStructCopy (stream, "ARE", &local_apu_registers, SnapAPURegisters, COUNT (SnapAPURegisters))) != SUCCESS)
				break;
			if ((result = UnfreezeBlockCopy (stream, "ARA", &local_apu_ram, 0x10000)) != SUCCESS)
				break;
			if ((result = UnfreezeStructCopy (stream, "SOU", &local_apu_sounddata, SnapSoundData, COUNT (SnapSoundData))) != SUCCESS)
				break;
		}

	#ifndef ZSNES_FX
		result = UnfreezeStructCopy(stream, "SFX", &local_superfx, SnapFX, COUNT(SnapFX));
		if (result != SUCCESS && Settings.SuperFX)
			break;
	#endif

		if ((result = UnfreezeStructCopy (stream, "SA1", &local_sa1, SnapSA1, COUNT(SnapSA1))) == SUCCESS)
		{
			if ((result = UnfreezeStructCopy (stream, "SAR", &local_sa1_registers, SnapSA1Registers, COUNT (SnapSA1Registers))) != SUCCESS)
				break;
		}

		result = UnfreezeStructCopy(stream, "DP1", &local_dsp1, SnapDSP1, COUNT(SnapDSP1));
		if (result != SUCCESS && Settings.DSP == 1)
			break;

		result = UnfreezeStructCopy(stream, "DP2", &local_dsp2, SnapDSP2, COUNT(SnapDSP2));
		if (result != SUCCESS && Settings.DSP == 2)
			break;

		result = UnfreezeStructCopy(stream, "DP4", &local_dsp4, SnapDSP4, COUNT(SnapDSP4));
		if (result != SUCCESS && Settings.DSP == 4)
			break;

		result = UnfreezeBlockCopy (stream, "CX4", &local_cx4_data, 8192);
		if (result != SUCCESS && Settings.C4)
			break;

		result = UnfreezeStructCopy(stream, "ST0", &local_st010, SnapST010, COUNT(SnapST010));
		if (result != SUCCESS && Settings.SETA == ST_010)
			break;

		result = UnfreezeStructCopy(stream, "OBC", &local_obc1, SnapOBC1, COUNT(SnapOBC1));
		if (result != SUCCESS && Settings.OBC1)
			break;

		result = UnfreezeBlockCopy (stream, "OBM", &local_obc1_data, 8192);
		if (result != SUCCESS && Settings.OBC1)
			break;

		result = UnfreezeStructCopy(stream, "S71", &local_spc7110, SnapSPC7110Snap, COUNT(SnapSPC7110Snap));
		if (result != SUCCESS && Settings.SPC7110)
			break;

		result = UnfreezeStructCopy(stream, "SRT", &local_srtc, SnapSRTCSnap, COUNT(SnapSRTCSnap));
		if (result != SUCCESS && Settings.SRTC)
			break;

		result = UnfreezeBlockCopy (stream, "CLK", &local_rtc_data, 20);
		if (result != SUCCESS && (Settings.SRTC || Settings.SPC7110RTC))
			break;

		result = UnfreezeStructCopy(stream, "BSX", &local_bsx_data, SnapBSX, COUNT(SnapBSX));
		if (result != SUCCESS && Settings.BS)
			break;

		if (S9xMovieActive ())
		{
			SnapshotMovieInfo mi;
			if ((result = UnfreezeStruct (stream, "MOV", &mi, SnapMovie, COUNT(SnapMovie))) != SUCCESS)
			{
				result = NOT_A_MOVIE_SNAPSHOT;
				break;
			}

			if ((result = UnfreezeBlockCopy (stream, "MID", &local_movie_data, mi.MovieInputDataSize)) != SUCCESS)
			{
				result = NOT_A_MOVIE_SNAPSHOT;
				break;
			}

			if (!S9xMovieUnfreeze(local_movie_data, mi.MovieInputDataSize))
			{
				result = WRONG_MOVIE_SNAPSHOT;
				break;
			}
		}

		UnfreezeBlockCopy (stream, "PSR", &local_bsram, 0x80000);
		UnfreezeStructCopy (stream, "EXT", &local_ext, SnapEXT, COUNT (SnapEXT));
		UnfreezeStructCopy(stream, "CTL", &local_control_data, SnapControls, COUNT(SnapControls));
		UnfreezeStructCopy (stream, "TIM", &local_timing_data, SnapTimings, COUNT (SnapTimings));

		result=SUCCESS;

	} while(false);
// ## end load ##

	if (result == SUCCESS)
	{
		uint32 old_flags = CPU.Flags;
		uint32 sa1_old_flags = SA1.Flags;
		Memory.BSXInitialized = BSX_LOAD_INITIALIZED;//for S9xResetBSX()
		S9xReset ();
		S9xSetSoundMute (TRUE);

		UnfreezeStructFromCopy (&CPU, SnapCPU, COUNT (SnapCPU), local_cpu);
		UnfreezeStructFromCopy (&Registers, SnapRegisters, COUNT (SnapRegisters), local_registers);

		UnfreezeStructFromCopy (&PPU, SnapPPU, COUNT (SnapPPU), local_ppu);

		UnfreezeStructFromCopy (DMA, SnapDMA, COUNT (SnapDMA), local_dma);
		memcpy (Memory.VRAM, local_vram, 0x10000);
		memcpy (Memory.RAM, local_ram, 0x20000);
		memcpy (Memory.SRAM, local_sram, 0x20000);
		memcpy (Memory.FillRAM, local_fillram, 0x8000);

		//Load new valuables
		NewPPUp npp = (NewPPUp)Memory.FillRAM;
		PPU.OAMWriteRegister = npp->OAMWriteRegister;
		PPU.VTimerPosition = npp->VTimerPosition;
		PPU.GunHLatch = npp->GunHLatch;
		PPU.GunVLatch = npp->GunVLatch;
		PPU.M7HOFS = npp->M7HOFS;
		PPU.M7VOFS = npp->M7VOFS;
		PPU.M7byte = npp->M7byte;
		PPU.HDMA = npp->HDMA;
		PPU.HDMAEnded = npp->HDMAEnded;
		PPU.OpenBus1 = npp->OpenBus1;
		PPU.OpenBus2 = npp->OpenBus2;

		APU.Cycles_err = READ_DWORD(&Memory.FillRAM[0x20]);
		APU.t_ticks_err = READ_DWORD(&Memory.FillRAM[0x24]);

		for(int i = 0; i < 3; i++)
		  {
		    APU.t_ticks[i] = READ_DWORD(&Memory.FillRAM[0x28 + i * 4]);
		    APU.t_stage1_ticks[i] = Memory.FillRAM[0x28 + 12 + i];
		    APU.t_current_line[i] = Memory.FillRAM[0x28 + 12 + 3 + i];
		  }

		APU.t64kCycles = READ_DWORD(&Memory.FillRAM[0x3A]);
		APU.t64kCycles_err = READ_DWORD(&Memory.FillRAM[0x3E]);

		if(local_apu)
		{
			UnfreezeStructFromCopy (&APU, SnapAPU, COUNT (SnapAPU), local_apu);
			UnfreezeStructFromCopy (&APURegisters, SnapAPURegisters, COUNT (SnapAPURegisters), local_apu_registers);
			memcpy (IAPU.RAM, local_apu_ram, 0x10000);
			
			//direct access iplrom
			memcpy(&IAPU.RAM[0xffc0], APU.ExtraRAM, 64);

			if(IAPU.RAM[0xf0] == 0xFF)
			  IAPU.RAM[0xf0] = 0x0A;//timers_enabled & ram_writable

			UnfreezeStructFromCopy (&SoundData, SnapSoundData, COUNT (SnapSoundData), local_apu_sounddata);
		}


	#ifndef ZSNES_FX
		if (local_superfx)
		{
			GSU.avRegAddr = (uint8 *) &GSU.avReg;
			UnfreezeStructFromCopy(&GSU, SnapFX, COUNT(SnapFX), local_superfx);
		}
	#endif


		if(local_sa1)
		{
			UnfreezeStructFromCopy (&SA1, SnapSA1, COUNT (SnapSA1), local_sa1);
			UnfreezeStructFromCopy (&SA1Registers, SnapSA1Registers, COUNT (SnapSA1Registers), local_sa1_registers);
		}

		if (local_dsp1)
			UnfreezeStructFromCopy(&DSP1, SnapDSP1, COUNT(SnapDSP1), local_dsp1);

		if (local_dsp2)
			UnfreezeStructFromCopy(&DSP2, SnapDSP2, COUNT(SnapDSP2), local_dsp2);

		if (local_dsp4)
			UnfreezeStructFromCopy(&DSP4, SnapDSP4, COUNT(SnapDSP4), local_dsp4);

		if (local_cx4_data)
	#ifndef ZSNES_C4
			memcpy(Memory.C4RAM, local_cx4_data, 8192);
	#else
			memcpy(C4Ram, local_cx4_data, 8192);
	#endif

		if (local_st010)
			UnfreezeStructFromCopy(&ST010, SnapST010, COUNT(SnapST010), local_st010);

		if (local_obc1)
			UnfreezeStructFromCopy(&OBC1, SnapOBC1, COUNT(SnapOBC1), local_obc1);

		if (local_obc1_data)
			memcpy(Memory.OBC1RAM, local_obc1_data, 8192);

		if (local_spc7110)
			UnfreezeStructFromCopy(&s7snap, SnapSPC7110Snap, COUNT(SnapSPC7110Snap), local_spc7110);

		if (local_srtc)
			UnfreezeStructFromCopy(&srtcsnap, SnapSRTCSnap, COUNT(SnapSRTCSnap), local_srtc);

		if (local_rtc_data)
			memcpy(RTCData.reg, local_rtc_data, 20);

		if (local_bsx_data)
			UnfreezeStructFromCopy(&BSX, SnapBSX, COUNT(SnapBSX), local_bsx_data);

		Memory.FixROMSpeed ();
		CPU.Flags |= old_flags & (DEBUG_MODE_FLAG | TRACE_FLAG |
			SINGLE_STEP_FLAG | FRAME_ADVANCE_FLAG);

	    IPPU.ColorsChanged = TRUE;
		IPPU.OBJChanged = TRUE;

		CPU.InDMA = CPU.InHDMA = FALSE;
		CPU.InDMAorHDMA = CPU.InWRAMDMAorHDMA = FALSE;
		CPU.HDMARanInDMA = 0;

		//old snapshot
		for (int d = 0; d < 8; d++)
		  {
		    DMA[d].ReverseTransfer = DMA[d].TransferDirection;
		    DMA[d].UnknownByte = Memory.FillRAM[0x430B + (d << 4)];
		    DMA[d].UnusedBit43x0 = (Memory.FillRAM[0x4300 + (d << 4)] & 0x20) ? TRUE : FALSE;

		    //seiken2 it's needed this value for old snapshot
		    DMA[d].DoTransfer = Memory.FillRAM[0x430F + (d << 4)];
		  }

		S9xFixColourBrightness ();
		IPPU.RenderThisFrame = FALSE;

		if (local_apu)
		{
			S9xSetSoundMute (FALSE);
			S9xAPUUnpackStatus ();
			Settings.APUEnabled = TRUE;
			IAPU.APUExecuting = TRUE;
		}
		else
		{
			Settings.APUEnabled = FALSE;
			IAPU.APUExecuting = FALSE;
			S9xSetSoundMute (TRUE);
		}

		if (local_sa1)
		{
			S9xSA1PostLoadState ();
			SA1.Flags |= sa1_old_flags & (TRACE_FLAG);
		}

		if(local_bsram)
			memcpy (Memory.BSRAM, local_bsram, 0x80000);

		if(local_ext) {
			UnfreezeStructFromCopy (&EXT, SnapEXT, COUNT (SnapEXT), local_ext);
			CPU.IRQPending = EXT.IRQCycleCount;
		}
		else {
			EXT.NextAPUTimerPos = APU.Cycles;
			EXT.NextAPUTimerPos += ((EXT.APUTimerCounter_err += EXT.APUTimerCounter) >> FIXED_POINT_SHIFT);
			EXT.APUTimerCounter_err &= FIXED_POINT_REMAINDER;
		}

		struct SControlSnapshot	ctl_snap;
		if(!local_control_data)
		  {
		    //old snapshot
		    S9xControlPreSaveState(&ctl_snap);
		    ctl_snap.port1_read_idx[0] = PPU.Joypad1ButtonReadPos;
		    ctl_snap.port2_read_idx[0] = PPU.Joypad2ButtonReadPos;
		    ctl_snap.port2_read_idx[1] = PPU.Joypad3ButtonReadPos;
		    ctl_snap.mouse_speed[0] = PPU.MouseSpeed[0];
		    ctl_snap.mouse_speed[1] = PPU.MouseSpeed[1];
		  }
		else
		    UnfreezeStructFromCopy(&ctl_snap, SnapControls, COUNT(SnapControls), local_control_data);

		if(local_timing_data)
		  UnfreezeStructFromCopy (&Timings, SnapTimings, COUNT (SnapTimings), local_timing_data);
		else
		  {
		    Timings.H_Max_Master = SNES_CYCLES_PER_SCANLINE;
		    Timings.H_Max        = Timings.H_Max_Master;
		    Timings.HBlankStart  = SNES_HBLANK_START_HC;
		    Timings.HBlankEnd    = SNES_HBLANK_END_HC;
		    Timings.HDMAInit     = SNES_HDMA_INIT_HC;
		    Timings.HDMAStart    = SNES_HDMA_START_HC;
		    Timings.RenderPos    = SNES_RENDER_START_HC;
		    Timings.V_Max_Master = Settings.PAL ? SNES_MAX_PAL_VCOUNTER : SNES_MAX_NTSC_VCOUNTER;
		    Timings.V_Max        = Timings.V_Max_Master;
		    Timings.DMACPUSync   = 18;
		    Timings.NMIDMADelay  = 24;
		    Timings.IRQPendCount = 0;
		    Timings.APUSpeedup = 0;
		    if(!Settings.OldTiming)
		      {
			Timings.HDMAStart   = SNES_HDMA_START_HC + Settings.HDMATimingHack - 100;
			Timings.HBlankStart = SNES_HBLANK_START_HC + Timings.HDMAStart - SNES_HDMA_START_HC;
		      }
		    else
		      {
			Timings.HDMAStart   = HBLANK_START + Settings.HDMATimingHack - 100;
			Timings.HBlankStart = Timings.HDMAStart;
		      }
		    Timings.InterlaceField = FALSE;
		    if (Model->_5A22 == 2)
		      Timings.WRAMRefreshPos = SNES_WRAM_REFRESH_HC_v2;
		    else
		      Timings.WRAMRefreshPos = SNES_WRAM_REFRESH_HC_v1;

		    if (CPU.Flags & NMI_FLAG)
		      Timings.NMITriggerPos = CPU.Cycles + 6 + 6;
		    else
		      Timings.NMITriggerPos = 0xffff;
		  }

		//old snapshot fix
		//Save & Load snapshot timing is (CPU.Flags & SCAN_KEYS_FLAG)
		if(CPU.WhichEvent != HC_HDMA_INIT_EVENT && CPU.WhichEvent != HC_IRQ_5_7_EVENT)
		  {
		    switch(CPU.WhichEvent)
		      {
		      case HBLANK_START_EVENT:
			CPU.WhichEvent = HC_HDMA_INIT_EVENT;
			CPU.NextEvent = Timings.HDMAInit;
			break;

		      case HTIMER_BEFORE_EVENT:
			if(CPU.NextEvent < Timings.HDMAInit)
			  {
			    CPU.NextEvent = HC_IRQ_5_7_EVENT;
			  }
			else
			  {
			    CPU.WhichEvent = HC_HDMA_INIT_EVENT;
			    CPU.NextEvent = Timings.HDMAInit;
			  }
			break;

		      default:
			break;
		      }

		    PPU.VTimerPosition = PPU.IRQVBeamPos;
		  }

		RestoreAPUDSP();
		memcpy(Loop, &Memory.FillRAM[0x100], FIRBUF * sizeof(int16));
		firCur = Memory.FillRAM[0x100 + FIRBUF * sizeof(int16)];
		firCur &= FIRBUF - 2;

		uint8 hdma_byte = Memory.FillRAM[0x420c];
		S9xSetCPU(hdma_byte, 0x420c);

		ICPU.ShiftedPB = Registers.PB << 16;
		ICPU.ShiftedDB = Registers.DB << 16;
		S9xSetPCBase (ICPU.ShiftedPB + Registers.PCw);
		S9xUnpackStatus ();
		S9xFixCycles ();
//		S9xReschedule ();				// <-- this causes desync when recording or playing movies

		S9xControlPostLoadState(&ctl_snap);

	#ifndef ZSNES_FX
		if (local_superfx)
		{
			GSU.pfPlot = fx_PlotTable[GSU.vMode];
			GSU.pfRpix = fx_PlotTable[GSU.vMode + 5];
		}
	#else
		if (Settings.SuperFX)
			S9xSuperFXPostLoadState();
	#endif

		if (local_spc7110)
			S9xSPC7110PostLoadState(version);

		if (local_srtc)
			S9xSRTCPostLoadState(version);

		if (local_bsx_data)
			S9xBSXPostLoadState();

		if (Settings.SDD1)
			S9xSDD1PostLoadState ();

		//EXT.NextAPUTimerPos = CPU.Cycles * 10000L;
		//EXT.APUTimerCounter = 0; 
	}

	if (local_cpu)            delete [] local_cpu;
	if (local_registers)      delete [] local_registers;
	if (local_ppu)            delete [] local_ppu;
	if (local_dma)            delete [] local_dma;
	if (local_vram)           delete [] local_vram;
	if (local_ram)            delete [] local_ram;
	if (local_sram)           delete [] local_sram;
	if (local_fillram)        delete [] local_fillram;
	if (local_apu)            delete [] local_apu;
	if (local_apu_registers)  delete [] local_apu_registers;
	if (local_apu_ram)        delete [] local_apu_ram;
	if (local_apu_sounddata)  delete [] local_apu_sounddata;
	if (local_superfx)			delete [] local_superfx;
	if (local_sa1)            delete [] local_sa1;
	if (local_sa1_registers)  delete [] local_sa1_registers;
	if (local_dsp1)				delete [] local_dsp1;
	if (local_dsp2)				delete [] local_dsp2;
	if (local_dsp4)				delete [] local_dsp4;
	if (local_cx4_data)			delete [] local_cx4_data;
	if (local_st010)			delete [] local_st010;
	if (local_obc1)				delete [] local_obc1;
	if (local_obc1_data)		delete [] local_obc1_data;
	if (local_spc7110)	  delete [] local_spc7110;
	if (local_srtc)           delete [] local_srtc;
	if (local_rtc_data)       delete [] local_rtc_data;
	if (local_bsx_data)			delete [] local_bsx_data;
	if (local_movie_data)     delete [] local_movie_data;
	if (local_bsram)          delete [] local_bsram;
	if (local_ext)            delete [] local_ext;
	if (local_control_data)		delete [] local_control_data;
	if (local_timing_data)    delete [] local_timing_data;

	return (result);
}

static int FreezeSize (int size, int type)
{
	switch (type)
	{
		case uint32_ARRAY_V:
		case uint32_INDIR_ARRAY_V:
			return (size * 4);

		case uint16_ARRAY_V:
		case uint16_INDIR_ARRAY_V:
			return (size * 2);

		default:
			return (size);
	}
}

static void FreezeStruct (STREAM stream, const char *name, void *base, FreezeData *fields,
				   int num_fields)
{
    // Work out the size of the required block
    int len = 0;
    int i;
    int j;
	
    for (i = 0; i < num_fields; i++)
    {
		int offset = fields [i].type == POINTER_V ? (fields [i].offset & 0xffff) : fields [i].offset;

		if (offset + FreezeSize (fields [i].size, 
			fields [i].type) > len)
			len = offset + FreezeSize (fields [i].size, 
			fields [i].type);
    }
	
    uint8 *block = new uint8 [len];
    uint8 *ptr = block;
    uint8 *addr;
    uint16 word;
    uint32 dword;
    int64  qword;
    int	relativeAddr;
	
    // Build the block ready to be streamed out
    for (i = 0; i < num_fields; i++)
    {

		addr = (uint8 *) base + fields[i].offset;

		// determine real address of indirect-type fields
		// (where the structure contains a pointer to an array rather than the array itself)
		if (fields[i].type == uint8_INDIR_ARRAY_V || fields[i].type == uint16_INDIR_ARRAY_V || fields[i].type == uint32_INDIR_ARRAY_V)
			addr = (uint8 *) (*((pint *) addr));

		// convert pointer-type saves from absolute to relative pointers
		if (fields[i].type == POINTER_V)
		{
			uint8	*pointer    = (uint8 *) *((pint *) ((uint8 *) base + (fields[i].offset & 0xffff)));
			uint8	*relativeTo = (uint8 *) *((pint *) ((uint8 *) base + ((fields[i].offset >> 16) & 0xffff)));
			relativeAddr = pointer - relativeTo;
			addr = (uint8 *) &relativeAddr;
		}

		switch (fields[i].type)
		{
			case INT_V:
			case POINTER_V:
				switch (fields[i].size)
				{
					case 1:
						*ptr++ = *(addr);
						break;

					case 2:
						word = *((uint16 *) (addr));
						*ptr++ = (uint8) (word >> 8);
						*ptr++ = (uint8) word;
						break;

					case 4:
						dword = *((uint32 *) (addr));
						*ptr++ = (uint8) (dword >> 24);
						*ptr++ = (uint8) (dword >> 16);
						*ptr++ = (uint8) (dword >> 8);
						*ptr++ = (uint8) dword;
						break;

					case 8:
						qword = *((int64 *) (addr));
						*ptr++ = (uint8) (qword >> 56);
						*ptr++ = (uint8) (qword >> 48);
						*ptr++ = (uint8) (qword >> 40);
						*ptr++ = (uint8) (qword >> 32);
						*ptr++ = (uint8) (qword >> 24);
						*ptr++ = (uint8) (qword >> 16);
						*ptr++ = (uint8) (qword >> 8);
						*ptr++ = (uint8) qword;
						break;
				}

				break;

			case uint8_ARRAY_V:
			case uint8_INDIR_ARRAY_V:
				memmove(ptr, addr, fields[i].size);
				ptr += fields[i].size;

				break;

			case uint16_ARRAY_V:
			case uint16_INDIR_ARRAY_V:
				for (j = 0; j < fields[i].size; j++)
				{
					word = *((uint16 *) (addr + j * 2));
					*ptr++ = (uint8) (word >> 8);
					*ptr++ = (uint8) word;
				}

				break;

			case uint32_ARRAY_V:
			case uint32_INDIR_ARRAY_V:
				for (j = 0; j < fields[i].size; j++)
				{
					dword = *((uint32 *) (addr + j * 4));
					*ptr++ = (uint8) (dword >> 24);
					*ptr++ = (uint8) (dword >> 16);
					*ptr++ = (uint8) (dword >> 8);
					*ptr++ = (uint8) dword;
				}

				break;
		}

    }
	
    FreezeBlock (stream, name, block, len);
    delete[] block;
}

static void FreezeBlock (STREAM stream, const char *name, uint8 *block, int size)
{
    char buffer [512];
    sprintf (buffer, "%s:%06d:", name, size);
    WRITE_STREAM (buffer, strlen (buffer), stream);
    WRITE_STREAM (block, size, stream);
    
}

static int UnfreezeBlock (STREAM stream, const char *name, uint8 *block, int size)
{
    char buffer [20];
    int len = 0;
    int rem = 0;
    int rew_len;
    if (READ_STREAM (buffer, 11, stream) != 11 ||
		strncmp (buffer, name, 3) != 0 || buffer [3] != ':' ||
		(len = atoi (&buffer [4])) == 0)
    {
		REVERT_STREAM(stream, FIND_STREAM(stream)-11, 0);
		return (WRONG_FORMAT);
    }

    if (len > size)
    {
		rem = len - size;
		len = size;
    }
    if ((rew_len=READ_STREAM (block, len, stream)) != len)
	{
		REVERT_STREAM(stream, FIND_STREAM(stream)-11-rew_len, 0);
		return (WRONG_FORMAT);
	}
    if (rem)
    {
		char *junk = new char [rem];
		READ_STREAM (junk, rem, stream);
		delete [] junk;
    }
	
    return (SUCCESS);
}

static int UnfreezeBlockCopy (STREAM stream, const char *name, uint8** block, int size)
{
    *block = new uint8 [size];
    int result;
	
    if ((result = UnfreezeBlock (stream, name, *block, size)) != SUCCESS)
    {
		delete [] (*block);
		*block = NULL;
		return (result);
    }
	
    return (result);
}


static int UnfreezeStruct (STREAM stream, const char *name, void *base, FreezeData *fields, int num_fields)
{
	int		result;
	uint8	*block = NULL;

	result = UnfreezeStructCopy(stream, name, &block, fields, num_fields);
	if (result != SUCCESS)
	{
		if (block != NULL)
			delete [] block;
		return (result);
	}

	UnfreezeStructFromCopy(base, fields, num_fields, block);
	delete [] block;

	return (SUCCESS);
}

static int UnfreezeStructCopy (STREAM stream, const char *name, uint8** block, FreezeData *fields, int num_fields)
{
    // Work out the size of the required block
    int len = 0;
    int i;
	
    for (i = 0; i < num_fields; i++)
    {
		int offset = fields [i].type == POINTER_V ? (fields [i].offset & 0xffff) : fields [i].offset;

		if (offset + FreezeSize (fields [i].size, 
			fields [i].type) > len)
			len = offset + FreezeSize (fields [i].size, 
			fields [i].type);
    }
	
    return (UnfreezeBlockCopy (stream, name, block, len));
}

static void UnfreezeStructFromCopy (void *sbase, FreezeData *fields, int num_fields, uint8* block)
{
	uint8	*ptr = block;
	uint16	word;
	uint32	dword;
	int64	qword;
	uint8	*addr;
	void	*base;
	int		relativeAddr;
	int		i, j;
	
    // Unpack the block of data into a C structure
    for (i = 0; i < num_fields; i++)
    {
		int offset = fields [i].type == POINTER_V ? (fields [i].offset & 0xffff) : fields [i].offset;

		base = sbase;
		addr = (uint8 *) base + offset;

		if (fields[i].type == uint8_INDIR_ARRAY_V || fields[i].type == uint16_INDIR_ARRAY_V || fields[i].type == uint32_INDIR_ARRAY_V)
			addr = (uint8 *) (*((pint *) addr));

		switch (fields[i].type)
		{
			case INT_V:
			case POINTER_V:
				switch (fields[i].size)
				{
					case 1:
						if (offset < 0)
						{
							ptr++; 
							break;
						}

						*(addr) = *ptr++;
						break;

					case 2:
						if (offset < 0)
						{
							ptr += 2;
							break;
						}

						word  = *ptr++ << 8;
						word |= *ptr++;
						*((uint16 *) (addr)) = word;
						break;

					case 4:
						if (offset < 0)
						{
							ptr += 4;
							break;
						}

						dword  = *ptr++ << 24;
						dword |= *ptr++ << 16;
						dword |= *ptr++ << 8;
						dword |= *ptr++;
						*((uint32 *) (addr)) = dword;
						break;

					case 8:
						if (offset < 0)
						{
							ptr += 8;
							break;
						}

						qword  = (int64) *ptr++ << 56;
						qword |= (int64) *ptr++ << 48;
						qword |= (int64) *ptr++ << 40;
						qword |= (int64) *ptr++ << 32;
						qword |= (int64) *ptr++ << 24;
						qword |= (int64) *ptr++ << 16;
						qword |= (int64) *ptr++ << 8;
						qword |= (int64) *ptr++;
						*((int64 *) (addr)) = qword;
						break;

					default:
						assert(0);
						break;
				}

				break;

			case uint8_ARRAY_V:
			case uint8_INDIR_ARRAY_V:
				if (offset >= 0)
					memmove(addr, ptr, fields[i].size);
				ptr += fields[i].size;

				break;

			case uint16_ARRAY_V:
			case uint16_INDIR_ARRAY_V:
				if (offset < 0)
				{
					ptr += fields[i].size * 2;
					break;
				}

				for (j = 0; j < fields[i].size; j++)
				{
					word  = *ptr++ << 8;
					word |= *ptr++;
					*((uint16 *) (addr + j * 2)) = word;
				}

				break;

			case uint32_ARRAY_V:
			case uint32_INDIR_ARRAY_V:
				if (offset < 0)
				{
					ptr += fields[i].size * 4;
					break;
				}

				for (j = 0; j < fields[i].size; j++)
				{
					dword  = *ptr++ << 24;
					dword |= *ptr++ << 16;
					dword |= *ptr++ << 8;
					dword |= *ptr++;
					*((uint32 *) (addr + j * 4)) = dword;
				}

				break;
		}

		if (fields[i].type == POINTER_V)
		{
			relativeAddr = (int) *((pint *) ((uint8 *) base + (fields[i].offset & 0xffff)));
			uint8	*relativeTo = (uint8 *) *((pint *) ((uint8 *) base + ((fields[i].offset >> 16) & 0xffff)));
			*((pint *) (addr)) = (pint) (relativeTo + relativeAddr);
		}
    }
}

extern uint8 spc_dump_dsp[0x100];

bool8 S9xSPCDump (const char *filename)
{
    static uint8 header [] = {
		'S', 'N', 'E', 'S', '-', 'S', 'P', 'C', '7', '0', '0', ' ',
			'S', 'o', 'u', 'n', 'd', ' ', 'F', 'i', 'l', 'e', ' ',
			'D', 'a', 't', 'a', ' ', 'v', '0', '.', '3', '0', 26, 26, 26
    };
    static uint8 version = {
		0x1e
    };
	
    FILE *fs;
	
    if (!(fs = fopen (filename, "wb")))
		return (FALSE);
	
    bool8 old = S9xSetSoundMute (TRUE);

	// SNESAPU
	StoreAPUDSP();

	// Store APURegisters
	S9xAPUPackStatus ();

	// copy APU.DSP[] to spc_dump_dsp[]
	memcpy(&spc_dump_dsp[0x00], APU.DSP, 0x80);
	memcpy(&spc_dump_dsp[0x80], APU.DSP, 0x80);
	memcpy(APU.ExtraRAM, &IAPU.RAM[0xffc0], 64);
	
    // The SPC file format:
    // 0000: header:	'SNES-SPC700 Sound File Data v0.30',26,26,26
    // 0036: version:	$1e
    // 0037: SPC700 PC:
    // 0039: SPC700 A:
    // 0040: SPC700 X:
    // 0041: SPC700 Y:
    // 0042: SPC700 P:
    // 0043: SPC700 S:
    // 0044: Reserved: 0, 0, 0, 0
    // 0048: Title of game: 32 bytes
    // 0000: Song name: 32 bytes
    // 0000: Name of dumper: 32 bytes
    // 0000: Comments: 32 bytes
    // 0000: Date of SPC dump: 4 bytes
    // 0000: Fade out time in milliseconds: 4 bytes
    // 0000: Fade out length in milliseconds: 2 bytes
    // 0000: Default channel enables: 1 bytes
    // 0000: Emulator used to dump .SPC files: 1 byte, 1 == ZSNES
    // 0000: Reserved: 36 bytes
    // 0256: SPC700 RAM: 64K
    // ----: DSP Registers: 256 bytes
	
    if (fwrite (header, sizeof (header), 1, fs) != 1 ||
		fputc (version, fs) == EOF ||
		fseek (fs, 37, SEEK_SET) == EOF ||
		fputc (APURegisters.PC & 0xff, fs) == EOF ||
		fputc (APURegisters.PC >> 8, fs) == EOF ||
		fputc (APURegisters.YA.B.A, fs) == EOF ||
		fputc (APURegisters.X, fs) == EOF ||
		fputc (APURegisters.YA.B.Y, fs) == EOF ||
		fputc (APURegisters.P, fs) == EOF ||
		fputc (APURegisters.S, fs) == EOF ||
		fseek (fs, 256, SEEK_SET) == EOF ||
		fwrite (IAPU.RAM, 0x10000, 1, fs) != 1 ||
		fwrite (spc_dump_dsp, 1, 256, fs) != 256 ||
		fwrite (APU.ExtraRAM, 64, 1, fs) != 1 ||
		fclose (fs) < 0)
    {
		S9xSetSoundMute (old);
		return (FALSE);
    }
    S9xSetSoundMute (old);
    return (TRUE);
}

