#ifndef _SNESAPU_H_
#define _SNESAPU_H_

//**************************************************************************************************
// Defines

//DSP options ----------------------------------
#define	DSP_ANALOG  0x01	//Simulate analog anomalies
#define	DSP_OLDSMP  0x02	//Traditional ADPCM routine
#define	DSP_SURND   0x04	//"Surround" sound
#define	DSP_REVERSE 0x08	//Reverse stereo samples
#define	DSP_NOECHO  0x10	//Disable echo (must not be bit 0 or 5, see disEcho)

//Mixing flags
#define	MFLG_MUTE   0x01	//Mute voice
#define	MFLG_KOFF   0x04	//Voice is in the process of keying off
#define	MFLG_OFF    0x08	//Voice is currently inactive
#define	MFLG_END    0x10	//End block was just played

// Envelope precision -------------------------
#define E_SHIFT     4		// Amount to shift envelope to get 4-bit signed value

typedef struct
{
  //Waveform --------
  uint16 bCur;			//-> current block
  uint8 bHdr;			//Block Header for current block
  uint8 mFlg;			//Mixing flags (see MixF)

  //Envelope --------
  uint8 eMode;			//[3-0] Current mode (see EnvM)
  //[6-4] ADSR mode to switch into from Gain
  //[7]   Envelope is idle
  uint8 eRIdx;			//Index in RateTab (0-31)
  uint32 eRate;			//Rate of envelope adjustment (16.16)
  uint32 eDec;			//Pitch Decimal (.16)
  int32 eVal;			//Current envelope value
  int32 eAdj;			//Amount to adjust envelope height
  int32 eDest;			//Envelope Destination

  //Samples ---------
  int16 sBuf[32];		// 32 + 32 bytes for decompressed sample blocks
  int8 sIdx;			//-> current sample in sBuf
  int32 sP1;			//Last sample decompressed (prev1)
  int32 sP2;			//Second to last sample (prev2)

  //Mixing ----------
  int32 mChnL;			//Channel Volume (-24.7)
  int32 mChnR;			// "  "
  uint32 mOrgP;			//Original pitch rate converted from the DSP (16.16)
  uint32 mOrgRate;		//Pitch Rate before modulation (16.16)
  uint32 mRate;			//Pitch Rate after modulation (16.16)
  uint32 mDec;			//Pitch Decimal (.16) (used as delta for interpolation)
  int32 mOut;			//Last sample output before chn vol (used for pitch mod)
} VoiceMix;

extern VoiceMix mix[8];		//Mixing structures for each voice
extern int16 Loop[];
extern uint8 firCur;

bool8 InitAPUDSP ();
void DeinitAPUDSP ();
void ResetAPUDSP ();
void SetAPUDSPOpt (int chn, int bits, int32 rate, int opts);
void StoreAPUDSP ();
void RestoreAPUDSP ();
void RestoreSPCFile ();
void SetAPUDSPAmp (int32 amp);
void EmuAPUDSP (uint8 *pBuf, int32 num);
void APUDSPIn (uint8 address, uint8 data);
void smpdsp_run();
void ClearSamples (void);
void SetBufferSize (int ms);
void WriteSamples0(int32 l, int32 r);
void WriteSamples(int32 l, int32 r);

#endif

