/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#include "snes9x.h"
#include "memmap.h"
#include "soundux.h"
#include "apu_orig.h"
#include "snesapu.h"

extern int8 FilterTaps [8];

void S9xSetFilterCoefficient (int tap, int8 value)
{
	FilterTaps [tap & 7] = value;
	SoundData.no_filter = (FilterTaps [0] == 127 || FilterTaps [0] == 0) &&
			   FilterTaps [1] == 0	 &&
			   FilterTaps [2] == 0	 &&
			   FilterTaps [3] == 0	 &&
			   FilterTaps [4] == 0	 &&
			   FilterTaps [5] == 0	 &&
			   FilterTaps [6] == 0	 &&
			   FilterTaps [7] == 0;
}

bool8 S9xSetSoundMute (bool8 mute)
{
    bool8 old = so.mute_sound;
    so.mute_sound = mute;
    return (old);
}

// For backwards compatibility with older port specific code
void S9xMixSamples (uint8 *buffer, int sample_count)
{
    S9xMixSamplesO (buffer, sample_count, 0);
}

void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
{
  if (so.mute_sound || Settings.Paused || Settings.StopEmulation || Settings.ForcedPause || Settings.Mute)
    memset(buffer + byte_offset,
	   so.sixteen_bit ? 0 : 128,
	   sample_count << (so.sixteen_bit ? 1 : 0));
  else
    EmuAPUDSP(buffer + byte_offset,
	      sample_count >> (so.stereo ? 1 : 0));
}

void S9xResetSound ()
{
	for (int i = 0; i < 8; i++)
	{
	SoundData.channels[i].state = SOUND_SILENT;
	SoundData.channels[i].mode = MODE_NONE;
	SoundData.channels[i].type = SOUND_SAMPLE;
	SoundData.channels[i].volume_left = 0;
	SoundData.channels[i].volume_right = 0;
	SoundData.channels[i].hertz = 0;
	SoundData.channels[i].count = 0;
	SoundData.channels[i].loop = FALSE;
	SoundData.channels[i].envx_target = 0;
	SoundData.channels[i].env_error = 0;
	SoundData.channels[i].erate = 0;
	SoundData.channels[i].envx = 0;
	SoundData.channels[i].envxx = 0;
	SoundData.channels[i].left_vol_level = 0;
	SoundData.channels[i].right_vol_level = 0;
	SoundData.channels[i].direction = 0;
	SoundData.channels[i].attack_rate = 0;
	SoundData.channels[i].decay_rate = 0;
	SoundData.channels[i].sustain_rate = 0;
	SoundData.channels[i].release_rate = 0;
	SoundData.channels[i].sustain_level = 0;
	SoundData.echo_ptr = 0;
	SoundData.echo_feedback = 0;
	SoundData.echo_buffer_size = 1;
	}
	FilterTaps [0] = 0;
	FilterTaps [1] = 0;
	FilterTaps [2] = 0;
	FilterTaps [3] = 0;
	FilterTaps [4] = 0;
	FilterTaps [5] = 0;
	FilterTaps [6] = 0;
	FilterTaps [7] = 0;
	so.mute_sound = TRUE;
	so.noise_gen = 1;
	S9xSetSoundControl (255);
	so.samples_mixed_so_far = 0;
	so.play_position = 0;
	so.err_counter = 0;

	SoundData.master_volume_left = 0;
	SoundData.master_volume_right = 0;
	SoundData.echo_volume_left = 0;
	SoundData.echo_volume_right = 0;
	SoundData.echo_enable = 0;
	SoundData.echo_write_enabled = 0;
	SoundData.echo_channel_enable = 0;
	SoundData.pitch_mod = 0;
	SoundData.dummy[0] = 0;
	SoundData.dummy[1] = 0;
	SoundData.dummy[2] = 0;
	SoundData.master_volume[0] = 0;
	SoundData.master_volume[1] = 0;
	SoundData.echo_volume[0] = 0;
	SoundData.echo_volume[1] = 0;
	SoundData.noise_hertz = 0;
	SoundData.no_filter = TRUE;
}

void S9xSetPlaybackRate (uint32 playback_rate)
{
	so.playback_rate = playback_rate;

	int opts =
		(Settings.ReverseStereo ? DSP_REVERSE : 0) |
		(Settings.DisableSoundEcho ? DSP_NOECHO : 0);

	//SNESAPU
	SetAPUDSPOpt(so.stereo ? 2 : 1,
		     so.sixteen_bit ? 16 : 8,
		     so.playback_rate,
		     opts);
}

bool8 S9xInitSound (int mode, bool8 stereo, int buffer_size)
{
    so.sound_fd = -1;
	S9xSetSoundControl (255);
	
    so.playback_rate = 0;
    so.buffer_size = 0;
    so.stereo = stereo;
    so.sixteen_bit = Settings.SixteenBitSound;
    so.encoded = FALSE;
    
    S9xResetSound ();
	
    if (!(mode & 7))
		return (1);
	
    S9xSetSoundMute (TRUE);
    if (!S9xOpenSoundDevice (mode, stereo, buffer_size))
    {
#ifdef NOSOUND
                S9xMessage (S9X_WARNING, S9X_SOUND_NOT_BUILT,
                            "No sound support compiled in");
#else
		S9xMessage (S9X_ERROR, S9X_SOUND_DEVICE_OPEN_FAILED,
                            "Sound device open failed");
#endif
		return (0);
    }
	
    return (1);
}

void S9xSetSoundControl (int sound_switch)
{
    so.sound_switch = sound_switch;
    for(int i=0; i<8; i++){
        if(sound_switch&(1<<i)) mix[i].mFlg &= ~MFLG_MUTE;
        else mix[i].mFlg |= MFLG_MUTE;
    }
}

int S9xGetSoundControl ()
{
  return so.sound_switch;
}
