/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/
#include "snes9x.h"
#include "memmap.h"
#include "spc700.h"
#include "apu_orig.h"

// SPC700/Sound DSP chips have a 24.57MHz crystal on their PCB.

#include "apumem.h"

#define OP1 S9xAPUGetByte (APURegisters.PC + 1)
#define OP2 S9xAPUGetByte (APURegisters.PC + 2)

#define bOP(OP, REL, COND, LEN) \
static void Apu##OP () \
{						\
  if ((COND))					\
    {						\
      APURegisters.PC = REL ();			\
      smp_add_clocks(2);			\
    }						\
  else						\
    {						\
      APURegisters.PC += LEN;			\
    }						\
}

static inline void Push(uint8 b)
{
  S9xAPUSetByte(b, 0x100 + APURegisters.S);
  APURegisters.S--;
}

static inline void Pop(uint8 &b)
{
  APURegisters.S++;
  b = S9xAPUGetByte(0x100 + APURegisters.S);
}

static inline void PushW(uint16 w)
{
  Push(w >> 8);
  Push(w);
}

static inline void PopW(uint16 &w)
{
  uint8 l, h;
  Pop(l);
  Pop(h);
  w = l | h << 8;
}

static inline void APUSetZN8(uint8 b)
{
  IAPU._Zero = (b);
}

static inline void APUSetZN16(uint16 w)
{
  IAPU._Zero = ((w) != 0) | ((w) >> 8);
}

static inline void TCALL(int n)
{
  PushW (APURegisters.PC + 1);
  APURegisters.PC = S9xAPUGetByte(0xffde - (n << 1)) | (S9xAPUGetByte(0xffde - (n << 1) + 1) << 8);
}

static inline void ADC(uint8 &a, uint8 b)
{
  int r = a + b + APUCheckCarry ();

  if(~(a ^ b) & (a ^ r) & 0x80)
    APUSetOverflow ();
  else
    APUClearOverflow ();

  if((a ^ b ^ r) & 0x10)
    APUSetHalfCarry ();
  else
    APUClearHalfCarry ();

  IAPU._Carry = r > 0xff;
  APUSetZN8 (r);

  a = r;
}

static inline void ADDW(uint16 &a, uint16 b)
{
  uint8 x, y;
  uint16 r;

  x = a;
  y = b;

  APUClearCarry ();
  ADC(x, y);
  r = x;

  x = a >> 8;
  y = b >> 8;
  ADC(x, y);
  r |= x << 8;

  APUSetZN16 (r);
  
  a = r;
}

static inline void SBC(uint8 &a, uint8 b)
{
  int r = a - b - !APUCheckCarry ();

  if((a ^ b) & (a ^ r) & 0x80)
    APUSetOverflow ();
  else
    APUClearOverflow ();

  if(!((a ^ b ^ r) & 0x10))
    APUSetHalfCarry ();
  else
    APUClearHalfCarry ();

  IAPU._Carry = r >= 0;
  APUSetZN8 (r);

  a = r;
}

static inline void SUBW(uint16 &a, uint16 b)
{
  uint8 x, y;
  uint16 r;

  x = a;
  y = b;

  APUSetCarry ();
  SBC(x, y);
  r = x;

  x = a >> 8;
  y = b >> 8;
  SBC(x, y);
  r |= x << 8;

  APUSetZN16 (r);
  
  a = r;
}

static inline void CMP(uint8 a, uint8 b)
{
  int r = a - b;
  IAPU._Carry = r >= 0;
  APUSetZN8 (r);
}

static inline void CMPW(uint16 a, uint16 b)
{
  int r = a - b;
  IAPU._Carry = r >= 0;
  APUSetZN16 (r);
}

static inline void ASL(uint8 &b)
{
  IAPU._Carry = (b & 0x80) != 0;
  b <<= 1;
  APUSetZN8 (b);
}

static inline void LSR(uint8 &b)
{
  IAPU._Carry = b & 1;
  b >>= 1;
  APUSetZN8 (b);
}

static inline void ROL(uint8 &b)
{
  unsigned carry = (unsigned)APUCheckCarry ();
  IAPU._Carry = (b & 0x80) != 0;
  b = (b << 1) | carry;
  APUSetZN8 (b);
}

static inline void ROR(uint8 &b)
{
  unsigned carry = (unsigned)APUCheckCarry () << 7;
  IAPU._Carry = (b & 0x01) != 0;
  b = carry | (b >> 1);
  APUSetZN8 (b);
}

static inline uint16 Relative()
{
  int8 a = OP1;
  return ((int)(APURegisters.PC + 2) + a);
}

static inline uint16 Relative2()
{
  int8 a = OP2;
  return ((int)(APURegisters.PC + 3) + a);
}

static inline uint16 IndexedXIndirect()
{
  uint8 a = OP1 + APURegisters.X;
  return (S9xAPUGetByteDP(a) | (S9xAPUGetByteDP(a + 1) << 8));
}

static inline uint16 IndirectIndexedY()
{
  uint8 a = OP1;
  //(0 | 1) + x  note, don't forget "()"
  return ((S9xAPUGetByteDP(a) | (S9xAPUGetByteDP(a + 1) << 8)) + APURegisters.YA.B.Y);
}

static inline uint16 Absolute()
{
  return (S9xAPUGetByte(APURegisters.PC + 1) | (S9xAPUGetByte(APURegisters.PC + 2) << 8));
}

static inline uint16 AbsoluteX()
{
  //(0 | 1) + x  note, don't forget "()"
  return ((S9xAPUGetByte(APURegisters.PC + 1) | (S9xAPUGetByte(APURegisters.PC + 2) << 8)) + APURegisters.X);
}

static inline uint16 AbsoluteY()
{
  //(0 | 1) + y  note, don't forget "()"
  return ((S9xAPUGetByte(APURegisters.PC + 1) | (S9xAPUGetByte(APURegisters.PC + 2) << 8)) + APURegisters.YA.B.Y);
}

static inline void MemBit(uint16 &a, uint8 &b)
{
  a = S9xAPUGetByte(APURegisters.PC + 1) | (S9xAPUGetByte(APURegisters.PC + 2) << 8);
  b = a >> 13;
  a &= 0x1fff;
}

static void
Apu00 ()
{
  // NOP
  APURegisters.PC++;
}

static void
Apu01 ()
{
  TCALL (0);
}

static void
Apu11 ()
{
  TCALL (1);
}

static void
Apu21 ()
{
  TCALL (2);
}

static void
Apu31 ()
{
  TCALL (3);
}

static void
Apu41 ()
{
  TCALL (4);
}

static void
Apu51 ()
{
  TCALL (5);
}

static void
Apu61 ()
{
  TCALL (6);
}

static void
Apu71 ()
{
  TCALL (7);
}

static void
Apu81 ()
{
  TCALL (8);
}

static void
Apu91 ()
{
  TCALL (9);
}

static void
ApuA1 ()
{
  TCALL (10);
}

static void
ApuB1 ()
{
  TCALL (11);
}

static void
ApuC1 ()
{
  TCALL (12);
}

static void
ApuD1 ()
{
  TCALL (13);
}

static void
ApuE1 ()
{
  TCALL (14);
}

static void
ApuF1 ()
{
  TCALL (15);
}

static void
Apu3F ()// CALL absolute
{
  PushW (APURegisters.PC + 3);
  APURegisters.PC = Absolute ();
}

static void
Apu4F ()// PCALL $XX
{
  PushW (APURegisters.PC + 2);
  APURegisters.PC = 0xff00 + OP1;
}

static inline void SET(int b)
{
  uint8 a = OP1;
  S9xAPUSetByteDP (S9xAPUGetByteDP (a) | (1 << b), a);
  APURegisters.PC += 2;
}

static void
Apu02 ()
{
  SET (0);
}

static void
Apu22 ()
{
  SET (1);
}

static void
Apu42 ()
{
  SET (2);
}

static void
Apu62 ()
{
  SET (3);
}

static void
Apu82 ()
{
  SET (4);
}

static void
ApuA2 ()
{
  SET (5);
}

static void
ApuC2 ()
{
  SET (6);
}

static void
ApuE2 ()
{
  SET (7);
}

static inline void CLR(int b)
{
  uint8 a = OP1;
  S9xAPUSetByteDP (S9xAPUGetByteDP (a) & ~(1 << b), a);
  APURegisters.PC += 2;
}

static void
Apu12 ()
{
  CLR (0);
}

static void
Apu32 ()
{
  CLR (1);
}

static void
Apu52 ()
{
  CLR (2);
}

static void
Apu72 ()
{
  CLR (3);
}

static void
Apu92 ()
{
  CLR (4);
}

static void
ApuB2 ()
{
  CLR (5);
}

static void
ApuD2 ()
{
  CLR (6);
}

static void
ApuF2 ()
{
  CLR (7);
}

//BBS
bOP(03, Relative2, S9xAPUGetByteDP (OP1) & (1 << 0), 3)
bOP(23, Relative2, S9xAPUGetByteDP (OP1) & (1 << 1), 3)
bOP(43, Relative2, S9xAPUGetByteDP (OP1) & (1 << 2), 3)
bOP(63, Relative2, S9xAPUGetByteDP (OP1) & (1 << 3), 3)
bOP(83, Relative2, S9xAPUGetByteDP (OP1) & (1 << 4), 3)
bOP(A3, Relative2, S9xAPUGetByteDP (OP1) & (1 << 5), 3)
bOP(C3, Relative2, S9xAPUGetByteDP (OP1) & (1 << 6), 3)
bOP(E3, Relative2, S9xAPUGetByteDP (OP1) & (1 << 7), 3)

//BBC
bOP(13, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 0)), 3)
bOP(33, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 1)), 3)
bOP(53, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 2)), 3)
bOP(73, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 3)), 3)
bOP(93, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 4)), 3)
bOP(B3, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 5)), 3)
bOP(D3, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 6)), 3)
bOP(F3, Relative2, !(S9xAPUGetByteDP (OP1) & (1 << 7)), 3)

static void
Apu04 ()
{
  // OR A,dp
  APURegisters.YA.B.A |= S9xAPUGetByteDP (OP1);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu05 ()
{
  // OR A,abs
  APURegisters.YA.B.A |= S9xAPUGetByte (Absolute ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu06 ()
{
  // OR A,(X)
  APURegisters.YA.B.A |= S9xAPUGetByteDP (APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu07 ()
{
  // OR A,(dp+X)
  APURegisters.YA.B.A |= S9xAPUGetByte (IndexedXIndirect ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu08 ()
{
  // OR A,#00
  APURegisters.YA.B.A |= OP1;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu09 ()
{
  // OR dp(dest),dp(src)
  uint8 v = S9xAPUGetByteDP (OP1);
  uint8 a = OP2;
  v |= S9xAPUGetByteDP (a);
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu14 ()
{
  // OR A,dp+X
  APURegisters.YA.B.A |= S9xAPUGetByteDP (OP1 + APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu15 ()
{
  // OR A,abs+X
  APURegisters.YA.B.A |= S9xAPUGetByte (AbsoluteX ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu16 ()
{
  // OR A,abs+Y
  APURegisters.YA.B.A |= S9xAPUGetByte (AbsoluteY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu17 ()
{
  // OR A,(dp)+Y
  APURegisters.YA.B.A |= S9xAPUGetByte (IndirectIndexedY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu18 ()
{
  // OR dp,#00
  uint8 v = OP1;
  uint8 a = OP2;
  v |= S9xAPUGetByteDP (a);
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu19 ()
{
  // OR (X),(Y)
  uint8 v = S9xAPUGetByteDP (APURegisters.X) | S9xAPUGetByteDP (APURegisters.YA.B.Y);
  APUSetZN8 (v);
  S9xAPUSetByteDP (v, APURegisters.X);
  APURegisters.PC++;
}

static void
Apu0A ()
{
  // OR1 C,membit
  if (!APUCheckCarry ())
    {
      uint16 a;
      uint8 b;
      MemBit(a, b);
      if (S9xAPUGetByte (a) & (1 << b))
	APUSetCarry ();
    }
  APURegisters.PC += 3;
}

static void
Apu2A ()
{
  // OR1 C,not membit
  if (!APUCheckCarry ())
    {
      uint16 a;
      uint8 b;
      MemBit(a, b);
      if (!(S9xAPUGetByte (a) & (1 << b)))
	APUSetCarry ();
    }
  APURegisters.PC += 3;
}

static void
Apu4A ()
{
  // AND1 C,membit
  if (APUCheckCarry ())
    {
      uint16 a;
      uint8 b;
      MemBit(a, b);
      if (!(S9xAPUGetByte (a) & (1 << b)))
	APUClearCarry ();
    }
  APURegisters.PC += 3;
}

static void
Apu6A ()
{
  // AND1 C, not membit
  if (APUCheckCarry ())
    {
      uint16 a;
      uint8 b;
      MemBit(a, b);
      if ((S9xAPUGetByte (a) & (1 << b)))
	APUClearCarry ();
    }
  APURegisters.PC += 3;
}

static void
Apu8A ()
{
  // EOR1 C, membit
  uint16 a;
  uint8 b;
  MemBit(a, b);
  if (APUCheckCarry ())
    {
      if (S9xAPUGetByte (a) & (1 << b))
	APUClearCarry ();
    }
  else
    {
      if (S9xAPUGetByte (a) & (1 << b))
	APUSetCarry ();
    }
  APURegisters.PC += 3;
}

static void
ApuAA ()
{
  // MOV1 C,membit
  uint16 a;
  uint8 b;
  MemBit(a, b);
  if (S9xAPUGetByte (a) & (1 << b))
    APUSetCarry ();
  else
    APUClearCarry ();
  APURegisters.PC += 3;
}

static void
ApuCA ()
{
  // MOV1 membit,C
  uint16 a;
  uint8 b;
  MemBit(a, b);
  if (APUCheckCarry ())
    {
      S9xAPUSetByte (S9xAPUGetByte (a) | (1 << b),
		     a);
    }
  else
    {
      S9xAPUSetByte (S9xAPUGetByte (a) & ~(1 << b),
		     a);
    }
  APURegisters.PC += 3;
}

static void
ApuEA ()
{
  // NOT1 membit
  uint16 a;
  uint8 b;
  MemBit(a, b);
  S9xAPUSetByte (S9xAPUGetByte (a) ^ (1 << b),
		 a);
  APURegisters.PC += 3;
}

static void
Apu0B ()
{
  // ASL dp
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a);
  ASL (v);
  S9xAPUSetByteDP (v, a);
  APURegisters.PC += 2;
}

static void
Apu0C ()
{
  // ASL abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a);
  ASL (v);
  S9xAPUSetByte (v, a);
  APURegisters.PC += 3;
}

static void
Apu1B ()
{
  // ASL dp+X
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a + APURegisters.X);
  ASL (v);
  S9xAPUSetByteDP (v, a + APURegisters.X);
  APURegisters.PC += 2;
}

static void
Apu1C ()
{
  // ASL A
  ASL (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu0D ()
{
  // PUSH PSW
  S9xAPUPackStatus ();
  Push (APURegisters.P);
  APURegisters.PC++;
}

static void
Apu2D ()
{
  // PUSH A
  Push (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu4D ()
{
  // PUSH X
  Push (APURegisters.X);
  APURegisters.PC++;
}

static void
Apu6D ()
{
  // PUSH Y
  Push (APURegisters.YA.B.Y);
  APURegisters.PC++;
}

static void
Apu8E ()
{
  // POP PSW
  Pop (APURegisters.P);
  S9xAPUUnpackStatus ();
  APURegisters.PC++;
}

static void
ApuAE ()
{
  // POP A
  Pop (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuCE ()
{
  // POP X
  Pop (APURegisters.X);
  APURegisters.PC++;
}

static void
ApuEE ()
{
  // POP Y
  Pop (APURegisters.YA.B.Y);
  APURegisters.PC++;
}

//SPC700 "TSET1" "TCLR1" opcode's NZ flag operation was changed.
//"Super Momotarou Dentetsu II (J)"
//"Super Momotarou Dentetsu III (J)"
//These games were fixed, These sound effects came to ring.

static void
Apu0E ()
{
  // TSET1 abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a);
  S9xAPUSetByte (v | APURegisters.YA.B.A, a);
  APUSetZN8 (APURegisters.YA.B.A - v);
  APURegisters.PC += 3;
}

static void
Apu4E ()
{
  // TCLR1 abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a);
  S9xAPUSetByte (v & ~APURegisters.YA.B.A, a);
  APUSetZN8 (APURegisters.YA.B.A - v);
  APURegisters.PC += 3;
}

static void
Apu0F ()
{
  // BRK
  PushW (APURegisters.PC + 1);
  S9xAPUPackStatus ();
  Push (APURegisters.P);
  APUSetBreak ();
  APUClearInterrupt ();
  APURegisters.PC = S9xAPUGetByte(0xffde) | (S9xAPUGetByte(0xffdf) << 8);
}

static void
ApuEF ()
{
  // SLEEP
  // XXX: sleep
  //IAPU.APUExecuting = FALSE;
  //APURegisters.PC++;//comment out ?? It doesn't restart.
}

static void
ApuFF ()
{
  // STOP
  //IAPU.APUExecuting = FALSE;
  //APURegisters.PC++;//comment out ?? It doesn't restart.
}

//BPL
bOP(10, Relative, !APUCheckNegative (), 2)
//BMI
bOP(30, Relative, APUCheckNegative (), 2)
//BCC
bOP(90, Relative, !APUCheckCarry (), 2)
//BCS
bOP(B0, Relative, APUCheckCarry (), 2)
//BNE
bOP(D0, Relative, !APUCheckZero (), 2)
//BEQ
bOP(F0, Relative, APUCheckZero (), 2)
//BVC
bOP(50, Relative, !APUCheckOverflow (), 2)
//BVS
bOP(70, Relative, APUCheckOverflow (), 2)
//BRA
bOP(2F, Relative, TRUE, 2)

static void
Apu80 ()
{
  // SETC
  APUSetCarry ();
  APURegisters.PC++;
}

static void
ApuED ()
{
  // NOTC
  IAPU._Carry ^= 1;
  APURegisters.PC++;
}

static void
Apu40 ()
{
  // SETP
  APUSetDirectPage ();
  APURegisters.PC++;
}

static void
Apu1A ()
{
  // DECW dp
  uint8 a = OP1;
  uint16 v = S9xAPUGetByteDP (a) + (S9xAPUGetByteDP (a + 1) << 8);
  v--;
  S9xAPUSetByteDP ((uint8) v, a);
  S9xAPUSetByteDP (v >> 8, a + 1);
  APUSetZN16 (v);
  APURegisters.PC += 2;
}

static void
Apu5A ()
{
  // CMPW YA,dp
  uint8 a = OP1;
  uint16 v = S9xAPUGetByteDP (a) + (S9xAPUGetByteDP (a + 1) << 8);
  CMPW(APURegisters.YA.W, v);
  APURegisters.PC += 2;
}

static void
Apu3A ()
{
  // INCW dp
  uint8 a = OP1;
  uint16 v = S9xAPUGetByteDP (a) + (S9xAPUGetByteDP (a + 1) << 8);
  v++;
  S9xAPUSetByteDP ((uint8) v, a);
  S9xAPUSetByteDP (v >> 8, a + 1);
  APUSetZN16 (v);
  APURegisters.PC += 2;
}

static void
Apu7A ()
{
  // ADDW YA,dp
  uint8 a = OP1;
  ADDW(APURegisters.YA.W, S9xAPUGetByteDP (a) + (S9xAPUGetByteDP (a + 1) << 8));
  APURegisters.PC += 2;
}

static void
Apu9A ()
{
  // SUBW YA,dp
  uint8 a = OP1;
  SUBW(APURegisters.YA.W, S9xAPUGetByteDP (a) + (S9xAPUGetByteDP (a + 1) << 8));
  APURegisters.PC += 2;
}

static void
ApuBA ()
{
  // MOVW YA,dp
  uint8 a = OP1;
  APURegisters.YA.B.A = S9xAPUGetByteDP (a);
  APURegisters.YA.B.Y = S9xAPUGetByteDP (a + 1);
  APUSetZN16 (APURegisters.YA.W);
  APURegisters.PC += 2;
}

static void
ApuDA ()
{
  // MOVW dp,YA
  uint8 a = OP1;
  S9xAPUSetByteDP (APURegisters.YA.B.A, a);
  S9xAPUSetByteDP (APURegisters.YA.B.Y, a + 1);
  APURegisters.PC += 2;
}

static void
Apu64 ()
{
  // CMP A,dp
  CMP (APURegisters.YA.B.A, S9xAPUGetByteDP (OP1));
  APURegisters.PC += 2;
}

static void
Apu65 ()
{
  // CMP A,abs
  CMP (APURegisters.YA.B.A, S9xAPUGetByte (Absolute ()));
  APURegisters.PC += 3;
}

static void
Apu66 ()
{
  // CMP A,(X)
  CMP (APURegisters.YA.B.A, S9xAPUGetByteDP (APURegisters.X));
  APURegisters.PC++;
}

static void
Apu67 ()
{
  // CMP A,(dp+X)
  CMP (APURegisters.YA.B.A, S9xAPUGetByte (IndexedXIndirect ()));
  APURegisters.PC += 2;
}

static void
Apu68 ()
{
  // CMP A,#00
  CMP (APURegisters.YA.B.A, OP1);
  APURegisters.PC += 2;
}

static void
Apu69 ()
{
  // CMP dp(dest), dp(src)
  uint8 v0 = S9xAPUGetByteDP (OP1);
  uint8 v1 = S9xAPUGetByteDP (OP2);
  CMP (v1, v0);
  APURegisters.PC += 3;
}

static void
Apu74 ()
{
  // CMP A, dp+X
  CMP (APURegisters.YA.B.A, S9xAPUGetByteDP (OP1 + APURegisters.X));
  APURegisters.PC += 2;
}

static void
Apu75 ()
{
  // CMP A,abs+X
  CMP (APURegisters.YA.B.A, S9xAPUGetByte (AbsoluteX ()));
  APURegisters.PC += 3;
}

static void
Apu76 ()
{
  // CMP A, abs+Y
  CMP (APURegisters.YA.B.A, S9xAPUGetByte (AbsoluteY ()));
  APURegisters.PC += 3;
}

static void
Apu77 ()
{
  // CMP A,(dp)+Y
  CMP (APURegisters.YA.B.A, S9xAPUGetByte (IndirectIndexedY ()));
  APURegisters.PC += 2;
}

static void
Apu78 ()
{
  // CMP dp,#00
  uint8 v0 = OP1;
  uint8 v1 = S9xAPUGetByteDP (OP2);
  CMP (v1, v0);
  APURegisters.PC += 3;
}

static void
Apu79 ()
{
  // CMP (X),(Y)
  CMP (S9xAPUGetByteDP (APURegisters.X), S9xAPUGetByteDP (APURegisters.YA.B.Y));
  APURegisters.PC++;
}

static void
Apu1E ()
{
  // CMP X,abs
  CMP (APURegisters.X, S9xAPUGetByte (Absolute ()));
  APURegisters.PC += 3;
}

static void
Apu3E ()
{
  // CMP X,dp
  CMP (APURegisters.X, S9xAPUGetByteDP (OP1));
  APURegisters.PC += 2;
}

static void
ApuC8 ()
{
  // CMP X,#00
  CMP (APURegisters.X, OP1);
  APURegisters.PC += 2;
}

static void
Apu5E ()
{
  // CMP Y,abs
  CMP (APURegisters.YA.B.Y, S9xAPUGetByte (Absolute ()));
  APURegisters.PC += 3;
}

static void
Apu7E ()
{
  // CMP Y,dp
  CMP (APURegisters.YA.B.Y, S9xAPUGetByteDP (OP1));
  APURegisters.PC += 2;
}

static void
ApuAD ()
{
  // CMP Y,#00
  CMP (APURegisters.YA.B.Y, OP1);
  APURegisters.PC += 2;
}

static void
Apu1F ()
{
  // JMP (abs+X)
  uint16 a = Absolute ();
  APURegisters.PC = S9xAPUGetByte (a + APURegisters.X) | (S9xAPUGetByte (a + APURegisters.X + 1) << 8);
}

static void
Apu5F ()
{
  // JMP abs
  APURegisters.PC = Absolute ();
}

static void
Apu20 ()
{
  // CLRP
  APUClearDirectPage ();
  APURegisters.PC++;
}

static void
Apu60 ()
{
  // CLRC
  APUClearCarry ();
  APURegisters.PC++;
}

static void
ApuE0 ()
{
  // CLRV
  APUClearHalfCarry ();
  APUClearOverflow ();
  APURegisters.PC++;
}

static void
Apu24 ()
{
  // AND A,dp
  APURegisters.YA.B.A &= S9xAPUGetByteDP (OP1);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu25 ()
{
  // AND A,abs
  APURegisters.YA.B.A &= S9xAPUGetByte (Absolute ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu26 ()
{
  // AND A,(X)
  APURegisters.YA.B.A &= S9xAPUGetByteDP (APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu27 ()
{
  // AND A,(dp+X)
  APURegisters.YA.B.A &= S9xAPUGetByte (IndexedXIndirect ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu28 ()
{
  // AND A,#00
  APURegisters.YA.B.A &= OP1;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu29 ()
{
  // AND dp(dest),dp(src)
  uint8 v = S9xAPUGetByteDP (OP1);
  uint8 a = OP2;
  v &= S9xAPUGetByteDP (a);
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu34 ()
{
  // AND A,dp+X
  APURegisters.YA.B.A &= S9xAPUGetByteDP (OP1 + APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu35 ()
{
  // AND A,abs+X
  APURegisters.YA.B.A &= S9xAPUGetByte (AbsoluteX ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu36 ()
{
  // AND A,abs+Y
  APURegisters.YA.B.A &= S9xAPUGetByte (AbsoluteY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu37 ()
{
  // AND A,(dp)+Y
  APURegisters.YA.B.A &= S9xAPUGetByte (IndirectIndexedY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu38 ()
{
  // AND dp,#00
  uint8 v = OP1;
  uint8 a = OP2;
  v &= S9xAPUGetByteDP (a);
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu39 ()
{
  // AND (X),(Y)
  uint8 v = S9xAPUGetByteDP (APURegisters.X) & S9xAPUGetByteDP (APURegisters.YA.B.Y);
  APUSetZN8 (v);
  S9xAPUSetByteDP (v, APURegisters.X);
  APURegisters.PC++;
}

static void
Apu2B ()
{
  // ROL dp
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a);
  ROL (v);
  S9xAPUSetByteDP (v, a);
  APURegisters.PC += 2;
}

static void
Apu2C ()
{
  // ROL abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a);
  ROL (v);
  S9xAPUSetByte (v, a);
  APURegisters.PC += 3;
}

static void
Apu3B ()
{
  // ROL dp+X
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a + APURegisters.X);
  ROL (v);
  S9xAPUSetByteDP (v, a + APURegisters.X);
  APURegisters.PC += 2;
}

static void
Apu3C ()
{
  // ROL A
  ROL (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu2E ()
{
  // CBNE dp,rel
  if (S9xAPUGetByteDP (OP1) != APURegisters.YA.B.A)
    {
      APURegisters.PC = Relative2 ();
      smp_add_clocks(2);
    }
  else
    {
      APURegisters.PC += 3;
    }
}

static void
ApuDE ()
{
  // CBNE dp+X,rel
  if (S9xAPUGetByteDP (OP1 + APURegisters.X) != APURegisters.YA.B.A)
    {
      APURegisters.PC = Relative2 ();
      smp_add_clocks(2);
    }
  else
    {
      APURegisters.PC += 3;
    }
}

static void
Apu3D ()
{
  // INC X
  APURegisters.X++;
  APUSetZN8 (APURegisters.X);
  APURegisters.PC++;
}

static void
ApuFC ()
{
  // INC Y
  APURegisters.YA.B.Y++;
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC++;
}

static void
Apu1D ()
{
  // DEC X
  APURegisters.X--;
  APUSetZN8 (APURegisters.X);
  APURegisters.PC++;
}

static void
ApuDC ()
{
  // DEC Y
  APURegisters.YA.B.Y--;
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC++;
}

static void
ApuAB ()
{
  // INC dp
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a) + 1;
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 2;
}

static void
ApuAC ()
{
  // INC abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a) + 1;
  S9xAPUSetByte (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
ApuBB ()
{
  // INC dp+X
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a + APURegisters.X) + 1;
  S9xAPUSetByteDP (v, a + APURegisters.X);
  APUSetZN8 (v);
  APURegisters.PC += 2;
}

static void
ApuBC ()
{
  // INC A
  APURegisters.YA.B.A++;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu8B ()
{
  // DEC dp
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a) - 1;
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 2;
}

static void
Apu8C ()
{
  // DEC abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a) - 1;
  S9xAPUSetByte (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu9B ()
{
  // DEC dp+X
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a + APURegisters.X) - 1;
  S9xAPUSetByteDP (v, a + APURegisters.X);
  APUSetZN8 (v);
  APURegisters.PC += 2;
}

static void
Apu9C ()
{
  // DEC A
  APURegisters.YA.B.A--;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu44 ()
{
  // EOR A,dp
  APURegisters.YA.B.A ^= S9xAPUGetByteDP (OP1);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu45 ()
{
  // EOR A,abs
  APURegisters.YA.B.A ^= S9xAPUGetByte (Absolute ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu46 ()
{
  // EOR A,(X)
  APURegisters.YA.B.A ^= S9xAPUGetByteDP (APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu47 ()
{
  // EOR A,(dp+X)
  APURegisters.YA.B.A ^= S9xAPUGetByte (IndexedXIndirect ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu48 ()
{
  // EOR A,#00
  APURegisters.YA.B.A ^= OP1;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu49 ()
{
  // EOR dp(dest),dp(src)
  uint8 v = S9xAPUGetByteDP (OP1);
  uint8 a = OP2;
  v ^= S9xAPUGetByteDP (a);
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu54 ()
{
  // EOR A,dp+X
  APURegisters.YA.B.A ^= S9xAPUGetByteDP (OP1 + APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu55 ()
{
  // EOR A,abs+X
  APURegisters.YA.B.A ^= S9xAPUGetByte (AbsoluteX ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu56 ()
{
  // EOR A,abs+Y
  APURegisters.YA.B.A ^= S9xAPUGetByte (AbsoluteY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
Apu57 ()
{
  // EOR A,(dp)+Y
  APURegisters.YA.B.A ^= S9xAPUGetByte (IndirectIndexedY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
Apu58 ()
{
  // EOR dp,#00
  uint8 v = OP1;
  uint8 a = OP2;
  v ^= S9xAPUGetByteDP (a);
  S9xAPUSetByteDP (v, a);
  APUSetZN8 (v);
  APURegisters.PC += 3;
}

static void
Apu59 ()
{
  // EOR (X),(Y)
  uint8 v = S9xAPUGetByteDP (APURegisters.X) ^ S9xAPUGetByteDP (APURegisters.YA.B.Y);
  APUSetZN8 (v);
  S9xAPUSetByteDP (v, APURegisters.X);
  APURegisters.PC++;
}

static void
Apu4B ()
{
  // LSR dp
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a);
  LSR (v);
  S9xAPUSetByteDP (v, a);
  APURegisters.PC += 2;
}

static void
Apu4C ()
{
  // LSR abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a);
  LSR (v);
  S9xAPUSetByte (v, a);
  APURegisters.PC += 3;
}

static void
Apu5B ()
{
  // LSR dp+X
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a + APURegisters.X);
  LSR (v);
  S9xAPUSetByteDP (v, a + APURegisters.X);
  APURegisters.PC += 2;
}

static void
Apu5C ()
{
  // LSR A
  LSR (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu7D ()
{
  // MOV A,X
  APURegisters.YA.B.A = APURegisters.X;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuDD ()
{
  // MOV A,Y
  APURegisters.YA.B.A = APURegisters.YA.B.Y;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu5D ()
{
  // MOV X,A
  APURegisters.X = APURegisters.YA.B.A;
  APUSetZN8 (APURegisters.X);
  APURegisters.PC++;
}

static void
ApuFD ()
{
  // MOV Y,A
  APURegisters.YA.B.Y = APURegisters.YA.B.A;
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC++;
}

static void
Apu9D ()
{
  //MOV X,SP
  APURegisters.X = APURegisters.S;
  APUSetZN8 (APURegisters.X);
  APURegisters.PC++;
}

static void
ApuBD ()
{
  // MOV SP,X
  APURegisters.S = APURegisters.X;
  APURegisters.PC++;
}

static void
Apu6B ()
{
  // ROR dp
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a);
  ROR (v);
  S9xAPUSetByteDP (v, a);
  APURegisters.PC += 2;
}

static void
Apu6C ()
{
  // ROR abs
  uint16 a = Absolute ();
  uint8 v = S9xAPUGetByte (a);
  ROR (v);
  S9xAPUSetByte (v, a);
  APURegisters.PC += 3;
}

static void
Apu7B ()
{
  // ROR dp+X
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a + APURegisters.X);
  ROR (v);
  S9xAPUSetByteDP (v, a + APURegisters.X);
  APURegisters.PC += 2;
}

static void
Apu7C ()
{
  // ROR A
  ROR (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu6E ()
{
  // DBNZ dp,rel
  uint8 a = OP1;
  uint8 v = S9xAPUGetByteDP (a) - 1;
  S9xAPUSetByteDP (v, a);
  if (v != 0)
    {
      APURegisters.PC = Relative2 ();
      smp_add_clocks(2);
    }
  else
    APURegisters.PC += 3;
}

static void
ApuFE ()
{
  // DBNZ Y,rel
  APURegisters.YA.B.Y--;
  if (APURegisters.YA.B.Y != 0)
    {
      APURegisters.PC = Relative ();
      smp_add_clocks(2);
    }
  else
    APURegisters.PC += 2;
}

static void
Apu6F ()
{
  // RET
  PopW (APURegisters.PC);
}

static void
Apu7F ()
{
  // RETI
  Pop (APURegisters.P);
  S9xAPUUnpackStatus ();
  PopW (APURegisters.PC);
}

static void
Apu84 ()
{
  // ADC A,dp
  ADC (APURegisters.YA.B.A, S9xAPUGetByteDP (OP1));
  APURegisters.PC += 2;
}

static void
Apu85 ()
{
  // ADC A, abs
  ADC (APURegisters.YA.B.A, S9xAPUGetByte (Absolute ()));
  APURegisters.PC += 3;
}

static void
Apu86 ()
{
  // ADC A,(X)
  ADC (APURegisters.YA.B.A, S9xAPUGetByteDP (APURegisters.X));
  APURegisters.PC++;
}

static void
Apu87 ()
{
  // ADC A,(dp+X)
  ADC (APURegisters.YA.B.A, S9xAPUGetByte (IndexedXIndirect ()));
  APURegisters.PC += 2;
}

static void
Apu88 ()
{
  // ADC A,#00
  ADC (APURegisters.YA.B.A, OP1);
  APURegisters.PC += 2;
}

static void
Apu89 ()
{
  // ADC dp(dest),dp(src)
  uint8 v0 = S9xAPUGetByteDP (OP1);
  uint8 a = OP2;
  uint8 v1 = S9xAPUGetByteDP (a);
  ADC (v1, v0);
  S9xAPUSetByteDP (v1, a);
  APURegisters.PC += 3;
}

static void
Apu94 ()
{
  // ADC A,dp+X
  ADC (APURegisters.YA.B.A, S9xAPUGetByteDP (OP1 + APURegisters.X));
  APURegisters.PC += 2;
}

static void
Apu95 ()
{
  // ADC A, abs+X
  ADC (APURegisters.YA.B.A, S9xAPUGetByte (AbsoluteX ()));
  APURegisters.PC += 3;
}

static void
Apu96 ()
{
  // ADC A, abs+Y
  ADC (APURegisters.YA.B.A, S9xAPUGetByte (AbsoluteY ()));
  APURegisters.PC += 3;
}

static void
Apu97 ()
{
  // ADC A, (dp)+Y
  ADC (APURegisters.YA.B.A, S9xAPUGetByte (IndirectIndexedY ()));
  APURegisters.PC += 2;
}

static void
Apu98 ()
{
  // ADC dp,#00
  uint8 v0 = OP1;
  uint8 a = OP2;
  uint8 v1 = S9xAPUGetByteDP (a);
  ADC (v1, v0);
  S9xAPUSetByteDP (v1, a);
  APURegisters.PC += 3;
}

static void
Apu99 ()
{
  // ADC (X),(Y)
  uint8 v = S9xAPUGetByteDP (APURegisters.X);
  ADC (v, S9xAPUGetByteDP (APURegisters.YA.B.Y));
  S9xAPUSetByteDP (v, APURegisters.X);
  APURegisters.PC++;
}

static void
Apu8D ()
{
  // MOV Y,#00
  APURegisters.YA.B.Y = OP1;
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC += 2;
}

static void
Apu8F ()
{
  // MOV dp,#00
  S9xAPUSetByteDP (OP1, OP2);
  APURegisters.PC += 3;
}

static void
Apu9E ()
{
  // DIV YA,X
  uint16 ya = APURegisters.YA.W;

  if(APURegisters.YA.B.Y >= APURegisters.X)
    APUSetOverflow ();
  else
    APUClearOverflow ();

  if((APURegisters.YA.B.Y & 15) >= (APURegisters.X & 15))
    APUSetHalfCarry ();
  else
    APUClearHalfCarry ();

  if(APURegisters.YA.B.Y < (APURegisters.X << 1))
    {
      APURegisters.YA.B.A = ya / APURegisters.X;
      APURegisters.YA.B.Y = ya % APURegisters.X;
    }
  else
    {
      APURegisters.YA.B.A = 255 - (ya - (APURegisters.X << 9)) / (256 - APURegisters.X);
      APURegisters.YA.B.Y = APURegisters.X + (ya - (APURegisters.X << 9)) % (256 - APURegisters.X);
    }
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
Apu9F ()
{
  // XCN A
  APURegisters.YA.B.A =
    (APURegisters.YA.B.A >> 4) | (APURegisters.YA.B.A << 4);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuA4 ()
{
  // SBC A, dp
  SBC (APURegisters.YA.B.A, S9xAPUGetByteDP (OP1));
  APURegisters.PC += 2;
}

static void
ApuA5 ()
{
  // SBC A, abs
  SBC (APURegisters.YA.B.A, S9xAPUGetByte (Absolute ()));
  APURegisters.PC += 3;
}

static void
ApuA6 ()
{
  // SBC A, (X)
  SBC (APURegisters.YA.B.A, S9xAPUGetByteDP (APURegisters.X));
  APURegisters.PC++;
}

static void
ApuA7 ()
{
  // SBC A,(dp+X)
  SBC (APURegisters.YA.B.A, S9xAPUGetByte (IndexedXIndirect ()));
  APURegisters.PC += 2;
}

static void
ApuA8 ()
{
  // SBC A,#00
  SBC (APURegisters.YA.B.A, OP1);
  APURegisters.PC += 2;
}

static void
ApuA9 ()
{
  // SBC dp(dest), dp(src)
  uint8 v0 = S9xAPUGetByteDP (OP1);
  uint8 a = OP2;
  uint8 v1 = S9xAPUGetByteDP (a);
  SBC (v1, v0);
  S9xAPUSetByteDP (v1, a);
  APURegisters.PC += 3;
}

static void
ApuB4 ()
{
  // SBC A, dp+X
  SBC (APURegisters.YA.B.A, S9xAPUGetByteDP (OP1 + APURegisters.X));
  APURegisters.PC += 2;
}

static void
ApuB5 ()
{
  // SBC A,abs+X
  SBC (APURegisters.YA.B.A, S9xAPUGetByte (AbsoluteX ()));
  APURegisters.PC += 3;
}

static void
ApuB6 ()
{
  // SBC A,abs+Y
  SBC (APURegisters.YA.B.A, S9xAPUGetByte (AbsoluteY ()));
  APURegisters.PC += 3;
}

static void
ApuB7 ()
{
  // SBC A,(dp)+Y
  SBC (APURegisters.YA.B.A, S9xAPUGetByte (IndirectIndexedY ()));
  APURegisters.PC += 2;
}

static void
ApuB8 ()
{
  // SBC dp,#00
  uint8 v0 = OP1;
  uint8 a = OP2;
  uint8 v1 = S9xAPUGetByteDP (a);
  SBC (v1, v0);
  S9xAPUSetByteDP (v1, a);
  APURegisters.PC += 3;
}

static void
ApuB9 ()
{
  // SBC (X),(Y)
  uint8 v = S9xAPUGetByteDP (APURegisters.X);
  SBC (v, S9xAPUGetByteDP (APURegisters.YA.B.Y));
  S9xAPUSetByteDP (v, APURegisters.X);
  APURegisters.PC++;
}

static void
ApuAF ()
{
  // MOV (X)+, A
  S9xAPUSetByteDP (APURegisters.YA.B.A, APURegisters.X++);
  APURegisters.PC++;
}

static void
ApuBE ()
{
  // DAS
  if(!APUCheckCarry () || APURegisters.YA.B.A > 0x99)
    {
      APURegisters.YA.B.A -= 0x60;
      APUClearCarry ();
    }

  if(!APUCheckHalfCarry () || (APURegisters.YA.B.A & 15) > 0x09)
      APURegisters.YA.B.A -= 0x06;

  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuBF ()
{
  // MOV A,(X)+
  APURegisters.YA.B.A = S9xAPUGetByteDP (APURegisters.X++);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuC0 ()
{
  // DI
  APUClearInterrupt ();
  APURegisters.PC++;
}

static void
ApuA0 ()
{
  // EI
  APUSetInterrupt ();
  APURegisters.PC++;
}

static void
ApuC4 ()
{
  // MOV dp,A
  S9xAPUSetByteDP (APURegisters.YA.B.A, OP1);
  APURegisters.PC += 2;
}

static void
ApuC5 ()
{
  // MOV abs,A
  S9xAPUSetByte (APURegisters.YA.B.A, Absolute ());
  APURegisters.PC += 3;
}

static void
ApuC6 ()
{
  // MOV (X), A
  S9xAPUSetByteDP (APURegisters.YA.B.A, APURegisters.X);
  APURegisters.PC++;
}

static void
ApuC7 ()
{
  // MOV (dp+X),A
  S9xAPUSetByte (APURegisters.YA.B.A, IndexedXIndirect ());
  APURegisters.PC += 2;
}

static void
ApuC9 ()
{
  // MOV abs,X
  S9xAPUSetByte (APURegisters.X, Absolute ());
  APURegisters.PC += 3;
}

static void
ApuCB ()
{
  // MOV dp,Y
  S9xAPUSetByteDP (APURegisters.YA.B.Y, OP1);
  APURegisters.PC += 2;
}

static void
ApuCC ()
{
  // MOV abs,Y
  S9xAPUSetByte (APURegisters.YA.B.Y, Absolute ());
  APURegisters.PC += 3;
}

static void
ApuCD ()
{
  // MOV X,#00
  APURegisters.X = OP1;
  APUSetZN8 (APURegisters.X);
  APURegisters.PC += 2;
}

static void
ApuCF ()
{
  // MUL YA
  APURegisters.YA.W = (uint16) APURegisters.YA.B.Y * APURegisters.YA.B.A;

  //result is set based on y (high-byte) only
  APUSetZN8 (APURegisters.YA.B.Y);

  APURegisters.PC++;
}

static void
ApuD4 ()
{
  // MOV dp+X, A
  S9xAPUSetByteDP (APURegisters.YA.B.A, OP1 + APURegisters.X);
  APURegisters.PC += 2;
}

static void
ApuD5 ()
{
  // MOV abs+X,A
  S9xAPUSetByte (APURegisters.YA.B.A, AbsoluteX ());
  APURegisters.PC += 3;
}

static void
ApuD6 ()
{
  // MOV abs+Y,A
  S9xAPUSetByte (APURegisters.YA.B.A, AbsoluteY ());
  APURegisters.PC += 3;
}

static void
ApuD7 ()
{
  // MOV (dp)+Y,A
  S9xAPUSetByte (APURegisters.YA.B.A, IndirectIndexedY ());
  APURegisters.PC += 2;
}

static void
ApuD8 ()
{
// MOV dp,X
  S9xAPUSetByteDP (APURegisters.X, OP1);
  APURegisters.PC += 2;
}

static void
ApuD9 ()
{
  // MOV dp+Y,X
  S9xAPUSetByteDP (APURegisters.X, OP1 + APURegisters.YA.B.Y);
  APURegisters.PC += 2;
}

static void
ApuDB ()
{
  // MOV dp+X,Y
  S9xAPUSetByteDP (APURegisters.YA.B.Y, OP1 + APURegisters.X);
  APURegisters.PC += 2;
}

static void
ApuDF ()
{
  // DAA
  if(APUCheckCarry () || APURegisters.YA.B.A > 0x99)
    {
      APURegisters.YA.B.A += 0x60;
      APUSetCarry ();
    }

  if(APUCheckHalfCarry () || (APURegisters.YA.B.A & 15) > 0x09)
      APURegisters.YA.B.A += 0x06;

  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuE4 ()
{
  // MOV A, dp
  APURegisters.YA.B.A = S9xAPUGetByteDP (OP1);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
ApuE5 ()
{
  // MOV A,abs
  APURegisters.YA.B.A = S9xAPUGetByte (Absolute ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
ApuE6 ()
{
  // MOV A,(X)
  APURegisters.YA.B.A = S9xAPUGetByteDP (APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC++;
}

static void
ApuE7 ()
{
  // MOV A,(dp+X)
  APURegisters.YA.B.A = S9xAPUGetByte (IndexedXIndirect ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
ApuE8 ()
{
  // MOV A,#00
  APURegisters.YA.B.A = OP1;
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
ApuE9 ()
{
  // MOV X, abs
  APURegisters.X = S9xAPUGetByte (Absolute ());
  APUSetZN8 (APURegisters.X);
  APURegisters.PC += 3;
}

static void
ApuEB ()
{
  // MOV Y,dp
  APURegisters.YA.B.Y = S9xAPUGetByteDP (OP1);
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC += 2;
}

static void
ApuEC ()
{
  // MOV Y,abs
  APURegisters.YA.B.Y = S9xAPUGetByte (Absolute ());
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC += 3;
}

static void
ApuF4 ()
{
  // MOV A, dp+X
  APURegisters.YA.B.A = S9xAPUGetByteDP (OP1 + APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
ApuF5 ()
{
  // MOV A, abs+X
  APURegisters.YA.B.A = S9xAPUGetByte (AbsoluteX ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
ApuF6 ()
{
  // MOV A, abs+Y
  APURegisters.YA.B.A = S9xAPUGetByte (AbsoluteY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 3;
}

static void
ApuF7 ()
{
  // MOV A, (dp)+Y
  APURegisters.YA.B.A = S9xAPUGetByte (IndirectIndexedY ());
  APUSetZN8 (APURegisters.YA.B.A);
  APURegisters.PC += 2;
}

static void
ApuF8 ()
{
  // MOV X,dp
  APURegisters.X = S9xAPUGetByteDP (OP1);
  APUSetZN8 (APURegisters.X);
  APURegisters.PC += 2;
}

static void
ApuF9 ()
{
  // MOV X,dp+Y
  APURegisters.X = S9xAPUGetByteDP (OP1 + APURegisters.YA.B.Y);
  APUSetZN8 (APURegisters.X);
  APURegisters.PC += 2;
}

static void
ApuFA ()
{
  // MOV dp(dest),dp(src)
  S9xAPUSetByteDP (S9xAPUGetByteDP (OP1), OP2);
  APURegisters.PC += 3;
}

static void
ApuFB ()
{
  // MOV Y,dp+X
  APURegisters.YA.B.Y = S9xAPUGetByteDP (OP1 + APURegisters.X);
  APUSetZN8 (APURegisters.YA.B.Y);
  APURegisters.PC += 2;
}

void (*S9xApuOpcodes[256]) (void) =
{
  Apu00, Apu01, Apu02, Apu03, Apu04, Apu05, Apu06, Apu07,
  Apu08, Apu09, Apu0A, Apu0B, Apu0C, Apu0D, Apu0E, Apu0F,
  Apu10, Apu11, Apu12, Apu13, Apu14, Apu15, Apu16, Apu17,
  Apu18, Apu19, Apu1A, Apu1B, Apu1C, Apu1D, Apu1E, Apu1F,
  Apu20, Apu21, Apu22, Apu23, Apu24, Apu25, Apu26, Apu27,
  Apu28, Apu29, Apu2A, Apu2B, Apu2C, Apu2D, Apu2E, Apu2F,
  Apu30, Apu31, Apu32, Apu33, Apu34, Apu35, Apu36, Apu37,
  Apu38, Apu39, Apu3A, Apu3B, Apu3C, Apu3D, Apu3E, Apu3F,
  Apu40, Apu41, Apu42, Apu43, Apu44, Apu45, Apu46, Apu47,
  Apu48, Apu49, Apu4A, Apu4B, Apu4C, Apu4D, Apu4E, Apu4F,
  Apu50, Apu51, Apu52, Apu53, Apu54, Apu55, Apu56, Apu57,
  Apu58, Apu59, Apu5A, Apu5B, Apu5C, Apu5D, Apu5E, Apu5F,
  Apu60, Apu61, Apu62, Apu63, Apu64, Apu65, Apu66, Apu67,
  Apu68, Apu69, Apu6A, Apu6B, Apu6C, Apu6D, Apu6E, Apu6F,
  Apu70, Apu71, Apu72, Apu73, Apu74, Apu75, Apu76, Apu77,
  Apu78, Apu79, Apu7A, Apu7B, Apu7C, Apu7D, Apu7E, Apu7F,
  Apu80, Apu81, Apu82, Apu83, Apu84, Apu85, Apu86, Apu87,
  Apu88, Apu89, Apu8A, Apu8B, Apu8C, Apu8D, Apu8E, Apu8F,
  Apu90, Apu91, Apu92, Apu93, Apu94, Apu95, Apu96, Apu97,
  Apu98, Apu99, Apu9A, Apu9B, Apu9C, Apu9D, Apu9E, Apu9F,
  ApuA0, ApuA1, ApuA2, ApuA3, ApuA4, ApuA5, ApuA6, ApuA7,
  ApuA8, ApuA9, ApuAA, ApuAB, ApuAC, ApuAD, ApuAE, ApuAF,
  ApuB0, ApuB1, ApuB2, ApuB3, ApuB4, ApuB5, ApuB6, ApuB7,
  ApuB8, ApuB9, ApuBA, ApuBB, ApuBC, ApuBD, ApuBE, ApuBF,
  ApuC0, ApuC1, ApuC2, ApuC3, ApuC4, ApuC5, ApuC6, ApuC7,
  ApuC8, ApuC9, ApuCA, ApuCB, ApuCC, ApuCD, ApuCE, ApuCF,
  ApuD0, ApuD1, ApuD2, ApuD3, ApuD4, ApuD5, ApuD6, ApuD7,
  ApuD8, ApuD9, ApuDA, ApuDB, ApuDC, ApuDD, ApuDE, ApuDF,
  ApuE0, ApuE1, ApuE2, ApuE3, ApuE4, ApuE5, ApuE6, ApuE7,
  ApuE8, ApuE9, ApuEA, ApuEB, ApuEC, ApuED, ApuEE, ApuEF,
  ApuF0, ApuF1, ApuF2, ApuF3, ApuF4, ApuF5, ApuF6, ApuF7,
  ApuF8, ApuF9, ApuFA, ApuFB, ApuFC, ApuFD, ApuFE, ApuFF
};

