/*
  Copyright (C) 2004 sanmaiwashi
*/

#include "snes9x.h"
#include "memmap.h"
#include "cpuexec.h"
#include "apu_orig.h"
#include "soundux.h"
#include "snesapu.h"
#include "gfx.h"
#include "spcplay.h"

static SAPURegisters BackupAPURegisters;
static uint8 BackupAPURAM[65536];
static uint8 BackupAPUExtraRAM[64];
static uint8 BackupDSPRAM[128];

void
SPCPlayMainLoop (void)
{
  CPU.Cycles = Timings.H_Max_Master*Timings.V_Max_Master - 2;
  S9xAPUMainLoop ();
  APU.Cycles -= Timings.H_Max_Master*Timings.V_Max_Master - 2;
#ifdef DEBUGGER
  if (!(CPU.Flags & FRAME_ADVANCE_FLAG))
#endif
    S9xSyncSpeed ();
}

// Restore SPC Load State
static void
RestoreSPC ()
{
  int i;

  APURegisters.PC = BackupAPURegisters.PC;
  APURegisters.YA.B.A = BackupAPURegisters.YA.B.A;
  APURegisters.X = BackupAPURegisters.X;
  APURegisters.YA.B.Y = BackupAPURegisters.YA.B.Y;
  APURegisters.P = BackupAPURegisters.P;
  APURegisters.S = BackupAPURegisters.S;
  memcpy (IAPU.RAM, BackupAPURAM, 65536);

  //direct access iplrom
  memcpy(&IAPU.RAM[0xffc0], BackupAPUExtraRAM, 64);

  memcpy (APU.DSP, BackupDSPRAM, 128);

  S9xAPUUnpackStatus ();
  S9xSetAPUControl (IAPU.RAM[0xf1]);

  if(IAPU.RAM[0xf0] == 0xFF)
    IAPU.RAM[0xf0] = 0x0A;//timers_enabled & ram_writable

  for (i = 0; i < 4; i++)
    {
      //APU.OutPorts [i] = BackupAPURAM[0xf4 + i];
      IAPU.RAM[0xf4 + i] = BackupAPURAM[0xf4 + i];
    }

  for (i = 0; i < 3; i++)
    IAPU.RAM[0xfd + i] = BackupAPURAM[0xfd + i] & 0xf;

  IAPU.Scanline = 0;
  RestoreSPCFile ();
}

// Load SPC file
bool8
LoadSPC (const char *pszFileName)
{
  FILE *file;
  char temp[64];

  if ((file = fopen (pszFileName, "rb")) == NULL)
    return FALSE;

  memset (&SNESGameFixes, 0, sizeof (SNESGameFixes));

  Settings.APURAMInitialValue = Settings.APURAMInitialValueMaster;
  Settings.APURAMInitialValueMaster = DefaultSettings.APURAMInitialValueMaster;

  Settings.SRAMInitialValue = Settings.SRAMInitialValueMaster;
  Settings.SRAMInitialValueMaster = DefaultSettings.SRAMInitialValueMaster;

  Settings.RAMInitialValue = Settings.RAMInitialValueMaster;
  Settings.RAMInitialValueMaster = DefaultSettings.RAMInitialValueMaster;

  Settings.SPCPlaying = TRUE;

  Settings.PAL = TRUE;
  Settings.FrameTime = Settings.FrameTimePAL;
  Memory.ROMFramesPerSecond = 50;

  Timings.H_Max_Master = SNES_CYCLES_PER_SCANLINE;
  Timings.V_Max_Master = Settings.PAL ? SNES_MAX_PAL_VCOUNTER : SNES_MAX_NTSC_VCOUNTER;

  IAPU.MasterClock = Settings.smp_master_clock;
  IAPU.OneCycle = Settings.APUOneCycle;

  S9xSetHCycles (Settings.H_MaxTimingHack, Settings.HDMATimingHack, Settings.OldTimingMaster);

  //S9xResetAPU();
  S9xReset ();

  fseek (file, 37, SEEK_SET);
  fread (&BackupAPURegisters.PC, 2, 1, file);
  fread (&BackupAPURegisters.YA.B.A, 1, 1, file);
  fread (&BackupAPURegisters.X, 1, 1, file);
  fread (&BackupAPURegisters.YA.B.Y, 1, 1, file);
  fread (&BackupAPURegisters.P, 1, 1, file);
  fread (&BackupAPURegisters.S, 1, 1, file);
  fseek (file, 256, SEEK_SET);
  fread (BackupAPURAM, 65536, 1, file);
  fread (BackupDSPRAM, 128, 1, file);
  fread (temp, 64, 1, file);
  fread (BackupAPUExtraRAM, 64, 1, file);

  fclose (file);
  RestoreSPC ();

  //APU.Flags |= TRACE_FLAG;

  return TRUE;
}

