/*
  Copyright (C) 2010 sanmaiwashi
*/

#include "snes9x.h"
#include "ppu.h"

#ifdef __WIN32__
struct SSurface
{
  PBYTE Surface;
  DWORD Pitch;
  DWORD Width;
  DWORD Height;
};
#endif

#include "render.h"
#include "blit.h"

#define READ_PIXEL_16(s) (*(uint16 *) (s))

static void
RenderNothing (SSurface * Src, SSurface * Dst, RECT * srcRect, RECT * dstRect)
{
  // Just an empty procedure, so that it never happens that something 'bad' is rendered =)
}

//store RenderInit()
static int ScreenDepth;
static int RedShift;
static int GreenShift;
static int BlueShift;

// Contains the pointer to the now active render method
TRenderMethod OldRenderMethod = RenderNothing;
TRenderMethod RenderMethod = RenderNothing;
bool8 DoubleWidth = FALSE;
bool8 DoubleHeight = FALSE;
bool8 ExtendHeight = FALSE;
bool8 ExtendHeight_previos = FALSE;
extern uint8 S9xPalette[];

static uint8 *depthconv_buf = NULL;
static int buffer_pitch;
static int buffer_size;

static uint8 color_table[0x10000 * 4];

#define SNES_MULBRIGHTNESS 140

#define SRC_PIXSIZE_8to8   1
#define SRC_PIXSIZE_8to16  1
#define SRC_PIXSIZE_8to24  1
#define SRC_PIXSIZE_8to32  1
#define SRC_PIXSIZE_16to8  2
#define SRC_PIXSIZE_16to16 2
#define SRC_PIXSIZE_16to24 2
#define SRC_PIXSIZE_16to32 2
#define DST_PIXSIZE_8to8   1
#define DST_PIXSIZE_8to16  2
#define DST_PIXSIZE_8to24  3
#define DST_PIXSIZE_8to32  4
#define DST_PIXSIZE_16to8  1
#define DST_PIXSIZE_16to16 2
#define DST_PIXSIZE_16to24 3
#define DST_PIXSIZE_16to32 4


/*******************
 put pixel macros */

#define Interp05(c1, c2)\
  (((c1) == (c2)) ? (c1) :						\
   ((((((c1) & SECOND_COLOR_MASK) + ((c2) & SECOND_COLOR_MASK)) >> 1) & SECOND_COLOR_MASK) + \
    (((((c1) & FIRST_THIRD_COLOR_MASK) + ((c2) & FIRST_THIRD_COLOR_MASK)) >> 1) & FIRST_THIRD_COLOR_MASK)))

#define PUT_PIXEL_W11_16to8()\
  *(pDst + dst_index) = S9xPalette [READ_PIXEL_16(pSrc + src_index)];

#define PUT_PIXEL_W12_16to8()\
  *(pDst + dst_index) =							\
    *(pDst + dst_index + 1) = S9xPalette [READ_PIXEL_16(pSrc + src_index)];

#define PUT_PIXEL_W13_16to8()\
  *(pDst + dst_index) =							\
    *(pDst + dst_index + 1) = \
    *(pDst + dst_index + 2) = S9xPalette [READ_PIXEL_16(pSrc + src_index)];

#define PUT_PIXEL_W14_16to8()\
  *(pDst + dst_index) =							\
    *(pDst + dst_index + 1) = \
    *(pDst + dst_index + 2) = \
    *(pDst + dst_index + 3) = S9xPalette [READ_PIXEL_16(pSrc + src_index)];

#define PUT_PIXEL_W23_16to8()\
  *(pDst + dst_index) =	S9xPalette [READ_PIXEL_16(pSrc + src_index)];\
  *(pDst + dst_index + 1) = S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + src_index + 2))]; \
  *(pDst + dst_index + 2) = S9xPalette [READ_PIXEL_16(pSrc + src_index + 2)];

#define PUT_PIXEL_W11H23_16to8()\
  *(pDst + dst_index) = S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index))];

#define PUT_PIXEL_W12H23_16to8()\
  *(pDst + dst_index) = *(pDst + dst_index + 1) = \
    S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index))];

#define PUT_PIXEL_W13H23_16to8()\
  *(pDst + dst_index) = *(pDst + dst_index + 1) = *(pDst + dst_index + 2) =\
    S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index))];

#define PUT_PIXEL_W23H23_16to8()\
  *(pDst + dst_index) =	S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index))];\
  *(pDst + dst_index + 1) = S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index + 2))]; \
  *(pDst + dst_index + 2) = S9xPalette [Interp05(READ_PIXEL_16(pSrc + src_index + 2), READ_PIXEL_16(pSrc + Src->Pitch + src_index + 2))];

#define PUT_PIXEL_W11_16to16()\
  *(uint16 *)(pDst + dst_index) = *(uint16 *)(pSrc + src_index);

#define PUT_PIXEL_W12_16to16()\
  PUT_PIXEL_W11_16to16();						\
  *(uint16 *)(pDst + dst_index + 2) = *(uint16 *)(pSrc + src_index);

#define PUT_PIXEL_W13_16to16()\
  PUT_PIXEL_W11_16to16();						\
  *(uint16 *)(pDst + dst_index + 2) = \
    *(uint16 *)(pDst + dst_index + 4) = *(uint16 *)(pSrc + src_index);

#define PUT_PIXEL_W14_16to16()\
  PUT_PIXEL_W11_16to16();						\
  *(uint16 *)(pDst + dst_index + 2) = \
    *(uint16 *)(pDst + dst_index + 4) = \
    *(uint16 *)(pDst + dst_index + 6) = *(uint16 *)(pSrc + src_index);

#define PUT_PIXEL_W23_16to16()\
  *(uint16 *)(pDst + dst_index) = *(uint16 *)(pSrc + src_index);\
  *(uint16 *)(pDst + dst_index + 2) = Interp05(*(uint16 *)(pSrc + src_index), *(uint16 *)(pSrc + src_index + 2));\
  *(uint16 *)(pDst + dst_index + 4) = *(uint16 *)(pSrc + src_index + 2);

#define PUT_PIXEL_W11H23_16to16()\
  *(uint16 *)(pDst + dst_index) = Interp05(*(uint16 *)(pSrc + src_index), *(uint16 *)(pSrc + Src->Pitch + src_index));

#define PUT_PIXEL_W12H23_16to16()\
  *(uint16 *)(pDst + dst_index) = *(uint16 *)(pDst + dst_index + 2) = \
    Interp05(*(uint16 *)(pSrc + src_index), *(uint16 *)(pSrc + Src->Pitch + src_index));

#define PUT_PIXEL_W13H23_16to16()\
  *(uint16 *)(pDst + dst_index) = *(uint16 *)(pDst + dst_index + 2) = *(uint16 *)(pDst + dst_index + 4) = \
    Interp05(*(uint16 *)(pSrc + src_index), *(uint16 *)(pSrc + Src->Pitch + src_index));

#define PUT_PIXEL_W23H23_16to16()\
  *(uint16 *)(pDst + dst_index) = Interp05(*(uint16 *)(pSrc + src_index), *(uint16 *)(pSrc + Src->Pitch + src_index));\
  *(uint16 *)(pDst + dst_index + 2) = Interp05(*(uint16 *)(pSrc + src_index), *(uint16 *)(pSrc + Src->Pitch + src_index + 2)); \
  *(uint16 *)(pDst + dst_index + 4) = Interp05(*(uint16 *)(pSrc + src_index + 2), *(uint16 *)(pSrc + Src->Pitch + src_index + 2));


#define PUT_PIXEL_W11_16to24()\
  int color_index = READ_PIXEL_16(pSrc + src_index) * 3;	\
  *(pDst + dst_index)     = color_table[color_index];		\
  *(pDst + dst_index + 1) = color_table[color_index + 1];	\
  *(pDst + dst_index + 2) = color_table[color_index + 2];

#define PUT_PIXEL_W12_16to24()\
  int color_index = READ_PIXEL_16(pSrc + src_index) * 3;		\
  *(pDst + dst_index)     = color_table[color_index];			\
  *(pDst + dst_index + 1) = color_table[color_index + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index];			\
  *(pDst + dst_index + 4) = color_table[color_index + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index + 2];

#define PUT_PIXEL_W13_16to24()\
  int color_index = READ_PIXEL_16(pSrc + src_index) * 3;		\
  *(pDst + dst_index)     = color_table[color_index];			\
  *(pDst + dst_index + 1) = color_table[color_index + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index];			\
  *(pDst + dst_index + 4) = color_table[color_index + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index + 2];		\
  *(pDst + dst_index + 6) = color_table[color_index];			\
  *(pDst + dst_index + 7) = color_table[color_index + 1];		\
  *(pDst + dst_index + 8) = color_table[color_index + 2];

#define PUT_PIXEL_W14_16to24()\
  int color_index = READ_PIXEL_16(pSrc + src_index) * 3;		\
  *(pDst + dst_index)     = color_table[color_index];			\
  *(pDst + dst_index + 1) = color_table[color_index + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index];			\
  *(pDst + dst_index + 4) = color_table[color_index + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index + 2];		\
  *(pDst + dst_index + 6) = color_table[color_index];			\
  *(pDst + dst_index + 7) = color_table[color_index + 1];		\
  *(pDst + dst_index + 8) = color_table[color_index + 2];		\
  *(pDst + dst_index + 9) = color_table[color_index];			\
  *(pDst + dst_index + 10) = color_table[color_index + 1];		\
  *(pDst + dst_index + 11) = color_table[color_index + 2];

#define PUT_PIXEL_W23_16to24()\
  int color_index0 = READ_PIXEL_16(pSrc + src_index) * 3;		\
  int color_index1 = Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + src_index + 2)) * 3; \
  int color_index2 = READ_PIXEL_16(pSrc + src_index + 2) * 3;		\
  *(pDst + dst_index)     = color_table[color_index0];			\
  *(pDst + dst_index + 1) = color_table[color_index0 + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index0 + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index1];			\
  *(pDst + dst_index + 4) = color_table[color_index1 + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index1 + 2];		\
  *(pDst + dst_index + 6) = color_table[color_index2];			\
  *(pDst + dst_index + 7) = color_table[color_index2 + 1];		\
  *(pDst + dst_index + 8) = color_table[color_index2 + 2];

#define PUT_PIXEL_W11H23_16to24()\
  int color_index = Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 3; \
  *(pDst + dst_index)     = color_table[color_index];			\
  *(pDst + dst_index + 1) = color_table[color_index + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index + 2];

#define PUT_PIXEL_W12H23_16to24()\
  int color_index = Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 3; \
  *(pDst + dst_index)     = color_table[color_index];			\
  *(pDst + dst_index + 1) = color_table[color_index + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index];			\
  *(pDst + dst_index + 4) = color_table[color_index + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index + 2];

#define PUT_PIXEL_W13H23_16to24()\
  int color_index = Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 3; \
  *(pDst + dst_index)     = color_table[color_index];			\
  *(pDst + dst_index + 1) = color_table[color_index + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index];			\
  *(pDst + dst_index + 4) = color_table[color_index + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index + 2];		\
  *(pDst + dst_index + 6) = color_table[color_index];			\
  *(pDst + dst_index + 7) = color_table[color_index + 1];		\
  *(pDst + dst_index + 8) = color_table[color_index + 2];

#define PUT_PIXEL_W23H23_16to24()\
  int color_index0 = Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 3; \
  int color_index1 = Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index + 2)) * 3; \
  int color_index2 = Interp05(READ_PIXEL_16(pSrc + src_index + 2), READ_PIXEL_16(pSrc + Src->Pitch + src_index + 2)) * 3; \
  *(pDst + dst_index)     = color_table[color_index0];			\
  *(pDst + dst_index + 1) = color_table[color_index0 + 1];		\
  *(pDst + dst_index + 2) = color_table[color_index0 + 2];		\
  *(pDst + dst_index + 3) = color_table[color_index1];			\
  *(pDst + dst_index + 4) = color_table[color_index1 + 1];		\
  *(pDst + dst_index + 5) = color_table[color_index1 + 2];		\
  *(pDst + dst_index + 6) = color_table[color_index2];			\
  *(pDst + dst_index + 7) = color_table[color_index2 + 1];		\
  *(pDst + dst_index + 8) = color_table[color_index2 + 2];

#define PUT_PIXEL_W11_16to32()\
  *(uint32 *)(pDst + dst_index) = *(uint32 *)(&color_table[READ_PIXEL_16(pSrc + src_index) * 4]);

#define PUT_PIXEL_W12_16to32()\
  *(uint32 *)(pDst + dst_index) =					\
    *(uint32 *)(pDst + dst_index + 4) =					\
    *(uint32 *)(&color_table[READ_PIXEL_16(pSrc + src_index) * 4]);

#define PUT_PIXEL_W13_16to32()\
  *(uint32 *)(pDst + dst_index) =					\
    *(uint32 *)(pDst + dst_index + 4) =					\
    *(uint32 *)(pDst + dst_index + 8) = *(uint32 *)(&color_table[READ_PIXEL_16(pSrc + src_index) * 4]);

#define PUT_PIXEL_W14_16to32()\
  *(uint32 *)(pDst + dst_index) =					\
    *(uint32 *)(pDst + dst_index + 4) =					\
    *(uint32 *)(pDst + dst_index + 8) =					\
    *(uint32 *)(pDst + dst_index + 12) = *(uint32 *)(&color_table[READ_PIXEL_16(pSrc + src_index) * 4]);

#define PUT_PIXEL_W23_16to32()\
  *(uint32 *)(pDst + dst_index) = *(uint32 *)(&color_table[READ_PIXEL_16(pSrc + src_index) * 4]);\
  *(uint32 *)(pDst + dst_index + 4) = *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + src_index + 2)) * 4]);\
  *(uint32 *)(pDst + dst_index + 8) = *(uint32 *)(&color_table[READ_PIXEL_16(pSrc + src_index + 2) * 4]);

#define PUT_PIXEL_W11H23_16to32()\
  *(uint32 *)(pDst + dst_index) = *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 4]);

#define PUT_PIXEL_W12H23_16to32()\
  *(uint32 *)(pDst + dst_index) = *(uint32 *)(pDst + dst_index + 4) = \
    *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 4]);

#define PUT_PIXEL_W13H23_16to32()\
  *(uint32 *)(pDst + dst_index) = *(uint32 *)(pDst + dst_index + 4) = *(uint32 *)(pDst + dst_index + 8) = \
    *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 4]);

#define PUT_PIXEL_W23H23_16to32()\
  *(uint32 *)(pDst + dst_index) = \
    *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index)) * 4]); \
  *(uint32 *)(pDst + dst_index + 4) = \
    *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index), READ_PIXEL_16(pSrc + Src->Pitch + src_index + 2)) * 4]);	\
  *(uint32 *)(pDst + dst_index + 8) = \
    *(uint32 *)(&color_table[Interp05(READ_PIXEL_16(pSrc + src_index + 2), READ_PIXEL_16(pSrc + Src->Pitch + src_index + 2)) * 4]);


/****************************
 Render line helper macros */

#define RENDER_LINE_H11(sw, dw, depth)					\
  int src_index = 0, dst_index = 0;					\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pSrc += Src->Pitch;							\
  pDst += Dst->Pitch;

#define RENDER_LINE_H12(sw, dw, depth)					\
  int src_index = 0, dst_index = 0;					\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pSrc += Src->Pitch;							\
  pDst += Dst->Pitch;

#define RENDER_LINE_H13(sw, dw, depth)					\
  int src_index = 0, dst_index = 0;					\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pSrc += Src->Pitch;							\
  pDst += Dst->Pitch;

#define RENDER_LINE_H14(sw, dw, depth)					\
  int src_index = 0, dst_index = 0;					\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pSrc += Src->Pitch;							\
  pDst += Dst->Pitch;

#define RENDER_LINE_H23(sw, dw, depth)					\
  int src_index = 0, dst_index = 0;					\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##depth();					\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##H23##depth();				\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pSrc += Src->Pitch;							\
  pDst += Dst->Pitch;							\
  src_index = dst_index = 0;						\
  for(width = (srcRect->right - srcRect->left); width; width -= (sw))	\
    {									\
      PUT_PIXEL_W##sw##dw##H23##depth();				\
      src_index += SRC_PIXSIZE##depth * (sw);				\
      dst_index += DST_PIXSIZE##depth * (dw);				\
    }									\
  pSrc += Src->Pitch;							\
  pDst += Dst->Pitch;

/*********************
 Render line macros */

#define COPY_LINE_W11H11(depth)\
  memcpy(pDst, pSrc, (size_t)width * DST_PIXSIZE##depth);	\
  pSrc += Src->Pitch;						\
  pDst += Dst->Pitch;

#define COPY_LINE_W11H12(depth)\
  memcpy(pDst, pSrc, (size_t)width * DST_PIXSIZE##depth);	\
  pDst += Dst->Pitch;						\
  COPY_LINE_W11H11(depth)

#define RENDER_LINE_W11H11(depth)\
  RENDER_LINE_H11(1, 1, depth);

#define RENDER_LINE_W12H11(depth)\
  RENDER_LINE_H11(1, 2, depth);

#define RENDER_LINE_W11H12(depth)\
  RENDER_LINE_H12(1, 1, depth);

#define RENDER_LINE_W12H12(depth)\
  RENDER_LINE_H12(1, 2, depth);

#define RENDER_LINE_W23H13(depth)\
  RENDER_LINE_H13(2, 3, depth);

#define RENDER_LINE_W13H23(depth)\
  RENDER_LINE_H23(1, 3, depth);

#define RENDER_LINE_W23H23(depth)\
  RENDER_LINE_H23(2, 3, depth);

#define RENDER_LINE_W12H14(depth)\
  RENDER_LINE_H14(1, 2, depth);

#define RENDER_LINE_W14H12(depth)\
  RENDER_LINE_H12(1, 4, depth);

#define RENDER_LINE_W13H13(depth)\
  RENDER_LINE_H13(1, 3, depth);

#define RENDER_LINE_W14H14(depth)\
  RENDER_LINE_H14(1, 4, depth);

#define COPY_SCANLINE_W11H12(depth)\
  COPY_LINE_W11H11(depth);						\
  memset(pDst, 0, (size_t)width * DST_PIXSIZE##depth);			\
  pDst += Dst->Pitch;

#define RENDER_SCANLINE_W11H12(depth)\
  RENDER_LINE_H11(1, 1, depth);						\
  width = (srcRect->right - srcRect->left);				\
  memset(pDst, 0, (size_t)width * DST_PIXSIZE##depth);			\
  pDst += Dst->Pitch;

#define RENDER_SCANLINE_W12H12(depth)\
  RENDER_LINE_H11(1, 2, depth);						\
  width = (srcRect->right - srcRect->left);				\
  memset(pDst, 0, (size_t)width * DST_PIXSIZE##depth * 2);		\
  pDst += Dst->Pitch;

#define CLEAR_DEST(depth)\
  uint8 *dp = pDst;							\
  int dw = (dstRect->right - dstRect->left);				\
  for(int dh = (dstRect->bottom - dstRect->top); dh; dh--)		\
    {									\
      memset(dp, 0, (size_t)dw * DST_PIXSIZE##depth);			\
      dp += Dst->Pitch;							\
    }

/***********************
 Render method macros */

#define RENDER_METHOD(name, w, h, sh, depth, render_line)		\
static void Render##name##W##w##H##h##depth(SSurface *Src, SSurface *Dst, RECT *srcRect, RECT *dstRect) \
{									\
  uint8 *pSrc = (Src->Surface + srcRect->top * Src->Pitch + srcRect->left * SRC_PIXSIZE##depth); \
  uint8 *pDst = (Dst->Surface + dstRect->top * Dst->Pitch + dstRect->left * DST_PIXSIZE##depth); \
  int width = (srcRect->right - srcRect->left);				\
  int height = (srcRect->bottom - srcRect->top);			\
  for(; height; height -= (sh))						\
    {									\
      render_line(depth);						\
    }									\
  ExtendHeight_previos = ExtendHeight;					\
}

// simply copy line render
RENDER_METHOD (, 11, 11, 1, _16to16, COPY_LINE_W11H11)
// simply copy H2 line render
RENDER_METHOD (, 11, 12, 1, _16to16, COPY_LINE_W11H12)
// simply W2 line render
RENDER_METHOD (, 12, 11, 1, _16to16, RENDER_LINE_W12H11)
// simply W2H2 line render
RENDER_METHOD (, 12, 12, 1, _16to16, RENDER_LINE_W12H12)
// simply copy scanline render
RENDER_METHOD (Scanline, 11, 12, 1, _16to16, COPY_SCANLINE_W11H12)
// simply W2H2 scanline render
RENDER_METHOD (Scanline, 12, 12, 1, _16to16, RENDER_SCANLINE_W12H12)
// depth convert render
RENDER_METHOD (, 11, 11, 1, _16to8, RENDER_LINE_W11H11)
RENDER_METHOD (, 11, 11, 1, _16to24, RENDER_LINE_W11H11)
RENDER_METHOD (, 11, 11, 1, _16to32, RENDER_LINE_W11H11)
// depth convert W2 render 
RENDER_METHOD (, 12, 11, 1, _16to8, RENDER_LINE_W12H11)
RENDER_METHOD (, 12, 11, 1, _16to24, RENDER_LINE_W12H11)
RENDER_METHOD (, 12, 11, 1, _16to32, RENDER_LINE_W12H11)
// depth convert H2 render 
RENDER_METHOD (, 11, 12, 1, _16to8, RENDER_LINE_W11H12)
RENDER_METHOD (, 11, 12, 1, _16to24, RENDER_LINE_W11H12)
RENDER_METHOD (, 11, 12, 1, _16to32, RENDER_LINE_W11H12)
// depth convert W2H2 render 
RENDER_METHOD (, 12, 12, 1, _16to8, RENDER_LINE_W12H12)
RENDER_METHOD (, 12, 12, 1, _16to24, RENDER_LINE_W12H12)
RENDER_METHOD (, 12, 12, 1, _16to32, RENDER_LINE_W12H12)
// depth convert scanline render
RENDER_METHOD (Scanline, 11, 12, 1, _16to8, RENDER_SCANLINE_W11H12)
RENDER_METHOD (Scanline, 11, 12, 1, _16to24, RENDER_SCANLINE_W11H12)
RENDER_METHOD (Scanline, 11, 12, 1, _16to32, RENDER_SCANLINE_W11H12)
// depth convert W2 scanline render
RENDER_METHOD (Scanline, 12, 12, 1, _16to8, RENDER_SCANLINE_W12H12)
RENDER_METHOD (Scanline, 12, 12, 1, _16to24, RENDER_SCANLINE_W12H12)
RENDER_METHOD (Scanline, 12, 12, 1, _16to32, RENDER_SCANLINE_W12H12)

RENDER_METHOD (, 23, 23, 2, _16to8, RENDER_LINE_W23H23)
RENDER_METHOD (, 13, 13, 1, _16to8, RENDER_LINE_W13H13)
RENDER_METHOD (, 14, 14, 1, _16to8, RENDER_LINE_W14H14)
RENDER_METHOD (, 23, 13, 1, _16to8, RENDER_LINE_W23H13)
RENDER_METHOD (, 13, 23, 2, _16to8, RENDER_LINE_W13H23)
RENDER_METHOD (, 14, 12, 1, _16to8, RENDER_LINE_W14H12)
RENDER_METHOD (, 12, 14, 1, _16to8, RENDER_LINE_W12H14)

RENDER_METHOD (, 23, 23, 2, _16to16, RENDER_LINE_W23H23)
RENDER_METHOD (, 13, 13, 1, _16to16, RENDER_LINE_W13H13)
RENDER_METHOD (, 14, 14, 1, _16to16, RENDER_LINE_W14H14)
RENDER_METHOD (, 23, 13, 1, _16to16, RENDER_LINE_W23H13)
RENDER_METHOD (, 13, 23, 2, _16to16, RENDER_LINE_W13H23)
RENDER_METHOD (, 14, 12, 1, _16to16, RENDER_LINE_W14H12)
RENDER_METHOD (, 12, 14, 1, _16to16, RENDER_LINE_W12H14)

RENDER_METHOD (, 23, 23, 2, _16to24, RENDER_LINE_W23H23)
RENDER_METHOD (, 13, 13, 1, _16to24, RENDER_LINE_W13H13)
RENDER_METHOD (, 14, 14, 1, _16to24, RENDER_LINE_W14H14)
RENDER_METHOD (, 23, 13, 1, _16to24, RENDER_LINE_W23H13)
RENDER_METHOD (, 13, 23, 2, _16to24, RENDER_LINE_W13H23)
RENDER_METHOD (, 14, 12, 1, _16to24, RENDER_LINE_W14H12)
RENDER_METHOD (, 12, 14, 1, _16to24, RENDER_LINE_W12H14)

RENDER_METHOD (, 23, 23, 2, _16to32, RENDER_LINE_W23H23)
RENDER_METHOD (, 13, 13, 1, _16to32, RENDER_LINE_W13H13)
RENDER_METHOD (, 14, 14, 1, _16to32, RENDER_LINE_W14H14)
RENDER_METHOD (, 23, 13, 1, _16to32, RENDER_LINE_W23H13)
RENDER_METHOD (, 13, 23, 2, _16to32, RENDER_LINE_W13H23)
RENDER_METHOD (, 14, 12, 1, _16to32, RENDER_LINE_W14H12)
RENDER_METHOD (, 12, 14, 1, _16to32, RENDER_LINE_W12H14)

/***********************
 filter defines       */
void TVMode (uint8 * srcPtr, uint32 srcPitch,
	     uint8 * dstPtr, uint32 dstPitch, int width, int height);

void hq2x (uint8 * srcPtr, uint32 srcPitch,
	   uint8 * dstPtr, uint32 dstPitch, int width, int height);

void lq2x (uint8 * srcPtr, uint32 srcPitch,
	   uint8 * dstPtr, uint32 dstPitch, int width, int height);


/***********************
 filter method macros */

#define RENDER_2xSaI(depth)\
  S9xBlitPix2xSaI16 (pSrc, Src->Pitch,		\
		     pDst, Dst->Pitch, width, height);

#define RENDER_Super2xSaI(depth)\
  S9xBlitPixSuper2xSaI16 (pSrc, Src->Pitch,	\
			  pDst, Dst->Pitch, width, height);

#define RENDER_SuperEagle(depth)\
  S9xBlitPixSuperEagle16 (pSrc, Src->Pitch,			\
			  pDst, Dst->Pitch, width, height);

#define RENDER_TVMode(depth)\
  TVMode (pSrc, Src->Pitch,			\
	  pDst, Dst->Pitch, width, height);

#define RENDER_hq2x(depth)\
  S9xBlitPixHQ2x16 (pSrc, Src->Pitch,			\
		    pDst, Dst->Pitch, width, height);

#define RENDER_hq3x(depth)\
  S9xBlitPixHQ3x16 (pSrc, Src->Pitch,			\
		    pDst, Dst->Pitch, width, height);

#define RENDER_hq4x(depth)\
  S9xBlitPixHQ4x16 (pSrc, Src->Pitch,			\
		    pDst, Dst->Pitch, width, height);

#define RENDER_HiRes_hq2x(depth)\
  hq2x (pSrc, Src->Pitch,			\
	pDst, Dst->Pitch, width, height);

#define RENDER_lq2x(depth)\
  lq2x (pSrc, Src->Pitch,			\
	pDst, Dst->Pitch, width, height);

#define RENDER_EPX(depth)\
  EPX_16 (pSrc, Src->Pitch,			\
	  pDst, Dst->Pitch, width, height);

#define RENDER_NTSC(depth)\
  S9xBlitPixNTSC16 (pSrc, Src->Pitch,		\
		    pDst, Dst->Pitch, width, height);

#define RENDER_HiRes_NTSC(depth)\
  S9xBlitPixHiResNTSC16 (pSrc, Src->Pitch,		\
			 pDst, Dst->Pitch, width, height);

#define RENDER_FILTER(depth, filter, convert_depth)\
  if(convert_depth)							\
    {									\
      uint8 *pdst = pDst;						\
      SSurface *dst = Dst;						\
      SSurface ddst;							\
      pDst = ddst.Surface = depthconv_buf;				\
      ddst.Width = dstRect->right - dstRect->left;			\
      ddst.Height = dstRect->bottom - dstRect->top;			\
      ddst.Pitch = buffer_pitch;					\
      Dst = &ddst;							\
      RECT *dstrect = dstRect;						\
      RECT ddstrect;							\
      ddstrect.top = 0;							\
      ddstrect.left = 0;						\
      ddstrect.right = ddst.Width;					\
      ddstrect.bottom = ddst.Height;					\
      dstRect = &ddstrect;						\
      filter(depth);							\
      Src = Dst;							\
      srcRect = dstRect;						\
      pSrc = pDst;							\
      Dst = dst;							\
      dstRect = dstrect;						\
      pDst = pdst;							\
      for(int height = (srcRect->bottom - srcRect->top); height; height--)\
	{								\
	  RENDER_LINE_W11H11(depth);					\
	}								\
    }									\
  else {								\
    filter(depth);							\
  }

#define FILTER_METHOD(name, w, h, depth, filter, convert_depth)\
static void Render##name##W##w##H##h##depth(SSurface *Src, SSurface *Dst, RECT *srcRect, RECT *dstRect) \
{									\
  uint8 *pSrc = (Src->Surface + srcRect->top * Src->Pitch + srcRect->left * SRC_PIXSIZE##depth); \
  uint8 *pDst = (Dst->Surface + dstRect->top * Dst->Pitch + dstRect->left * DST_PIXSIZE##depth); \
  int width = (srcRect->right - srcRect->left);				\
  int height = (srcRect->bottom - srcRect->top);			\
  RENDER_FILTER(depth, filter, convert_depth);				\
  ExtendHeight_previos = ExtendHeight;					\
}

// 2xSaI
  FILTER_METHOD (_2xSaI, 12, 12, _16to8, RENDER_2xSaI, TRUE)
FILTER_METHOD (_2xSaI, 12, 12, _16to16, RENDER_2xSaI, FALSE)
FILTER_METHOD (_2xSaI, 12, 12, _16to24, RENDER_2xSaI, TRUE)
FILTER_METHOD (_2xSaI, 12, 12, _16to32, RENDER_2xSaI, TRUE)
// Super2xSaI
  FILTER_METHOD (Super2xSaI, 12, 12, _16to8, RENDER_Super2xSaI, TRUE)
FILTER_METHOD (Super2xSaI, 12, 12, _16to16, RENDER_Super2xSaI, FALSE)
FILTER_METHOD (Super2xSaI, 12, 12, _16to24, RENDER_Super2xSaI, TRUE)
FILTER_METHOD (Super2xSaI, 12, 12, _16to32, RENDER_Super2xSaI, TRUE)
// SuperEagle
FILTER_METHOD (SuperEagle, 12, 12, _16to8, RENDER_SuperEagle, TRUE)
FILTER_METHOD (SuperEagle, 12, 12, _16to16, RENDER_SuperEagle, FALSE)
FILTER_METHOD (SuperEagle, 12, 12, _16to24, RENDER_SuperEagle, TRUE)
FILTER_METHOD (SuperEagle, 12, 12, _16to32, RENDER_SuperEagle, TRUE)
// TV
  FILTER_METHOD (TV, 12, 12, _16to8, RENDER_TVMode, TRUE)
FILTER_METHOD (TV, 12, 12, _16to16, RENDER_TVMode, FALSE)
FILTER_METHOD (TV, 12, 12, _16to24, RENDER_TVMode, TRUE)
FILTER_METHOD (TV, 12, 12, _16to32, RENDER_TVMode, TRUE)
FILTER_METHOD (TV, 11, 12, _16to8, RENDER_TVMode, TRUE)
FILTER_METHOD (TV, 11, 12, _16to16, RENDER_TVMode, FALSE)
FILTER_METHOD (TV, 11, 12, _16to24, RENDER_TVMode, TRUE)
FILTER_METHOD (TV, 11, 12, _16to32, RENDER_TVMode, TRUE)

// HQ2X
FILTER_METHOD (HQ2X, 12, 12, _16to8, RENDER_hq2x, TRUE)
FILTER_METHOD (HQ2X, 12, 12, _16to16, RENDER_hq2x, FALSE)
FILTER_METHOD (HQ2X, 12, 12, _16to24, RENDER_hq2x, TRUE)
FILTER_METHOD (HQ2X, 12, 12, _16to32, RENDER_hq2x, TRUE)
FILTER_METHOD (HiRes_HQ2X, 11, 12, _16to8, RENDER_HiRes_hq2x, TRUE)
FILTER_METHOD (HiRes_HQ2X, 11, 12, _16to16, RENDER_HiRes_hq2x, FALSE)
FILTER_METHOD (HiRes_HQ2X, 11, 12, _16to24, RENDER_HiRes_hq2x, TRUE)
FILTER_METHOD (HiRes_HQ2X, 11, 12, _16to32, RENDER_HiRes_hq2x, TRUE)
FILTER_METHOD (HQ3X, 13, 13, _16to8, RENDER_hq3x, TRUE)
FILTER_METHOD (HQ3X, 13, 13, _16to16, RENDER_hq3x, FALSE)
FILTER_METHOD (HQ3X, 13, 13, _16to24, RENDER_hq3x, TRUE)
FILTER_METHOD (HQ3X, 13, 13, _16to32, RENDER_hq3x, TRUE)
FILTER_METHOD (HQ4X, 14, 14, _16to8, RENDER_hq4x, TRUE)
FILTER_METHOD (HQ4X, 14, 14, _16to16, RENDER_hq4x, FALSE)
FILTER_METHOD (HQ4X, 14, 14, _16to24, RENDER_hq4x, TRUE)
FILTER_METHOD (HQ4X, 14, 14, _16to32, RENDER_hq4x, TRUE)

// LQ2X
  FILTER_METHOD (LQ2X, 12, 12, _16to8, RENDER_lq2x, TRUE)
FILTER_METHOD (LQ2X, 12, 12, _16to16, RENDER_lq2x, FALSE)
FILTER_METHOD (LQ2X, 12, 12, _16to24, RENDER_lq2x, TRUE)
FILTER_METHOD (LQ2X, 12, 12, _16to32, RENDER_lq2x, TRUE)
FILTER_METHOD (LQ2X, 11, 12, _16to8, RENDER_lq2x, TRUE)
FILTER_METHOD (LQ2X, 11, 12, _16to16, RENDER_lq2x, FALSE)
FILTER_METHOD (LQ2X, 11, 12, _16to24, RENDER_lq2x, TRUE)
FILTER_METHOD (LQ2X, 11, 12, _16to32, RENDER_lq2x, TRUE)

// EPX
FILTER_METHOD (EPX, 12, 12, _16to8, RENDER_EPX, TRUE)
FILTER_METHOD (EPX, 12, 12, _16to16, RENDER_EPX, FALSE)
FILTER_METHOD (EPX, 12, 12, _16to24, RENDER_EPX, TRUE)
FILTER_METHOD (EPX, 12, 12, _16to32, RENDER_EPX, TRUE)

//NTSC
FILTER_METHOD (NTSC, 12, 11, _16to8, RENDER_NTSC, TRUE)
FILTER_METHOD (NTSC, 12, 11, _16to16, RENDER_NTSC, FALSE)
FILTER_METHOD (NTSC, 12, 11, _16to24, RENDER_NTSC, TRUE)
FILTER_METHOD (NTSC, 12, 11, _16to32, RENDER_NTSC, TRUE)
FILTER_METHOD (HiRes_NTSC, 11, 11, _16to8, RENDER_HiRes_NTSC, TRUE)
FILTER_METHOD (HiRes_NTSC, 11, 11, _16to16, RENDER_HiRes_NTSC, FALSE)
FILTER_METHOD (HiRes_NTSC, 11, 11, _16to24, RENDER_HiRes_NTSC, TRUE)
FILTER_METHOD (HiRes_NTSC, 11, 11, _16to32, RENDER_HiRes_NTSC, TRUE)

static inline void
CreatePalette16to24 ()
{
  uint32 pixel;
  if (RedShift < BlueShift)
    {
      // Order is RGB
      for (pixel = 0; pixel < 0x10000; pixel++)
	{
	  color_table[pixel * 3] = (pixel >> (11 - 3)) & 0xf8;
	  color_table[pixel * 3 + 1] = (pixel >> (6 - 3)) & 0xf8;
	  color_table[pixel * 3 + 2] = (pixel & 0x1f) << 3;
	}
    }
  else
    {
      // Order is BGR
      for (pixel = 0; pixel < 0x10000; pixel++)
	{
	  color_table[pixel * 3] = (pixel & 0x1f) << 3;
	  color_table[pixel * 3 + 1] = (pixel >> (6 - 3)) & 0xf8;
	  color_table[pixel * 3 + 2] = (pixel >> (11 - 3)) & 0xf8;
	}
    }
}

static inline void
CreatePalette16to32 ()
{
  uint32 pixel;
  if (RedShift < BlueShift)
    {
      // Order is RGB
      for (pixel = 0; pixel < 0x10000; pixel++)
	{
	  color_table[pixel * 4] = (pixel >> (11 - 3)) & 0xf8;
	  color_table[pixel * 4 + 1] = (pixel >> (6 - 3)) & 0xf8;
	  color_table[pixel * 4 + 2] = (pixel & 0x1f) << 3;
	  color_table[pixel * 4 + 3] = 0;
	}
    }
  else
    {
      // Order is BGR
      for (pixel = 0; pixel < 0x10000; pixel++)
	{
	  color_table[pixel * 4] = (pixel & 0x1f) << 3;
	  color_table[pixel * 4 + 1] = (pixel >> (6 - 3)) & 0xf8;
	  color_table[pixel * 4 + 2] = (pixel >> (11 - 3)) & 0xf8;
	  color_table[pixel * 4 + 3] = 0;
	}
    }
}

#define RENDER_METHOD_CASE0(depth)\
  case 0:								\
  RenderMethod = RenderW11H11##depth;					\
  imageWidth = srcWidth;						\
  imageHeight = srcHeight;						\
  break;

#define RENDER_METHOD_CASE1(depth)\
  case 1:								\
  if(dox >= 5)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW12H14##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW14H12##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H12##depth;				\
      else								\
	RenderMethod = RenderW14H14##depth;				\
      imageWidth = 1024;						\
      imageHeight = (!ExtendHeight ? 896 : 956);			\
    }									\
  else if(dox >= 4)							\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW23H13##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW13H23##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW23H23##depth;				\
      else								\
	RenderMethod = RenderW13H13##depth;				\
      imageWidth = 768;							\
      imageHeight = (!ExtendHeight ? 672 : 717);			\
    }									\
  else if(dox >= 2)							\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderW12H12##depth;				\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE2(depth)\
  case 2:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderScanlineW11H12##depth;			\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderScanlineW12H12##depth;			\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE3(depth)\
  case 3:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = Render_2xSaIW12H12##depth;			\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE4(depth)\
  case 4:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderSuper2xSaIW12H12##depth;			\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE5(depth)\
  case 5:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderSuperEagleW12H12##depth;			\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE6(depth)\
  case 6:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderTVW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderTVW12H12##depth;				\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE7(depth)\
  case 7:								\
  if(dox >= 5)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW12H14##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW14H12##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderHQ2XW12H12##depth;				\
      else								\
	RenderMethod = RenderHQ4XW14H14##depth;				\
      imageWidth = 1024;						\
      imageHeight = (!ExtendHeight ? 896 : 956);			\
    }									\
  else if(dox >= 4)							\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW23H13##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW13H23##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW23H23##depth;				\
      else								\
	RenderMethod = RenderHQ3XW13H13##depth;				\
      imageWidth = 768;							\
      imageHeight = (!ExtendHeight ? 672 : 717);			\
    }									\
  else if(dox >= 2)							\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderHiRes_HQ2XW11H12##depth;			\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderHQ2XW12H12##depth;				\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE8(depth)\
  case 8:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderLQ2XW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderLQ2XW12H12##depth;				\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE9(depth)\
  case 9:								\
  if(dox >= 2)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderW11H12##depth;				\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderW11H11##depth;				\
      else								\
	RenderMethod = RenderEPXW12H12##depth;				\
      imageWidth = 512;							\
      imageHeight = (!ExtendHeight ? 448 : 478);			\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE10_12(depth)\
  case 10:								\
  case 11:								\
  case 12:								\
  if(dox >= 3)								\
    {									\
      if(DoubleWidth && !DoubleHeight)					\
	RenderMethod = RenderHiRes_NTSCW11H11##depth;			\
      else if(!DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderNTSCW12H11##depth;				\
      else if(DoubleWidth && DoubleHeight)				\
	RenderMethod = RenderHiRes_NTSCW11H11##depth;			\
      else								\
	RenderMethod = RenderNTSCW12H11##depth;				\
      imageWidth = 602;							\
      imageHeight = srcHeight;						\
    }									\
  else									\
    {									\
      RenderMethod = RenderW11H11##depth;				\
      imageWidth = srcWidth;						\
      imageHeight = srcHeight;						\
    }									\
  break;

#define RENDER_METHOD_CASE(depth)\
  RENDER_METHOD_CASE0(depth);						\
  RENDER_METHOD_CASE1(depth);						\
  RENDER_METHOD_CASE2(depth);						\
  RENDER_METHOD_CASE3(depth);						\
  RENDER_METHOD_CASE4(depth);						\
  RENDER_METHOD_CASE5(depth);						\
  RENDER_METHOD_CASE6(depth);						\
  RENDER_METHOD_CASE7(depth);						\
  RENDER_METHOD_CASE8(depth);						\
  RENDER_METHOD_CASE9(depth);						\
  RENDER_METHOD_CASE10_12(depth);

#define RENDER_METHOD_SWITCH(depth)\
  switch(FilterMethod)					\
    {							\
      RENDER_METHOD_CASE(depth);			\
    default:						\
      RenderMethod = RenderNothing;			\
      break;						\
    }

bool8
RenderInit (int filtermethod,
	    int screendepth,
	    int redshift,
	    int greenshift,
	    int blueshift)
{
  ScreenDepth = screendepth;
  RedShift = redshift;
  GreenShift = greenshift;
  BlueShift = blueshift;

  int scale = GetFilterScale(filtermethod);
  buffer_pitch = SNES_WIDTH * sizeof(uint16) * scale;
  buffer_size = buffer_pitch * SNES_HEIGHT_EXTENDED * scale;

  depthconv_buf = new uint8 [buffer_size];

  if(!depthconv_buf ||
     !S9xBlitFilterInit () ||
     !S9xBlit2xSaIFilterInit () ||
     !S9xBlitHQ2xFilterInit() ||
     !S9xBlitNTSCFilterInit())
    {
      RenderDeinit();
      S9xBlitFilterDeinit();
      S9xBlit2xSaIFilterDeinit();
      S9xBlitHQ2xFilterDeinit();
      S9xBlitNTSCFilterDeinit();
      return FALSE;
    }

  switch (screendepth)
    {

    case 24:
      CreatePalette16to24 ();
      break;

    case 32:
      CreatePalette16to32 ();
      break;

    default:
      break;
    }

  switch (filtermethod)
    {
    case 10:
      S9xBlitNTSCFilterSet(&snes_ntsc_composite);
      break;
    case 11:
      S9xBlitNTSCFilterSet(&snes_ntsc_svideo);
      break;
    case 12:
      S9xBlitNTSCFilterSet(&snes_ntsc_rgb);
      break;
    }

  return TRUE;
}

void RenderDeinit()
{
  if(depthconv_buf)
    {
      delete [] depthconv_buf;
      depthconv_buf = NULL;
    }

  S9xBlitFilterDeinit();
  S9xBlit2xSaIFilterDeinit();
  S9xBlitHQ2xFilterDeinit();
  S9xBlitNTSCFilterDeinit();
}

void
SelectRenderMethod (int FilterMethod, int srcWidth, int srcHeight, int dstWidth, int dstHeight, int &imageWidth, int &imageHeight)
{
  OldRenderMethod = RenderMethod;

  DoubleWidth = srcWidth == 512;
  DoubleHeight = srcHeight > 239;
  ExtendHeight = srcHeight == 239 || srcHeight == 478;

  //256x224or239 == 1, 512x448or478 == 2, 602x448or478 == 3, 768x672or717 == 4, 1024x896or956 == 5
  int dox;

  if(dstWidth >= 1024 && dstHeight >= (!ExtendHeight ? 896 : 956))
    dox = 5;
  else if(dstWidth >= 768 && dstHeight >= (!ExtendHeight ? 672 : 717))
    dox = 4;
  else if(dstWidth >= 602 && dstHeight >= (!ExtendHeight ? 448 : 478))
    dox = 3;
  else if(dstWidth >= 512 && dstHeight >= (!ExtendHeight ? 448 : 478))
    dox = 2;
  else
    dox = 1;

  switch (ScreenDepth)
    {

    case 8:
      RENDER_METHOD_SWITCH (_16to8);
      break;

    case 16:
      RENDER_METHOD_SWITCH (_16to16);
      break;

    case 24:
      RENDER_METHOD_SWITCH (_16to24);
      break;

    case 32:
      RENDER_METHOD_SWITCH (_16to32);
      break;

    default:
      RenderMethod = RenderNothing;
      break;
    }
}

void
ResetRenderMethod ()
{
  OldRenderMethod = RenderNothing;
  RenderMethod = RenderNothing;
}

int GetFilterScale(int FilterMethod)
{
  int scale = 2;

  switch(FilterMethod)
    {
    case 0:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 8:
    case 9:
      scale = 2;
      break;

    case 10: 
    case 11:
    case 12:
      scale = 3;
      break;

    case 1: 
    case 7:
      scale = 4;
      break;

    default:
      scale = 2;
      break;
    }

  return scale;
}

