/*
  Copyright (C) 2010 sanmaiwashi
*/

extern "C" {
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
}
#include "snes9x.h"
#include "memmap.h"
#include "uosnes_avcodec.h"

static bool8 init = FALSE;
static FILE *fp;
static int PixelFormat = PIX_FMT_RGB565LE;
static int frame_rate;
static AVCodecContext *c;

static int get_pix_fmt()
{
  int r;

  switch (GFX.PixelFormat)
    {
    case RGB565: r = PIX_FMT_RGB565LE;
    case RGB555: r = PIX_FMT_RGB555LE;
    case BGR565: r = PIX_FMT_BGR565LE;
    case BGR555: r = PIX_FMT_BGR555LE;
    default: r = PIX_FMT_NONE;
    }
  return r;
}

bool8 movie_open (const char *filename)
{
  if(!init)
    {
      /* must be called before using avcodec lib */
      avcodec_init();

      /* register all the codecs */
      avcodec_register_all();

      init = TRUE;
    }

  AVCodec *codec;

  if((PixelFormat = get_pix_fmt()) == PIX_FMT_NONE ||
     !(codec = avcodec_find_encoder(CODEC_ID_MPEG4)))
    return FALSE;

  frame_rate = Memory.ROMFramesPerSecond;

  c = avcodec_alloc_context();

  c->bit_rate = 400000;
  c->width = 256;
  c->height = 224;
  c->gop_size = 10;
  c->max_b_frames = 1;
  c->time_base.den = frame_rate / 2;
  c->time_base.num = 1;
  c->pix_fmt = (enum PixelFormat)PixelFormat;

  /* open it */
  if (avcodec_open(c, codec) < 0 || !(fp = fopen(filename, "wb")))
    {
      av_free(c);
      c = NULL;
      return FALSE;
    }

  return TRUE;
}

int movie_encode (uint8 *p, int width, int height, int pitch)
{
  return AVCENC_SUCCESS;
}

void movie_close ()
{
  if(!fp)
    return;

  fclose(fp);
  fp = NULL;
}

