/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu_orig.h"
#include "display.h"
#include "gfx.h"
#include "controls.h"
#include "controls2.h"
#include "soundux.h"
#include "snesapu.h"
#include "spc700.h"
#include "spc7110.h"

#include "x11.h"

#ifdef _NETPLAY_SUPPORT
#include "netplay.h"
#endif

#include <signal.h>
#include <errno.h>

#include <stdlib.h>
#include <string.h>
#ifdef HAVE_IO_H
#include <io.h>
#elif HAVE_SYS_IO_H
#include <sys/io.h>
#endif
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#ifdef HAVE_ICONV_H
#include <iconv.h>
#endif
#ifdef HAVE_LANGINFO_H
#include <langinfo.h>
#endif

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>

#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif

#include <sys/soundcard.h>
#include <sys/mman.h>

#include <locale.h>

static const char	*s9x_base_dir        = NULL,
					*rom_filename        = NULL,
					*snapshot_filename   = NULL,
					*play_smv_filename   = NULL,
					*record_smv_filename = NULL;

static char		default_dir[PATH_MAX + 1];

static const char	dirNames[13][32] =
{
	"",				// DEFAULT_DIR
	"",				// HOME_DIR
	"",				// ROMFILENAME_DIR
	"rom",			// ROM_DIR
	"sram",			// SRAM_DIR
	"savestate",	// SNAPSHOT_DIR
	"screenshot",	// SCREENSHOT_DIR
	"spc",			// SPC_DIR
	"cheat",		// CHEAT_DIR
	"patch",		// IPS_DIR
	"bios",			// BIOS_DIR
	"log",			// LOG_DIR
	""
};

uint32 joypads[8] = { 0 };

int NumControllers = 8;

#ifdef JOYSTICK_SUPPORT
#include <linux/joystick.h>
int js_fd[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };

int js_map_button[8][16] = {
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0},
  {
   SNES_B_MASK, SNES_A_MASK, SNES_A_MASK,
   SNES_Y_MASK, SNES_X_MASK, SNES_B_MASK,
   SNES_TL_MASK, SNES_TR_MASK,
   SNES_START_MASK, SNES_SELECT_MASK, 0, 0, 0, 0, 0, 0}
};

const char *js_device[8] = {
  "/dev/input/js0",
  "/dev/input/js1",
  "/dev/input/js2",
  "/dev/input/js3",
  "/dev/input/js4",
  "/dev/input/js5",
  "/dev/input/js6",
  "/dev/input/js7"
};

void InitJoysticks ();
void ReadJoysticks ();
#endif

void InitTimer ();
void *S9xProcessSound (void *);

#ifdef DEBUGGER
static void
sigbrkhandler (int)
{
  CPU.Flags |= DEBUG_MODE_FLAG;
  signal (SIGINT, (SIG_PF) sigbrkhandler);
}
#endif

void
OutOfMemory ()
{
  fprintf (stderr, "Snes9X: Memory allocation failure -"
	   " not enough RAM/virtual memory available.\n" "S9xExiting...\n");
  Memory.Deinit ();
  S9xDeinitAPU ();

  exit (1);
}

void
S9xParseArg (char **argv, int &i, int argc)
{
#ifdef JOYSTICK_SUPPORT
  if (strcmp (argv[i], "-j") == 0 || strcasecmp (argv[i], "-nojoy") == 0)
    Settings.JoystickEnabled = FALSE;
  else if (strcasecmp (argv[i], "-joydev1") == 0)
    {
      if (i + 1 < argc)
	js_device[0] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev2") == 0)
    {
      if (i + 1 < argc)
	js_device[1] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev3") == 0)
    {
      if (i + 1 < argc)
	js_device[2] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev4") == 0)
    {
      if (i + 1 < argc)
	js_device[3] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev5") == 0)
    {
      if (i + 1 < argc)
	js_device[4] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev6") == 0)
    {
      if (i + 1 < argc)
	js_device[5] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev7") == 0)
    {
      if (i + 1 < argc)
	js_device[6] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joydev8") == 0)
    {
      if (i + 1 < argc)
	js_device[7] = argv[++i];
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap1") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[0][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[0][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap2") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[1][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[1][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap3") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[2][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[2][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap4") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[3][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[3][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap5") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[4][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[4][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap6") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[5][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[5][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap7") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[6][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[6][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else if (strcasecmp (argv[i], "-joymap8") == 0)
    {
      if (i + 8 < argc)
	{
	  int t;

	  for (t = 0; t < 15; t++)
	    js_map_button[7][t] = 0;

	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_A_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_B_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_X_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_Y_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_TL_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_TR_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_START_MASK;
	  if ((t = atoi (argv[++i])) < 15)
	    js_map_button[7][t] = SNES_SELECT_MASK;
	}
      else
	S9xUsage ();
    }
  else
#endif
  if (strcasecmp (argv[i], "-b") == 0 ||
	strcasecmp (argv[i], "-bs") == 0 ||
	strcasecmp (argv[i], "-buffersize") == 0)
    {
      if (i + 1 < argc)
	Settings.SoundBufferSize = atoi (argv[++i]);
      else
	S9xUsage ();
    }
  else if (strcmp (argv[i], "-l") == 0 ||
	   strcasecmp (argv[i], "-loadsnapshot") == 0)
    {
      if (i + 1 < argc)
	snapshot_filename = argv[++i];
      else
	S9xUsage ();
    }
  else
    S9xParseDisplayArg (argv, i, argc);
}

#include "cheats.h"

int
main (int argc, char **argv)
{
  __argc = argc;
  __argv = argv;

  setlocale (LC_ALL, "");

  if (argc < 2)
    S9xUsage();

  snprintf(default_dir, PATH_MAX + 1, "%s%s%s", getenv("HOME"), SLASH_STR, ".uosnes");

  LoadDefaultSettings ();
  S9xLoadDefaultSettings ();

#ifdef JOYSTICK_SUPPORT
  Settings.JoystickEnabled = TRUE;
#else
  Settings.JoystickEnabled = FALSE;
#endif
  Settings.SoundPlaybackRate = 4;	//32000Hz
  Settings.Stereo = TRUE;
  Settings.SoundBufferSize = 0;
  Settings.DisableSoundEcho = FALSE;
  Settings.APUEnabled = Settings.NextAPUEnabled = TRUE;
  Settings.SkipFrames = AUTO_FRAMERATE;
  Settings.ShutdownMaster = TRUE;
  Settings.FrameTimePAL = 19997;
  Settings.FrameTimeNTSC = 16639;
  Settings.FrameTime = Settings.FrameTimeNTSC;
  Settings.DisableSampleCaching = FALSE;
  Settings.DisableMasterVolume = FALSE;
  Settings.Transparency = FALSE;
  Settings.SixteenBit = FALSE;
  Settings.SupportHiRes = TRUE;
  Settings.NetPlay = FALSE;
  Settings.ServerName[0] = 0;
  Settings.ThreadSound = FALSE;
  Settings.AutoSaveDelay = 30;
#ifdef _NETPLAY_SUPPORT
  Settings.Port = NP_DEFAULT_PORT;
#endif
  Settings.ApplyCheats = FALSE;
  Settings.TurboMode = FALSE;
  Settings.TurboSkipFrames = 40;
  Settings.StretchScreenshots = 1;

  GUI.FilterMethod = 0;

  rom_filename = S9xParseArgs (argv, argc);

  Settings.Transparency = Settings.ForceTransparency;
  if (Settings.ForceNoTransparency)
    Settings.Transparency = FALSE;

  if (Settings.Transparency)
    Settings.SixteenBit = TRUE;

  S9xUnmapAllControls();
  S9xSetupDefaultKeymap();
  ChangeInputDevice ();

  if (!Memory.Init () || !S9xInitAPU ())
    OutOfMemory ();

#ifdef GFX_MULTI_FORMAT
  S9xSetRenderPixelFormat (RGB565);
#endif

  S9xInitDisplay (argc, argv);
  if (!S9xGraphicsInit ())
    OutOfMemory ();

  S9xInitSound (Settings.SoundPlaybackRate, Settings.Stereo,
		Settings.SoundBufferSize);

  if (!Settings.APUEnabled)
    S9xSetSoundMute (TRUE);

  S9xInitInputDevices ();

  uint32 saved_flags = CPU.Flags;
  bool8	loaded = FALSE;

  //APU.Flags |= TRACE_FLAG;

  if (rom_filename)
    {
      loaded = Memory.LoadROM(rom_filename);
      
      if (!loaded && rom_filename[0])
	{
	  char	s[PATH_MAX + 1];
	  char	drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	  _splitpath(rom_filename, drive, dir, fname, ext);
	  snprintf(s, PATH_MAX + 1, "%s%s%s", S9xGetDirectory(ROM_DIR), SLASH_STR, fname);
	  if (ext[0] && (strlen(s) <= PATH_MAX - 1 - strlen(ext)))
	    {
	      strcat(s, ".");
	      strcat(s, ext);
	    }
	  
	  loaded = Memory.LoadROM(s);
	}
    }
  
  if (!loaded)
    {
      fprintf(stderr, "Error opening the ROM file.\n");
      exit(1);
    }

  Memory.LoadSRAM(S9xGetFilename(".srm", SRAM_DIR));
  S9xLoadCheatFile(S9xGetFilename(".cht", CHEAT_DIR));

  CPU.Flags = saved_flags;
  Settings.StopEmulation = FALSE;

#ifdef DEBUGGER
  struct sigaction sa;
  memset (&sa, 0, sizeof (struct sigaction));
  sa.sa_handler = sigbrkhandler;
  sa.sa_flags = SA_RESTART;
  sigemptyset (&sa.sa_mask);
  sigaction (SIGINT, &sa, NULL);
#endif

  S9xTextMode ();

#ifdef _NETPLAY_SUPPORT
  if (strlen (Settings.ServerName) == 0)
    {
      char *server = getenv ("S9XSERVER");
      if (server)
	{
	  strncpy (Settings.ServerName, server, 127);
	  Settings.ServerName[127] = 0;
	}
    }
  char *port = getenv ("S9XPORT");
  if (Settings.Port >= 0 && port)
    Settings.Port = atoi (port);
  else if (Settings.Port < 0)
    Settings.Port = -Settings.Port;

  if (Settings.NetPlay)
    {
      int player;

      if (!S9xNetPlayConnectToServer (Settings.ServerName, Settings.Port,
				      Memory.ROMName, player))
	{
	  fprintf (stderr, "Failed to connected to Snes9x netplay"
		   " server \"%s\" on port %d.\n",
		   Settings.ServerName, Settings.Port);
	  S9xExit ();
	}
      fprintf (stderr, "Connected to \"%s\" on port %d as"
	       " player #%d playing \"%s\"\n",
	       Settings.ServerName, Settings.Port, player, Memory.ROMName);
    }

#endif

  if (snapshot_filename)
    {
      int Flags = CPU.Flags & (DEBUG_MODE_FLAG | TRACE_FLAG);
      if (!S9xLoadSnapshot (snapshot_filename))
	exit (1);
      CPU.Flags |= Flags;
    }

  S9xGraphicsMode ();

#if 1
  size_t inbytesleft, outbytesleft;
  char *inbuf_p, *outbuf_p;
  char romname[256];
  inbytesleft = strlen ((const char *) Memory.ROMName) + 1;
  outbytesleft = 256;
  inbuf_p = Memory.ROMName;
  outbuf_p = romname;
  iconv_t cd = iconv_open (nl_langinfo (CODESET), "CP932");
  if (cd != (iconv_t) (-1))
    {
      size_t status =
	iconv (cd, &inbuf_p, &inbytesleft, &outbuf_p, &outbytesleft);
      iconv_close (cd);
      if (status == (size_t) (-1))
	romname[0] = 0;
    }
  else
    romname[0] = 0;
  sprintf (String, "\"%s\" %s: %s", romname, TITLE, VERSION);
#else
  sprintf (String, "\"%s\" %s: %s", Memory.ROMName, TITLE, VERSION);
#endif
  S9xSetTitle (String);

#ifdef JOYSTICK_SUPPORT
  uint32 JoypadSkip = 0;
#endif

  InitTimer ();
  if (!Settings.APUEnabled)
    S9xSetSoundMute (FALSE);

  while (1)
    {
      if (Settings.Paused
#ifdef DEBUGGER
	  || (CPU.Flags & DEBUG_MODE_FLAG)
#endif
	)
	{
	  S9xSetSoundMute (TRUE);
	}

#ifdef DEBUGGER
      if (CPU.Flags & DEBUG_MODE_FLAG)
	{
	  S9xDoDebug ();
	}
      else
#endif
      if (Settings.Paused)
	{
	  S9xProcessEvents (FALSE);
	  usleep (100000);
	}

#ifdef JOYSTICK_SUPPORT
      if (Settings.JoystickEnabled && (JoypadSkip++ & 1) == 0)
	ReadJoysticks ();
#endif
      S9xProcessEvents (FALSE);
      ProcessInput();

      if (!Settings.Paused
#ifdef DEBUGGER
	  && !(CPU.Flags & DEBUG_MODE_FLAG)
#endif
	)
	{
	  S9xSetSoundMute (FALSE);
	}
      
      if (!Settings.Paused
#ifdef DEBUGGER
	  || (CPU.Flags & (DEBUG_MODE_FLAG | SINGLE_STEP_FLAG))
#endif
	)
	{
	  S9xMainLoop ();
	}
    }
  return (0);
}

void
S9xAutoSaveSRAM ()
{
  Memory.SaveSRAM (S9xGetFilename (".srm", SRAM_DIR));
}

void
S9xExit ()
{
  S9xSetSoundMute (TRUE);
  S9xGraphicsDeinit ();
  S9xDeinitDisplay ();
  Memory.SaveSRAM (S9xGetFilename (".srm", SRAM_DIR));
  S9xSaveCheatFile (S9xGetFilename (".cht", CHEAT_DIR));
  Memory.Deinit ();
  S9xDeinitAPU ();

#ifdef _NETPLAY_SUPPORT
  if (Settings.NetPlay)
    S9xNetPlayDisconnect ();
#endif
  exit (0);
}

void
S9xInitInputDevices ()
{
#ifdef JOYSTICK_SUPPORT
  InitJoysticks ();
#endif
}

#ifdef JOYSTICK_SUPPORT
void
InitJoysticks ()
{
#ifdef JSIOCGVERSION
  int version;
  unsigned char axes, buttons;

  if ((js_fd[0] = open (js_device[0], O_RDONLY | O_NONBLOCK)) < 0)
    {
#ifdef DEBUGGER
      perror (js_device[0]);
#endif
      fprintf (stderr, "joystick: No joystick found.\n");
      return;
    }

  if (ioctl (js_fd[0], JSIOCGVERSION, &version))
    {
      fprintf (stderr, "joystick: You need at least driver version 1.0"
	       " for joystick support.\n");
      close (js_fd[0]);
      return;
    }
  js_fd[1] = open (js_device[1], O_RDONLY | O_NONBLOCK);
  js_fd[2] = open (js_device[2], O_RDONLY | O_NONBLOCK);
  js_fd[3] = open (js_device[3], O_RDONLY | O_NONBLOCK);
  js_fd[4] = open (js_device[4], O_RDONLY | O_NONBLOCK);
  js_fd[5] = open (js_device[5], O_RDONLY | O_NONBLOCK);
  js_fd[6] = open (js_device[6], O_RDONLY | O_NONBLOCK);
  js_fd[7] = open (js_device[7], O_RDONLY | O_NONBLOCK);

#ifdef JSIOCGNAME
  char name[130];
  bzero (name, 128);
  if (ioctl (js_fd[0], JSIOCGNAME (128), name) > 0)
    {
      printf ("Using %s (%s) as pad1", name, js_device[0]);
      if (js_fd[1] > 0)
	{
	  ioctl (js_fd[1], JSIOCGNAME (128), name);
	  printf ("and %s (%s) as pad2", name, js_device[1]);
	}
    }
  else
#endif
    {
      ioctl (js_fd[0], JSIOCGAXES, &axes);
      ioctl (js_fd[0], JSIOCGBUTTONS, &buttons);
      printf ("Using %d-axis %d-button joystick (%s) as pad1", axes, buttons,
	      js_device[0]);
      if (js_fd[1] > 0)
	{
	  ioctl (js_fd[0], JSIOCGAXES, &axes);
	  ioctl (js_fd[0], JSIOCGBUTTONS, &buttons);
	  printf (" and %d-axis %d-button (%s) as pad2", axes, buttons,
		  js_device[1]);
	}
    }

  puts (".");
#endif
}

void
ReadJoysticks ()
{
#ifdef JSIOCGVERSION
  struct js_event js_ev;
  int i;

  for (i = 0; i < 8 && js_fd[i] >= 0; i++)
    {
      while (read (js_fd[i], &js_ev, sizeof (struct js_event)) ==
	     sizeof (struct js_event))
	{
	  switch (js_ev.type & ~JS_EVENT_INIT)
	    {
	    case JS_EVENT_AXIS:
	      if (js_ev.number == 0)
		{
		  if (js_ev.value < -16384)
		    {
		      joypads[i] |= SNES_LEFT_MASK;
		      joypads[i] &= ~SNES_RIGHT_MASK;
		      break;
		    }
		  if (js_ev.value > 16384)
		    {
		      joypads[i] &= ~SNES_LEFT_MASK;
		      joypads[i] |= SNES_RIGHT_MASK;
		      break;
		    }
		  joypads[i] &= ~SNES_LEFT_MASK;
		  joypads[i] &= ~SNES_RIGHT_MASK;
		  break;
		}

	      if (js_ev.number == 1)
		{
		  if (js_ev.value < -16384)
		    {
		      joypads[i] |= SNES_UP_MASK;
		      joypads[i] &= ~SNES_DOWN_MASK;
		      break;
		    }
		  if (js_ev.value > 16384)
		    {
		      joypads[i] &= ~SNES_UP_MASK;
		      joypads[i] |= SNES_DOWN_MASK;
		      break;
		    }
		  joypads[i] &= ~SNES_UP_MASK;
		  joypads[i] &= ~SNES_DOWN_MASK;
		  break;
		}

	      break;

	    case JS_EVENT_BUTTON:
	      if (js_ev.number > 15)
		break;

	      if (js_ev.value)
		joypads[i] |= js_map_button[i][js_ev.number];
	      else
		joypads[i] &= ~js_map_button[i][js_ev.number];

	      break;
	    }
	}
    }
#endif
}
#endif // defined (JOYSTICK_SUPPORT)

const char * S9xGetDirectory (enum s9x_getdirtype dirtype)
{
	static char	s[PATH_MAX + 1];

	if (dirNames[dirtype][0])
		snprintf(s, PATH_MAX + 1, "%s%s%s", s9x_base_dir, SLASH_STR, dirNames[dirtype]);
	else
	{
		switch (dirtype)
		{
			case DEFAULT_DIR:
				strncpy(s, s9x_base_dir, PATH_MAX + 1);
				s[PATH_MAX] = 0;
				break;

			case HOME_DIR:
				strncpy(s, getenv("HOME"), PATH_MAX + 1);
				s[PATH_MAX] = 0;
				break;

			case ROMFILENAME_DIR:
				strncpy(s, Memory.ROMFilename, PATH_MAX + 1);
				s[PATH_MAX] = 0;

				for (int i = strlen(s); i >= 0; i--)
				{
					if (s[i] == SLASH_CHAR)
					{
						s[i] = 0;
						break;
					}
				}

				break;

			default:
				s[0] = 0;
				break;
		}
	}

	return (s);
}

const char * S9xGetFilename (const char *ex, enum s9x_getdirtype dirtype)
{
	static char	s[PATH_MAX + 1];
	char		drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	snprintf(s, PATH_MAX + 1, "%s%s%s%s", S9xGetDirectory(dirtype), SLASH_STR, fname, ex);

	return (s);
}

const char * S9xGetFilenameInc (const char *ex, enum s9x_getdirtype dirtype)
{
	static char	s[PATH_MAX + 1];
	char		drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	unsigned int	i = 0;
	const char		*d;
	struct stat		buf;

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	d = S9xGetDirectory(dirtype);

	do
		snprintf(s, PATH_MAX + 1, "%s%s%s.%03d%s", d, SLASH_STR, fname, i++, ex);
	while (stat(s, &buf) == 0 && i < 1000);

	return (s);
}

const char * S9xBasename (const char *f)
{
	const char	*p;

	if ((p = strrchr(f, '/')) != NULL || (p = strrchr(f, '\\')) != NULL)
		return (p + 1);

	return (f);
}

const char * S9xChooseFilename (bool8 read_only)
{
	char	s[PATH_MAX + 1];
	char	drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	const char	*filename;
	char		title[64];

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	snprintf(s, PATH_MAX + 1, "%s.frz", fname);
	sprintf(title, "%s snapshot filename", read_only ? "Select load" : "Choose save");

	S9xSetSoundMute(TRUE);
	filename = S9xSelectFilename(s, S9xGetDirectory(SNAPSHOT_DIR), "frz", title);
	S9xSetSoundMute(FALSE);

	return (filename);
}

const char * S9xChooseMovieFilename (bool8 read_only)
{
	char	s[PATH_MAX + 1];
	char	drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	const char	*filename;
	char		title[64];

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	snprintf(s, PATH_MAX + 1, "%s.smv", fname);
	sprintf(title, "Choose movie %s filename", read_only ? "playback" : "record");

	S9xSetSoundMute(TRUE);
	filename = S9xSelectFilename(s, S9xGetDirectory(HOME_DIR), "smv", title);
	S9xSetSoundMute(FALSE);

	return (filename);
}

bool8 S9xOpenSnapshotFile (const char *filename, bool8 read_only, STREAM *file)
{
	char	s[PATH_MAX + 1];
	char	drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	_splitpath(filename, drive, dir, fname, ext);

	if (*drive || *dir == SLASH_CHAR || (strlen(dir) > 1 && *dir == '.' && *(dir + 1) == SLASH_CHAR))
	{
		strncpy(s, filename, PATH_MAX + 1);
		s[PATH_MAX] = 0;
	}
	else
		snprintf(s, PATH_MAX + 1, "%s%s%s", S9xGetDirectory(SNAPSHOT_DIR), SLASH_STR, fname);

	if (!*ext && strlen(s) <= PATH_MAX - 4)
		strcat(s, ".frz");

	if ((*file = OPEN_STREAM(s, read_only ? "rb" : "wb")))
		return (TRUE);

	return (FALSE);
}

void S9xCloseSnapshotFile (STREAM file)
{
	CLOSE_STREAM(file);
}

const char *
S9xGetFilenameIncNumExt ()	//uosnes
{
  return NULL;
}

bool8
S9xInitUpdate ()
{
  return (TRUE);
}

bool8
S9xDeinitUpdate (int Width, int Height)
{
  S9xPutImage (Width, Height);
  return (TRUE);
}

static unsigned long
now ()
{
  static unsigned long seconds_base = 0;
  struct timeval tp;
  gettimeofday (&tp, NULL);
  if (!seconds_base)
    seconds_base = tp.tv_sec;

  return ((tp.tv_sec - seconds_base) * 1000 + tp.tv_usec / 1000);
}

void
OutputFrameRate ()
{
  static int FrameCount = 0;
  static unsigned long then = now ();

  if (++FrameCount % 60 == 0)
    {
      unsigned long here = now ();
//      printf ("\rFrame rate: %.2lfms", (double) (here - then) / 60);
//      fflush (stdout);
      then = here;
    }
}

void
_makepath (char *path, const char *, const char *dir,
	   const char *fname, const char *ext)
{
  if (dir && *dir)
    {
      strcpy (path, dir);
      strcat (path, "/");
    }
  else
    *path = 0;
  strcat (path, fname);
  if (ext && *ext)
    {
      strcat (path, ".");
      strcat (path, ext);
    }
}

void
_splitpath (const char *path, char *drive, char *dir, char *fname, char *ext)
{
  *drive = 0;

  const char *slash = strrchr (path, '/');
  if (!slash)
    slash = strrchr (path, '\\');

  const char *dot = strrchr (path, '.');

  if (dot && slash && dot < slash)
    dot = NULL;

  if (!slash)
    {
      strcpy (dir, "");
      strcpy (fname, path);
      if (dot)
	{
	  *(fname + (dot - path)) = 0;
	  strcpy (ext, dot + 1);
	}
      else
	strcpy (ext, "");
    }
  else
    {
      strcpy (dir, path);
      *(dir + (slash - path)) = 0;
      strcpy (fname, slash + 1);
      if (dot)
	{
	  *(fname + (dot - slash) - 1) = 0;
	  strcpy (ext, dot + 1);
	}
      else
	strcpy (ext, "");
    }
}

void
S9xToggleSoundChannel (int c)
{
  if (c == 8)
    so.sound_switch = 255;
  else
    so.sound_switch ^= 1 << c;
  S9xSetSoundControl (so.sound_switch);
}

static void
SoundTrigger ()
{
  if (Settings.APUEnabled && !so.mute_sound)
    S9xProcessSound (NULL);
}

void
StopTimer ()
{
  struct itimerval timeout;

  timeout.it_interval.tv_sec = 0;
  timeout.it_interval.tv_usec = 0;
  timeout.it_value.tv_sec = 0;
  timeout.it_value.tv_usec = 0;
  if (setitimer (ITIMER_REAL, &timeout, NULL) < 0)
    perror ("setitimer");
}

void
InitTimer ()
{
  struct itimerval timeout;
  struct sigaction sa;

  memset (&sa, 0, sizeof (struct sigaction));
  sa.sa_handler = (SIG_PF) SoundTrigger;
  sa.sa_flags = SA_RESTART;
  sigemptyset (&sa.sa_mask);
  sigaction (SIGALRM, &sa, NULL);

  timeout.it_interval.tv_sec = 0;
  timeout.it_interval.tv_usec = 10000;
  timeout.it_value.tv_sec = 0;
  timeout.it_value.tv_usec = 10000;
  if (setitimer (ITIMER_REAL, &timeout, NULL) < 0)
    perror ("setitimer");
}

void
S9xSyncSpeed ()
{
  if (!Settings.SPCPlaying && IPPU.RenderThisFrame && !(GFX.DoInterlace && GFX.InterlaceFrame == 0))
    S9xDeinitUpdate (IPPU.RenderedScreenWidth,
		     IPPU.RenderedScreenHeight);

#ifdef _NETPLAY_SUPPORT
  if (Settings.NetPlay)
    {
      // XXX: Send joypad position update to server
      // XXX: Wait for heart beat from server
      S9xNetPlaySendJoypadUpdate (joypads[0]);
      if (!S9xNetPlayCheckForHeartBeat ())
	{
	  do
	    {
	      CHECK_SOUND ();
	      S9xProcessEvents (FALSE);
	    }
	  while (!S9xNetPlayCheckForHeartBeat ());
	  IPPU.RenderThisFrame = TRUE;
	  IPPU.SkippedFrames = 0;
	}
      else
	{
	  if (IPPU.SkippedFrames < 10)
	    {
	      IPPU.SkippedFrames++;
	      IPPU.RenderThisFrame = FALSE;
	    }
	  else
	    {
	      IPPU.RenderThisFrame = TRUE;
	      IPPU.SkippedFrames = 0;
	    }
	}
    }
  else
#endif

  if (Settings.TurboMode)
    {
      if (++IPPU.FrameSkip >= Settings.TurboSkipFrames)
	{
	  IPPU.FrameSkip = 0;
	  IPPU.SkippedFrames = 0;
	  IPPU.RenderThisFrame = TRUE;
	}
      else
	{
	  ++IPPU.SkippedFrames;
	  IPPU.RenderThisFrame = FALSE;
	}
      return;
    }

  /* Check events */

  static struct timeval next1 = { 0, 0 };
  struct timeval now;

  S9xProcessEvents (FALSE);

  while (gettimeofday (&now, NULL) < 0);

  /* If there is no known "next" frame, initialize it now */
  if (next1.tv_sec == 0)
    {
      next1 = now;
      ++next1.tv_usec;
    }

  /* If we're on AUTO_FRAMERATE, we'll display frames always
   * only if there's excess time.
   * Otherwise we'll display the defined amount of frames.
   */
  unsigned limit = Settings.SkipFrames == AUTO_FRAMERATE
    ? (timercmp (&next1, &now, <)? 10 : 1) : Settings.SkipFrames;

  IPPU.RenderThisFrame = ++IPPU.SkippedFrames >= limit;
  if (IPPU.RenderThisFrame)
    {
      IPPU.SkippedFrames = 0;
    }
  else
    {
      /* If we were behind the schedule, check how much it is */
      if (timercmp (&next1, &now, <))
	{
	  unsigned lag =
	    (now.tv_sec - next1.tv_sec) * 1000000
	    + now.tv_usec - next1.tv_usec;
	  if (lag >= 1000000)
	    {
	      /* More than a second behind means probably
	       * pause. The next line prevents the magic
	       * fast-forward effect.
	       */
	      next1 = now;
	    }
	}
    }

  /* Delay until we're completed this frame */

  /* Can't use setitimer because the sound code already could
   * be using it. We don't actually need it either.
   */

  while (timercmp (&next1, &now, >))
    {
      /* If we're ahead of time, sleep a while */
      unsigned timeleft =
	(next1.tv_sec - now.tv_sec) * 1000000 + next1.tv_usec - now.tv_usec;
      //fprintf(stderr, "<%u>", timeleft);
      usleep (timeleft);

      S9xProcessEvents (FALSE);

      while (gettimeofday (&now, NULL) < 0);
      /* Continue with a while-loop because usleep()
       * could be interrupted by a signal
       */
    }

  /* Calculate the timestamp of the next frame. */
  next1.tv_usec += Settings.FrameTime;
  if (next1.tv_usec >= 1000000)
    {
      next1.tv_sec += next1.tv_usec / 1000000;
      next1.tv_usec %= 1000000;
    }
}

static long
log2 (long num)
{
  long n = 0;

  while (num >>= 1)
    n++;

  return (n);
}

#define SOUND_BUFFER_SIZE (1024 * 16)
#define MAX_BUFFER_SIZE SOUND_BUFFER_SIZE
#define SOUND_BUFFER_SIZE_MASK (SOUND_BUFFER_SIZE - 1)

static int Rates[7] = {
  0, 8000, 11025, 22050, 32000, 44100, 48000
};

static int BufferSizes[7] = {
  0, 256, 256, 512, 512, 1024, 1024
};

bool8
S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
{
  /* Linux version (OSS) */
  int J, K;

  if ((so.sound_fd = open ("/dev/dsp", O_WRONLY)) < 0)
    {
      perror ("/dev/dsp");
      return (FALSE);
    }

  //Test and check back to http://snes9x.com/forum/topic.asp?TOPIC_ID=8512
//    J = AFMT_U8;
  J = AFMT_S16_NE;
  if (ioctl (so.sound_fd, SNDCTL_DSP_SETFMT, &J) < 0)
    {
      perror ("ioctl SNDCTL_DSP_SETFMT");
      return (FALSE);
    }

  if (J != AFMT_S16_NE)
    {
      so.sixteen_bit = FALSE;
      J = AFMT_U8;
      if (ioctl (so.sound_fd, SNDCTL_DSP_SETFMT, &J) < 0)
	{
	  perror ("ioctl SNDCTL_DSP_SETFMT");
	  return (FALSE);
	}
    }
  else
    so.sixteen_bit = TRUE;

  so.stereo = stereo;
  if (ioctl (so.sound_fd, SNDCTL_DSP_STEREO, &so.stereo) < 0)
    {
      perror ("ioctl SNDCTL_DSP_STEREO");
      return (FALSE);
    }

  so.playback_rate = Rates[mode & 0x07];
  if (ioctl (so.sound_fd, SNDCTL_DSP_SPEED, &so.playback_rate) < 0)
    {
      perror ("ioctl SNDCTL_DSP_SPEED");
      return (FALSE);
    }

  S9xSetPlaybackRate (so.playback_rate);
  //SetBufferSize (BufferSizes[mode & 7] / (Rates[mode & 0x07] / 1000));

  if (buffer_size == 0)
    buffer_size = BufferSizes[mode & 7];

  if (buffer_size > MAX_BUFFER_SIZE / 4)
    buffer_size = MAX_BUFFER_SIZE / 4;
  if (so.sixteen_bit)
    buffer_size *= 2;
  if (so.stereo)
    buffer_size *= 2;

  int power2 = log2 (buffer_size);
  J = K = power2 | (3 << 16);
  if (ioctl (so.sound_fd, SNDCTL_DSP_SETFRAGMENT, &J) < 0)
    {
      perror ("ioctl SNDCTL_DSP_SETFRAGMENT");
      return (FALSE);
    }
  ioctl (so.sound_fd, SNDCTL_DSP_GETBLKSIZE, &so.buffer_size);

  printf ("Rate: %d, Buffer size: %d, 16-bit: %s, Stereo: %s, Encoded: %s\n",
	  so.playback_rate, so.buffer_size, so.sixteen_bit ? "yes" : "no",
	  so.stereo ? "yes" : "no", so.encoded ? "yes" : "no");

  return (TRUE);
}

static uint8 Buf[MAX_BUFFER_SIZE];

static volatile bool8 block_signal = FALSE;
static volatile bool8 block_generate_sound = FALSE;
static volatile bool8 pending_signal = FALSE;

void WriteSamples(int32 l, int32 r)
{
  WriteSamples0(l, r);
}

void
S9xGenerateSound ()
{
}

void *
S9xProcessSound (void *)
{

  /* If threads in use, this is to loop indefinitely */
  /* If not, this will be called by timer */

  audio_buf_info info;

  if (!Settings.ThreadSound &&
      (ioctl (so.sound_fd, SNDCTL_DSP_GETOSPACE, &info) == -1 ||
       info.bytes < so.buffer_size))
    {
      return (NULL);
    }

  /* Number of samples to generate now */
  int sample_count = so.buffer_size;

  if (so.sixteen_bit)
    {
      /* to prevent running out of buffer space,
       * create less samples
       */
      sample_count >>= 1;
    }

  unsigned byte_offset = so.play_position;
  S9xMixSamplesO (Buf, sample_count, byte_offset & SOUND_BUFFER_SIZE_MASK);

  unsigned bytes_to_write = sample_count;
  if (so.sixteen_bit)
    bytes_to_write <<= 1;

  so.play_position += bytes_to_write;
  so.play_position &= SOUND_BUFFER_SIZE_MASK;	/* wrap to beginning */

  /* Feed the samples to the soundcard until nothing is left */
  for (;;)
    {
      int I = bytes_to_write;
      if (byte_offset + I > SOUND_BUFFER_SIZE)
	{
	  I = SOUND_BUFFER_SIZE - byte_offset;
	}
      if (I == 0)
	break;

      I = write (so.sound_fd, (char *) Buf + byte_offset, I);
      if (I > 0)
	{
	  bytes_to_write -= I;
	  byte_offset += I;
	  byte_offset &= SOUND_BUFFER_SIZE_MASK;	/* wrap */
	}
      /* give up if an unrecoverable error happened */
      if (I < 0 && errno != EINTR)
	break;
    }
  /* All data sent. */

  return (NULL);
}

static uint32 JoyReadCount;

uint32
S9xReadJoypad (int which)
{
  if (which > 7)
    return 0;

  if (which == 0)
    {
      ++JoyReadCount;
    }

  return ((Settings.JoystickConnected[which] ? 0x80000000 : 0) | joypads[which]);
}

uint32 S9xReadPseudoPointer ()
{
  uint32 j = 0;
  //j |= FunctionCodeStates[MF_LG2_UP] ? 0x800 : 0;
  //j |= FunctionCodeStates[MF_LG2_DOWN] ? 0x400 : 0;
  //j |= FunctionCodeStates[MF_LG2_LEFT] ? 0x200 : 0;
  //j |= FunctionCodeStates[MF_LG2_RIGHT] ? 0x100 : 0;
  return j;
}

