/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "ppu.h"
#include "snapshot.h"
#include "gfx.h"
#include "display.h"
#include "apu_orig.h"
#include "soundux.h"
#include "controls.h"
#include "controls2.h"
#include "x11.h"

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

GUIData GUI;
extern uint32 joypads[8];

void Scale8 (int width, int height);
void Scale16 (int width, int height);
void Convert8To16 (int width, int height);
void Convert16To8 (int width, int height);
void Convert8To24 (int width, int height);
void Convert8To24Packed (int width, int height);
void Convert16To24 (int width, int height);
void Convert16To24Packed (int width, int height);
void SetupImage ();
int ErrorHandler (Display *, XErrorEvent *);
void S9xDisplayStateChange (const char *str, bool8 on);

uint8* ScreenBuf1 = NULL;
uint8* ScreenBuffer = NULL;

uint8 S9xPalette[0x10000];

uint8 snes9x_clear_change_log = 0;
void
S9xDeinitDisplay ()
{
  S9xTextMode ();
  uint32 i;

  for (i = 0; i < sizeof (GUI.to_free) / sizeof (GUI.to_free[0]); i++)
    if (GUI.to_free[i])
      {
	free (GUI.to_free[i]);
	GUI.to_free[i] = NULL;
      }

  if (GUI.image)
    {
      XDestroyImage (GUI.image);
      GUI.image = NULL;
    }
  XSync (GUI.display, False);
  XCloseDisplay (GUI.display);

  if(ScreenBuf1)
    {
      delete [] ScreenBuf1;
      ScreenBuf1 = NULL;
    }
}

void
S9xInitDisplay (int, char **)
{
  GUI.FlipCounter = 0;
  GUI.NumFlipFrames = 1;

  ScreenBuf1 = new uint8 [EXT_PITCH * EXT_HEIGHT];

  if(!ScreenBuf1)
    {
      fprintf (stderr, "Failed to allocate ScreenBuffer.\n");
      exit (1);
    }

  ScreenBuffer = ScreenBuf1 + EXT_OFFSET;
  GFX.Screen = (uint16*)ScreenBuffer;
  GFX.Pitch = EXT_PITCH;


  if (!(GUI.display = XOpenDisplay (NULL)))
    {
      fprintf (stderr, "Failed to connect to X server.\n");
      exit (1);
    }
  GUI.screen = DefaultScreenOfDisplay (GUI.display);
  GUI.screen_num = XScreenNumberOfScreen (GUI.screen);
  GUI.visual = DefaultVisualOfScreen (GUI.screen);
  GUI.window_width = IMAGE_WIDTH;
  GUI.window_height = IMAGE_HEIGHT;

  XVisualInfo plate;
  XVisualInfo *matches;
  int count;

  plate.visualid = XVisualIDFromVisual (GUI.visual);
  matches = XGetVisualInfo (GUI.display, VisualIDMask, &plate, &count);

  if (!count)
    {
      fprintf (stderr, "Your X Window System server is unwell!\n");
      exit (1);
    }
  GUI.ScreenDepth = matches[0].depth;
  if ((GUI.ScreenDepth != 8 && GUI.ScreenDepth != 15 && GUI.ScreenDepth != 16
       && GUI.ScreenDepth != 24) || (matches[0].c_class != PseudoColor
			       && matches[0].c_class != TrueColor
			       && matches[0].c_class != GrayScale))
    {
      fprintf (stderr, "\
Snes9x needs an X Window System server set to 8, 15, 16, 24 or 32-bit colour GUI.ScreenDepth\n\
supporting PseudoColor, TrueColor or GrayScale.\n");
      exit (1);
    }

  if (GUI.ScreenDepth >= 15 && !Settings.ForceNoTransparency)
    {
      Settings.Transparency = TRUE;
      Settings.SixteenBit = TRUE;
    }

  GUI.pseudo = matches[0].c_class == PseudoColor ||
    matches[0].c_class == GrayScale;
  GUI.grayscale = matches[0].c_class == GrayScale;

  if (GUI.ScreenDepth != 8 || !GUI.pseudo)
    {
      GUI.RedShift = ffs (matches[0].red_mask) - 1;
      GUI.GreenShift = ffs (matches[0].green_mask) - 1;
      GUI.BlueShift = ffs (matches[0].blue_mask) - 1;
      GUI.red_size = matches[0].red_mask >> GUI.RedShift;
      GUI.green_size = matches[0].green_mask >> GUI.GreenShift;
      GUI.blue_size = matches[0].blue_mask >> GUI.BlueShift;
      if (GUI.ScreenDepth == 16 && GUI.green_size == 63)
	GUI.GreenShift++;

#ifdef GFX_MULTI_FORMAT
      switch (GUI.ScreenDepth)
	{
	default:
	case 24:
	case 8:
	  S9xSetRenderPixelFormat (RGB565);
	  break;
	case 16:
	  if (GUI.red_size != GUI.green_size
	      || GUI.blue_size != GUI.green_size)
	    {
	      // 565 format
	      if (GUI.GreenShift > GUI.BlueShift
		  && GUI.GreenShift > GUI.RedShift)
		S9xSetRenderPixelFormat (GBR565);
	      else if (GUI.RedShift > GUI.BlueShift)
		S9xSetRenderPixelFormat (RGB565);
	      else
		S9xSetRenderPixelFormat (BGR565);
	      break;
	    }
	  /* FALL ... */
	case 15:
	  if (GUI.GreenShift > GUI.BlueShift
	      && GUI.GreenShift > GUI.RedShift)
	    S9xSetRenderPixelFormat (GBR555);
	  else if (GUI.RedShift > GUI.BlueShift)
	    S9xSetRenderPixelFormat (RGB555);
	  else
	    S9xSetRenderPixelFormat (BGR555);
	  break;
	}
#endif
    }
  XFree ((char *) matches);

  int l = 0;
  int i;

  for (i = 0; i < 6; i++)
    {
      int r = (i * 31) / (6 - 1);
      for (int j = 0; j < 6; j++)
	{
	  int g = (j * 31) / (6 - 1);
	  for (int k = 0; k < 6; k++)
	    {
	      int b = (k * 31) / (6 - 1);

	      GUI.fixed_colours[l].red = r;
	      GUI.fixed_colours[l].green = g;
	      GUI.fixed_colours[l++].blue = b;
	    }
	}
    }

  int *color_diff = new int[0x10000];
  int diffr, diffg, diffb, maxdiff = 0, won = 0, lost;
  int r, d = 8;
  for (r = 0; r <= (int) MAX_RED; r++)
    {
      int cr, g, q;

      int k = 6 - 1;
      cr = (r * k) / MAX_RED;
      q = (r * k) % MAX_RED;
      if (q > d && cr < k)
	cr++;
      diffr = abs (cr * k - r);
      for (g = 0; g <= (int) MAX_GREEN; g++)
	{
	  int cg, b;

	  k = 6 - 1;
	  cg = (g * k) / MAX_GREEN;
	  q = (g * k) % MAX_GREEN;
	  if (q > d && cg < k)
	    cg++;
	  diffg = abs (cg * k - g);
	  for (b = 0; b <= (int) MAX_BLUE; b++)
	    {
	      int cb;
	      int rgb = BUILD_PIXEL2 (r, g, b);

	      k = 6 - 1;
	      cb = (b * k) / MAX_BLUE;
	      q = (b * k) % MAX_BLUE;
	      if (q > d && cb < k)
		cb++;
	      diffb = abs (cb * k - b);
	      GUI.palette[rgb] = (cr * 6 + cg) * 6 + cb;
	      color_diff[rgb] = diffr + diffg + diffb;
	      if (color_diff[rgb] > maxdiff)
		maxdiff = color_diff[rgb];
	    }
	}
    }

  while (maxdiff > 0 && l < 256)
    {
      int newmaxdiff = 0;
      lost = 0;
      won++;
      for (r = MAX_RED; r >= 0; r--)
	{
	  int g;

	  for (g = MAX_GREEN; g >= 0; g--)
	    {
	      int b;

	      for (b = MAX_BLUE; b >= 0; b--)
		{
		  int rgb = BUILD_PIXEL2 (r, g, b);

		  if (color_diff[rgb] == maxdiff)
		    {
		      if (l >= 256)
			lost++;
		      else
			{
			  GUI.fixed_colours[l].red = r;
			  GUI.fixed_colours[l].green = g;
			  GUI.fixed_colours[l].blue = b;
			  GUI.palette[rgb] = l;
			  l++;
			}
		      color_diff[rgb] = 0;
		    }
		  else if (color_diff[rgb] > newmaxdiff)
		    newmaxdiff = color_diff[rgb];

		}
	    }
	}
      maxdiff = newmaxdiff;
    }
  delete[]color_diff;

  XSetWindowAttributes attrib;

  attrib.background_pixel = BlackPixelOfScreen (GUI.screen);
  GUI.window = XCreateWindow (GUI.display, RootWindowOfScreen (GUI.screen),
			      (WidthOfScreen (GUI.screen) -
			       GUI.window_width) / 2,
			      (HeightOfScreen (GUI.screen) -
			       GUI.window_height) / 2, GUI.window_width,
			      GUI.window_height, 0, GUI.ScreenDepth, InputOutput,
			      GUI.visual, CWBackPixel, &attrib);

  Atom atom1 = XInternAtom (GUI.display, "WM_PROTOCOLS", False);
  Atom atom2 = XInternAtom (GUI.display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols (GUI.display, GUI.window, &atom2, 1);

  static XColor bg;
  static XColor fg;
  static char data[8] = { 0x01 };

  Pixmap bitmap = XCreateBitmapFromData (GUI.display, GUI.window, data, 8, 8);
  GUI.point_cursor =
    XCreatePixmapCursor (GUI.display, bitmap, bitmap, &fg, &bg, 0, 0);
  XDefineCursor (GUI.display, GUI.window, GUI.point_cursor);

  GUI.cross_hair_cursor = XCreateFontCursor (GUI.display, XC_crosshair);
  GUI.gc = DefaultGCOfScreen (GUI.screen);
  {
    XSizeHints Hints;
    XWMHints WMHints;

    memset ((void *) &Hints, 0, sizeof (XSizeHints));
    memset ((void *) &WMHints, 0, sizeof (XWMHints));

    Hints.flags = PSize | PMinSize;
    Hints.min_width = Hints.base_width = SNES_WIDTH;
    Hints.min_height = Hints.base_height = SNES_HEIGHT_EXTENDED;
    WMHints.input = True;
    WMHints.flags = InputHint;
    XSetWMHints (GUI.display, GUI.window, &WMHints);
    XSetWMNormalHints (GUI.display, GUI.window, &Hints);
  }
  XSelectInput (GUI.display, GUI.window, FocusChangeMask | ExposureMask |
		KeyPressMask | KeyReleaseMask | StructureNotifyMask |
		ButtonPressMask | ButtonReleaseMask);

  if (GUI.pseudo)
    {
      GUI.cmap = XCreateColormap (GUI.display, GUI.window, GUI.visual, True);
      XSetWindowColormap (GUI.display, GUI.window, GUI.cmap);
      for (i = 0; i < 256; i++)
	{
	  GUI.colors[i].red = GUI.colors[i].green = GUI.colors[i].blue = 0;
	  GUI.colors[i].pixel = i;
	  GUI.colors[i].flags = DoRed | DoGreen | DoBlue;
	}
      XStoreColors (GUI.display, GUI.cmap, GUI.colors, 256);
    }
  XMapRaised (GUI.display, GUI.window);
  XClearWindow (GUI.display, GUI.window);
  SetupImage ();

  switch (GUI.ScreenDepth)
    {
    case 8:
      GUI.bytes_per_pixel = 1;
      break;

    case 15:
    case 16:
      GUI.bytes_per_pixel = 2;
      break;

    case 24:
      if (GUI.image->bits_per_pixel == 24)
	GUI.bytes_per_pixel = 3;
      else
	GUI.bytes_per_pixel = 4;
      break;

    case 32:
      GUI.bytes_per_pixel = 4;
      break;
    }
}

void
SetupImage ()
{
  int tf = 0;
  int image_width = GUI.window_width;
  int image_height = GUI.window_height;

  if (image_width < IMAGE_WIDTH)
    image_width = IMAGE_WIDTH;
  if (image_height < IMAGE_HEIGHT)
    image_height = IMAGE_HEIGHT;

  GUI.image_needs_scaling = GUI.window_width != IMAGE_WIDTH ||
    GUI.window_height != IMAGE_HEIGHT || GUI.scale || Settings.SupportHiRes;

  uint32 i;

  for (i = 0; i < sizeof (GUI.to_free) / sizeof (GUI.to_free[0]); i++)
    if (GUI.to_free[i])
      {
	free (GUI.to_free[i]);
	GUI.to_free[i] = NULL;
      }

  if (GUI.image)
    {
      XDestroyImage (GUI.image);
      GUI.image = NULL;
    }
  GUI.image = XCreateImage (GUI.display, GUI.visual, GUI.ScreenDepth, ZPixmap, 0,
			    (char *) NULL, image_width, image_height,
			    BitmapUnit (GUI.display), 0);
  GUI.image->data = (char *) malloc (image_height *
				     GUI.image->bytes_per_line);
  GUI.image->byte_order = LSBFirst;

  int h = IMAGE_HEIGHT;

  ZeroMemory (ScreenBuf1, EXT_PITCH * EXT_HEIGHT);
  if ((uint8 *) GUI.image->data != (uint8 *) GFX.Screen)
    ZeroMemory (GUI.image->data,
		GUI.image->bytes_per_line * GUI.image->height);

  GUI.image_date = (uint8 *) GUI.image->data;
  GUI.bytes_per_line = GUI.image->bytes_per_line;
}

int
ErrorHandler (Display *, XErrorEvent *)
{
  return (0);
}

void
S9xSetTitle (const char *string)
{
  XmbSetWMProperties (GUI.display, GUI.window, string,
		      NULL, __argv, __argc, NULL, NULL, NULL);
  XFlush (GUI.display);
}

bool8
S9xReadMousePosition (int which1, int &x, int &y, uint32 & buttons)
{
  if (which1 == 0)
    {
      x = GUI.mouse_x;
      y = GUI.mouse_y;
      buttons = GUI.mouse_buttons;
      return (TRUE);
    }
  return (FALSE);
}

bool8
S9xReadSuperScopePosition (int &x, int &y, uint32 & buttons)
{
  x = (int) ((GUI.mouse_x - GUI.box.x) * (256.0 / (double) GUI.box.width));
  y = (int) ((GUI.mouse_y - GUI.box.y) * (224.0 / (double) GUI.box.height));
  buttons = (GUI.mouse_buttons & 3) | (GUI.superscope_turbo << 2) |
    (GUI.superscope_pause << 3);

  return (TRUE);
}

bool
JustifierOffscreen ()
{
  return (bool) (GUI.mouse_buttons & 2);
}

void
JustifierButtons (uint32 & justifiers)
{
  if (Settings.ControllerOption == SNES_JUSTIFIER_2)
    {
      if ((GUI.mouse_buttons & 1) || (GUI.mouse_buttons & 2))
	{
	  justifiers |= 0x00200;
	}
      if (GUI.mouse_buttons & 4)
	{
	  justifiers |= 0x00800;
	}
    }
  else
    {
      if ((GUI.mouse_buttons & 1) || (GUI.mouse_buttons & 2))
	{
	  justifiers |= 0x00100;
	}
      if (GUI.mouse_buttons & 4)
	{
	  justifiers |= 0x00400;
	}
    }
}

#include "movie.h"

static void
S9xInfoMessage (const char *msg)
{
  S9xSetInfoString (msg);
  if (Settings.Paused)
    puts (msg);
}

static uint32
GetNormalSpeed ()
{
  return Settings.PAL ? Settings.FrameTimePAL : Settings.FrameTimeNTSC;
}

static unsigned
GetSpeedPercentage ()
{
  const uint32 normalspeed = GetNormalSpeed ();
  return normalspeed * 100 / Settings.FrameTime;
}

static void
SetSpeedPercentage (unsigned p)
{
  const uint32 normalspeed = GetNormalSpeed ();
  Settings.FrameTime = normalspeed * 100 / p;
}

#include <map>
using namespace std;
static class KeyboardSetup
{
public:
  /* List of all possible keyboard-accessible functions */
  enum functiontype
  {
    FUNC_NONE = 0,
    PLAYING_KEY,
    TOGGLE_SOUND,
    WRITE_SAVE_NUM,
    WRITE_SAVE_ASK,
    LOAD_SAVE_NUM,
    LOAD_SAVE_ASK,
    TOGGLE_LAYER,
    WRITE_SPC_ASK,
    FRAMETIME_INC,
    FRAMETIME_DEC,
    FRAMESKIP_INC,
    FRAMESKIP_DEC,
    PAUSE,
    DGA_FULLSCREEN,
    SCREENSHOT,
    SPC7110LOG,
    TOGGLE_HDMA,
    TOGGLE_JOYSWAP,
    TOGGLE_GL_CUBE,
    TOGGLE_BG_HACK,
    TOGGLE_TRANSP,
    TOGGLE_CLIPWIN,
    TOGGLE_CONTROLLER,
    TOGGLE_INTERPSOUND,
    TOGGLE_SYNCSOUND,
    TOGGLE_MODE7INTERP,
    TURBO_ENABLE,
    SUPERSCOPE_TURBO,
    SUPERSCOPE_PAUSE_ENABLE,
    EXIT,
    WRITE_MOVIE_ASK,
    LOAD_MOVIE_ASK,
    STOP_MOVIE,

    FUNC_LAST = STOP_MOVIE	/* update this to match the last token */
  };
  static const char *GetFuncName (functiontype func)
  {
    switch (func)
      {
      case PLAYING_KEY:
	return "PLAYING_KEY";
	case TOGGLE_SOUND:return "TOGGLE_SOUND";
	case WRITE_SAVE_NUM:return "WRITE_SAVE_NUM";
	case WRITE_SAVE_ASK:return "WRITE_SAVE_ASK";
	case LOAD_SAVE_NUM:return "LOAD_SAVE_NUM";
	case LOAD_SAVE_ASK:return "LOAD_SAVE_ASK";
	case TOGGLE_LAYER:return "TOGGLE_LAYER";
	case WRITE_SPC_ASK:return "WRITE_SPC_ASK";
	case FRAMETIME_INC:return "FRAMETIME_INC";
	case FRAMETIME_DEC:return "FRAMETIME_DEC";
	case FRAMESKIP_INC:return "FRAMESKIP_INC";
	case FRAMESKIP_DEC:return "FRAMESKIP_DEC";
	case PAUSE:return "PAUSE";
	case DGA_FULLSCREEN:return "DGA_FULLSCREEN";
	case SCREENSHOT:return "SCREENSHOT";
	case SPC7110LOG:return "SPC7110LOG";
	case TOGGLE_HDMA:return "TOGGLE_HDMA";
	case TOGGLE_JOYSWAP:return "TOGGLE_JOYSWAP";
	case TOGGLE_GL_CUBE:return "TOGGLE_GL_CUBE";
	case TOGGLE_BG_HACK:return "TOGGLE_BG_HACK";
	case TOGGLE_TRANSP:return "TOGGLE_TRANSP";
	case TOGGLE_CLIPWIN:return "TOGGLE_CLIPWIN";
	case TOGGLE_CONTROLLER:return "TOGGLE_CONTROLLER";
	case TOGGLE_INTERPSOUND:return "TOGGLE_INTERPSOUND";
	case TOGGLE_SYNCSOUND:return "TOGGLE_SYNCSOUND";
	case TOGGLE_MODE7INTERP:return "TOGGLE_MODE7INTERP";
	case TURBO_ENABLE:return "TURBO_ENABLE";
	case SUPERSCOPE_TURBO:return "SUPERSCOPE_TURBO";
	case SUPERSCOPE_PAUSE_ENABLE:return "SUPERSCOPE_PAUSE_ENABLE";
	case EXIT:return "EXIT";
	case WRITE_MOVIE_ASK:return "WRITE_MOVIE_ASK";
	case LOAD_MOVIE_ASK:return "LOAD_MOVIE_ASK";
	case STOP_MOVIE:return "STOP_MOVIE";
	default:return "unknown";
      }
  }
  static const char *GetKeySymName (int sym)
  {
    static char Buf[64];
    sprintf (Buf, "'%c'(%d)", sym, sym);
    return Buf;
  }
  struct Keyfunction
  {
    functiontype func;
    unsigned param;
    bool inited;
    int sym;
    const char *prefix;
  private:
    /* ensure nobody uses the assignment operator on us */
    void operator= (const Keyfunction & b);
  public:
      Keyfunction ():func (FUNC_NONE), inited (false)
    {
    }
    Keyfunction (functiontype f, unsigned p):func (f), param (p),
      inited (true)
    {
    }
    void Assign (const Keyfunction & b)
    {
      if (inited)
	{
	  fprintf (stderr,
		   "[Input] Warning: Key %s%s had %s[0x%X], overwriting by %s[0x%X]\n",
		   prefix, GetKeySymName (sym),
		   GetFuncName (func), param, GetFuncName (b.func), b.param);
	}
      func = b.func;
      param = b.param;
      inited = b.inited;
    }
    void SetName (int s, const char *p)
    {
      sym = s;
      prefix = p;
    }
  };
  struct Keydata
  {
    Keyfunction plain, shift, ctrl, alt;
  public:
      Keydata ()
    {
    }
    void DefineAs (int sym)
    {
      plain.SetName (sym, "");
      shift.SetName (sym, "shift+");
      ctrl.SetName (sym, "ctrl+");
      alt.SetName (sym, "alt+");
    }
  };
private:
  enum masktype
  { MaskNone, MaskCtrl, MaskShift, MaskAlt };
  map < int, Keydata > keydata;

  void Define (int sym, masktype mask, functiontype func, unsigned param)
  {
    Keyfunction result (func, param);
    Keydata & target = keydata[sym];
    target.DefineAs (sym);
    switch (mask)
      {
      case MaskCtrl:
	target.ctrl.Assign (result);
	break;
      case MaskShift:
	target.shift.Assign (result);
	break;
      case MaskAlt:
	target.alt.Assign (result);
	break;
      case MaskNone:
	{
	  target.plain.Assign (result);

	  result.inited = false;
	  /* so we'll not get warnings when overwriting
	   * these extra fields later: */
	  if (!target.shift.inited)
	    target.shift.Assign (result);
	  if (!target.ctrl.inited)
	    target.ctrl.Assign (result);
	  if (!target.alt.inited)
	    target.alt.Assign (result);
	  break;
	}
      }
  }
  void VerifySetup ()
  {
    bool8 Defined[(unsigned) FUNC_LAST + 1] = { FALSE };
    map < int, Keydata >::const_iterator i;
    for (i = keydata.begin (); i != keydata.end (); ++i)
      {
	Defined[i->second.plain.func] = TRUE;
	Defined[i->second.shift.func] = TRUE;
	Defined[i->second.ctrl.func] = TRUE;
	Defined[i->second.alt.func] = TRUE;
      }
    for (unsigned a = 0; a < sizeof (Defined); ++a)
      if (a != FUNC_NONE && !Defined[a])
	{
	  fprintf (stderr, "[Input] Warning: No key assigned for %s\n",
		   GetFuncName ((functiontype) a));
	}
  }
private:
  struct Key
  {
    int symbol;
    masktype mask;
  public:
      Key ():symbol (0), mask (MaskNone)
    {
    }
    Key (int sym):symbol (sym), mask (MaskNone)
    {
    }
    Key (int sym, masktype m):symbol (sym), mask (m)
    {
    }
  };
  static Key Shift (int symbol)
  {
    return Key (symbol, MaskShift);
  }
  static Key Ctrl (int symbol)
  {
    return Key (symbol, MaskCtrl);
  }
  static Key Alt (int symbol)
  {
    return Key (symbol, MaskAlt);
  }
private:
  void Define (functiontype func, unsigned param,
	       Key k1, Key k2 = Key (), Key k3 = Key (), Key k4 = Key ())
  {
    if (k1.symbol)
      Define (k1.symbol, k1.mask, func, param);
    if (k2.symbol)
      Define (k2.symbol, k2.mask, func, param);
    if (k3.symbol)
      Define (k3.symbol, k3.mask, func, param);
    if (k4.symbol)
      Define (k4.symbol, k4.mask, func, param);
  }
public:
  /* Contruct the keyboard setup upon program start. */
  KeyboardSetup ()
  {
    // player 1: tr,tl, x,a
    Define (PLAYING_KEY, 0x00000010, XK_z, XK_w, XK_b);
    Define (PLAYING_KEY, 0x00000020, XK_a, XK_q, XK_v);
    Define (PLAYING_KEY, 0x00000040, XK_s, XK_e, XK_m);
    Define (PLAYING_KEY, 0x00000080, XK_d, XK_t, XK_period);
    // player 1: right,left,down,up
    Define (PLAYING_KEY, 0x00000100, XK_k, XK_Right);
    Define (PLAYING_KEY, 0x00000200, XK_h, XK_Left);
    Define (PLAYING_KEY, 0x00000400, XK_j, XK_n, XK_Down);
    Define (PLAYING_KEY, 0x00000800, XK_u, XK_Up);
    // player 1: start,select, y,b
    Define (PLAYING_KEY, 0x00001000, XK_Return);
    Define (PLAYING_KEY, 0x00002000, XK_space);
    Define (PLAYING_KEY, 0x00004000, XK_x, XK_r, XK_comma);
    Define (PLAYING_KEY, 0x00008000, XK_y, XK_c);
    // player 2: tr,tl, x,a
    Define (PLAYING_KEY, 0x00100000, XK_Delete);
    Define (PLAYING_KEY, 0x00200000, XK_Insert);
    Define (PLAYING_KEY, 0x00400000, XK_Home);
    Define (PLAYING_KEY, 0x00800000, XK_Prior);
    // player 2: right,left,down,up
    Define (PLAYING_KEY, 0x01000000, XK_KP_4);
    Define (PLAYING_KEY, 0x02000000, XK_KP_6);
    Define (PLAYING_KEY, 0x04000000, XK_KP_2);
    Define (PLAYING_KEY, 0x08000000, XK_KP_8);
    // player 2: start,select,y,b
    Define (PLAYING_KEY, 0x10000000, XK_KP_Enter);
    Define (PLAYING_KEY, 0x20000000, XK_KP_Add);
    Define (PLAYING_KEY, 0x40000000, XK_End);
    Define (PLAYING_KEY, 0x80000000, XK_Next);
    // toggle sound channels
    // Careful: alt + f-keys might be bound to the window manager!
    Define (TOGGLE_SOUND, 0, Alt (XK_F4), Ctrl (XK_F4));
    Define (TOGGLE_SOUND, 1, Alt (XK_F5), Ctrl (XK_F5));
    Define (TOGGLE_SOUND, 2, Alt (XK_F6), Ctrl (XK_F6));
    Define (TOGGLE_SOUND, 3, Alt (XK_F7), Ctrl (XK_F7));
    Define (TOGGLE_SOUND, 4, Alt (XK_F8), Ctrl (XK_F8));
    Define (TOGGLE_SOUND, 5, Alt (XK_F9), Ctrl (XK_F9));
    Define (TOGGLE_SOUND, 6, Alt (XK_F10), Ctrl (XK_F10));
    Define (TOGGLE_SOUND, 7, Alt (XK_F11), Ctrl (XK_F11));
    // re-enable all sound channels
    Define (TOGGLE_SOUND, 8, Alt (XK_F12), Ctrl (XK_F12));
    // numbered quicksaves
    Define (WRITE_SAVE_NUM, 0, XK_F1);
    Define (WRITE_SAVE_NUM, 1, XK_F2);
    Define (WRITE_SAVE_NUM, 2, XK_F3);
    Define (WRITE_SAVE_NUM, 3, XK_F4);
    Define (WRITE_SAVE_NUM, 4, XK_F5);
    Define (WRITE_SAVE_NUM, 5, XK_F6);
    Define (WRITE_SAVE_NUM, 6, XK_F7);
    Define (WRITE_SAVE_NUM, 7, XK_F8);
    Define (WRITE_SAVE_NUM, 8, XK_F9);
    Define (WRITE_SAVE_NUM, 9, XK_F10);
    Define (LOAD_SAVE_NUM, 0, Shift (XK_F1));
    Define (LOAD_SAVE_NUM, 1, Shift (XK_F2));
    Define (LOAD_SAVE_NUM, 2, Shift (XK_F3));
    Define (LOAD_SAVE_NUM, 3, Shift (XK_F4));
    Define (LOAD_SAVE_NUM, 4, Shift (XK_F5));
    Define (LOAD_SAVE_NUM, 5, Shift (XK_F6));
    Define (LOAD_SAVE_NUM, 6, Shift (XK_F7));
    Define (LOAD_SAVE_NUM, 7, Shift (XK_F8));
    Define (LOAD_SAVE_NUM, 8, Shift (XK_F9));
    Define (LOAD_SAVE_NUM, 9, Shift (XK_F10));
    // layer toggles
    Define (TOGGLE_LAYER, 0, XK_1);
    Define (TOGGLE_LAYER, 1, XK_2);
    Define (TOGGLE_LAYER, 2, XK_3);
    Define (TOGGLE_LAYER, 3, XK_4);
    // sprite layer toggle
    Define (TOGGLE_LAYER, 4, XK_5);
    // named saves
    Define (LOAD_SAVE_ASK, 0, Alt (XK_F2), Ctrl (XK_F2), XK_F11);
    Define (WRITE_SAVE_ASK, 0, Alt (XK_F3), Ctrl (XK_F3), XK_F12);
    Define (WRITE_SPC_ASK, 0, Alt (XK_F1), Ctrl (XK_F1));
    // timings
    Define (FRAMETIME_INC, 0, Shift (XK_equal), Shift (XK_plus));
    Define (FRAMETIME_DEC, 0, Shift (XK_minus));
    Define (FRAMESKIP_INC, 0, XK_equal, XK_plus);
    Define (FRAMESKIP_DEC, 0, XK_minus);
    // pause
    Define (PAUSE, 0, XK_Pause, XK_Break, XK_Scroll_Lock);
    // misc functions
    Define (DGA_FULLSCREEN, 0, Alt (XK_Return));
    Define (SCREENSHOT, 0, XK_Print);
    Define (SPC7110LOG, 0, XK_Sys_Req);
    Define (TOGGLE_HDMA, 0, XK_0);
    Define (TOGGLE_JOYSWAP, 0, XK_6);
    Define (TOGGLE_GL_CUBE, 0, Shift (XK_6));
    Define (TOGGLE_BG_HACK, 0, XK_8);
    Define (TOGGLE_TRANSP, 0, XK_9);
    Define (TOGGLE_CLIPWIN, 0, XK_BackSpace);
    Define (TOGGLE_CONTROLLER, 0, XK_7);
    Define (TOGGLE_INTERPSOUND, 0, XK_bracketleft, Alt (XK_8));
    Define (TOGGLE_SYNCSOUND, 0, XK_bracketright, Alt (XK_9));
    Define (TOGGLE_MODE7INTERP, 0, Shift (XK_9));
    Define (TURBO_ENABLE, 0, XK_Tab);
    Define (SUPERSCOPE_TURBO, 0, XK_grave, XK_asciitilde, XK_numbersign);
    Define (SUPERSCOPE_PAUSE_ENABLE, 0, XK_slash);
    // escape from fullscreen, escape from emulator
    Define (EXIT, 0, XK_Escape);
    // movie functions
    Define (WRITE_MOVIE_ASK, 0, Shift (XK_1));
    Define (LOAD_MOVIE_ASK, 0, Shift (XK_2));
    Define (STOP_MOVIE, 0, Shift (XK_3));

    VerifySetup ();
  }
public:
  const Keyfunction & GetKeyFunction (int keycode, unsigned keymask) const
  {
    static Keyfunction notfound;
      map < int, Keydata >::const_iterator i = keydata.find (keycode);
    //fprintf(stderr, "Pressed key %d (mask %X)\n", keycode, keymask);
    if (i != keydata.end ())
      {
	// ShiftMask: shift ($0001)
	// LockMask:  capslock ($0002)
	// Mod1Mask:  alt ($0008)
	// Mod4Mask:  numlock ($0010)
	//         :  altgr ($2000)
	// Mod5Mask:  scrolllock ($0080)
	// Button1Mask: left mouse button ($0100)
	// Button2Mask: middle mouse button ($0200)
	// Button3Mask: right mouse button ($0400)

	if (keymask & (Mod1Mask | 0x2000))
	  return i->second.alt;
	if (keymask & (ControlMask))
	  return i->second.ctrl;
	if (keymask & (ShiftMask))
	  return i->second.shift;
	return i->second.plain;
      }
    return notfound;
  }
}

KBSetup;

void
S9xProcessEvents (bool8 block)
{
  while (block || XPending (GUI.display))
    {
      XEvent event;

      XNextEvent (GUI.display, &event);
      block = FALSE;
      switch (event.type)
	{
	case KeyPress:
	case KeyRelease:
	  {
	    int key = XKeycodeToKeysym (GUI.display, event.xkey.keycode, 0);

	    const KeyboardSetup::Keyfunction & func
	      = KBSetup.GetKeyFunction (key, event.xkey.state);
	    switch (func.func)
	      {
	      case KeyboardSetup::FUNC_NONE:
		{
		  break;
		}
	      case KeyboardSetup::PLAYING_KEY:
		{
		  uint16 word1 = (func.param) & 0xFFFF;
		  uint16 word2 = (func.param >> 16) & 0xFFFF;
		  if (event.type == KeyPress)
		    {
		      joypads[0] |= word1;
		      joypads[1] |= word2;
		    }
		  else
		    {
		      joypads[0] &= ~word1;
		      joypads[1] &= ~word2;
		    }
		  break;
		}
	      case KeyboardSetup::TOGGLE_SOUND:
		{
		  if (event.type != KeyPress)
		    break;
		  if (func.param > 8)
		    break;

		  S9xToggleSoundChannel (func.param);
		  if (func.param == 8)
		    S9xInfoMessage ("All sound channels on");
		  else
		    {
		      sprintf (GUI.info_string, "Sound channel %d %s",
			       func.param,
			       (so.
				sound_switch & (1 << (func.param))) ? "on" :
			       "off");
		      S9xInfoMessage (GUI.info_string);
		    }
		  break;
		}
	      case KeyboardSetup::WRITE_SAVE_NUM:
		{
		  if (event.type != KeyPress)
		    break;
		  if (func.param > 999)
		    break;

		  char def[PATH_MAX];
		  char filename[PATH_MAX];
		  char drive[_MAX_DRIVE];
		  char dir[_MAX_DIR];
		  char ext[_MAX_EXT];

		  _splitpath (Memory.ROMFilename, drive, dir, def, ext);
		  sprintf (filename, "%s%s%s.%03d",
			   S9xGetDirectory(SNAPSHOT_DIR), SLASH_STR, def,
			   func.param);
		  sprintf (GUI.info_string, "%s.%03d saved", def, func.param);
		  S9xInfoMessage (GUI.info_string);
		  S9xFreezeGame (filename);
		  break;
		}
	      case KeyboardSetup::WRITE_SAVE_ASK:
		{
		  if (event.type != KeyPress)
		    break;

		  S9xFreezeGame (S9xChooseFilename (FALSE));
		  break;
		}
	      case KeyboardSetup::LOAD_SAVE_NUM:
		{
		  if (event.type != KeyPress)
		    break;
		  if (func.param > 999)
		    break;

		  char def[PATH_MAX];
		  char filename[PATH_MAX];
		  char drive[_MAX_DRIVE];
		  char dir[_MAX_DIR];
		  char ext[_MAX_EXT];

		  _splitpath (Memory.ROMFilename, drive, dir, def, ext);
		  sprintf (filename, "%s%s%s.%03d",
			   S9xGetDirectory(SNAPSHOT_DIR), SLASH_STR, def,
			   func.param);
		  if (S9xUnfreezeGame (filename))
		    {
		      sprintf (GUI.info_string, "%s.%03d loaded", def,
			       func.param);
		      S9xInfoMessage (GUI.info_string);
		    }
		  else
		    {
		      static const char *digits = "t123456789";
		      _splitpath (Memory.ROMFilename, drive, dir, def, ext);
		      sprintf (filename, "%s%s%s.zs%c",
			       S9xGetDirectory(SNAPSHOT_DIR), SLASH_STR,
			       def, digits[func.param]);
		      if (S9xUnfreezeGame (filename))
			{
			  sprintf (GUI.info_string,
				   "Loaded ZSNES freeze file %s.zs%c", def,
				   digits[func.param]);
			  S9xInfoMessage (GUI.info_string);
			}
		      else
			{
			  sprintf (GUI.info_string,
				   "Freeze file %u not found", func.param);
			  S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND,
				      GUI.info_string);
			}
		    }
		  break;
		}
	      case KeyboardSetup::LOAD_SAVE_ASK:
		{
		  if (event.type != KeyPress)
		    break;

		  S9xUnfreezeGame (S9xChooseFilename (TRUE));
		  break;
		}
	      case KeyboardSetup::TOGGLE_LAYER:
		{
		  if (event.type != KeyPress)
		    break;
		  if (func.param > 4)
		    break;

		  unsigned mask = 1 << func.param;
		  Settings.BG_Forced ^= mask;
		  if (func.param == 4)
		    sprintf (GUI.info_string, "Sprites");
		  else
		    sprintf (GUI.info_string, "BG#%d", func.param);
		  S9xDisplayStateChange (GUI.info_string,
					 !(Settings.BG_Forced & mask));
		  break;
		}
	      case KeyboardSetup::WRITE_SPC_ASK:
		{
		  if (event.type != KeyPress)
		    break;

		  char def[PATH_MAX];
		  char filename[PATH_MAX];
		  char drive[_MAX_DRIVE];
		  char dir[_MAX_DIR];
		  char ext[_MAX_EXT];

		  _splitpath (Memory.ROMFilename, drive, dir, def, ext);
		  strcpy (ext, "spc");
		  _makepath (filename, drive, S9xGetDirectory(SNAPSHOT_DIR),
			     def, ext);
		  if (S9xSPCDump (filename))
		    sprintf (GUI.info_string, "%s.%s saved", def, ext);
		  else
		    sprintf (GUI.info_string, "%s.%s not saved (%s)", def,
			     ext, strerror (errno));

		  S9xInfoMessage (GUI.info_string);
		  break;
		}
	      case KeyboardSetup::FRAMETIME_INC:
		{
		  if (event.type != KeyPress)
		    break;

		  unsigned CurSpeed = GetSpeedPercentage ();
		  if (CurSpeed < 10)
		    ++CurSpeed;
		  else if (CurSpeed < 150)
		    CurSpeed += 5;
		  else if (CurSpeed < 200)
		    CurSpeed += 10;
		  else if (CurSpeed < 400)
		    CurSpeed += 20;
		  else if (CurSpeed < 500)
		    CurSpeed += 50;
		  else if (CurSpeed < 1500)
		    CurSpeed += 100;

		  SetSpeedPercentage (CurSpeed);
		  sprintf (GUI.info_string, "Emulated speed: %d%% (%.1f FPS)",
			   CurSpeed, 1e6 / Settings.FrameTime);
		  S9xInfoMessage (GUI.info_string);
		  break;
		}
	      case KeyboardSetup::FRAMETIME_DEC:
		{
		  if (event.type != KeyPress)
		    break;

		  unsigned CurSpeed = GetSpeedPercentage ();
		  if (CurSpeed <= 1)
		    {
		    }
		  else if (CurSpeed <= 10)
		    --CurSpeed;
		  else if (CurSpeed <= 150)
		    CurSpeed -= 5;
		  else if (CurSpeed <= 200)
		    CurSpeed -= 10;
		  else if (CurSpeed <= 400)
		    CurSpeed -= 20;
		  else if (CurSpeed <= 500)
		    CurSpeed -= 50;
		  else
		    CurSpeed -= 100;

		  SetSpeedPercentage (CurSpeed);
		  sprintf (GUI.info_string, "Emulated speed: %d%% (%.1f FPS)",
			   CurSpeed, 1e6 / Settings.FrameTime);
		  S9xInfoMessage (GUI.info_string);
		  break;
		}
	      case KeyboardSetup::FRAMESKIP_INC:
		{
		  if (event.type != KeyPress)
		    break;

		  if (Settings.SkipFrames == AUTO_FRAMERATE)
		    Settings.SkipFrames = 1;
		  else if (Settings.SkipFrames < 10)
		    Settings.SkipFrames++;

		  if (Settings.SkipFrames == AUTO_FRAMERATE)
		    S9xInfoMessage ("Auto frame skip");
		  else
		    {
		      sprintf (GUI.info_string, "Frame skip: %d",
			       Settings.SkipFrames - 1);
		      S9xInfoMessage (GUI.info_string);
		    }
		  break;
		}
	      case KeyboardSetup::FRAMESKIP_DEC:
		{
		  if (event.type != KeyPress)
		    break;

		  if (Settings.SkipFrames <= 1)
		    Settings.SkipFrames = AUTO_FRAMERATE;
		  else if (Settings.SkipFrames != AUTO_FRAMERATE)
		    Settings.SkipFrames--;

		  if (Settings.SkipFrames == AUTO_FRAMERATE)
		    S9xInfoMessage ("Auto frame skip");
		  else
		    {
		      sprintf (GUI.info_string, "Frame skip: %d",
			       Settings.SkipFrames - 1);
		      S9xInfoMessage (GUI.info_string);
		    }
		  break;
		}
	      case KeyboardSetup::PAUSE:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.Paused ^= 1;
		  S9xDisplayStateChange ("Pause", Settings.Paused);

		  break;
		}
	      case KeyboardSetup::DGA_FULLSCREEN:
		{
		  if (event.type != KeyPress)
		    break;
		  break;
		}
	      case KeyboardSetup::SCREENSHOT:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.TakeScreenshot = TRUE;
		  break;
		}
	      case KeyboardSetup::SPC7110LOG:
		{
		  if (event.type != KeyPress)
		    break;
		  break;
		}
	      case KeyboardSetup::TOGGLE_HDMA:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.DisableHDMA = !Settings.DisableHDMA;
		  S9xDisplayStateChange ("HDMA emulation",
					 !Settings.DisableHDMA);
		  break;
		}
	      case KeyboardSetup::TOGGLE_JOYSWAP:
		{
		  if (event.type != KeyPress)
		    break;

		  S9xApplyCommand(S9xGetCommandT("SwapJoypads"),1,0);
		  if(GFX.InfoString)
		    {
		      const char *s = GFX.InfoString;
		      GFX.InfoString = NULL;
		      S9xInfoMessage (s);
		    }
		  break;
		}
	      case KeyboardSetup::TOGGLE_GL_CUBE:
		{
		  if (event.type != KeyPress)
		    break;
		  break;
		}
	      case KeyboardSetup::TOGGLE_BG_HACK:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.BGLayering = !Settings.BGLayering;
		  S9xDisplayStateChange ("Background layering hack",
					 Settings.BGLayering);
		  break;
		}
	      case KeyboardSetup::TOGGLE_TRANSP:
		{
		  if (event.type != KeyPress)
		    break;

		  if (Settings.SixteenBit)
		    {
		      Settings.Transparency = !Settings.Transparency;
		      S9xDisplayStateChange ("Transparency effects",
					     Settings.Transparency);
		    }
		  break;
		}
	      case KeyboardSetup::TOGGLE_CLIPWIN:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.DisableGraphicWindows =
		    !Settings.DisableGraphicWindows;
		  S9xDisplayStateChange ("Graphic clip windows",
					 !Settings.DisableGraphicWindows);

		  break;
		}
	      case KeyboardSetup::TOGGLE_CONTROLLER:
		{
		  if (event.type != KeyPress)
		    break;

		  ++Settings.ControllerOption %= SNES_MAX_CONTROLLER_OPTIONS;
		  ChangeInputDevice();

		  S9xReportControllers();
		  if(GFX.InfoString)
		    {
		      const char *s = GFX.InfoString;
		      GFX.InfoString = NULL;
		      S9xInfoMessage (s);
		    }
		  break;
		}
	      case KeyboardSetup::TOGGLE_INTERPSOUND:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.InterpolatedSound ^= 1;
		  S9xDisplayStateChange ("Interpolated sound",
					 Settings.InterpolatedSound);
		  break;
		}
	      case KeyboardSetup::TOGGLE_SYNCSOUND:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.SoundSync ^= 1;
		  S9xDisplayStateChange ("Synchronised sound",
					 Settings.SoundSync);
		  break;
		}
	      case KeyboardSetup::TOGGLE_MODE7INTERP:
		{
		  if (event.type != KeyPress)
		    break;

		  Settings.Mode7Interpolate ^= TRUE;
		  S9xDisplayStateChange ("Mode 7 Interpolation",
					 Settings.Mode7Interpolate);

		  break;
		}
	      case KeyboardSetup::TURBO_ENABLE:
		{
		  Settings.TurboMode = event.type == KeyPress;
		  break;
		}
	      case KeyboardSetup::SUPERSCOPE_TURBO:
		{
		  if (event.type != KeyPress)
		    break;

		  GUI.superscope_turbo = !GUI.superscope_turbo;

		  break;
		}
	      case KeyboardSetup::SUPERSCOPE_PAUSE_ENABLE:
		{
		  GUI.superscope_pause = event.type == KeyPress;
		  break;
		}
	      case KeyboardSetup::EXIT:
		{
		  if (event.type != KeyPress)
		    break;
		  S9xExit ();

		  break;
		}
	      case KeyboardSetup::WRITE_MOVIE_ASK:
		{
		  if (event.type != KeyPress)
		    break;

		  wchar_t name[MOVIE_MAX_METADATA] = { 0 };
		  if (S9xMovieActive ())
		    S9xMovieStop (FALSE);
		  S9xMovieCreate (S9xChooseMovieFilename (FALSE), 0x1F,
				  //MOVIE_OPT_FROM_SNAPSHOT
				  MOVIE_OPT_FROM_RESET, name, 0);
		  break;
		}
	      case KeyboardSetup::LOAD_MOVIE_ASK:
		{
		  if (event.type != KeyPress)
		    break;

		  if (S9xMovieActive ())
		    S9xMovieStop (FALSE);
		  S9xMovieOpen (S9xChooseMovieFilename (TRUE), FALSE);
		  break;
		}
	      case KeyboardSetup::STOP_MOVIE:
		{
		  if (event.type != KeyPress)
		    break;

		  if (S9xMovieActive ())
		    S9xMovieStop (FALSE);
		  break;
		}

		/* Do not put a "default" case here, or you'll
		 * miss the compiler warning about unhandled
		 * enumeration values
		 */
	      }

	    break;
	  }
	case FocusIn:
	  //XAutoRepeatOff (GUI.display);
	  XFlush (GUI.display);
	  //Settings.Paused &= ~2;
	  break;
	case FocusOut:
	  XAutoRepeatOn (GUI.display);
	  XFlush (GUI.display);
	  //Settings.Paused |= 2;
	  break;
	case ConfigureNotify:
	  if (event.xany.window == GUI.window &&
	      (GUI.window_width != event.xconfigure.width ||
	       GUI.window_height != event.xconfigure.height))
	    {
	      GUI.window_width = event.xconfigure.width;
	      GUI.window_height = event.xconfigure.height;
	      IPPU.RenderThisFrame = TRUE;
	      IPPU.FrameSkip = Settings.SkipFrames;
	      SetupImage ();
	    }
	  break;

	case ClientMessage:
	  if (event.xclient.message_type ==
	      XInternAtom (GUI.display, "WM_PROTOCOLS", True)
	      && event.xclient.data.l[0] == XInternAtom (GUI.display,
							 "WM_DELETE_WINDOW",
							 True))
	    {
	      S9xExit ();
	    }
	  break;
	}
    }
}

void
S9xPutImage (int snes_width, int snes_height)
{
  bool8 done = FALSE;
  int width, height, cheight;

  width = snes_width;
  height = snes_height;
  cheight =
    (height >
     SNES_HEIGHT_EXTENDED) ? SNES_HEIGHT_EXTENDED * 2 : SNES_HEIGHT_EXTENDED;

  GUI.output_screen = (uint8 *)GFX.Screen;
  GUI.output_pitch = GFX.Pitch;
  width = snes_width;
  height = snes_height;
  cheight =
    (height >
     SNES_HEIGHT_EXTENDED) ? SNES_HEIGHT_EXTENDED << 1 :
    SNES_HEIGHT_EXTENDED;

  if (GUI.image_needs_scaling)
    {
      GUI.box.x = 0;
      GUI.box.y = 0;
    }
  else
    {
      GUI.box.x = (GUI.window_width - width) / 2;
      GUI.box.y = (GUI.window_height - cheight) / 2;
    }

  if ((Settings.SixteenBit && GUI.ScreenDepth != 15 && GUI.ScreenDepth != 16) ||
      (!Settings.SixteenBit && (!GUI.pseudo || GUI.ScreenDepth != 8)) ||
      (GUI.image_needs_scaling
       && !(Settings.SixteenBit && GUI.interpolate == 5)))
    {
      done = TRUE;
      switch (GUI.ScreenDepth)
	{
	case 8:
	  if (Settings.SixteenBit)
	    Convert16To8 (width, height);
	  else
	    Scale8 (width, height);
	  break;

	case 15:
	case 16:
	  if (!Settings.SixteenBit)
	    Convert8To16 (width, height);
	  else
	    Scale16 (width, height);
	  break;

	case 32:
	case 24:
	  if (Settings.SixteenBit)
	    {
	      if (GUI.image->bits_per_pixel == 32)
		Convert16To24 (width, height);
	      else
		Convert16To24Packed (width, height);
	    }
	  else
	    {
	      if (GUI.image->bits_per_pixel == 32)
		Convert8To24 (width, height);
	      else
		Convert8To24Packed (width, height);
	    }
	  break;
	}
    }

  if (GUI.image_needs_scaling)
    {
      GUI.box.width = GUI.window_width;
      GUI.box.height = GUI.window_height;
    }
  else
    {
      GUI.box.width = width;
      GUI.box.height = height;
    }

  XPutImage (GUI.display, GUI.window, GUI.gc, GUI.image,
	     0, 0, GUI.box.x, GUI.box.y, GUI.box.width, GUI.box.height);

  GUI.last_snes_width = snes_width;
  GUI.last_snes_height = snes_height;

  if (GUI.box.x != GUI.old_box.x || GUI.box.y != GUI.old_box.y ||
      GUI.box.width != GUI.old_box.width
      || GUI.box.height != GUI.old_box.height)
    {
      // If the rendered image has changed size/position clear any areas of the
      // screen that should now be border
      Region old_box = XCreateRegion ();
      Region new_box = XCreateRegion ();

      XUnionRectWithRegion (&GUI.old_box, old_box, old_box);
      XUnionRectWithRegion (&GUI.box, new_box, new_box);
      XSubtractRegion (old_box, new_box, old_box);
      if (!XEmptyRegion (old_box))
	{
	  Window window = GUI.window;
	  XRectangle clip;

	  XClipBox (old_box, &clip);
	  XSetRegion (GUI.display, GUI.gc, old_box);
	  XSetForeground (GUI.display, GUI.gc, GUI.ScreenDepth == 8 ? 0 :
			  BlackPixelOfScreen (GUI.screen));
	  XFillRectangle (GUI.display, window, GUI.gc,
			  clip.x, clip.y, clip.width, clip.height);
	  XSetClipMask (GUI.display, GUI.gc, None);
	}

      XDestroyRegion (new_box);
      XDestroyRegion (old_box);
      GUI.old_box = GUI.box;
    }

  Window root, child;
  int root_x, root_y;
  int x, y;
  unsigned int mask;

  // Use QueryPointer to sync X server and as a side effect also gets
  // current pointer position for SNES mouse emulation.
  XQueryPointer (GUI.display, GUI.window, &root, &child, &root_x, &root_y,
		 &x, &y, &mask);

  if (Settings.ControllerOption == SNES_SUPERSCOPE)
    {
      if (!GUI.superscope)
	{
	  XDefineCursor (GUI.display, GUI.window, GUI.cross_hair_cursor);
	  GUI.superscope = TRUE;
	}
    }
  else if (GUI.superscope)
    {
      XDefineCursor (GUI.display, GUI.window, GUI.point_cursor);
      GUI.superscope = FALSE;
    }
  if (x >= 0 && y >= 0 && GUI.window_width && y < GUI.window_height)
    {
      GUI.mouse_x = x * width / GUI.window_width;	/* Scale to SNES's. */
      GUI.mouse_y = y * height / GUI.window_height;
      if (mask & Mod1Mask)
	{
	  //IPPU.PrevMouseX[0] = IPPU.PrevMouseX[1] = GUI.mouse_x;
	  //IPPU.PrevMouseY[0] = IPPU.PrevMouseY[1] = GUI.mouse_y;
	  if (!GUI.mod1_pressed)
	    {
	      GUI.mod1_pressed = TRUE;
	      XDefineCursor (GUI.display, GUI.window, GUI.cross_hair_cursor);
	    }
	}
      else if (GUI.mod1_pressed)
	{
	  GUI.mod1_pressed = FALSE;
	  if (!GUI.superscope)
	    {
	      XDefineCursor (GUI.display, GUI.window, GUI.point_cursor);
	    }
	}
      GUI.mouse_buttons = ((mask & 0x100) >> 8) | ((mask & 0x200) >> 7) |
	((mask & 0x400) >> 9) | ((mask & 0x800) >> 8);
    }
}

void
S9xSetPalette ()
{
  int i;

  if (GUI.grayscale)
    {
      uint16 Brightness = IPPU.MaxBrightness;

      for (i = 0; i < 256; i++)
	{
	  GUI.colors[i].flags = DoRed | DoGreen | DoBlue;
	  GUI.colors[i].red = GUI.colors[i].green = GUI.colors[i].blue =
	    (uint16) (((((PPU.CGDATA[i] >> 0) & 0x1F) * Brightness * 50) +
		       (((PPU.CGDATA[i] >> 5) & 0x1F) * Brightness * 69) +
		       (((PPU.CGDATA[i] >> 10) & 0x1F) * Brightness * 21)) *
		      1.40935);
	}
      XStoreColors (GUI.display, GUI.cmap, GUI.colors, 256);
    }
  else if (GUI.pseudo)
    {
      if (Settings.SixteenBit)
	{
	  for (i = 0; i < 256; i++)
	    {
	      GUI.colors[i].flags = DoRed | DoGreen | DoBlue;
	      GUI.colors[i].red = GUI.fixed_colours[i].red << 11;
	      GUI.colors[i].green = GUI.fixed_colours[i].green << 11;
	      GUI.colors[i].blue = GUI.fixed_colours[i].blue << 11;
	    }
	}
      else
	{
	  uint16 Brightness = (IPPU.MaxBrightness) * 140;

	  for (i = 0; i < 256; i++)
	    {
	      GUI.colors[i].flags = DoRed | DoGreen | DoBlue;
	      GUI.colors[i].red = ((PPU.CGDATA[i] >> 0) & 0x1F) * Brightness;
	      GUI.colors[i].green =
		((PPU.CGDATA[i] >> 5) & 0x1F) * Brightness;
	      GUI.colors[i].blue =
		((PPU.CGDATA[i] >> 10) & 0x1F) * Brightness;
	    }
	}
      XStoreColors (GUI.display, GUI.cmap, GUI.colors, 256);
    }
}

const char *
S9xSelectFilename (const char *def, const char *dir1,
		   const char *ext1, const char *title)
{
  static char path[PATH_MAX];
  char buffer[PATH_MAX];

  XAutoRepeatOn (GUI.display);

  printf ("\n%s (default: %s): ", title, def);
  fflush (stdout);
  if (fgets (buffer, sizeof (buffer) - 1, stdin))
    {
      //XAutoRepeatOff (GUI.display);

      char *p = buffer;
      while (isspace (*p))
	p++;
      if (!*p)
	{
	  strcpy (buffer, def);
	  p = buffer;
	}

      char *q = strrchr (p, '\n');
      if (q)
	*q = 0;

      char fname[PATH_MAX];
      char drive[_MAX_DRIVE];
      char dir[_MAX_DIR];
      char ext[_MAX_EXT];

      _splitpath (p, drive, dir, fname, ext);
      _makepath (path, drive, *dir ? dir : dir1, fname, *ext ? ext : ext1);
      return (path);
    }

  //XAutoRepeatOff (GUI.display);

  return (NULL);
}

void
Scale8 (int width, int height)
{
  register uint32 x_error;
  register uint32 x_fraction;
  uint32 y_error = 0;
  uint32 y_fraction;
  int yy = height - 1;

  x_fraction = (width * 0x10000) / GUI.window_width;
  y_fraction = (height * 0x10000) / GUI.window_height;

  for (int y = GUI.window_height - 1; y >= 0; y--)
    {
      register uint8 *d = (uint8 *) GUI.image_date + y * GUI.bytes_per_line +
	GUI.window_width - 1;
      register uint8 *s =
	GUI.output_screen + yy * GUI.output_pitch + width - 1;
      y_error += y_fraction;
      while (y_error >= 0x10000)
	{
	  yy--;
	  y_error -= 0x10000;
	}
      x_error = 0;
      for (register int x = GUI.window_width - 1; x >= 0; x--)
	{
	  *d-- = *s;
	  x_error += x_fraction;

	  while (x_error >= 0x10000)
	    {
	      s--;
	      x_error -= 0x10000;
	    }
	}
    }
}

void
Scale16 (int width, int height)
{
  register uint32 x_error;
  register uint32 x_fraction;
  uint32 y_error = 0;
  uint32 y_fraction;
  int yy = height - 1;

  x_fraction = (width * 0x10000) / GUI.window_width;
  y_fraction = (height * 0x10000) / GUI.window_height;

  for (int y = GUI.window_height - 1; y >= 0; y--)
    {
      register uint16 *d =
	(uint16 *) (GUI.image_date + y * GUI.bytes_per_line) +
	GUI.window_width - 1;
      register uint16 *s =
	(uint16 *) (GUI.output_screen + yy * GUI.output_pitch) + width - 1;
      y_error += y_fraction;
      while (y_error >= 0x10000)
	{
	  yy--;
	  y_error -= 0x10000;
	}
      x_error = 0;
      for (register int x = GUI.window_width - 1; x >= 0; x--)
	{
	  *d-- = *s;
	  x_error += x_fraction;

	  while (x_error >= 0x10000)
	    {
	      s--;
	      x_error -= 0x10000;
	    }
	}
    }
}

void
Convert8To24 (int width, int height)
{
  uint32 brightness = IPPU.MaxBrightness >> 1;

  if (!GUI.image_needs_scaling)
    {
      // Convert
      for (register int y = 0; y < height; y++)
	{
	  register uint32 *d = (uint32 *) (GUI.image_date +
					   y * GUI.bytes_per_line);
	  register uint8 *s = GUI.output_screen + y * GUI.output_pitch;

	  for (register int x = 0; x < width; x++)
	    {
	      uint32 pixel = PPU.CGDATA[*s++];
	      *d++ = (((pixel & 0x1f) * brightness) << GUI.RedShift) |
		((((pixel >> 5) & 0x1f) * brightness) << GUI.GreenShift) |
		((((pixel >> 10) & 0x1f) * brightness) << GUI.BlueShift);
	    }
	}
    }
  else
    {
      // Scale and convert
      register uint32 x_error;
      register uint32 x_fraction;
      uint32 y_error = 0;
      uint32 y_fraction;
      int yy = 0;

      x_fraction = (width * 0x10000) / GUI.window_width;
      y_fraction = (height * 0x10000) / GUI.window_height;

      for (int y = 0; y < GUI.window_height; y++)
	{
	  register uint32 *d = (uint32 *) (GUI.image_date +
					   y * GUI.bytes_per_line);
	  register uint8 *s = GUI.output_screen + yy * GUI.output_pitch;
	  y_error += y_fraction;
	  while (y_error >= 0x10000)
	    {
	      yy++;
	      y_error -= 0x10000;
	    }
	  x_error = 0;
	  for (register int x = 0; x < GUI.window_width; x++)
	    {
	      uint32 pixel = PPU.CGDATA[*s];
	      *d++ = (((pixel & 0x1f) * brightness) << GUI.RedShift) |
		((((pixel >> 5) & 0x1f) * brightness) << GUI.GreenShift) |
		((((pixel >> 10) & 0x1f) * brightness) << GUI.BlueShift);

	      x_error += x_fraction;
	      while (x_error >= 0x10000)
		{
		  s++;
		  x_error -= 0x10000;
		}
	    }
	}
    }
}

void
Convert16To24 (int width, int height)
{
  if (!GUI.image_needs_scaling)
    {
      // Convert
      for (register int y = 0; y < height; y++)
	{
	  register uint32 *d = (uint32 *) (GUI.image_date +
					   y * GUI.bytes_per_line);
	  register uint16 *s =
	    (uint16 *) (GUI.output_screen + y * GUI.output_pitch);

	  for (register int x = 0; x < width; x++)
	    {
	      uint32 pixel = *s++;
	      *d++ = (((pixel >> 11) & 0x1f) << (GUI.RedShift + 3)) |
		(((pixel >> 6) & 0x1f) << (GUI.GreenShift + 3)) |
		((pixel & 0x1f) << (GUI.BlueShift + 3));
	    }
	}
    }
  else
    {
      // Scale and convert
      register uint32 x_error;
      register uint32 x_fraction;
      uint32 y_error = 0;
      uint32 y_fraction;
      int yy = 0;

      x_fraction = (width * 0x10000) / GUI.window_width;
      y_fraction = (height * 0x10000) / GUI.window_height;

      for (int y = 0; y < GUI.window_height; y++)
	{
	  register uint32 *d = (uint32 *) (GUI.image_date +
					   y * GUI.bytes_per_line);
	  register uint16 *s =
	    (uint16 *) (GUI.output_screen + yy * GUI.output_pitch);
	  y_error += y_fraction;
	  while (y_error >= 0x10000)
	    {
	      yy++;
	      y_error -= 0x10000;
	    }
	  x_error = 0;
	  for (register int x = 0; x < GUI.window_width; x++)
	    {
	      uint32 pixel = *s;
	      *d++ = (((pixel >> 11) & 0x1f) << (GUI.RedShift + 3)) |
		(((pixel >> 6) & 0x1f) << (GUI.GreenShift + 3)) |
		((pixel & 0x1f) << (GUI.BlueShift + 3));

	      x_error += x_fraction;
	      while (x_error >= 0x10000)
		{
		  s++;
		  x_error -= 0x10000;
		}
	    }
	}
    }
}

void
Convert8To24Packed (int width, int height)
{
  uint32 brightness = IPPU.MaxBrightness >> 1;
  uint8 levels[32];

  for (int l = 0; l < 32; l++)
    levels[l] = l * brightness;

  if (!GUI.image_needs_scaling)
    {
      // Convert
      for (register int y = 0; y < height; y++)
	{
	  register uint8 *d =
	    (uint8 *) (GUI.image_date + y * GUI.bytes_per_line);
	  register uint8 *s = GUI.output_screen + y * GUI.output_pitch;

	  if (GUI.RedShift < GUI.BlueShift)
	    {
	      // Order is RGB
	      for (register int x = 0; x < width; x++)
		{
		  uint16 pixel = PPU.CGDATA[*s++];
		  *d++ = levels[(pixel & 0x1f)];
		  *d++ = levels[((pixel >> 5) & 0x1f)];
		  *d++ = levels[((pixel >> 10) & 0x1f)];
		}
	    }
	  else
	    {
	      // Order is BGR
	      for (register int x = 0; x < width; x++)
		{
		  uint16 pixel = PPU.CGDATA[*s++];
		  *d++ = levels[((pixel >> 10) & 0x1f)];
		  *d++ = levels[((pixel >> 5) & 0x1f)];
		  *d++ = levels[(pixel & 0x1f)];
		}
	    }
	}
    }
  else
    {
      // Scale and convert
      register uint32 x_error;
      register uint32 x_fraction;
      uint32 y_error = 0;
      uint32 y_fraction;
      int yy = 0;

      x_fraction = (width * 0x10000) / GUI.window_width;
      y_fraction = (height * 0x10000) / GUI.window_height;

      for (int y = 0; y < GUI.window_height; y++)
	{
	  register uint8 *d = (uint8 *) (GUI.image_date +
					 y * GUI.bytes_per_line);
	  register uint8 *s = GUI.output_screen + yy * GUI.output_pitch;
	  y_error += y_fraction;
	  while (y_error >= 0x10000)
	    {
	      yy++;
	      y_error -= 0x10000;
	    }
	  x_error = 0;
	  if (GUI.RedShift < GUI.BlueShift)
	    {
	      // Order is RGB
	      for (register int x = 0; x < GUI.window_width; x++)
		{
		  uint16 pixel = PPU.CGDATA[*s];
		  *d++ = levels[(pixel & 0x1f)];
		  *d++ = levels[((pixel >> 5) & 0x1f)];
		  *d++ = levels[((pixel >> 10) & 0x1f)];

		  x_error += x_fraction;
		  while (x_error >= 0x10000)
		    {
		      s++;
		      x_error -= 0x10000;
		    }
		}
	    }
	  else
	    {
	      // Order is BGR
	      for (register int x = 0; x < GUI.window_width; x++)
		{
		  uint16 pixel = PPU.CGDATA[*s];
		  *d++ = levels[((pixel >> 10) & 0x1f)];
		  *d++ = levels[((pixel >> 5) & 0x1f)];
		  *d++ = levels[(pixel & 0x1f)];

		  x_error += x_fraction;
		  while (x_error >= 0x10000)
		    {
		      s++;
		      x_error -= 0x10000;
		    }
		}
	    }
	}
    }
}

void
Convert16To24Packed (int width, int height)
{
  if (!GUI.image_needs_scaling)
    {
      // Convert
      for (register int y = 0; y < height; y++)
	{
	  register uint8 *d = (uint8 *) (GUI.image_date +
					 y * GUI.bytes_per_line);
	  register uint16 *s =
	    (uint16 *) (GUI.output_screen + y * GUI.output_pitch);

	  if (GUI.RedShift < GUI.BlueShift)
	    {
	      // Order is RGB
	      for (register int x = 0; x < width; x++)
		{
		  uint32 pixel = *s++;
		  *d++ = (pixel >> (11 - 3)) & 0xf8;
		  *d++ = (pixel >> (6 - 3)) & 0xf8;
		  *d++ = (pixel & 0x1f) << 3;
		}
	    }
	  else
	    {
	      // Order is BGR
	      for (register int x = 0; x < width; x++)
		{
		  uint32 pixel = *s++;
		  *d++ = (pixel & 0x1f) << 3;
		  *d++ = (pixel >> (6 - 3)) & 0xf8;
		  *d++ = (pixel >> (11 - 3)) & 0xf8;
		}
	    }
	}
    }
  else
    {
      // Scale and convert
      register uint32 x_error;
      register uint32 x_fraction;
      uint32 y_error = 0;
      uint32 y_fraction;
      int yy = 0;

      x_fraction = (width * 0x10000) / GUI.window_width;
      y_fraction = (height * 0x10000) / GUI.window_height;

      for (int y = 0; y < GUI.window_height; y++)
	{
	  register uint8 *d = (uint8 *) (GUI.image_date +
					 y * GUI.bytes_per_line);
	  register uint16 *s =
	    (uint16 *) (GUI.output_screen + yy * GUI.output_pitch);
	  y_error += y_fraction;
	  while (y_error >= 0x10000)
	    {
	      yy++;
	      y_error -= 0x10000;
	    }
	  x_error = 0;
	  if (GUI.RedShift < GUI.BlueShift)
	    {
	      // Order is RGB
	      for (register int x = 0; x < GUI.window_width; x++)
		{
		  uint32 pixel = *s;
		  *d++ = (pixel >> (11 - 3)) & 0xf8;
		  *d++ = (pixel >> (6 - 3)) & 0xf8;
		  *d++ = (pixel & 0x1f) << 3;

		  x_error += x_fraction;
		  while (x_error >= 0x10000)
		    {
		      s++;
		      x_error -= 0x10000;
		    }
		}
	    }
	  else
	    {
	      // Order is BGR
	      for (register int x = 0; x < GUI.window_width; x++)
		{
		  uint32 pixel = *s;
		  *d++ = (pixel & 0x1f) << 3;
		  *d++ = (pixel >> (6 - 3)) & 0xf8;
		  *d++ = (pixel >> (11 - 3)) & 0xf8;

		  x_error += x_fraction;
		  while (x_error >= 0x10000)
		    {
		      s++;
		      x_error -= 0x10000;
		    }
		}
	    }
	}
    }
}

void
Convert16To8 (int width, int height)
{
  if (!GUI.image_needs_scaling)
    {
      // Convert
      for (register int y = 0; y < height; y++)
	{
	  register uint8 *d =
	    (uint8 *) GUI.image_date + y * GUI.bytes_per_line;
	  register uint16 *s =
	    (uint16 *) (GUI.output_screen + y * GUI.output_pitch);

	  for (register int x = 0; x < width; x++)
	    *d++ = GUI.palette[*s++];
	}
    }
  else
    {
      // Scale and convert
      register uint32 x_error;
      register uint32 x_fraction;
      uint32 y_error = 0;
      uint32 y_fraction;
      int yy = 0;

      x_fraction = (width * 0x10000) / GUI.window_width;
      y_fraction = (height * 0x10000) / GUI.window_height;

      for (int y = 0; y < GUI.window_height; y++)
	{
	  register uint8 *d =
	    (uint8 *) GUI.image_date + y * GUI.bytes_per_line;
	  register uint16 *s =
	    (uint16 *) (GUI.output_screen + yy * GUI.output_pitch);
	  y_error += y_fraction;
	  while (y_error >= 0x10000)
	    {
	      yy++;
	      y_error -= 0x10000;
	    }
	  x_error = 0;
	  for (register int x = 0; x < GUI.window_width; x++)
	    {
	      *d++ = GUI.palette[*s];

	      x_error += x_fraction;
	      while (x_error >= 0x10000)
		{
		  s++;
		  x_error -= 0x10000;
		}
	    }
	}
    }
}

void
Convert8To16 (int width, int height)
{
  uint32 levels[32];

  for (int l = 0; l < 32; l++)
    levels[l] = (l * IPPU.MaxBrightness) >> 4;

  if (!GUI.image_needs_scaling)
    {
      // Convert
      for (register int y = 0; y < height; y++)
	{
	  register uint16 *d =
	    (uint16 *) (GUI.image_date + y * GUI.bytes_per_line);
	  register uint8 *s = GUI.output_screen + y * GUI.output_pitch;

	  for (register int x = 0; x < width; x++)
	    {
	      uint32 pixel = PPU.CGDATA[*s++];
	      *d++ = (levels[pixel & 0x1f] << GUI.RedShift) |
		(levels[(pixel >> 5) & 0x1f] << GUI.GreenShift) |
		(levels[(pixel >> 10) & 0x1f] << GUI.BlueShift);
	    }
	}
    }
  else
    {
      // Scale and convert
      register uint32 x_error;
      register uint32 x_fraction;
      uint32 y_error = 0;
      uint32 y_fraction;
      int yy = 0;

      x_fraction = (width * 0x10000) / GUI.window_width;
      y_fraction = (height * 0x10000) / GUI.window_height;

      for (int y = 0; y < GUI.window_height; y++)
	{
	  register uint16 *d = (uint16 *) (GUI.image_date +
					   y * GUI.bytes_per_line);
	  register uint8 *s = GUI.output_screen + yy * GUI.output_pitch;
	  y_error += y_fraction;
	  while (y_error >= 0x10000)
	    {
	      yy++;
	      y_error -= 0x10000;
	    }
	  x_error = 0;
	  for (register int x = 0; x < GUI.window_width; x++)
	    {
	      uint32 pixel = PPU.CGDATA[*s];
	      *d++ = (levels[pixel & 0x1f] << GUI.RedShift) |
		(levels[(pixel >> 5) & 0x1f] << GUI.GreenShift) |
		(levels[(pixel >> 10) & 0x1f] << GUI.BlueShift);

	      x_error += x_fraction;
	      while (x_error >= 0x10000)
		{
		  s++;
		  x_error -= 0x10000;
		}
	    }
	}
    }
}

void
S9xTextMode ()
{
  XAutoRepeatOn (GUI.display);
}

void
S9xGraphicsMode ()
{
  //XAutoRepeatOff (GUI.display);
}

void
S9xParseDisplayArg (char **argv, int &ind, int)
{
  if (strncasecmp (argv[ind], "-y", 2) == 0)
    {
      Settings.SixteenBit = TRUE;
      Settings.SupportHiRes = TRUE;
      Settings.ForceTransparency = TRUE;
      switch (argv[ind][2])
	{
	case 0:
	  GUI.interpolate = TRUE;
	  break;
	case '1':
	  GUI.interpolate = TRUE;
	  break;
	case '2':
	  GUI.interpolate = 2;
	  break;
	case '3':
	  GUI.interpolate = 3;
	  break;
	case '4':
	  GUI.interpolate = 4;
	  break;
	case '5':
	  GUI.interpolate = 5;
	  break;
	}
    }
  else if (strncasecmp (argv[ind], "-GUI.interpolate", 12) == 0)
    {
      Settings.SixteenBit = TRUE;
      Settings.SupportHiRes = TRUE;
      Settings.ForceTransparency = TRUE;
      switch (argv[ind][12])
	{
	case 0:
	  GUI.interpolate = TRUE;
	  break;
	case '1':
	  GUI.interpolate = TRUE;
	  break;
	case '2':
	  GUI.interpolate = 2;
	  break;
	case '3':
	  GUI.interpolate = 3;
	  break;
	case '4':
	  GUI.interpolate = 4;
	  break;
	case '5':
	  GUI.interpolate = 5;
	  break;
	}
    }
  else
    if (strcasecmp (argv[ind], "-scale") == 0 ||
	strcasecmp (argv[ind], "-sc") == 0)
    {
      GUI.scale = TRUE;
    }
  else
    S9xUsage ();
}

void
S9xExtraUsage ()
{
}

void
S9xMessage (int /*type */ , int /*number */ , const char *message)
{
#define MAX_MESSAGE_LEN (36 * 3)

  static char buffer[MAX_MESSAGE_LEN + 1];

  fprintf (stdout, "%s\n", message);
  strncpy (buffer, message, MAX_MESSAGE_LEN);
  buffer[MAX_MESSAGE_LEN] = 0;
  S9xSetInfoString (buffer);
}

// MYO
void S9xHandlePortCommand(s9xcommand_t cmd, int16 data1, int16 data2)
{
	return;
}

const char * S9xStringInput (const char *message)
{
	static char	buffer[256];

	printf("%s: ", message);
	fflush(stdout);

	if (fgets(buffer, sizeof(buffer) - 2, stdin))
		return (buffer);

	return (NULL);
}

bool S9xPollButton(uint32 id, bool *pressed)
{
  return FALSE;
}

bool S9xPollPointer(uint32 id, int16 *x, int16 *y)
{
  return FALSE;
}

bool S9xPollAxis(uint32 id, int16 *value)
{
  return FALSE;
}

bool8 S9xContinueUpdate(int Width, int Height)
{
  return TRUE;
}

void
S9xDisplayStateChange (const char *str, bool8 on)
{
  static char string[100];

  sprintf (string, "%s %s", str, on ? "on" : "off");
  S9xInfoMessage (string);
}
