#include "snes9x.h"
#include "gfx.h"
#include "gfxutil.h"
#include "snesapu.h"
#include "uosnesw.h"
#include "windraw.h"
#include "render.h"
#include "commutil.h"
#include "cfgdlg.h"
#include "uow_draw.h"
#include "uow_sound.h"
#include <ctype.h>

static char tmpcharbuf[256];

//proc
static void SettingsDlgCtrlSet (HWND hWnd);
static void Scale2X16RadioCtrlSet (HWND hWnd);
static void Scale2X8RadioCtrlSet (HWND hWnd);
static void SettingsDlgSetSubClass (HWND hWnd);
static void SettingsDlgReleaseSubClass ();

//sound settings
static int Options_SoundRate_cpy;
static int Options_SoundBit_cpy;
static int Options_SoundChn_cpy;
static int Settings_SoundBufferSize_cpy;
static int Settings_SoundMixInterval_cpy;
static BOOL GUI_DSuse_cpy;
static BOOL GUI_DirectSoundNotify_cpy;
static BOOL GUI_SoundGlobalFocus_cpy;

//Display settings
static int GUI_FilterMethod_cpy;
static int GUI_FullscreenDisplayModeIndex_cpy;
static BOOL Renderer_ForceWindowedDirectDrawClipper_cpy;
static BOOL GUI_DoubleBuffered_cpy;
static BOOL Renderer_OffScreenVRAM_cpy;
static BOOL Renderer_BackBufferVRAM_cpy;
static int Renderer_wait_for_vsync_cpy;

//Thread
static DWORD GUI_PriorityClass_cpy;
static int GUI_ThreadPriority_cpy;

//Window
static BOOL GUI_WindowTopMost_cpy;
static BOOL GUI_WindowTopMostAlways_cpy;
static BOOL GUI_AdjustWindow43_cpy;
static int GUI_statusbar_cpy;


//Valuable extern
extern BOOL GUI_SupportHiRes_cpy;
extern bool8 Settings_SixteenBit_cpy;
extern bool8 Settings_GFXengine2_cpy;
extern bool8 Settings_Transparency_cpy;

/*****************
 default settings
*****************/
extern int GUI_FilterMethod_def;
extern BOOL GUI_DoubleBuffered_def;
extern DWORD GUI_PriorityClass_def;
extern int GUI_ThreadPriority_def;
extern BOOL GUI_WindowTopMost_def;
extern BOOL GUI_WindowTopMostAlways_def;
extern BOOL GUI_AdjustWindow43_def;
extern int GUI_statusbar_def;
extern BOOL GUI_DSuse_def;
extern BOOL GUI_DirectSoundNotify_def;
extern BOOL GUI_SoundGlobalFocus_def;

static void SetVideoMode()
{
  GUI.FilterMethod = GUI_FilterMethod_cpy;
  Renderer.OffScreenVRAM = Renderer_OffScreenVRAM_cpy;

  if (GUI.FullScreen)
    {
      if (!(errConfigSwitch = !SwitchScreenMode (TRUE)))
	GUI.FullScreen = TRUE;
    }
  else
    {
      if (!(errConfigSwitch = !SwitchScreenMode (TRUE)))
	GUI.FullScreen = FALSE;
    }

  if (!errConfigSwitch && (Settings.StopEmulation || Settings.SPCPlaying))
    S9xDrawScreenData ();
}


void
S9xwSettingsLoad ()
{
  //sound settings
  Options_SoundRate_cpy = Options.SoundRate;
  Options_SoundBit_cpy = Options.SoundBit;
  Options_SoundChn_cpy = Options.SoundChn;
  GUI_SoundGlobalFocus_cpy = GUI.SoundGlobalFocus;
  Settings_SoundBufferSize_cpy = Settings.SoundBufferSize;
  Settings_SoundMixInterval_cpy = Settings.SoundMixInterval;

  //Display settings
  GUI_FilterMethod_cpy = GUI.FilterMethod;
  GUI_FullscreenDisplayModeIndex_cpy = GUI.FullscreenDisplayModeIndex;
  Renderer_ForceWindowedDirectDrawClipper_cpy =
    Renderer.ForceWindowedDirectDrawClipper;
  GUI_DoubleBuffered_cpy = GUI.DoubleBuffered;
  Renderer_wait_for_vsync_cpy = Renderer.wait_for_vsync;
  Renderer_OffScreenVRAM_cpy = Renderer.OffScreenVRAM;
  Renderer_BackBufferVRAM_cpy = Renderer.BackBufferVRAM;
  GUI_AdjustWindow43_cpy = GUI.AdjustWindow43;

  //Thread
  GUI_PriorityClass_cpy = GUI.PriorityClass;
  GUI_ThreadPriority_cpy = GUI.ThreadPriority;

  //Stream API
  GUI_DSuse_cpy = GUI.DSuse;
  GUI_DirectSoundNotify_cpy = GUI.DirectSoundNotify;

  //Window
  GUI_WindowTopMost_cpy = GUI.WindowTopMost;
  GUI_WindowTopMostAlways_cpy = GUI.WindowTopMostAlways;
  GUI_statusbar_cpy = GUI.statusbar;
}

static HWND hSettingsDlg;

void
S9xwDefaultSettingsLoad ()
{
  //sound settings
  Options_SoundRate_cpy = DefaultOptions.SoundRate;
  Options_SoundBit_cpy = DefaultOptions.SoundBit;
  Options_SoundChn_cpy = DefaultOptions.SoundChn;
  GUI_SoundGlobalFocus_cpy = GUI_SoundGlobalFocus_def;
  Settings_SoundBufferSize_cpy = DefaultSettings.SoundBufferSize;
  Settings_SoundMixInterval_cpy = DefaultSettings.SoundMixInterval;

  //Display settings
  GUI_FilterMethod_cpy = GUI_FilterMethod_def;
  GUI_FullscreenDisplayModeIndex_cpy = GetDefaultDisplayMode();
  Renderer_ForceWindowedDirectDrawClipper_cpy =
    Renderer_ForceWindowedDirectDrawClipper_def;
  GUI_DoubleBuffered_cpy = GUI_DoubleBuffered_def;
  Renderer_wait_for_vsync_cpy = Renderer_wait_for_vsync_def;
  Renderer_OffScreenVRAM_cpy = Renderer_OffScreenVRAM_def;
  Renderer_BackBufferVRAM_cpy = Renderer_BackBufferVRAM_def;
  GUI_AdjustWindow43_cpy = GUI_AdjustWindow43_def;

  //Stream API
  GUI_DSuse_cpy = GUI_DSuse_def;
  GUI_DirectSoundNotify_cpy = GUI_DirectSoundNotify_def;

  //Thread
  GUI_PriorityClass_cpy = GUI_PriorityClass_def;
  GUI_ThreadPriority_cpy = GUI_ThreadPriority_def;

  //Window
  GUI_WindowTopMost_cpy = GUI_WindowTopMost_def;
  GUI_WindowTopMostAlways_cpy = GUI_WindowTopMostAlways_def;
  GUI_statusbar_cpy = GUI_statusbar_def;

  SettingsDlgCtrlSet (hSettingsDlg);
}

void
S9xwSettingsSave ()
{
  int tmpnumval;

  Renderer.ForceWindowedDirectDrawClipper =
    Renderer_ForceWindowedDirectDrawClipper_cpy;
  GUI.DoubleBuffered = GUI_DoubleBuffered_cpy;
  Renderer.BackBufferVRAM = Renderer_BackBufferVRAM_cpy;
  Settings.Transparency = Settings_Transparency_cpy;

  if(Renderer.wait_for_vsync != Renderer_wait_for_vsync_cpy)
    {
      Renderer.wait_for_vsync = Renderer_wait_for_vsync_cpy;
      if(!SetupVsyncTimer(Renderer.wait_for_vsync, VsyncTimerProc, NULL))
	{
	  MessageBox (GUI.hWnd,
		      "The vsync timer failed to initialize, and set to default(None).",
		      "uosnes - vsync timer", MB_OK | MB_ICONINFORMATION);
	  Renderer.wait_for_vsync = Renderer_wait_for_vsync_cpy = Renderer_wait_for_vsync_def;

	  if(hSettingsDlg)
	    SendMessage (GetDlgItem (hSettingsDlg, IDC_COMBO_VSYNCMETHOD),
			 CB_SETCURSEL, (WPARAM) Renderer_wait_for_vsync_cpy, 0);
	}
    }

  if (GUI.FilterMethod != GUI_FilterMethod_cpy ||
      Renderer.OffScreenVRAM != Renderer_OffScreenVRAM_cpy)
    SetVideoMode();

  Settings.SixteenBit = Settings_SixteenBit_cpy;
  GUI.SupportHiRes = GUI_SupportHiRes_cpy;

  if (!GUI.DoubleRendering)
    Settings.SupportHiRes = FALSE;
  else
    Settings.SupportHiRes = GUI.SupportHiRes;

  Settings.GFXengine2 = Settings_GFXengine2_cpy;

  GUI.FullscreenDisplayModeIndex = GUI_FullscreenDisplayModeIndex_cpy;

  if (Options.SoundRate != Options_SoundRate_cpy
      ||
      Options.SoundBit != Options_SoundBit_cpy
      ||
      Options.SoundChn != Options_SoundChn_cpy
      ||
      Settings.SoundBufferSize != Settings_SoundBufferSize_cpy
      ||
      Settings.SoundMixInterval != Settings_SoundMixInterval_cpy
      ||
      GUI.DSuse != GUI_DSuse_cpy
      ||
      GUI.SoundGlobalFocus != GUI_SoundGlobalFocus_cpy
      || GUI.DirectSoundNotify != GUI_DirectSoundNotify_cpy)
    {
      Options.SoundRate = Options_SoundRate_cpy;
      Options.SoundBit = Options_SoundBit_cpy;
      Options.SoundChn = Options_SoundChn_cpy;
      Settings.SoundBufferSize = Settings_SoundBufferSize_cpy;
      Settings.SoundMixInterval = Settings_SoundMixInterval_cpy;
      GUI.DSuse = GUI_DSuse_cpy;
      GUI.SoundGlobalFocus = GUI_SoundGlobalFocus_cpy;
      GUI.DirectSoundNotify = GUI_DirectSoundNotify_cpy;
      uowSetupSound (SoundRates[Options.SoundRate],
		     Options.SoundBit == 16, Options.SoundChn == 2, GUI.hWnd);
      GUI_DSuse_cpy = GUI.DSuse;
      GUI_DirectSoundNotify_cpy = GUI.DirectSoundNotify;
    }

  GUI.PriorityClass = GUI_PriorityClass_cpy;
  GUI.ThreadPriority = GUI_ThreadPriority_cpy;
  GUI.WindowTopMost = GUI_WindowTopMost_cpy;
  GUI.WindowTopMostAlways = GUI_WindowTopMostAlways_cpy;
  GUI.AdjustWindow43 = GUI_AdjustWindow43_cpy;

  if (GUI.statusbar != GUI_statusbar_cpy)
    {
      GUI.statusbar = GUI_statusbar_cpy;
      uowSetupStatusBar (GUI.statusbar);
    }
}

static const char *FilterMethodStr[] = {
  "None",
  "Software Scale 2X 3X 4X",
  "Scanline 2X",
  "2xSaI 2X",
  "Super 2xSaI 2X",
  "SuperEagle 2X",
  "TV Mode 2X",
  "HQ2X 3X 4X",
  "LQ2X",
  "EPX 2X",
  "Blargg's NTSC (Composite) 4:3 2X",
  "Blargg's NTSC (S-Video) 4:3 2X",
  "Blargg's NTSC (RGB) 4:3 2X"
};

static const char *vsyncMethodStr[] = {
  "None",
  "WaitForVerticalBlank()",
  "GetScanLine() and Timer",
  "Thread Render"
};

static const char *SoundRateStr[] = {
  "8000Hz",
  "11025Hz",
  "22050Hz",
  "32000Hz",
  "44100Hz",
  "48000Hz" 
};

static const char *SoundBitStr[] = { "8bit", "16bit" };

static const char *SoundChnStr[] = { "mono", "stereo" };

static const char *SoundBufferLengthStr[] = {
  "2 interval",
  "3 interval",
  "4 interval",
  "5 interval",
  "6 interval",
  "7 interval",
  "8 interval",
  "9 interval"
};

static const char *SoundAPI[] = {
  "waveOutSound",
  "DirectSound"
};

static const char *PriorityClassStr[] = {
  "IDLE_PRIORITY_CLASS",
  "NORMAL_PRIORITY_CLASS",
  "HIGH_PRIORITY_CLASS",
  "REALTIME_PRIORITY_CLASS"
};

static const char *ThreadPriorityStr[] = {
  "THREAD_PRIORITY_LOWEST",
  "THREAD_PRIORITY_BELOW_NORMAL",
  "THREAD_PRIORITY_NORMAL",
  "THREAD_PRIORITY_ABOVE_NORMAL",
  "THREAD_PRIORITY_HIGHEST",
  "THREAD_PRIORITY_IDLE",
  "THREAD_PRIORITY_TIME_CRITICAL"
};


LRESULT CALLBACK
SettingsDlgProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
    {
    case WM_INITDIALOG:
      {
	int i;
	hSettingsDlg = hWnd;

	//FilterMethod
	for (i = 0; i < 13; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_FILTERMETHOD),
		       CB_INSERTSTRING, (WPARAM) i, (LPARAM) FilterMethodStr[i]);

	//Display Mode
	for (i = 0; i < Renderer.dm.size(); i++)
	  {
	    wsprintf(tmpcharbuf, "%dx%d %dbpp %dHz",
		     Renderer.dm[i].Width, Renderer.dm[i].Height,
		     Renderer.dm[i].Depth, Renderer.dm[i].Rate);
	    SendMessage (GetDlgItem (hWnd, IDC_COMBO_FULLSCREEN),
			 CB_INSERTSTRING, (WPARAM) i, (LPARAM) tmpcharbuf);
	  }

	//vsync Method
	for (i = 0; i < 4; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_VSYNCMETHOD),
		       CB_INSERTSTRING, (WPARAM) i, (LPARAM) vsyncMethodStr[i]);

	//Sound Rate
	for (i = 0; i < 6; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDRATE),
		       CB_INSERTSTRING, (WPARAM) i, (LPARAM) SoundRateStr[i]);

	//Sound Bit
	for (i = 0; i < 2; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDBIT),
		       CB_INSERTSTRING, (WPARAM) i, (LPARAM) SoundBitStr[i]);

	//Sound Chn
	for (i = 0; i < 2; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDCHN),
		       CB_INSERTSTRING, (WPARAM) i, (LPARAM) SoundChnStr[i]);

	//SoundBufferLengthR{{bNXɃeLXgZbg
	for (i = 0; i < 8; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDBUFFERLENGTH),
		       CB_INSERTSTRING,
		       (WPARAM) i, (LPARAM) SoundBufferLengthStr[i]);

	//Stream API
	for (i = 0; i < 2; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDAPI),
		       CB_INSERTSTRING, (WPARAM) i, (LPARAM) SoundAPI[i]);

	//Thread
	for (i = 0; i < 4; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_PRIORITYCLASS),
		       CB_INSERTSTRING,
		       (WPARAM) i, (LPARAM) PriorityClassStr[i]);


	for (i = 0; i < 7; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_THREADPRIORITY),
		       CB_INSERTSTRING,
		       (WPARAM) i, (LPARAM) ThreadPriorityStr[i]);

	SettingsDlgCtrlSet (hWnd);
      }
      break;

    case WM_COMMAND:
      switch (LOWORD (wParam))
	{
	case IDC_COMBO_FULLSCREEN:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      GUI_FullscreenDisplayModeIndex_cpy = i;
	      if(GUI_FullscreenDisplayModeIndex_cpy != GUI.FullscreenDisplayModeIndex)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_FILTERMETHOD:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      GUI_FilterMethod_cpy = i;
	      if(GUI_FilterMethod_cpy != GUI.FilterMethod)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_VSYNCMETHOD:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      Renderer_wait_for_vsync_cpy = i;
	      if(Renderer_wait_for_vsync_cpy != Renderer.wait_for_vsync)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Renderer_ForceWindowedDirectDrawClipper_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_DOUBLEBUFFERING:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_DoubleBuffered_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_OFFSCREENVRAM:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Renderer_OffScreenVRAM_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_BACKBUFFERVRAM:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Renderer_BackBufferVRAM_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_WINDOWTOPMOST:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_WindowTopMost_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      EnableWindow (GetDlgItem
			    (hWnd, IDC_CHECK_WINDOWTOPMOST_ALWAYS),
			    GUI_WindowTopMost_cpy ? TRUE : FALSE);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_WINDOWTOPMOST_ALWAYS:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_WindowTopMostAlways_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_ADJUSTWINDOW43:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_AdjustWindow43_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_STATUSBAR:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      BOOL tmpVal = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      GUI_statusbar_cpy = tmpVal ? 1 : 0;
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_DIRECTSOUNDNOTIFY:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_DirectSoundNotify_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_SOUNDGLOBALFOCUS:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_SoundGlobalFocus_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;


	case IDC_COMBO_SOUNDRATE:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      Options_SoundRate_cpy = i;
	      if(Options_SoundRate_cpy != Options.SoundRate)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_SOUNDBIT:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      switch(i)
		{
		case 0:
		  Options_SoundBit_cpy = 8;
		  break;

		case 1:
		  Options_SoundBit_cpy = 16;
		  break;

		default:
		  Options_SoundBit_cpy = Options.SoundBit;
		  break;
		}

	      if(Options_SoundBit_cpy != Options.SoundBit)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_SOUNDCHN:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      switch(i)
		{
		case 0:
		  Options_SoundChn_cpy = 1;
		  break;

		case 1:
		  Options_SoundChn_cpy = 2;
		  break;

		default:
		  Options_SoundChn_cpy = Options.SoundChn;
		  break;
		}

	      if(Options_SoundChn_cpy != Options.SoundChn)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_SOUNDBUFFERLENGTH:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      Settings_SoundBufferSize_cpy = i + 2;
	      if(Settings_SoundBufferSize_cpy != Settings.SoundBufferSize)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_SOUNDAPI:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      GUI_DSuse_cpy = (BOOL) i;
	      if(GUI_DSuse_cpy != GUI.DSuse)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_PRIORITYCLASS:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      switch(i)
		{
		case 0:
		  GUI_PriorityClass_cpy = IDLE_PRIORITY_CLASS;
		  break;

		case 1:
		  GUI_PriorityClass_cpy = NORMAL_PRIORITY_CLASS;
		  break;

		case 2:
		  GUI_PriorityClass_cpy = HIGH_PRIORITY_CLASS;
		  break;

		case 3:
		  GUI_PriorityClass_cpy = REALTIME_PRIORITY_CLASS;
		  break;
		}
	      if(GUI_PriorityClass_cpy != GUI.PriorityClass)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_THREADPRIORITY:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i =
		(int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
				   CB_GETCURSEL,
				   0,
				   0);
	      switch(i)
		{
		case 0:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_LOWEST;
		  break;

		case 1:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_BELOW_NORMAL;
		  break;

		case 2:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_NORMAL;
		  break;

		case 3:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_ABOVE_NORMAL;
		  break;

		case 4:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_HIGHEST;
		  break;

		case 5:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_IDLE;
		  break;

		case 6:
		  GUI_ThreadPriority_cpy = THREAD_PRIORITY_TIME_CRITICAL;
		  break;
		}
	      if(GUI_ThreadPriority_cpy != GUI.ThreadPriority)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;


	case IDC_EDIT_SOUNDMIXINTERVAL:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_SOUNDMIXINTERVAL),
			     tmpcharbuf, 255);
	      int gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 1 && gv <= 999 &&
		 gv != Settings_SoundMixInterval_cpy)
		{
		  Settings_SoundMixInterval_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	case IDC_BUTTON_SETVIDEOMODE:
	  SetVideoMode();
	  break;

	default:
	  break;
	}
      break;

    case WM_NOTIFY:
      switch (((NMHDR *) lParam)->code)
	{
	case PSN_APPLY:
	  ConfigDlgSettingsSave ();
	  break;

	case PSN_HELP:
	  ConfigDlgDefaultSettingsLoad (hWnd);
	  break;

	case PSN_QUERYCANCEL:
	  SettingsDlgReleaseSubClass ();
	  break;

	case PSN_SETACTIVE:
	  StartPageIndex = 0;
	  SettingsDlgSetSubClass (hWnd);
	  break;


	case PSN_KILLACTIVE:
	  SettingsDlgReleaseSubClass ();
	  break;

	default:
	  break;
	}
      break;

    case WM_DESTROY:
      //note, PSN_KILLACTIVE is not sent.
      SettingsDlgReleaseSubClass ();
      hSettingsDlg = NULL;
      break;
      
    default:
      break;
    }
  return FALSE;
}



static void
SettingsDlgCtrlSet (HWND hWnd)
{
  if (!hWnd)
    return;

  int i, c;

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_FULLSCREEN),
	       CB_SETCURSEL, (WPARAM) GUI_FullscreenDisplayModeIndex_cpy, 0);

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_FILTERMETHOD),
	       CB_SETCURSEL, (WPARAM) GUI_FilterMethod_cpy, 0);

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_VSYNCMETHOD),
	       CB_SETCURSEL, (WPARAM) Renderer_wait_for_vsync_cpy, 0);

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDRATE),
	       CB_SETCURSEL, (WPARAM) Options_SoundRate_cpy, 0);

  switch (Options_SoundBit_cpy)
    {
    case 8:
      i = 0;
      break;

    case 16:
      i = 1;
      break;

    default:
      i = 1;
      break;
    }

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDBIT),
	       CB_SETCURSEL, (WPARAM) i, 0);

  switch (Options_SoundChn_cpy)
    {
    case 1:
      i = 0;
      break;

    case 2:
      i = 1;
      break;

    default:
      i = 1;
      break;
    }

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDCHN),
	       CB_SETCURSEL, (WPARAM) i, 0);

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDBUFFERLENGTH),
	       CB_SETCURSEL, (WPARAM) (Settings_SoundBufferSize_cpy - 2), 0);

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDAPI),
	       CB_SETCURSEL, (WPARAM) GUI_DSuse_cpy, 0);

  //MixInterval
  SendMessage (GetDlgItem (hWnd, IDC_EDIT_SOUNDMIXINTERVAL),
	       EM_SETLIMITTEXT, 3, 0);
  wsprintf (tmpcharbuf, "%d", Settings_SoundMixInterval_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_SOUNDMIXINTERVAL),
		 tmpcharbuf);

  SendMessage (GetDlgItem
	       (hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER),
	       BM_SETCHECK,
	       (WPARAM) (Renderer_ForceWindowedDirectDrawClipper_cpy ?
			 BST_CHECKED : BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_DOUBLEBUFFERING),
	       BM_SETCHECK,
	       (WPARAM) (GUI_DoubleBuffered_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_OFFSCREENVRAM),
	       BM_SETCHECK,
	       (WPARAM) (Renderer_OffScreenVRAM_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_BACKBUFFERVRAM),
	       BM_SETCHECK,
	       (WPARAM) (Renderer_BackBufferVRAM_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_WINDOWTOPMOST),
	       BM_SETCHECK,
	       (WPARAM) (GUI_WindowTopMost_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_WINDOWTOPMOST_ALWAYS),
	       BM_SETCHECK,
	       (WPARAM) (GUI_WindowTopMostAlways_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_STATUSBAR),
	       BM_SETCHECK,
	       (WPARAM) (GUI_statusbar_cpy ==
			 1 ? BST_CHECKED : BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_ADJUSTWINDOW43),
	       BM_SETCHECK,
	       (WPARAM) (GUI_AdjustWindow43_cpy ? BST_CHECKED : BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_DIRECTSOUNDNOTIFY),
	       BM_SETCHECK,
	       (WPARAM) (GUI_DirectSoundNotify_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_SOUNDGLOBALFOCUS),
	       BM_SETCHECK,
	       (WPARAM) (GUI_SoundGlobalFocus_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  EnableWindow (GetDlgItem (hWnd, IDC_CHECK_DIRECTSOUNDNOTIFY),
		GUI_DSuse_cpy == 0 ? FALSE : TRUE);
  EnableWindow (GetDlgItem (hWnd, IDC_CHECK_WINDOWTOPMOST_ALWAYS),
		GUI_WindowTopMost_cpy ? TRUE : FALSE);

  if (GUI_PriorityClass_cpy == IDLE_PRIORITY_CLASS)
    i = 0;
  else if (GUI_PriorityClass_cpy == NORMAL_PRIORITY_CLASS)
    i = 1;
  else if (GUI_PriorityClass_cpy == HIGH_PRIORITY_CLASS)
    i = 2;
  else if (GUI_PriorityClass_cpy == REALTIME_PRIORITY_CLASS)
    i = 3;
  else
    i = 1;

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_PRIORITYCLASS),
	       CB_SETCURSEL, (WPARAM) i, 0);

  if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_LOWEST)
    i = 0;
  else if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_BELOW_NORMAL)
    i = 1;
  else if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_NORMAL)
    i = 2;
  else if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_ABOVE_NORMAL)
    i = 3;
  else if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_HIGHEST)
    i = 4;
  else if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_IDLE)
    i = 5;
  else if (GUI_ThreadPriority_cpy == THREAD_PRIORITY_TIME_CRITICAL)
    i = 6;
  else
    i = 2;

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_THREADPRIORITY),
	       CB_SETCURSEL, (WPARAM) i, 0);
}


static WNDPROC Edit_MixInterval;
static HWND hEdit_MixInterval;

static LRESULT CALLBACK
SettingsEditSubProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
    {
    case WM_CHAR:
      if ((LOWORD (wParam) >= '0' && LOWORD (wParam) <= '9')
	  || LOWORD (wParam) == VK_BACK)
	break;

      return 0;

    default:
      break;
    }

  if (hWnd == hEdit_MixInterval && Edit_MixInterval)
    return (CallWindowProc (Edit_MixInterval, hWnd, uMsg, wParam, lParam));

  return 0;
}

static void
SettingsDlgSetSubClass (HWND hWnd)
{
  hEdit_MixInterval = GetDlgItem (hWnd, IDC_EDIT_SOUNDMIXINTERVAL);
  Edit_MixInterval = (WNDPROC) GetWindowLong (hEdit_MixInterval, GWL_WNDPROC);
  SetWindowLong (hEdit_MixInterval, GWL_WNDPROC, (LONG) SettingsEditSubProc);
}

static void
SettingsDlgReleaseSubClass ()
{
  SetWindowLong (hEdit_MixInterval, GWL_WNDPROC, (LONG) Edit_MixInterval);
}

