#include "snes9x.h"
#include "soundux.h"
#include "ppu.h"
#include "apu_orig.h"
#include "snesapu.h"
#include "display.h"
#include "uosnesw.h"
#include "windraw.h"
#include "cfgdlg.h"
#include <ctype.h>

static char tmpcharbuf[256];

//proc
static void SNESEmulationDlgCtrlSet (HWND hWnd);
static void SneseDlgSetSubClass (HWND hWnd);
static void SneseDlgReleaseSubClass ();

//graphics
BOOL GUI_SupportHiRes_cpy;
bool8 Settings_SixteenBit_cpy;
bool8 Settings_Transparency_cpy;
bool8 Settings_GFXengine2_cpy;
static bool8 Settings_Mode7Interpolate_cpy;
static bool8 Settings_BGLayering_cpy;
static bool8 Settings_DisableGraphicWindows_cpy;
static bool8 Settings_DisableHDMA_cpy;
static bool8 Settings_BlockInvalidVRAMAccessMaster_cpy;

//sound
static bool8 Settings_AltSampleDecode_cpy;
static bool8 Settings_FixFrequency_cpy;
static int Settings_InterpolatedSound_cpy;
static bool8 Settings_LowPassFilter_cpy;
static bool8 Settings_DisableSampleCaching_cpy;
static bool8 Settings_NextAPUEnabled_cpy;
static bool8 Settings_DisableSoundEcho_cpy;
static bool8 Settings_ReverseStereo_cpy;
static bool8 Settings_DisableMasterVolume_cpy;
static bool8 Settings_SoundEnvelopeHeightReading_cpy;
static uint8 Settings_SoundSkipMethod_cpy;

//CPU
static bool8 Settings_ShutdownMaster_cpy;
static bool8 Settings_OldTimingMaster_cpy;
static bool8 Settings_DisableIRQCycleCount_cpy;
static int32 Settings_HDMATimingHack_cpy;
static int32 Settings_H_MaxTimingHack_cpy;
static int32 Settings_smp_master_clock_cpy;
static int32 Settings_APUOneCycle_cpy;

//timing
static uint32 Settings_SkipFrames_cpy;
static uint32 Settings_AutoMaxSkipFrames_cpy;
static uint32 Settings_FrameTimePAL_cpy;
static uint32 Settings_FrameTimeNTSC_cpy;
static bool8 Settings_SoundConstantTempo_cpy;

//auto save sram
static int32 Settings_AutoSaveDelay_cpy;	// Original

//default settings
extern BOOL GUI_SupportHiRes_def;

void
SNESEmulationSettingsLoad ()
{
  //graphics
  GUI_SupportHiRes_cpy = GUI.SupportHiRes;
  Settings_Mode7Interpolate_cpy = Settings.Mode7Interpolate;
  Settings_SixteenBit_cpy = Settings.SixteenBit;
  Settings_Transparency_cpy = Settings.Transparency;
  Settings_BGLayering_cpy = Settings.BGLayering;
  Settings_DisableGraphicWindows_cpy = Settings.DisableGraphicWindows;
  Settings_DisableHDMA_cpy = Settings.DisableHDMA;
  Settings_GFXengine2_cpy = Settings.GFXengine2;
  Settings_BlockInvalidVRAMAccessMaster_cpy = Settings.BlockInvalidVRAMAccessMaster;

  //sound
  Settings_InterpolatedSound_cpy = Settings.InterpolatedSound;
  Settings_LowPassFilter_cpy = Settings.LowPassFilter;
  Settings_DisableSampleCaching_cpy = Settings.DisableSampleCaching;
  Settings_FixFrequency_cpy = Settings.FixFrequency;
  Settings_AltSampleDecode_cpy = Settings.AltSampleDecode;
  Settings_NextAPUEnabled_cpy = Settings.NextAPUEnabled;
  Settings_DisableSoundEcho_cpy = Settings.DisableSoundEcho;
  Settings_ReverseStereo_cpy = Settings.ReverseStereo;
  Settings_DisableMasterVolume_cpy = Settings.DisableMasterVolume;
  Settings_SoundEnvelopeHeightReading_cpy =
    Settings.SoundEnvelopeHeightReading;
  Settings_SoundSkipMethod_cpy = Settings.SoundSkipMethod;

  //cpu
  Settings_ShutdownMaster_cpy = Settings.ShutdownMaster;
  Settings_OldTimingMaster_cpy = Settings.OldTimingMaster;
  Settings_DisableIRQCycleCount_cpy = Settings.DisableIRQCycleCount;	// Original
  Settings_HDMATimingHack_cpy = Settings.HDMATimingHack;
  Settings_H_MaxTimingHack_cpy = Settings.H_MaxTimingHack;
  Settings_smp_master_clock_cpy = Settings.smp_master_clock;
  Settings_APUOneCycle_cpy = Settings.APUOneCycle;

  //timing
  Settings_SkipFrames_cpy = Settings.SkipFrames;
  Settings_AutoMaxSkipFrames_cpy = Settings.AutoMaxSkipFrames;
  Settings_FrameTimePAL_cpy = Settings.FrameTimePAL;
  Settings_FrameTimeNTSC_cpy = Settings.FrameTimeNTSC;
  Settings_SoundConstantTempo_cpy = Settings.SoundConstantTempo;

  //AutoSave
  Settings_AutoSaveDelay_cpy = Settings.AutoSaveDelay;

}

static HWND hSNESEmulationDlg;

void
SNESEmulationDefaultSettingsLoad ()
{
  //graphics
  GUI_SupportHiRes_cpy = GUI_SupportHiRes_def;
  Settings_Mode7Interpolate_cpy = DefaultSettings.Mode7Interpolate;
  Settings_SixteenBit_cpy = DefaultSettings.SixteenBit;
  Settings_Transparency_cpy = DefaultSettings.Transparency;
  Settings_BGLayering_cpy = DefaultSettings.BGLayering;
  Settings_DisableGraphicWindows_cpy = DefaultSettings.DisableGraphicWindows;
  Settings_DisableHDMA_cpy = DefaultSettings.DisableHDMA;
  Settings_GFXengine2_cpy = DefaultSettings.GFXengine2;
  Settings_BlockInvalidVRAMAccessMaster_cpy = DefaultSettings.BlockInvalidVRAMAccessMaster;

  //sound
  Settings_InterpolatedSound_cpy = DefaultSettings.InterpolatedSound;
  Settings_LowPassFilter_cpy = DefaultSettings.LowPassFilter;
  Settings_DisableSampleCaching_cpy = DefaultSettings.DisableSampleCaching;
  Settings_FixFrequency_cpy = DefaultSettings.FixFrequency;
  Settings_AltSampleDecode_cpy = DefaultSettings.AltSampleDecode;
  Settings_NextAPUEnabled_cpy = DefaultSettings.NextAPUEnabled;
  Settings_DisableSoundEcho_cpy = DefaultSettings.DisableSoundEcho;
  Settings_ReverseStereo_cpy = DefaultSettings.ReverseStereo;
  Settings_DisableMasterVolume_cpy = DefaultSettings.DisableMasterVolume;
  Settings_SoundEnvelopeHeightReading_cpy =
    DefaultSettings.SoundEnvelopeHeightReading;
  Settings_SoundSkipMethod_cpy = DefaultSettings.SoundSkipMethod;

  //cpu
  Settings_ShutdownMaster_cpy = DefaultSettings.ShutdownMaster;
  Settings_OldTimingMaster_cpy = DefaultSettings.OldTimingMaster;
  Settings_DisableIRQCycleCount_cpy = DefaultSettings.DisableIRQCycleCount;
  Settings_HDMATimingHack_cpy = DefaultSettings.HDMATimingHack;
  Settings_H_MaxTimingHack_cpy = DefaultSettings.H_MaxTimingHack;
  Settings_smp_master_clock_cpy = DefaultSettings.smp_master_clock;
  Settings_APUOneCycle_cpy = DefaultSettings.APUOneCycle;

  //timing
  Settings_SkipFrames_cpy = DefaultSettings.SkipFrames;
  Settings_AutoMaxSkipFrames_cpy = DefaultSettings.AutoMaxSkipFrames;
  Settings_FrameTimePAL_cpy = DefaultSettings.FrameTimePAL;
  Settings_FrameTimeNTSC_cpy = DefaultSettings.FrameTimeNTSC;
  Settings_SoundConstantTempo_cpy = DefaultSettings.SoundConstantTempo;

  //AutoSave Original
  Settings_AutoSaveDelay_cpy = DefaultSettings.AutoSaveDelay;

  SNESEmulationDlgCtrlSet (hSNESEmulationDlg);
}


static const char APUEnabledMessage_US[] =
  "Sound emulation setting chenged to enabled\nsound will generate after Reset.";
static const char APUEnabledMessage_JP[] =
  "TEhG~[VLɐݒύX܂\nZbgɏԂɔf܂";
static const char APUDisabledMessage_US[] =
  "Sound emulation setting chenged to disabled\nsound will generate after Reset in case the setting will be enabled again, would you continue?";
static const char APUDisabledMessage_JP[] =
  "TEhG~[VɐݒύX܂\n߂ėLɂꍇ̓ZbgɏԂɔfɂȂ܂Aɂ܂H";

void
SNESEmulationSettingsSave ()
{
  short tmpVolume;
  uint32 tmpnumval;

  Settings.Mode7Interpolate = Settings_Mode7Interpolate_cpy;
  Settings.BGLayering = Settings_BGLayering_cpy;
  Settings.DisableGraphicWindows = Settings_DisableGraphicWindows_cpy;
  Settings.DisableHDMA = Settings_DisableHDMA_cpy;
  Settings.InterpolatedSound = Settings_InterpolatedSound_cpy;
  Settings.LowPassFilter = Settings_LowPassFilter_cpy;
  Settings.FixFrequency = Settings_FixFrequency_cpy;
  Settings.DisableSampleCaching = Settings_DisableSampleCaching_cpy;
  Settings.AltSampleDecode = Settings_AltSampleDecode_cpy;
  Settings.AutoMaxSkipFrames = Settings_AutoMaxSkipFrames_cpy;

  if (!Settings.StopEmulation && !Settings.BlockInvalidVRAMAccessLock)
    Settings.BlockInvalidVRAMAccess = Settings.BlockInvalidVRAMAccessMaster = Settings_BlockInvalidVRAMAccessMaster_cpy;
  else
    Settings.BlockInvalidVRAMAccessMaster = Settings_BlockInvalidVRAMAccessMaster_cpy;

  if (Settings_NextAPUEnabled_cpy != Settings.NextAPUEnabled)
    {
      if (!Settings.StopEmulation && !Settings.SPCPlaying
	  && Settings_NextAPUEnabled_cpy)
	{
	  MessageBox (hSNESEmulationDlg,
		      GUI.
		      Language == 1 ? APUEnabledMessage_JP : APUEnabledMessage_US,
		      "uosnes - Enabled sound emulation",
		      MB_OK | MB_ICONINFORMATION);
	  Settings.NextAPUEnabled = TRUE;
	}
      else if (!Settings.StopEmulation && !Settings.SPCPlaying
	       && !Settings_NextAPUEnabled_cpy)
	{
	  if (IDOK == MessageBox (hSNESEmulationDlg,
				  GUI.
				  Language == 1 ? APUDisabledMessage_JP :
				  APUDisabledMessage_US,
				  "uosnes - Disabled sound emulation",
				  MB_OKCANCEL | MB_ICONQUESTION))
	    {
	      //kill APU executing
	      IAPU.APUExecuting = Settings.APUEnabled =
		Settings.NextAPUEnabled = FALSE;
	      S9xResetAPU ();
	    }
	  else
	    {
	      Settings_NextAPUEnabled_cpy = TRUE;
	      SendMessage (GetDlgItem
			   (hSNESEmulationDlg, IDC_CHECK_APUENABLED),
			   BM_SETCHECK,
			   (WPARAM) (Settings_NextAPUEnabled_cpy ? BST_CHECKED
				     : BST_UNCHECKED), 0);
	    }
	}
      else
	Settings.APUEnabled = Settings.NextAPUEnabled =
	  Settings_NextAPUEnabled_cpy;
    }

  // SNESAPU options
  int opts_cpy =
    (Settings_ReverseStereo_cpy ? DSP_REVERSE : 0) |
    (Settings_DisableSoundEcho_cpy ? DSP_NOECHO : 0);

  int opts =
    (Settings.ReverseStereo ? DSP_REVERSE : 0) |
    (Settings.DisableSoundEcho ? DSP_NOECHO : 0);

  if (opts != opts_cpy)
    {
      Settings.DisableSoundEcho = Settings_DisableSoundEcho_cpy;
      if (Settings_ReverseStereo_cpy != Settings.ReverseStereo)
	{
	  tmpVolume = SoundData.master_volume[0];
	  SoundData.master_volume[0] = SoundData.master_volume[1];
	  SoundData.master_volume[1] = tmpVolume;
	  tmpVolume = SoundData.echo_volume[0];
	  SoundData.echo_volume[0] = SoundData.echo_volume[1];
	  SoundData.echo_volume[1] = tmpVolume;
	  Settings.ReverseStereo = Settings_ReverseStereo_cpy;
	}

      opts = opts_cpy;

      //SNESAPU
      SetAPUDSPOpt (so.stereo ? 2 : 1,
		    so.sixteen_bit ? 16 : 8,
		    so.playback_rate,
		    opts);
    }

  Settings.DisableMasterVolume = Settings_DisableMasterVolume_cpy;
  Settings.SoundEnvelopeHeightReading =
    Settings_SoundEnvelopeHeightReading_cpy;

  if (!Settings.StopEmulation && !Settings.ShutdownLock)
    Settings.Shutdown = Settings.ShutdownMaster = Settings_ShutdownMaster_cpy;
  else
    Settings.ShutdownMaster = Settings_ShutdownMaster_cpy;

  Settings.OldTimingMaster = Settings_OldTimingMaster_cpy;
  if(!Settings.StopEmulation && !Settings.HCyclesLock)
      S9xSetHCycles (Settings.H_MaxTimingHack, Settings.HDMATimingHack, Settings.OldTimingMaster);

  //HDMATimingHack
  Settings.HDMATimingHack = Settings_HDMATimingHack_cpy;
  if (!Settings.StopEmulation && !Settings.HCyclesLock)
    S9xSetHCycles (Settings.H_MaxTimingHack, Settings.HDMATimingHack, Settings.OldTimingMaster);

  //H_MaxTimingHack
  if(Settings_H_MaxTimingHack_cpy < Settings_HDMATimingHack_cpy - 100)
    Settings_H_MaxTimingHack_cpy += ((Settings_HDMATimingHack_cpy - 100) - Settings_H_MaxTimingHack_cpy);
  Settings.H_MaxTimingHack = Settings_H_MaxTimingHack_cpy;
  if (!Settings.StopEmulation && !Settings.HCyclesLock)
    S9xSetHCycles (Settings.H_MaxTimingHack, Settings.HDMATimingHack, Settings.OldTimingMaster);

  Settings.smp_master_clock = Settings_smp_master_clock_cpy;
  Settings.APUOneCycle = Settings_APUOneCycle_cpy;
  if (!Settings.StopEmulation && !Settings.SPCPlaying && !Settings.APUCycleLock)
    {
      IAPU.MasterClock = Settings.smp_master_clock;
      S9xSetAPUCycle (Settings.APUOneCycle);
    }

  Settings.DisableIRQCycleCount = Settings_DisableIRQCycleCount_cpy;	// Original

  Settings.SoundSkipMethod = Settings_SoundSkipMethod_cpy;

  Settings.SkipFrames = Settings_SkipFrames_cpy;

  //PAL timing ms
  if (!Settings.StopEmulation && !Settings.SPCPlaying
      && Settings.FrameTimePAL == Settings.FrameTime)
    {
      Settings.FrameTime = Settings_FrameTimePAL_cpy;
      if (!uowSetupFrameTimer ((DWORD) Settings.FrameTime))
	{
	  MessageBox (GUI.hWnd,
		      Languages[GUI.Language].errFrameTimer,
		      "uosnes - Frame Timer", MB_OK | MB_ICONINFORMATION);
	}
    }
  Settings.FrameTimePAL = Settings_FrameTimePAL_cpy;

  //NTSC timing ms
  if (!Settings.StopEmulation && !Settings.SPCPlaying
      && Settings.FrameTimeNTSC == Settings.FrameTime)
    {
      Settings.FrameTime = Settings_FrameTimeNTSC_cpy;
      if (!uowSetupFrameTimer ((DWORD) Settings.FrameTime))
	{
	  MessageBox (GUI.hWnd,
		      Languages[GUI.Language].errFrameTimer,
		      "uosnes - Frame Timer", MB_OK | MB_ICONINFORMATION);
	}
    }
  Settings.FrameTimeNTSC = Settings_FrameTimeNTSC_cpy;

  Settings.SoundConstantTempo = Settings_SoundConstantTempo_cpy;

  Settings.AutoSaveDelay = Settings_AutoSaveDelay_cpy;
}

static const char *InterpolatedSound_US[] = {
  "None",
  "Linear",
  "Cubic",
  "Gaussian",
  "Sinc"
};

static const char *InterpolatedSound_JP[] = {
  "",
  "`",
  "L[rbN",
  "KEX",
  "Sinc֐"
};

static const char *SoundSkipMethodStr[] =
  { "Method 0", "Method 1", "Method 2", "Method 3" };
static const char *JoyReadRateStr_US[] =
  { "1 frame", "2 frame", "4 frame", "8 frame" };
static const char *JoyReadRateStr_JP[] =
  { "1 t[", "2 t[", "4 t[", "8 t[" };

static const char *SkipFramesStr_US[] = {
  "Auto FrameSkip", "0 FrameSkip",
  "1 FrameSkip", "2 FrameSkip",
  "3 FrameSkip", "4 FrameSkip",
  "5 FrameSkip", "6 FrameSkip",
  "7 FrameSkip", "8 FrameSkip",
  "9 FrameSkip"
};

static const char *SkipFramesStr_JP[] =
  { "t[XLbv", "0 t[XLbv",
  "1 t[XLbv", "2 t[XLbv",
  "3 t[XLbv", "4 t[XLbv",
  "5 t[XLbv", "6 t[XLbv",
  "7 t[XLbv", "8 t[XLbv",
  "9 t[XLbv"
};

static const char *AutoMaxSkipFramesStr[] =
  { "Max 0", "Max 1", "Max 2", "Max 3", "Max 4", "Max 5",
  "Max 6", "Max 7", "Max 8", "Max 9", "Max 10"
};

LRESULT CALLBACK
SNESEmulationDlgProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
    {
    case WM_INITDIALOG:
      {
	hSNESEmulationDlg = hWnd;

	int i;
	for (i = 0; i < 5; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_INTERPOLATEDSOUND),
		       CB_INSERTSTRING,
		       (WPARAM) i,
		       (LPARAM) (GUI.
				 Language == 1 ? InterpolatedSound_JP[i] :
				 InterpolatedSound_US[i]));

	for (i = 0; i < 4; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDSKIPMETHOD),
		       CB_INSERTSTRING,
		       (WPARAM) i, (LPARAM) SoundSkipMethodStr[i]);

	//timing
	for (i = 0; i < 11; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_FRAMESKIP),
		       CB_INSERTSTRING,
		       (WPARAM) i,
		       (LPARAM) (GUI.
				 Language == 1 ? SkipFramesStr_JP[i] :
				 SkipFramesStr_US[i]));

	for (i = 0; i < 11; i++)
	  SendMessage (GetDlgItem (hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES),
		       CB_INSERTSTRING,
		       (WPARAM) i, (LPARAM) (AutoMaxSkipFramesStr[i]));

	SNESEmulationDlgCtrlSet (hWnd);
      }
      break;

    case WM_COMMAND:
      switch (LOWORD (wParam))
	{
	case IDC_CHECK_MODE7INTERPOLATE:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_Mode7Interpolate_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_HIRES:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      GUI_SupportHiRes_cpy = (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_GFX16:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_SixteenBit_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_TRANSPARENCY:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_Transparency_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_BGLAYERING:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_BGLayering_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_GFXW:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_DisableGraphicWindows_cpy = !(bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_HDMA:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_DisableHDMA_cpy = !(bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_BLOCKINVALIDVRAMACCESS:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_BlockInvalidVRAMAccessMaster_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;


	case IDC_CHECK_GFXENGINE2:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_GFXengine2_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_LOWPASSFILTER:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_LowPassFilter_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_FIXFREQUENCY:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_FixFrequency_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_DISABLESAMPLECACHING:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_DisableSampleCaching_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_APUENABLED:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_NextAPUEnabled_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      EnableWindow (GetDlgItem (hWnd, IDC_COMBO_SOUNDSKIPMETHOD),
			    !Settings_NextAPUEnabled_cpy ? TRUE : FALSE);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_SOUNDECHO:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_DisableSoundEcho_cpy = !(bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_REVERSESTEREO:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_ReverseStereo_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_MASTERVOLUME:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_DisableMasterVolume_cpy = !(bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_SOUNDENVELOPEHEIGHTREADING:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_SoundEnvelopeHeightReading_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_SHUTDOWN:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_ShutdownMaster_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_OLDTIMING:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_OldTimingMaster_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_CHECK_DISABLE_IRQ_CYCLECOUNT:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_DisableIRQCycleCount_cpy = (bool8) (IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;
	  // Original

	case IDC_CHECK_SOUNDCONSTANTTEMPO:
	  if (HIWORD (wParam) == BN_CLICKED)
	    {
	      Settings_SoundConstantTempo_cpy = (bool8)(IsDlgButtonChecked (hWnd, (int) LOWORD (wParam)) == BST_CHECKED);
	      ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_SOUNDSKIPMETHOD:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i = (int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
					 CB_GETCURSEL, 0, 0);
	      Settings_SoundSkipMethod_cpy = (uint8) i;
	      if (Settings_SoundSkipMethod_cpy != Settings.SoundSkipMethod)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_FRAMESKIP:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i = (int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
					 CB_GETCURSEL, 0, 0);
	      if (i == 0)
		Settings_SkipFrames_cpy = AUTO_FRAMERATE;
	      else if (i >= 1 && i <= 10)
		Settings_SkipFrames_cpy = (uint32) i;
	      else
		Settings_SkipFrames_cpy = AUTO_FRAMERATE;
	      EnableWindow (GetDlgItem
			    (hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES),
			    (Settings_SkipFrames_cpy ==
			     AUTO_FRAMERATE) ? TRUE : FALSE);
	      if (Settings_SkipFrames_cpy != Settings.SkipFrames)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_AUTOMAXSKIPFRAMES:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i = (int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
					 CB_GETCURSEL, 0, 0);
	      Settings_AutoMaxSkipFrames_cpy = (uint32) i;
	      if (Settings_AutoMaxSkipFrames_cpy != Settings.AutoMaxSkipFrames)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_COMBO_INTERPOLATEDSOUND:
	  if (HIWORD (wParam) == CBN_SELENDOK)
	    {
	      int i = (int) SendMessage (GetDlgItem (hWnd, (int) LOWORD (wParam)),
					 CB_GETCURSEL, 0, 0);
	      Settings_InterpolatedSound_cpy = i;
	      if (Settings_InterpolatedSound_cpy != Settings.InterpolatedSound)
		ConfigDlgSettingsChanged (hWnd);
	    }
	  break;

	case IDC_EDIT_HDMATIMINGHACK:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_HDMATIMINGHACK),
			     tmpcharbuf, 255);
	      int32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 0 && gv <= 350 &&
		 gv != Settings_HDMATimingHack_cpy)
		{
		  Settings_HDMATimingHack_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	case IDC_EDIT_HMAXTIMINGHACK:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_HMAXTIMINGHACK),
			     tmpcharbuf, 255);
	      int32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 0 && gv <= 999 &&
		 gv != Settings_H_MaxTimingHack_cpy)
		{
		  Settings_H_MaxTimingHack_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	case IDC_EDIT_SMP_MASTER_CLOCK:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_SMP_MASTER_CLOCK),
			     tmpcharbuf, 255);
	      int32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 1 && gv <= 99999999 &&
		 gv != Settings_smp_master_clock_cpy)
		{
		  Settings_smp_master_clock_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;


	case IDC_EDIT_APUONECYCLE:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_APUONECYCLE),
			     tmpcharbuf, 255);
	      int32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 1 && gv <= 999 &&
		 gv != Settings_APUOneCycle_cpy)
		{
		  Settings_APUOneCycle_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	case IDC_EDIT_NTSC:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_NTSC),
			     tmpcharbuf, 255);
	      uint32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 1 && gv <= 999 &&
		 gv != Settings_FrameTimeNTSC_cpy)
		{
		  Settings_FrameTimeNTSC_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	case IDC_EDIT_PAL:
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_PAL),
			     tmpcharbuf, 255);
	      uint32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 1 && gv <= 999 &&
		 gv != Settings_FrameTimePAL_cpy)
		{
		  Settings_FrameTimePAL_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	case IDC_EDIT_AUTOSAVESRAM:	// Original
	  if (HIWORD (wParam) == EN_CHANGE)
	    {
	      GetWindowText (GetDlgItem (hWnd, IDC_EDIT_AUTOSAVESRAM),
			     tmpcharbuf, 255);
	      int32 gv;
	      if(*tmpcharbuf && isdigit(*tmpcharbuf) &&
		 sscanf(tmpcharbuf, "%d", &gv) == 1 &&
		 gv >= 0 && gv <= 9999999 &&
		 gv != Settings_AutoSaveDelay_cpy)
		{
		  Settings_AutoSaveDelay_cpy = gv;
		  ConfigDlgSettingsChanged (hWnd);
		}
	    }
	  break;

	default:
	  break;
	}
      break;

    case WM_NOTIFY:
      switch (((NMHDR *) lParam)->code)
	{
	case PSN_APPLY:
	  ConfigDlgSettingsSave ();
	  break;

	case PSN_HELP:
	  ConfigDlgDefaultSettingsLoad (hWnd);
	  break;

	case PSN_QUERYCANCEL:
	  SneseDlgReleaseSubClass ();
	  break;

	case PSN_SETACTIVE:
	  StartPageIndex = 1;
	  SneseDlgSetSubClass (hWnd);
	  break;

	case PSN_KILLACTIVE:
	  SneseDlgReleaseSubClass ();
	  break;

	default:
	  break;
	}
      break;

    case WM_DESTROY:
      //note, PSN_KILLACTIVE is not sent.
      SneseDlgReleaseSubClass ();
      hSNESEmulationDlg = NULL;
      break;

    default:
      break;
    }
  return FALSE;
}

static void
SNESEmulationDlgCtrlSet (HWND hWnd)
{
  if (!hWnd)
    return;

  //sound
  SendMessage (GetDlgItem (hWnd, IDC_COMBO_INTERPOLATEDSOUND), CB_SETCURSEL,
	       (WPARAM) Settings_InterpolatedSound_cpy, 0);

  SendMessage (GetDlgItem (hWnd, IDC_CHECK_LOWPASSFILTER), BM_SETCHECK,
	       (WPARAM) (Settings_LowPassFilter_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  EnableWindow (GetDlgItem (hWnd, IDC_CHECK_LOWPASSFILTER), FALSE);

  SendMessage (GetDlgItem (hWnd, IDC_CHECK_FIXFREQUENCY), BM_SETCHECK,
	       (WPARAM) (Settings_FixFrequency_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_DISABLESAMPLECACHING), BM_SETCHECK,
	       (WPARAM) (Settings_DisableSampleCaching_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  EnableWindow (GetDlgItem (hWnd, IDC_CHECK_DISABLESAMPLECACHING),
		(!Settings_AltSampleDecode_cpy ? TRUE : FALSE));
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_APUENABLED), BM_SETCHECK,
	       (WPARAM) (Settings_NextAPUEnabled_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_SOUNDECHO), BM_SETCHECK,
	       (WPARAM) (Settings_DisableSoundEcho_cpy ? BST_UNCHECKED :
			 BST_CHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_REVERSESTEREO), BM_SETCHECK,
	       (WPARAM) (Settings_ReverseStereo_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_MASTERVOLUME), BM_SETCHECK,
	       (WPARAM) (Settings_DisableMasterVolume_cpy ? BST_UNCHECKED :
			 BST_CHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_SOUNDENVELOPEHEIGHTREADING),
	       BM_SETCHECK,
	       (WPARAM) (Settings_SoundEnvelopeHeightReading_cpy ? BST_CHECKED
			 : BST_UNCHECKED), 0);

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_SOUNDSKIPMETHOD), CB_SETCURSEL,
	       (WPARAM) Settings_SoundSkipMethod_cpy, 0);
  EnableWindow (GetDlgItem (hWnd, IDC_COMBO_SOUNDSKIPMETHOD),
		!Settings_NextAPUEnabled_cpy ? TRUE : FALSE);

  //cpu
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_SHUTDOWN), BM_SETCHECK,
	       (WPARAM) (Settings_ShutdownMaster_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);

  SendMessage (GetDlgItem (hWnd, IDC_CHECK_OLDTIMING), BM_SETCHECK,
	       (WPARAM) (Settings_OldTimingMaster_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);

  SendMessage (GetDlgItem (hWnd, IDC_CHECK_DISABLE_IRQ_CYCLECOUNT),
	       BM_SETCHECK,
	       (WPARAM) (Settings_DisableIRQCycleCount_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);

  SendMessage (GetDlgItem (hWnd, IDC_EDIT_HDMATIMINGHACK), EM_SETLIMITTEXT,
	       3, 0);
  wsprintf (tmpcharbuf, "%d", Settings_HDMATimingHack_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_HDMATIMINGHACK), tmpcharbuf);

  SendMessage (GetDlgItem (hWnd, IDC_EDIT_HMAXTIMINGHACK), EM_SETLIMITTEXT,
	       3, 0);
  wsprintf (tmpcharbuf, "%d", Settings_H_MaxTimingHack_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_HMAXTIMINGHACK), tmpcharbuf);

  SendMessage (GetDlgItem (hWnd, IDC_EDIT_SMP_MASTER_CLOCK), EM_SETLIMITTEXT,
	       8, 0);
  wsprintf (tmpcharbuf, "%d", Settings_smp_master_clock_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_SMP_MASTER_CLOCK), tmpcharbuf);

  SendMessage (GetDlgItem (hWnd, IDC_EDIT_APUONECYCLE), EM_SETLIMITTEXT,
	       3, 0);
  wsprintf (tmpcharbuf, "%d", Settings_APUOneCycle_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_APUONECYCLE), tmpcharbuf);


  //graphics
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_MODE7INTERPOLATE), BM_SETCHECK,
	       (WPARAM) (Settings_Mode7Interpolate_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_HIRES), BM_SETCHECK,
	       (WPARAM) (GUI_SupportHiRes_cpy ? BST_CHECKED : BST_UNCHECKED),
	       0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_GFXENGINE2), BM_SETCHECK,
	       (WPARAM) (Settings_GFXengine2_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_GFX16), BM_SETCHECK,
	       (WPARAM) (Settings_SixteenBit_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_TRANSPARENCY), BM_SETCHECK,
	       (WPARAM) (Settings_Transparency_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_BGLAYERING), BM_SETCHECK,
	       (WPARAM) (Settings_BGLayering_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_GFXW), BM_SETCHECK,
	       (WPARAM) (Settings_DisableGraphicWindows_cpy ? BST_UNCHECKED :
			 BST_CHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_BLOCKINVALIDVRAMACCESS), BM_SETCHECK,
	       (WPARAM) (Settings_BlockInvalidVRAMAccessMaster_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);
  SendMessage (GetDlgItem (hWnd, IDC_CHECK_HDMA), BM_SETCHECK,
	       (WPARAM) (Settings_DisableHDMA_cpy ? BST_UNCHECKED :
			 BST_CHECKED), 0);

  if (Settings_SkipFrames_cpy == AUTO_FRAMERATE)
    SendMessage (GetDlgItem (hWnd, IDC_COMBO_FRAMESKIP), CB_SETCURSEL, 0, 0);
  else if (Settings_SkipFrames_cpy >= 1 && Settings_SkipFrames_cpy <= 10)
    SendMessage (GetDlgItem (hWnd, IDC_COMBO_FRAMESKIP), CB_SETCURSEL,
		 (WPARAM) (Settings_SkipFrames_cpy), 0);
  else
    {
      SendMessage (GetDlgItem (hWnd, IDC_COMBO_FRAMESKIP), CB_SETCURSEL, 0,
		   0);
      Settings_SkipFrames_cpy = Settings.SkipFrames = AUTO_FRAMERATE;
    }

  SendMessage (GetDlgItem (hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES), CB_SETCURSEL,
	       (WPARAM) Settings_AutoMaxSkipFrames_cpy, 0);
  EnableWindow (GetDlgItem (hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES),
		(Settings_SkipFrames_cpy == AUTO_FRAMERATE) ? TRUE : FALSE);

  SendMessage (GetDlgItem (hWnd, IDC_EDIT_PAL), EM_SETLIMITTEXT, 3, 0);
  SendMessage (GetDlgItem (hWnd, IDC_EDIT_NTSC), EM_SETLIMITTEXT, 3, 0);

  wsprintf (tmpcharbuf, "%d", Settings_FrameTimePAL_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_PAL), tmpcharbuf);
  wsprintf (tmpcharbuf, "%d", Settings_FrameTimeNTSC_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_NTSC), tmpcharbuf);

  SendMessage (GetDlgItem (hWnd, IDC_CHECK_SOUNDCONSTANTTEMPO), BM_SETCHECK,
	       (WPARAM) (Settings_SoundConstantTempo_cpy ? BST_CHECKED :
			 BST_UNCHECKED), 0);

  // Original
  SendMessage (GetDlgItem (hWnd, IDC_EDIT_AUTOSAVESRAM), EM_SETLIMITTEXT, 6,
	       0);
  wsprintf (tmpcharbuf, "%d", Settings_AutoSaveDelay_cpy);
  SetWindowText (GetDlgItem (hWnd, IDC_EDIT_AUTOSAVESRAM), tmpcharbuf);


  //not suported yet
  //EnableWindow(GetDlgItem(hWnd, IDC_CHECK_GFXENGINE2), FALSE);
}

static WNDPROC Edit_HDMATimingHack;
static WNDPROC Edit_H_MaxTimingHack;
static WNDPROC Edit_FrameTimePAL;
static WNDPROC Edit_FrameTimeNTSC;
static WNDPROC Edit_smp_master_clock;
static WNDPROC Edit_APUOneCycle;
static HWND hEdit_HDMATimingHack;
static HWND hEdit_H_MaxTimingHack;
static HWND hEdit_FrameTimePAL;
static HWND hEdit_FrameTimeNTSC;
static HWND hEdit_smp_master_clock;
static HWND hEdit_APUOneCycle;

// Original
static WNDPROC Edit_AutoSaveDelay;
static HWND hEdit_AutoSaveDelay;


static LRESULT CALLBACK
SneseEditSubProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
    {
    case WM_CHAR:
      if ((LOWORD (wParam) >= '0' && LOWORD (wParam) <= '9')
	  || LOWORD (wParam) == VK_BACK)
	break;
      return 0;

    default:
      break;

    }
  if (hWnd == hEdit_HDMATimingHack)
    return (CallWindowProc (Edit_HDMATimingHack, hWnd, uMsg, wParam, lParam));
  if (hWnd == hEdit_H_MaxTimingHack)
    return (CallWindowProc (Edit_H_MaxTimingHack, hWnd, uMsg, wParam, lParam));
  if (hWnd == hEdit_FrameTimePAL)
    return (CallWindowProc (Edit_FrameTimePAL, hWnd, uMsg, wParam, lParam));
  if (hWnd == hEdit_FrameTimeNTSC)
    return (CallWindowProc (Edit_FrameTimeNTSC, hWnd, uMsg, wParam, lParam));
  if (hWnd == hEdit_smp_master_clock)
    return (CallWindowProc (Edit_smp_master_clock, hWnd, uMsg, wParam, lParam));
  if (hWnd == hEdit_APUOneCycle)
    return (CallWindowProc (Edit_APUOneCycle, hWnd, uMsg, wParam, lParam));
  // Original
  if (hWnd == hEdit_AutoSaveDelay)
    return (CallWindowProc (Edit_AutoSaveDelay, hWnd, uMsg, wParam, lParam));
  return 0;
}

static void
SneseDlgSetSubClass (HWND hWnd)
{
  hEdit_HDMATimingHack = GetDlgItem (hWnd, IDC_EDIT_HDMATIMINGHACK);
  hEdit_H_MaxTimingHack = GetDlgItem (hWnd, IDC_EDIT_HMAXTIMINGHACK);
  hEdit_FrameTimePAL = GetDlgItem (hWnd, IDC_EDIT_PAL);
  hEdit_FrameTimeNTSC = GetDlgItem (hWnd, IDC_EDIT_NTSC);
  hEdit_smp_master_clock = GetDlgItem (hWnd, IDC_EDIT_SMP_MASTER_CLOCK);
  hEdit_APUOneCycle = GetDlgItem (hWnd, IDC_EDIT_APUONECYCLE);

  Edit_HDMATimingHack =
    (WNDPROC) GetWindowLong (hEdit_HDMATimingHack, GWL_WNDPROC);
  Edit_H_MaxTimingHack =
    (WNDPROC) GetWindowLong (hEdit_H_MaxTimingHack, GWL_WNDPROC);
  Edit_FrameTimePAL =
    (WNDPROC) GetWindowLong (hEdit_FrameTimePAL, GWL_WNDPROC);
  Edit_FrameTimeNTSC =
    (WNDPROC) GetWindowLong (hEdit_FrameTimeNTSC, GWL_WNDPROC);
  Edit_smp_master_clock =
    (WNDPROC) GetWindowLong (hEdit_smp_master_clock, GWL_WNDPROC);
  Edit_APUOneCycle =
    (WNDPROC) GetWindowLong (hEdit_APUOneCycle, GWL_WNDPROC);

  SetWindowLong (hEdit_HDMATimingHack, GWL_WNDPROC, (LONG) SneseEditSubProc);
  SetWindowLong (hEdit_H_MaxTimingHack, GWL_WNDPROC, (LONG) SneseEditSubProc);
  SetWindowLong (hEdit_FrameTimePAL, GWL_WNDPROC, (LONG) SneseEditSubProc);
  SetWindowLong (hEdit_FrameTimeNTSC, GWL_WNDPROC, (LONG) SneseEditSubProc);
  SetWindowLong (hEdit_smp_master_clock, GWL_WNDPROC, (LONG) SneseEditSubProc);
  SetWindowLong (hEdit_APUOneCycle, GWL_WNDPROC, (LONG) SneseEditSubProc);

  // Original
  hEdit_AutoSaveDelay = GetDlgItem (hWnd, IDC_EDIT_AUTOSAVESRAM);
  Edit_AutoSaveDelay =
    (WNDPROC) GetWindowLong (hEdit_AutoSaveDelay, GWL_WNDPROC);
  SetWindowLong (hEdit_AutoSaveDelay, GWL_WNDPROC, (LONG) SneseEditSubProc);
}

static void
SneseDlgReleaseSubClass ()
{
  SetWindowLong (hEdit_HDMATimingHack, GWL_WNDPROC, (LONG) Edit_HDMATimingHack);
  SetWindowLong (hEdit_H_MaxTimingHack, GWL_WNDPROC, (LONG) Edit_H_MaxTimingHack);
  SetWindowLong (hEdit_FrameTimePAL, GWL_WNDPROC, (LONG) Edit_FrameTimePAL);
  SetWindowLong (hEdit_FrameTimeNTSC, GWL_WNDPROC, (LONG) Edit_FrameTimeNTSC);
  SetWindowLong (hEdit_smp_master_clock, GWL_WNDPROC, (LONG) Edit_smp_master_clock);
  SetWindowLong (hEdit_APUOneCycle, GWL_WNDPROC, (LONG) Edit_APUOneCycle);
  SetWindowLong (hEdit_AutoSaveDelay, GWL_WNDPROC, (LONG) Edit_AutoSaveDelay);	// Original
}

