/*
  Copyright (C) 2010 sanmaiwashi
*/

#include "snes9x.h"
#include "memmap.h"
#include "soundux.h"
#include "uosnesw.h"
#include "windraw.h"
#include "uow_input.h"
#include "explorer.h"
#include <io.h>
#include <direct.h>
#include "cfgfile.h"
#include "commutil.h"

extern char ROMFileName[];
extern char InitialROMFileDirectory[];
extern char RecentROMFileName[];
extern OPENFILENAMEEXPLORERPOS uosxplorerposition;

static void
ChangeWinHomeDirectory (char *fn, const char *hname, const char *last_hname,
			int bufsize)
{
  int len = strlen (hname);
  int last_len = strlen (last_hname);

  char drive[_MAX_DRIVE + 1];
  char dir[_MAX_DIR + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];

  int slash_len;

  _splitpath (hname, drive, dir, fname, ext);
  if (dir[0] && fname[0] == 0)
    slash_len = 1;
  else
    slash_len = 0;

  _splitpath (last_hname, drive, dir, fname, ext);
  if (dir[0] && fname[0] == 0)
    {
      if (slash_len == 1)
	slash_len = 0;
      else
	slash_len = -1;
    }

  if (last_len < bufsize && strncmp (fn, last_hname, (size_t) last_len) == 0
      && ((int) strlen (fn) + len - slash_len - last_len) < bufsize)
    {
      memmove (&fn[len], &fn[last_len + slash_len],
	       (size_t) (strlen (fn) - (last_len + slash_len) + 1));
      memcpy (&fn[0], hname, (size_t) len);
    }
}

void
read_port_config ()
{
  char drive[_MAX_DRIVE + 1];
  char dir[_MAX_DIR + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];
  char tmpCharBuf[_MAX_PATH + 1];
  const char *tmpcharptr;
  int i;

  /* Read [windows] */
  i = get_config_int (1, cs.windows, "PriorityClass");
  switch (i)
    {
    case 0:
      GUI.PriorityClass = IDLE_PRIORITY_CLASS;
      break;
    case 1:
      GUI.PriorityClass = NORMAL_PRIORITY_CLASS;
      break;
    case 2:
      GUI.PriorityClass = HIGH_PRIORITY_CLASS;
      break;
    case 3:
      GUI.PriorityClass = REALTIME_PRIORITY_CLASS;
      break;
    default:
      GUI.PriorityClass = NORMAL_PRIORITY_CLASS;
      break;
    }

  i = get_config_int (2, cs.windows, "ThreadPriority");
  switch (i)
    {
    case 0:
      GUI.ThreadPriority = THREAD_PRIORITY_LOWEST;
      break;
    case 1:
      GUI.ThreadPriority = THREAD_PRIORITY_BELOW_NORMAL;
      break;
    case 2:
      GUI.ThreadPriority = THREAD_PRIORITY_NORMAL;
      break;
    case 3:
      GUI.ThreadPriority = THREAD_PRIORITY_ABOVE_NORMAL;
      break;
    case 4:
      GUI.ThreadPriority = THREAD_PRIORITY_HIGHEST;
      break;
    case 5:
      GUI.ThreadPriority = THREAD_PRIORITY_IDLE;
      break;
    case 6:
      GUI.ThreadPriority = THREAD_PRIORITY_TIME_CRITICAL;
      break;
    default:
      GUI.ThreadPriority = THREAD_PRIORITY_NORMAL;
      break;
    }

  GUI.WindowTopMost =
    (BOOL) get_config_int (GUI.WindowTopMost, cs.windows,
			   "WindowTopMost");
  GUI.WindowTopMostAlways =
    (BOOL) get_config_int (GUI.WindowTopMostAlways, cs.windows,
			   "WindowTopMostAlways");
  GUI.AdjustWindow43 =
    (BOOL) get_config_int (GUI.AdjustWindow43, cs.windows,
			   "AdjustWindow43");

  Options.inactive_x =
    (b8) get_config_int (Options.inactive_x, cs.windows, "Inactive_X");

  GUI.statusbar =
    get_config_int (GUI.statusbar, cs.windows, "statusbar");

  /* Read [graphics] */
  i = get_config_int (GUI.FullScreen, cs.graphics, "Fullscreen");
  GUI.FullScreen = (i ? 1 : 0);
  GUI.DoubleRendering =
    (BOOL) get_config_int (GUI.DoubleRendering, cs.graphics,
			   "DoubleRendering");

  i =
    get_config_int (GUI.FilterMethod, cs.graphics,
		    "FilterMethod");
  GUI.FilterMethod = (i >= 0 && i <= 12) ? i : GUI.FilterMethod;

  GUI.FullscreenDisplayModeIndex =
    get_config_int (GUI.FullscreenDisplayModeIndex, cs.graphics,
		    "FullscreenDisplayModeIndex");

  tmpcharptr = get_config_string (cs.graphics, "DisplayModeString");
  if (tmpcharptr[0] && strlen ((const char *) tmpcharptr) <= 1024)
    strcpy (GUI.DisplayModeString, (const char *) tmpcharptr);

  i =
    get_config_int (GUI.DoubleBuffered, cs.graphics, "DoubleBuffering");
  GUI.DoubleBuffered = i ? 1 : 0;
  Renderer.ForceWindowedDirectDrawClipper =
    (BOOL) get_config_int (Renderer.
			   ForceWindowedDirectDrawClipper,
			   cs.graphics,
			   "ForceWindowedDirectDrawClipper");

  i =
    get_config_int (Renderer.wait_for_vsync, cs.graphics,
		    "wait_for_vsync");
  Renderer.wait_for_vsync = (i >= 0 && i <= 3) ? i : Renderer.wait_for_vsync;

  i =
    get_config_int (Renderer.OffScreenVRAM, cs.graphics,
		    "OffScreenVRAM");
  Renderer.OffScreenVRAM = i ? TRUE : FALSE;
  i =
    get_config_int (Renderer.BackBufferVRAM, cs.graphics,
		    "BackBufferVRAM");
  Renderer.BackBufferVRAM = i ? TRUE : FALSE;


  Settings.GFXengine2 =
    (bool8) get_config_int (Settings.GFXengine2, cs.graphics,
			    "GFXengine2");
  Settings.SixteenBit =
    (bool8) get_config_int (Settings.SixteenBit, cs.graphics,
			    "SixteenBit");

  i =
    get_config_int (Settings.ForceTransparency, cs.graphics,
		    "Transparency");
  if (i == 0)
    {
      Settings.ForceTransparency = 0;
      Settings.ForceNoTransparency = 1;
    }
  else
    {
      Settings.ForceTransparency = 1;
      Settings.ForceNoTransparency = 0;
    }

  Settings.BGLayering =
    (bool8) get_config_int (Settings.BGLayering, cs.graphics,
			    "BGLayering");
  Settings.DisableGraphicWindows =
    (bool8) get_config_int (Settings.DisableGraphicWindows, cs.graphics,
			    "DisableGraphicWindows");
  Settings.DisableHDMA =
    (bool8) get_config_int (Settings.DisableHDMA, cs.graphics,
			    "DisableHDMA");
  Settings.Mode7Interpolate =
    (bool8) get_config_int (Settings.Mode7Interpolate, cs.graphics,
			    "Mode7Interpolate");
  Settings.BlockInvalidVRAMAccessMaster =
    (bool8) get_config_int (Settings.BlockInvalidVRAMAccessMaster, cs.graphics,
			    "BlockInvalidVRAMAccess");
  GUI.SupportHiRes = (BOOL) get_config_int (GUI.SupportHiRes, cs.graphics,
					    "hires");

  i = get_config_int (GUI.Stretch, cs.graphics, "stretch");
  GUI.Stretch = (i >= 0 && i <= 3) ? i : GUI.Stretch;

  /* Read [sound] */
  GUI.DSuse =
    (BOOL) get_config_int (GUI.DSuse, cs.sound, "DirectSound");
  GUI.DirectSoundNotify =
    (BOOL) get_config_int (GUI.DirectSoundNotify, cs.sound,
			   "DirectSoundNotify");
  Settings.NextAPUEnabled =
    (bool8) get_config_int (Settings.NextAPUEnabled, cs.sound,
			    "APUEnabled");

  i = get_config_int (Options.SoundRate, cs.sound, "SoundRate");
  Options.SoundRate = (char) ((0 <= i && i <= 5) ? i : Options.SoundRate);

  i = get_config_int (1, cs.sound, "SoundBit");
  switch (i)
    {
    case 0:
      Options.SoundBit = 8;
      break;
    case 1:
      Options.SoundBit = 16;
      break;
    default:
      Options.SoundBit = DefaultOptions.SoundBit;
      break;
    }

  i = get_config_int (1, cs.sound, "Stereo");
  switch (i)
    {
    case 0:
      Options.SoundChn = 1;
      break;
    case 1:
      Options.SoundChn = 2;
      break;
    default:
      Options.SoundChn = DefaultOptions.SoundChn;
      break;
    }

  i =
    get_config_int (Settings.InterpolatedSound, cs.sound,
		    "InterpolatedSound");
  Settings.InterpolatedSound = (0 <= i
				&& i <= 4) ? i : Settings.InterpolatedSound;
  //Settings.LowPassFilter = (bool8)get_config_int(Settings.LowPassFilter, cs.sound, "LowPassFilter");
  //Settings.SimpleFilter = (bool8)get_config_int(Settings.SimpleFilter, cs.sound, "SimpleFilter");
  i =
    get_config_int (Settings.SoundBufferSize, cs.sound,
		    "SoundBufferLength");
  Settings.SoundBufferSize = (i >= 2
			      && i <= 9) ? i : Settings.SoundBufferSize;
  i =
    get_config_int (Settings.SoundMixInterval, cs.sound, "SoundMixInterval");
  Settings.SoundMixInterval = (i >= 1
			       && i <= 999) ? i : Settings.SoundMixInterval;
  GUI.SoundGlobalFocus =
    (BOOL) get_config_int (GUI.SoundGlobalFocus, cs.sound,
			   "SoundGlobalFocus");
  Settings.SoundEnvelopeHeightReading =
    (bool8) get_config_int (Settings.SoundEnvelopeHeightReading,
			    cs.sound, "SoundEnvelopeHeightReading");
  Settings.DisableSoundEcho =
    (bool8) get_config_int (Settings.DisableSoundEcho, cs.sound,
			    "DisableSoundEcho");
  Settings.DisableMasterVolume =
    (bool8) get_config_int (Settings.DisableMasterVolume, cs.sound,
			    "DisableMasterVolume");
  i = get_config_int (Options.VolumeLevel, cs.sound, "VolumeLevel");
  Options.VolumeLevel = (i >= 1 && i <= 5) ? i : Options.VolumeLevel;
#if 0
  Settings.DisableSampleCaching =
    (bool8) get_config_int (Settings.DisableSampleCaching, cs.sound,
			    "DisableSampleCaching");
  i =
    get_config_int (Settings.AltSampleDecode, cs.sound,
		    "SampleDecodingMethod");
  Settings.AltSampleDecode =
    (uint8) ((0 <= i && i <= 5) ? i : Settings.AltSampleDecode);
#endif
  i =
    get_config_int (Settings.SoundSkipMethod, cs.sound,
		    "SoundSkipMethod");
  Settings.SoundSkipMethod =
    (uint8) ((i >= 0 && i <= 3) ? i : Settings.SoundSkipMethod);
  Settings.ReverseStereo =
    (uint8) get_config_int (Settings.ReverseStereo, cs.sound,
			    "ReverseStereo");
#if 0
  Settings.FixFrequency =
    (bool8) get_config_int (Settings.FixFrequency, cs.sound,
			    "FixFrequency");
#endif

  /* Read [cpu] */
  Settings.ShutdownMaster =
    (bool8) ! get_config_int (!Settings.ShutdownMaster, cs.cpu,
			      "Shutdown");

  int32 hdth = get_config_int (Settings.HDMATimingHack, cs.cpu, "HDMATimingHack");
  hdth = (hdth >= 0 && hdth <= 350) ? hdth : Settings.HDMATimingHack;

  int32 hmth = get_config_int (Settings.H_MaxTimingHack, cs.cpu, "H_MaxTimingHack");
  hmth = (hmth >= 0 && hmth <= 999) ? hmth : Settings.H_MaxTimingHack;

  if(hmth < hdth - 100)
    hmth += ((hdth - 100) - hmth);

  Settings.HDMATimingHack = hdth;
  Settings.H_MaxTimingHack = hmth;

  Settings.OldTimingMaster =
    (bool8) get_config_int (Settings.OldTimingMaster, cs.cpu,
			    "OldTiming");

  Settings.DisableIRQCycleCount =
    (bool8) get_config_int (Settings.DisableIRQCycleCount, cs.cpu,
			    "DisableIRQCycleCount");

  Settings.smp_master_clock =
    (int32) get_config_int (Settings.smp_master_clock, cs.cpu,
			    "smp_master_clock");

  Settings.APUOneCycle =
    (int32) get_config_int (Settings.APUOneCycle, cs.cpu,
			    "smp_one_cycle");

  /* Read [timer] */
  i = get_config_int ((Settings.SkipFrames == AUTO_FRAMERATE ?
			     AUTO_FRAMERATE : (Settings.SkipFrames - 1)),
		      cs.timer, "SkipFrames");
  if (i != AUTO_FRAMERATE && i >= 0 && i <= 9)
    Settings.SkipFrames = (uint32) i + 1;
  else
    Settings.SkipFrames = AUTO_FRAMERATE;
  i =
    get_config_int (Settings.AutoMaxSkipFrames, cs.timer,
		    "AutoMaxSkipFrames");
  Settings.AutoMaxSkipFrames =
    (uint32) ((i >= 0 && i <= 10) ? i : Settings.AutoMaxSkipFrames);
  i =
    get_config_int (Settings.TurboSkipFrames, cs.timer,
		    "TurboSkipFrames");
  Settings.TurboSkipFrames =
    (uint32) ((i >= 0 && i <= 15) ? i : Settings.TurboSkipFrames);
  Settings.FrameTimeNTSC =
    (uint32) get_config_int (Settings.FrameTimeNTSC, cs.timer,
			     "NTSCtiming");
  Settings.FrameTimePAL =
    (uint32) get_config_int (Settings.FrameTimePAL, cs.timer,
			     "PALtiming");

  Settings.SoundConstantTempo =
    (uint32) get_config_int (Settings.SoundConstantTempo, cs.timer,
			     "SoundConstantTempo");

  /* Read [FileDirectory] */
  tmpcharptr = get_config_string (cs.filedirectory, "HomeDirectory");

  if (tmpcharptr[0] && strlen ((const char *) tmpcharptr) <= _MAX_PATH)
    strcpy (Options.LastHomeDirectory, (const char *) tmpcharptr);

  if (strcmp (Options.HomeDirectory, Options.LastHomeDirectory) == 0)
    Options.HomeDirectoryMoved = FALSE;
  else
    Options.HomeDirectoryMoved = TRUE;

  tmpcharptr = get_config_string (cs.filedirectory, "LastROMFileName");
  if (tmpcharptr[0])
    {
      strcpy (ROMFileName, tmpcharptr);
    }
  _splitpath (ROMFileName, drive, dir, fname, ext);
  _makepath (InitialROMFileDirectory, drive, dir, NULL, NULL);

  strcpy (tmpCharBuf, "RecentROMFileName1");
  for (i = 0; i < 10; i++)
    {
      sprintf (&tmpCharBuf[17], "%X", i + 1);
      tmpcharptr =
	get_config_string (cs.filedirectory, (const char *) tmpCharBuf);
      if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
	strcpy (&RecentROMFileName[_MAX_PATH * i], tmpcharptr);
      else
	RecentROMFileName[_MAX_PATH * i] = 0;
    }

  tmpcharptr = get_config_string (cs.filedirectory, "SnapShotDirectory");
  if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
    strcpy (tmpCharBuf, tmpcharptr);
  else
    strcpy (tmpCharBuf, GUI.FreezeFileDir);

  if (Options.HomeDirectoryMoved)
    ChangeWinHomeDirectory (tmpCharBuf, (const char *) Options.HomeDirectory,
			    (const char *) Options.LastHomeDirectory,
			    _MAX_PATH);

  if (access ((const char *) tmpCharBuf, 0) == 0)
    strcpy (GUI.FreezeFileDir, (const char *) tmpCharBuf);
  _splitpath (GUI.FreezeFileDir, drive, dir, fname, ext);
  if (dir[0] && fname[0] == 0)
    {
      dir[strlen (dir) - 1] = 0;
    }
  _makepath (GUI.FreezeFileDir, drive, dir, fname, ext);

  tmpcharptr = get_config_string (cs.filedirectory, "ScreenShotDirectory");
  if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
    strcpy (tmpCharBuf, tmpcharptr);
  else
    strcpy (tmpCharBuf, (const char *) Memory.ScreenShotDirectory);

  if (Options.HomeDirectoryMoved)
    ChangeWinHomeDirectory (tmpCharBuf, (const char *) Options.HomeDirectory,
			    (const char *) Options.LastHomeDirectory,
			    _MAX_PATH);

  if (access ((const char *) tmpCharBuf, 0) == 0)
    strcpy (Memory.ScreenShotDirectory, (const char *) tmpCharBuf);
  _splitpath (Memory.ScreenShotDirectory, drive, dir, fname, ext);
  if (dir[0] && fname[0] == 0)
    {
      dir[strlen (dir) - 1] = 0;
    }
  _makepath (Memory.ScreenShotDirectory, drive, dir, fname, ext);

  tmpcharptr = get_config_string (cs.filedirectory, "SPCDumpDirectory");
  if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
    strcpy (tmpCharBuf, tmpcharptr);
  else
    strcpy (tmpCharBuf, (const char *) Memory.SPCDumpDirectory);

  if (Options.HomeDirectoryMoved)
    ChangeWinHomeDirectory (tmpCharBuf, (const char *) Options.HomeDirectory,
			    (const char *) Options.LastHomeDirectory,
			    _MAX_PATH);

  if (access ((const char *) tmpCharBuf, 0) == 0)
    strcpy (Memory.SPCDumpDirectory, (const char *) tmpCharBuf);
  _splitpath (Memory.SPCDumpDirectory, drive, dir, fname, ext);
  if (dir[0] && fname[0] == 0)
    {
      dir[strlen (dir) - 1] = 0;
    }
  _makepath (Memory.SPCDumpDirectory, drive, dir, fname, ext);

  tmpcharptr = get_config_string (cs.filedirectory, "BinaryDumpDirectory");
  if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
    strcpy (tmpCharBuf, tmpcharptr);
  else
    strcpy (tmpCharBuf, (const char *) Options.BinaryDumpDirectory);

  if (Options.HomeDirectoryMoved)
    ChangeWinHomeDirectory (tmpCharBuf, (const char *) Options.HomeDirectory,
			    (const char *) Options.LastHomeDirectory,
			    _MAX_PATH);

  if (access ((const char *) tmpCharBuf, 0) == 0)
    strcpy (Options.BinaryDumpDirectory, (const char *) tmpCharBuf);
  _splitpath (Options.BinaryDumpDirectory, drive, dir, fname, ext);
  if (dir[0] && fname[0] == 0)
    {
      dir[strlen (dir) - 1] = 0;
    }
  _makepath (Options.BinaryDumpDirectory, drive, dir, fname, ext);

  tmpcharptr = get_config_string (cs.filedirectory, "BSX");
  if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
    strcpy (tmpCharBuf, tmpcharptr);
  else
    strcpy (tmpCharBuf, (const char *) Memory.BSXBIOSROMFileName);

  if (Options.HomeDirectoryMoved)
    ChangeWinHomeDirectory (tmpCharBuf, (const char *) Options.HomeDirectory,
			    (const char *) Options.LastHomeDirectory,
			    _MAX_PATH);

  strcpy (Memory.BSXBIOSROMFileName, (const char *) tmpCharBuf);

  tmpcharptr = get_config_string (cs.filedirectory, "SFT");
  if (tmpcharptr[0] && strlen (tmpcharptr) < _MAX_PATH)
    strcpy (tmpCharBuf, tmpcharptr);
  else
    strcpy (tmpCharBuf, (const char *) Memory.SFTBIOSROMFileName);

  if (Options.HomeDirectoryMoved)
    ChangeWinHomeDirectory (tmpCharBuf, (const char *) Options.HomeDirectory,
			    (const char *) Options.LastHomeDirectory,
			    _MAX_PATH);

  strcpy (Memory.SFTBIOSROMFileName, (const char *) tmpCharBuf);


  /* Read [ControllerOption] */
  strcpy (tmpCharBuf, "JoystickConnected0");
  for (i = 0; i < 8; i++)
    {
      sprintf (&tmpCharBuf[17], "%d", i);
      Settings.JoystickConnected[i] =
	(bool8) get_config_int (Settings.JoystickConnected[i],
				cs.ControllerOption,
				(const char *) tmpCharBuf);
    }

  i = get_config_int (Settings.ControllerOption, cs.ControllerOption, "ControllerOption");
  Settings.ControllerOption = (i >= SNES_JOYPAD && i < SNES_MAX_CONTROLLER_OPTIONS) ? i : Settings.ControllerOption;

  GUI.DIuse =
    (BOOL) get_config_int (GUI.DIuse, cs.ControllerOption,
			   "DirectInput");

  Options.TurboShotEnable =
    (b8) get_config_int (Options.TurboShotEnable, cs.ControllerOption,
			 "TurboShotEnable");

  /* Read [joypadA1 2 3 4 5 6 7 8] */
  for (i = 0; i < 8; i++)
    {
      JoypadA[i].Up =
	(WORD) get_config_int (JoypadA[i].Up, cs.joypadA[i],
			       cs.button[0]);
      JoypadA[i].Down =
	(WORD) get_config_int (JoypadA[i].Down, cs.joypadA[i],
			       cs.button[1]);
      JoypadA[i].Left =
	(WORD) get_config_int (JoypadA[i].Left, cs.joypadA[i],
			       cs.button[2]);
      JoypadA[i].Right =
	(WORD) get_config_int (JoypadA[i].Right, cs.joypadA[i],
			       cs.button[3]);
      JoypadA[i].A =
	(WORD) get_config_int (JoypadA[i].A, cs.joypadA[i],
			       cs.button[4]);
      JoypadA[i].B =
	(WORD) get_config_int (JoypadA[i].B, cs.joypadA[i],
			       cs.button[5]);
      JoypadA[i].X =
	(WORD) get_config_int (JoypadA[i].X, cs.joypadA[i],
			       cs.button[6]);
      JoypadA[i].Y =
	(WORD) get_config_int (JoypadA[i].Y, cs.joypadA[i],
			       cs.button[7]);
      JoypadA[i].Select =
	(WORD) get_config_int (JoypadA[i].Select, cs.joypadA[i],
			       cs.button[8]);
      JoypadA[i].Start =
	(WORD) get_config_int (JoypadA[i].Start, cs.joypadA[i],
			       cs.button[9]);
      JoypadA[i].L =
	(WORD) get_config_int (JoypadA[i].L, cs.joypadA[i],
			       cs.button[10]);
      JoypadA[i].R =
	(WORD) get_config_int (JoypadA[i].R, cs.joypadA[i],
			       cs.button[11]);
      JoypadA[i].Turbo_Up =
	(get_config_int
	 (JoypadA[i].Turbo_Up, cs.joypadA[i],
	  cs.button[12]) ? TRUE : FALSE);
      JoypadA[i].Turbo_Down =
	(get_config_int
	 (JoypadA[i].Turbo_Down, cs.joypadA[i],
	  cs.button[13]) ? TRUE : FALSE);
      JoypadA[i].Turbo_Left =
	(get_config_int
	 (JoypadA[i].Turbo_Left, cs.joypadA[i],
	  cs.button[14]) ? TRUE : FALSE);
      JoypadA[i].Turbo_Right =
	(get_config_int
	 (JoypadA[i].Turbo_Right, cs.joypadA[i],
	  cs.button[15]) ? TRUE : FALSE);
      JoypadA[i].Turbo_A =
	(get_config_int
	 (JoypadA[i].Turbo_A, cs.joypadA[i],
	  cs.button[16]) ? TRUE : FALSE);
      JoypadA[i].Turbo_B =
	(get_config_int
	 (JoypadA[i].Turbo_B, cs.joypadA[i],
	  cs.button[17]) ? TRUE : FALSE);
      JoypadA[i].Turbo_X =
	(get_config_int
	 (JoypadA[i].Turbo_X, cs.joypadA[i],
	  cs.button[18]) ? TRUE : FALSE);
      JoypadA[i].Turbo_Y =
	(get_config_int
	 (JoypadA[i].Turbo_Y, cs.joypadA[i],
	  cs.button[19]) ? TRUE : FALSE);
      JoypadA[i].Turbo_Select =
	(get_config_int
	 (JoypadA[i].Turbo_Select, cs.joypadA[i],
	  cs.button[20]) ? TRUE : FALSE);
      JoypadA[i].Turbo_Start =
	(get_config_int
	 (JoypadA[i].Turbo_Start, cs.joypadA[i],
	  cs.button[21]) ? TRUE : FALSE);
      JoypadA[i].Turbo_L =
	(get_config_int
	 (JoypadA[i].Turbo_L, cs.joypadA[i],
	  cs.button[22]) ? TRUE : FALSE);
      JoypadA[i].Turbo_R =
	(get_config_int
	 (JoypadA[i].Turbo_R, cs.joypadA[i],
	  cs.button[23]) ? TRUE : FALSE);
    }

  /* Read [FunctionCodeA] */
  strcpy (tmpCharBuf, "Code0");
  for (i = 0; i < 9; i++)
    {
      sprintf (&tmpCharBuf[4], "%d", i);
      FunctionCodeA[i] =
	(DWORD) get_config_int (FunctionCodeA[i], cs.functioncodeA,
				(const char *) tmpCharBuf);
    }

  /* Read [joypadB1 2 3 4 5 6 7 8] */
  for (i = 0; i < 8; i++)
    {
      JoypadB[i].Up =
	(WORD) get_config_int (JoypadB[i].Up, cs.joypadB[i],
			       cs.button[0]);
      JoypadB[i].Down =
	(WORD) get_config_int (JoypadB[i].Down, cs.joypadB[i],
			       cs.button[1]);
      JoypadB[i].Left =
	(WORD) get_config_int (JoypadB[i].Left, cs.joypadB[i],
			       cs.button[2]);
      JoypadB[i].Right =
	(WORD) get_config_int (JoypadB[i].Right, cs.joypadB[i],
			       cs.button[3]);
      JoypadB[i].A =
	(WORD) get_config_int (JoypadB[i].A, cs.joypadB[i],
			       cs.button[4]);
      JoypadB[i].B =
	(WORD) get_config_int (JoypadB[i].B, cs.joypadB[i],
			       cs.button[5]);
      JoypadB[i].X =
	(WORD) get_config_int (JoypadB[i].X, cs.joypadB[i],
			       cs.button[6]);
      JoypadB[i].Y =
	(WORD) get_config_int (JoypadB[i].Y, cs.joypadB[i],
			       cs.button[7]);
      JoypadB[i].Select =
	(WORD) get_config_int (JoypadB[i].Select, cs.joypadB[i],
			       cs.button[8]);
      JoypadB[i].Start =
	(WORD) get_config_int (JoypadB[i].Start, cs.joypadB[i],
			       cs.button[9]);
      JoypadB[i].L =
	(WORD) get_config_int (JoypadB[i].L, cs.joypadB[i],
			       cs.button[10]);
      JoypadB[i].R =
	(WORD) get_config_int (JoypadB[i].R, cs.joypadB[i],
			       cs.button[11]);
      JoypadB[i].Turbo_Up =
	(get_config_int
	 (JoypadB[i].Turbo_Up, cs.joypadB[i],
	  cs.button[12]) ? TRUE : FALSE);
      JoypadB[i].Turbo_Down =
	(get_config_int
	 (JoypadB[i].Turbo_Down, cs.joypadB[i],
	  cs.button[13]) ? TRUE : FALSE);
      JoypadB[i].Turbo_Left =
	(get_config_int
	 (JoypadB[i].Turbo_Left, cs.joypadB[i],
	  cs.button[14]) ? TRUE : FALSE);
      JoypadB[i].Turbo_Right =
	(get_config_int
	 (JoypadB[i].Turbo_Right, cs.joypadB[i],
	  cs.button[15]) ? TRUE : FALSE);
      JoypadB[i].Turbo_A =
	(get_config_int
	 (JoypadB[i].Turbo_A, cs.joypadB[i],
	  cs.button[16]) ? TRUE : FALSE);
      JoypadB[i].Turbo_B =
	(get_config_int
	 (JoypadB[i].Turbo_B, cs.joypadB[i],
	  cs.button[17]) ? TRUE : FALSE);
      JoypadB[i].Turbo_X =
	(get_config_int
	 (JoypadB[i].Turbo_X, cs.joypadB[i],
	  cs.button[18]) ? TRUE : FALSE);
      JoypadB[i].Turbo_Y =
	(get_config_int
	 (JoypadB[i].Turbo_Y, cs.joypadB[i],
	  cs.button[19]) ? TRUE : FALSE);
      JoypadB[i].Turbo_Select =
	(get_config_int
	 (JoypadB[i].Turbo_Select, cs.joypadB[i],
	  cs.button[20]) ? TRUE : FALSE);
      JoypadB[i].Turbo_Start =
	(get_config_int
	 (JoypadB[i].Turbo_Start, cs.joypadB[i],
	  cs.button[21]) ? TRUE : FALSE);
      JoypadB[i].Turbo_L =
	(get_config_int
	 (JoypadB[i].Turbo_L, cs.joypadB[i],
	  cs.button[22]) ? TRUE : FALSE);
      JoypadB[i].Turbo_R =
	(get_config_int
	 (JoypadB[i].Turbo_R, cs.joypadB[i],
	  cs.button[23]) ? TRUE : FALSE);
    }

  /* Read [FunctionCodeB] */
  strcpy (tmpCharBuf, "Code0");
  for (i = 0; i < 9; i++)
    {
      sprintf (&tmpCharBuf[4], "%d", i);
      FunctionCodeB[i] =
	(DWORD) get_config_int (FunctionCodeB[i], cs.functioncodeB,
				(const char *) tmpCharBuf);
    }

  /* Read [gui] */
  i = get_config_int (GUI.Language, cs.gui, "Language");
  GUI.Language = (i == 1 || i == 0) ? i : GUI.Language;

  GUI.WindowRectSetting.left =
    get_config_int (-0x7fffffff, cs.gui, "WindowRectLeft");
  GUI.WindowRectSetting.right =
    get_config_int (-0x7fffffff, cs.gui, "WindowRectRight");
  GUI.WindowRectSetting.top =
    get_config_int (-0x7fffffff, cs.gui, "WindowRectTop");
  GUI.WindowRectSetting.bottom =
    get_config_int (-0x7fffffff, cs.gui, "WindowRectBottom");
  if (GUI.WindowRectSetting.left == -0x7fffffff
      || GUI.WindowRectSetting.right == -0x7fffffff
      || GUI.WindowRectSetting.top == -0x7fffffff
      || GUI.WindowRectSetting.bottom == -0x7fffffff)
    GUI.WindowRectSetting.left = GUI.WindowRectSetting.right =
      GUI.WindowRectSetting.top = GUI.WindowRectSetting.bottom = -1;
  else
    {
      if ((GUI.WindowRectSetting.right - GUI.WindowRectSetting.left) <
	  GUI.DefaultWindowSizeW)
	{
	  GUI.WindowRectSetting.right =
	    (GUI.WindowRectSetting.left + GUI.DefaultWindowSizeW);
	}
      if ((GUI.WindowRectSetting.bottom - GUI.WindowRectSetting.top) <
	  GUI.DefaultWindowSizeH)
	{
	  GUI.WindowRectSetting.bottom =
	    (GUI.WindowRectSetting.top + GUI.DefaultWindowSizeH);
	  GUI.NeedWindowSizeAddition = 2;
	}
    }

  GUI.FileBrowseWindow = get_config_int (GUI.FileBrowseWindow, cs.gui,
					 "FileBrowseWindow");

  uosxplorerposition.WindowRect.left =
    get_config_int (uosxplorerposition.WindowRect.left, cs.gui,
		    "uosXpWindowRectLeft");
  uosxplorerposition.WindowRect.right =
    get_config_int (uosxplorerposition.WindowRect.right, cs.gui,
		    "uosXpWindowRectRight");
  uosxplorerposition.WindowRect.top =
    get_config_int (uosxplorerposition.WindowRect.top, cs.gui,
		    "uosXpWindowRectTop");
  uosxplorerposition.WindowRect.bottom =
    get_config_int (uosxplorerposition.WindowRect.bottom, cs.gui,
		    "uosXpWindowRectBottom");
  uosxplorerposition.Maximize =
    (BOOL) get_config_int (uosxplorerposition.Maximize, cs.gui,
			   "uosXpMaximize");
  uosxplorerposition.VSplitBarPos0 =
    get_config_int (uosxplorerposition.VSplitBarPos0, cs.gui,
		    "uosXpVSplitBarPos0");
  uosxplorerposition.ColumnWidth0 =
    get_config_int (uosxplorerposition.ColumnWidth0, cs.gui,
		    "uosXpColumnWidth0");
  uosxplorerposition.ColumnWidth1 =
    get_config_int (uosxplorerposition.ColumnWidth1, cs.gui,
		    "uosXpColumnWidth1");
  uosxplorerposition.ColumnWidth2 =
    get_config_int (uosxplorerposition.ColumnWidth2, cs.gui,
		    "uosXpColumnWidth2");
  uosxplorerposition.ColumnWidth3 =
    get_config_int (uosxplorerposition.ColumnWidth3, cs.gui,
		    "uosXpColumnWidth3");

  /* Read [screenshot] */
  i =
    get_config_int (Options.SaveImagefileFormat, cs.screenshot,
		    "SaveImagefileFormat");
  Options.SaveImagefileFormat = (0 <= i
				 && i <= 3) ? i : Options.SaveImagefileFormat;
  Options.ReduceTheColorTo256 =
    (b8) get_config_int (Options.ReduceTheColorTo256,
			 cs.screenshot, "ReduceTheColorTo256");
  Options.ReduceTheColorTo256IfWithin =
    (b8) get_config_int (Options.ReduceTheColorTo256IfWithin,
			 cs.screenshot, "ReduceTheColorTo256IfWithin");
  Options.ImagefileHiresStretch =
    (b8) get_config_int (Options.ImagefileHiresStretch, cs.screenshot,
			 "ImagefileHiresStretch");
  Options.RLEcompression256bitmap =
    (b8) get_config_int (Options.RLEcompression256bitmap, cs.screenshot,
			 "RLEcompression256bitmap");

  /* Read [debug] */
  GUI.AttachConsole =
    (BOOL) get_config_int (GUI.AttachConsole, cs.debug,
			 "AttachConsole");

  Settings.DisableGameSpecificHacks =
    (bool8) get_config_int (Settings.DisableGameSpecificHacks, cs.debug,
			 "DisableGameSpecificHacks");

  Options.DisplayScreenBufferAll =
    (b8) get_config_int (Options.DisplayScreenBufferAll, cs.debug,
			 "DisplayScreenBufferAll");

  /* Read [option] */
  Settings.ApplyCheats =
    (bool8) get_config_int (Settings.ApplyCheats, cs.option,
			    "ApplyCheats");
  Settings.DisplayFrameRate =
    (bool8) get_config_int (Settings.DisplayFrameRate, cs.option,
			    "ShowFPS");
  Settings.Mute =
    (bool8) get_config_int (Settings.Mute, cs.option, "MuteSound");
  Settings.AutoSaveDelay =
    (int32) get_config_int (Settings.AutoSaveDelay, cs.option,
			    "AutoSaveDelay");
  Xxxcdda.Enable =
    (bool8) get_config_int (Xxxcdda.Enable, cs.option, "UseCDDA");
  i = get_config_int (Xxxcdda.Drv, cs.option, "CDDrv");
  Xxxcdda.Drv = (uint32) ((0 <= i && i <= 26) ? i : Xxxcdda.Drv);

  return;
}

void
write_port_config ()
{
  int i, c;
  char drive[_MAX_DRIVE + 1];
  char dir[_MAX_DIR + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];
  char tmpCharBuf[_MAX_PATH + 1];

  /* Write [windows] */
  switch (GUI.PriorityClass)
    {
    case IDLE_PRIORITY_CLASS:
      i = 0;
      break;
    case NORMAL_PRIORITY_CLASS:
      i = 1;
      break;
    case HIGH_PRIORITY_CLASS:
      i = 2;
      break;
    case REALTIME_PRIORITY_CLASS:
      i = 3;
      break;
    default:
      i = 1;
      break;
    }
  set_config_int (cs.windows, "PriorityClass", i);

  switch (GUI.ThreadPriority)
    {
    case THREAD_PRIORITY_LOWEST:
      i = 0;
      break;
    case THREAD_PRIORITY_BELOW_NORMAL:
      i = 1;
      break;
    case THREAD_PRIORITY_NORMAL:
      i = 2;
      break;
    case THREAD_PRIORITY_ABOVE_NORMAL:
      i = 3;
      break;
    case THREAD_PRIORITY_HIGHEST:
      i = 4;
      break;
    case THREAD_PRIORITY_IDLE:
      i = 5;
      break;
    case THREAD_PRIORITY_TIME_CRITICAL:
      i = 6;
      break;
    default:
      i = 2;
      break;
    }
  set_config_int (cs.windows, "ThreadPriority", i);

  set_config_int (cs.windows, "WindowTopMost", GUI.WindowTopMost);
  set_config_int (cs.windows, "WindowTopMostAlways", GUI.WindowTopMostAlways);
  set_config_int (cs.windows, "AdjustWindow43", GUI.AdjustWindow43);
  set_config_int (cs.windows, "Inactive_X", Options.inactive_x);
  set_config_int (cs.windows, "statusbar", GUI.statusbar);

  /* Write [graphics] */
  set_config_int (cs.graphics, "Fullscreen", GUI.FullScreen);
  set_config_int (cs.graphics, "DoubleRendering", GUI.DoubleRendering);

  set_config_int (cs.graphics, "FilterMethod",
		  GUI.FilterMethod);
  set_config_int (cs.graphics, "FullscreenDisplayModeIndex",
		  GUI.FullscreenDisplayModeIndex);
  set_config_string (cs.graphics, "DisplayModeString",
		     GUI.DisplayModeString);

  set_config_int (cs.graphics, "DoubleBuffering", GUI.DoubleBuffered);
  set_config_int (cs.graphics, "ForceWindowedDirectDrawClipper",
		  Renderer.ForceWindowedDirectDrawClipper);
  set_config_int (cs.graphics, "wait_for_vsync",
		  Renderer.wait_for_vsync);
  set_config_int (cs.graphics, "OffScreenVRAM", Renderer.OffScreenVRAM);
  set_config_int (cs.graphics, "BackBufferVRAM", Renderer.BackBufferVRAM);
  set_config_int (cs.graphics, "GFXengine2", Settings.GFXengine2);
  set_config_int (cs.graphics, "stretch", GUI.Stretch);
  set_config_int (cs.graphics, "SixteenBit", Settings.SixteenBit);
  set_config_int (cs.graphics, "Transparency", Settings.Transparency);
  set_config_int (cs.graphics, "BGLayering", Settings.BGLayering);
  set_config_int (cs.graphics, "DisableGraphicWindows",
		  Settings.DisableGraphicWindows);
  set_config_int (cs.graphics, "DisableHDMA", Settings.DisableHDMA);
  set_config_int (cs.graphics, "hires", GUI.SupportHiRes);
  set_config_int (cs.graphics, "Mode7Interpolate",
		  Settings.Mode7Interpolate);
  set_config_int (cs.graphics, "BlockInvalidVRAMAccess",
		  Settings.BlockInvalidVRAMAccessMaster);


  /* Write [sound] */
  set_config_int (cs.sound, "SoundRate", Options.SoundRate);

  switch (Options.SoundBit)
    {
    case 8:
      i = 0;
      break;
    case 16:
      i = 1;
      break;
    default:
      i = 1;
      break;
    }
  set_config_int (cs.sound, "SoundBit", i);


  switch (Options.SoundChn)
    {
    case 1:
      i = 0;
      break;
    case 2:
      i = 1;
      break;
    default:
      i = 1;
      break;
    }
  set_config_int (cs.sound, "Stereo", i);

  set_config_int (cs.sound, "DirectSound", GUI.DSuse);
  set_config_int (cs.sound, "DirectSoundNotify", GUI.DirectSoundNotify);
  set_config_int (cs.sound, "APUEnabled", Settings.NextAPUEnabled);
  set_config_int (cs.sound, "InterpolatedSound", Settings.InterpolatedSound);
  set_config_int (cs.sound, "LowPassFilter", Settings.LowPassFilter);
  set_config_int (cs.sound, "SimpleFilter", Settings.SimpleFilter);
  set_config_int (cs.sound, "SoundBufferLength",
		  Settings.SoundBufferSize);
  set_config_int (cs.sound, "SoundMixInterval",
		  Settings.SoundMixInterval);
  set_config_int (cs.sound, "SoundGlobalFocus", GUI.SoundGlobalFocus);
  set_config_int (cs.sound, "SoundEnvelopeHeightReading",
		  Settings.SoundEnvelopeHeightReading);
  set_config_int (cs.sound, "DisableSoundEcho",
		  Settings.DisableSoundEcho);
  set_config_int (cs.sound, "DisableMasterVolume",
		  Settings.DisableMasterVolume);
  set_config_int (cs.sound, "VolumeLevel", Options.VolumeLevel);
#if 0
  set_config_int (cs.sound, "DisableSampleCaching",
		  Settings.DisableSampleCaching);
  set_config_int (cs.sound, "SampleDecodingMethod",
		  Settings.AltSampleDecode);
#endif
  set_config_int (cs.sound, "SoundSkipMethod",
		  Settings.SoundSkipMethod);
  set_config_int (cs.sound, "ReverseStereo", Settings.ReverseStereo);
#if 0
  set_config_int (cs.sound, "FixFrequency", Options.FixFrequency);
#endif

  /* Write [cpu] */
  set_config_int (cs.cpu, "Shutdown", (Settings.ShutdownMaster ? 0 : 1));
  set_config_int (cs.cpu, "HDMATimingHack", Settings.HDMATimingHack);
  set_config_int (cs.cpu, "H_MaxTimingHack", Settings.H_MaxTimingHack);
  set_config_int (cs.cpu, "OldTiming", Settings.OldTimingMaster);
  set_config_int (cs.cpu, "DisableIRQCycleCount", Settings.DisableIRQCycleCount);
  set_config_int (cs.cpu, "smp_master_clock", Settings.smp_master_clock);
  set_config_int (cs.cpu, "smp_one_cycle", Settings.APUOneCycle);


  /* Write [timer] */
  if (Settings.SkipFrames != AUTO_FRAMERATE)
    set_config_int (cs.timer, "SkipFrames", (Settings.SkipFrames - 1));
  else
    set_config_string (cs.timer, "SkipFrames", "");
  set_config_int (cs.timer, "AutoMaxSkipFrames",
		  Settings.AutoMaxSkipFrames);
  set_config_int (cs.timer, "TurboSkipFrames",
		  Settings.TurboSkipFrames);
  set_config_int (cs.timer, "NTSCtiming", Settings.FrameTimeNTSC);
  set_config_int (cs.timer, "PALtiming", Settings.FrameTimePAL);

  set_config_int (cs.timer, "SoundConstantTempo", Settings.SoundConstantTempo);


  /* Write [FileDirectory] */
  set_config_string (cs.filedirectory, "HomeDirectory",
		     Options.HomeDirectory);

  set_config_string (cs.filedirectory, "LastROMFileName", ROMFileName);
  strcpy (tmpCharBuf, "RecentROMFileName1");
  for (i = 0; i < 10; i++)
    {
      sprintf (&tmpCharBuf[17], "%X", i + 1);
      set_config_string (cs.filedirectory, (const char *) tmpCharBuf,
			 &RecentROMFileName[_MAX_PATH * i]);
    }

  _splitpath (GUI.FreezeFileDir, drive, dir, fname, ext);
  if (dir[0] == 0)
    _makepath (GUI.FreezeFileDir, drive, "\\", NULL, NULL);
  else
    _makepath (GUI.FreezeFileDir, drive, dir, fname, ext);
  set_config_string (cs.filedirectory, "SnapShotDirectory",
		     GUI.FreezeFileDir);

  _splitpath (Memory.ScreenShotDirectory, drive, dir, fname, ext);
  if (dir[0] == 0)
    _makepath (Memory.ScreenShotDirectory, drive, "\\", NULL, NULL);
  else
    _makepath (Memory.ScreenShotDirectory, drive, dir, fname, ext);
  set_config_string (cs.filedirectory, "ScreenShotDirectory",
		     Memory.ScreenShotDirectory);

  _splitpath (Memory.SPCDumpDirectory, drive, dir, fname, ext);
  if (dir[0] == 0)
    _makepath (Memory.SPCDumpDirectory, drive, "\\", NULL, NULL);
  else
    _makepath (Memory.SPCDumpDirectory, drive, dir, fname, ext);
  set_config_string (cs.filedirectory, "SPCDumpDirectory",
		     Memory.SPCDumpDirectory);

  _splitpath (Options.BinaryDumpDirectory, drive, dir, fname, ext);
  if (dir[0] == 0)
    _makepath (Options.BinaryDumpDirectory, drive, "\\", NULL, NULL);
  else
    _makepath (Options.BinaryDumpDirectory, drive, dir, fname, ext);
  set_config_string (cs.filedirectory, "BinaryDumpDirectory",
		     Options.BinaryDumpDirectory);

  set_config_string (cs.filedirectory, "BSX", Memory.BSXBIOSROMFileName);
  set_config_string (cs.filedirectory, "SFT", Memory.SFTBIOSROMFileName);

  /* Write [ControllerOption] */
  strcpy (tmpCharBuf, "JoystickConnected0");
  for (i = 0; i < 8; i++)
    {
      sprintf (&tmpCharBuf[17], "%d", i);
      set_config_int (cs.ControllerOption, (const char *) tmpCharBuf,
		      Settings.JoystickConnected[i]);
    }

  set_config_int (cs.ControllerOption, "ControllerOption", Settings.ControllerOption);
  set_config_int (cs.ControllerOption, "DirectInput", GUI.DIuse);
  set_config_int (cs.ControllerOption, "TurboShotEnable", Options.TurboShotEnable);

  /* Write [joypadA1 2 3 4 5 6 7 8] */
  for (i = 0; i < 8; i++)
    {
      set_config_int (cs.joypadA[i], cs.button[0], JoypadA[i].Up);
      set_config_int (cs.joypadA[i], cs.button[1], JoypadA[i].Down);
      set_config_int (cs.joypadA[i], cs.button[2], JoypadA[i].Left);
      set_config_int (cs.joypadA[i], cs.button[3], JoypadA[i].Right);
      set_config_int (cs.joypadA[i], cs.button[4], JoypadA[i].A);
      set_config_int (cs.joypadA[i], cs.button[5], JoypadA[i].B);
      set_config_int (cs.joypadA[i], cs.button[6], JoypadA[i].X);
      set_config_int (cs.joypadA[i], cs.button[7], JoypadA[i].Y);
      set_config_int (cs.joypadA[i], cs.button[8], JoypadA[i].Select);
      set_config_int (cs.joypadA[i], cs.button[9], JoypadA[i].Start);
      set_config_int (cs.joypadA[i], cs.button[10], JoypadA[i].L);
      set_config_int (cs.joypadA[i], cs.button[11], JoypadA[i].R);
      set_config_int (cs.joypadA[i], cs.button[12], JoypadA[i].Turbo_Up);
      set_config_int (cs.joypadA[i], cs.button[13], JoypadA[i].Turbo_Down);
      set_config_int (cs.joypadA[i], cs.button[14], JoypadA[i].Turbo_Left);
      set_config_int (cs.joypadA[i], cs.button[15], JoypadA[i].Turbo_Right);
      set_config_int (cs.joypadA[i], cs.button[16], JoypadA[i].Turbo_A);
      set_config_int (cs.joypadA[i], cs.button[17], JoypadA[i].Turbo_B);
      set_config_int (cs.joypadA[i], cs.button[18], JoypadA[i].Turbo_X);
      set_config_int (cs.joypadA[i], cs.button[19], JoypadA[i].Turbo_Y);
      set_config_int (cs.joypadA[i], cs.button[20], JoypadA[i].Turbo_Select);
      set_config_int (cs.joypadA[i], cs.button[21], JoypadA[i].Turbo_Start);
      set_config_int (cs.joypadA[i], cs.button[22], JoypadA[i].Turbo_L);
      set_config_int (cs.joypadA[i], cs.button[23], JoypadA[i].Turbo_R);
    }

  /* Write [FunctionCodeA] */
  strcpy (tmpCharBuf, "Code0");
  for (i = 0; i < 9; i++)
    {
      sprintf (&tmpCharBuf[4], "%d", i);
      set_config_int (cs.functioncodeA, (const char *) tmpCharBuf,
		      FunctionCodeA[i]);
    }

  /* Write [joypadB1 2 3 4 5 6 7 8] */
  for (i = 0; i < 8; i++)
    {
      set_config_int (cs.joypadB[i], cs.button[0], JoypadB[i].Up);
      set_config_int (cs.joypadB[i], cs.button[1], JoypadB[i].Down);
      set_config_int (cs.joypadB[i], cs.button[2], JoypadB[i].Left);
      set_config_int (cs.joypadB[i], cs.button[3], JoypadB[i].Right);
      set_config_int (cs.joypadB[i], cs.button[4], JoypadB[i].A);
      set_config_int (cs.joypadB[i], cs.button[5], JoypadB[i].B);
      set_config_int (cs.joypadB[i], cs.button[6], JoypadB[i].X);
      set_config_int (cs.joypadB[i], cs.button[7], JoypadB[i].Y);
      set_config_int (cs.joypadB[i], cs.button[8], JoypadB[i].Select);
      set_config_int (cs.joypadB[i], cs.button[9], JoypadB[i].Start);
      set_config_int (cs.joypadB[i], cs.button[10], JoypadB[i].L);
      set_config_int (cs.joypadB[i], cs.button[11], JoypadB[i].R);
      set_config_int (cs.joypadB[i], cs.button[12], JoypadB[i].Turbo_Up);
      set_config_int (cs.joypadB[i], cs.button[13], JoypadB[i].Turbo_Down);
      set_config_int (cs.joypadB[i], cs.button[14], JoypadB[i].Turbo_Left);
      set_config_int (cs.joypadB[i], cs.button[15], JoypadB[i].Turbo_Right);
      set_config_int (cs.joypadB[i], cs.button[16], JoypadB[i].Turbo_A);
      set_config_int (cs.joypadB[i], cs.button[17], JoypadB[i].Turbo_B);
      set_config_int (cs.joypadB[i], cs.button[18], JoypadB[i].Turbo_X);
      set_config_int (cs.joypadB[i], cs.button[19], JoypadB[i].Turbo_Y);
      set_config_int (cs.joypadB[i], cs.button[20], JoypadB[i].Turbo_Select);
      set_config_int (cs.joypadB[i], cs.button[21], JoypadB[i].Turbo_Start);
      set_config_int (cs.joypadB[i], cs.button[22], JoypadB[i].Turbo_L);
      set_config_int (cs.joypadB[i], cs.button[23], JoypadB[i].Turbo_R);
    }

  /* Write [FunctionCodeB] */
  strcpy (tmpCharBuf, "Code0");
  for (i = 0; i < 9; i++)
    {
      sprintf (&tmpCharBuf[4], "%d", i);
      set_config_int (cs.functioncodeB, (const char *) tmpCharBuf,
		      FunctionCodeB[i]);
    }

  /* Write [gui] */
  set_config_int (cs.gui, "Language", GUI.Language);

  set_config_int (cs.gui, "WindowRectLeft", GUI.WindowRectSetting.left);
  set_config_int (cs.gui, "WindowRectRight",
		  GUI.WindowRectSetting.right);
  set_config_int (cs.gui, "WindowRectTop", GUI.WindowRectSetting.top);
  set_config_int (cs.gui, "WindowRectBottom",
		  GUI.WindowRectSetting.bottom);

  set_config_int (cs.gui, "FileBrowseWindow", GUI.FileBrowseWindow);

  set_config_int (cs.gui, "uosXpWindowRectLeft",
		  uosxplorerposition.WindowRect.left);
  set_config_int (cs.gui, "uosXpWindowRectRight",
		  uosxplorerposition.WindowRect.right);
  set_config_int (cs.gui, "uosXpWindowRectTop",
		  uosxplorerposition.WindowRect.top);
  set_config_int (cs.gui, "uosXpWindowRectBottom",
		  uosxplorerposition.WindowRect.bottom);
  set_config_int (cs.gui, "uosXpMaximize", uosxplorerposition.Maximize);
  set_config_int (cs.gui, "uosXpVSplitBarPos0",
		  uosxplorerposition.VSplitBarPos0);
  set_config_int (cs.gui, "uosXpColumnWidth0",
		  uosxplorerposition.ColumnWidth0);
  set_config_int (cs.gui, "uosXpColumnWidth1",
		  uosxplorerposition.ColumnWidth1);
  set_config_int (cs.gui, "uosXpColumnWidth2",
		  uosxplorerposition.ColumnWidth2);
  set_config_int (cs.gui, "uosXpColumnWidth3",
		  uosxplorerposition.ColumnWidth3);


  /* Write [screenshot] */
  set_config_int (cs.screenshot, "SaveImagefileFormat",
		  Options.SaveImagefileFormat);
  set_config_int (cs.screenshot, "ReduceTheColorTo256",
		  Options.ReduceTheColorTo256);
  set_config_int (cs.screenshot, "ReduceTheColorTo256IfWithin",
		  Options.ReduceTheColorTo256IfWithin);
  set_config_int (cs.screenshot, "ImagefileHiresStretch",
		  Options.ImagefileHiresStretch);
  set_config_int (cs.screenshot, "RLEcompression256bitmap",
		  Options.RLEcompression256bitmap);

  /* Write [debug] */
  set_config_int (cs.debug, "AttachConsole",
		  GUI.AttachConsole);

  set_config_int (cs.debug, "DisableGameSpecificHacks",
		  Settings.DisableGameSpecificHacks);

  set_config_int (cs.debug, "DisplayScreenBufferAll",
		  Options.DisplayScreenBufferAll);


  /* Write [option] */
  set_config_int (cs.option, "ApplyCheats", Settings.ApplyCheats);
  set_config_int (cs.option, "ShowFPS", Settings.DisplayFrameRate);
  set_config_int (cs.option, "MuteSound", Settings.Mute);
  set_config_int (cs.option, "AutoSaveDelay", Settings.AutoSaveDelay);
  set_config_int (cs.option, "UseCDDA", Xxxcdda.Enable);
  set_config_int (cs.option, "CDDrv", Xxxcdda.Drv);

  return;
}

