#include "snes9x.h"
#include "ppu.h"
#include "uosnesw.h"
#include "windraw.h"
#include "uow_draw.h"
#include "render.h"

#ifndef max
#define max(a, b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a, b) (((a) < (b)) ? (a) : (b))
#endif

typedef struct
{
  uint8 red;
  uint8 green;
  uint8 blue;
} Colour;

static Colour FixedColours[256];
static int color_diff[0x10000];
static PALETTEENTRY S9xPaletteEntry[256];
uint8 S9xPalette[0x10000];

void
S9xSetPalette (void)
{
  LPDIRECTDRAWPALETTE lpDDTemp;

  // Only update the palette structures if needed
  if (GUI.ScreenDepth == 8)
    {
      for (int i = 0; i < 256; i++)
	{
	  S9xPaletteEntry[i].peRed = FixedColours[i].red;
	  S9xPaletteEntry[i].peGreen = FixedColours[i].green;
	  S9xPaletteEntry[i].peBlue = FixedColours[i].blue;
	}

      Renderer.lpDDSPrimary2->GetPalette (&lpDDTemp);
      if (lpDDTemp != Renderer.lpDDPalette)
	Renderer.lpDDSPrimary2->SetPalette (Renderer.lpDDPalette);
      Renderer.lpDDPalette->SetEntries (0, 0, 256, S9xPaletteEntry);
    }
}

PBYTE ScreenBuf1 = NULL;
PBYTE ScreenBuffer = NULL;
BOOL sb_avail = FALSE;

bool8
S9xInitUpdate (void)
{
  return (TRUE);
}

static inline void
ClearRenderScreen (int w, int h)
{
  if (ExtendHeight_previos != ExtendHeight &&
      !ExtendHeight && GUI.Stretch != 1 &&
      (GUI.ClientRect.right - GUI.ClientRect.left) > 0 &&
      (GUI.ClientRect.bottom - GUI.ClientRect.top) > h)
    {

      RECT clrRect;

      if (Renderer.Windowed)
	{
	  HDC clrhdc;

	  clrhdc = GetDC (GUI.hWnd);

	  clrRect.top = 0;
	  clrRect.bottom =
	    ((GUI.ClientRect.bottom - GUI.ClientRect.top) - h) / 2;
	  clrRect.left = GUI.ClientRect.left;
	  clrRect.right = GUI.ClientRect.right;

	  if ((clrRect.bottom - clrRect.top) > 0)
	    FillRect (clrhdc, &clrRect,
		      (HBRUSH) GetStockObject (BLACK_BRUSH));

	  clrRect.top =
	    ((GUI.ClientRect.bottom - GUI.ClientRect.top) - h) / 2 + h;
	  clrRect.bottom = GUI.ClientRect.bottom;
	  clrRect.left = GUI.ClientRect.left;
	  clrRect.right = GUI.ClientRect.right;

	  if ((clrRect.bottom - clrRect.top) > 0)
	    FillRect (clrhdc, &clrRect,
		      (HBRUSH) GetStockObject (BLACK_BRUSH));

	  ReleaseDC (GUI.hWnd, clrhdc);
	}
      else
	{
	  clrRect.top = 0;
	  clrRect.bottom =
	    ((GUI.ClientRect.bottom - GUI.ClientRect.top) - h) / 2;
	  clrRect.left = GUI.ClientRect.left;
	  clrRect.right = GUI.ClientRect.right;

	  if ((clrRect.bottom - clrRect.top) > 0)
	    {
	      DirectX_FillPrimarySurface (&clrRect);
	      if (!Renderer.Windowed && Renderer.DoubleBuffered)
		DirectX_FillBackbufferSurface (&clrRect);
	    }

	  clrRect.top =
	    ((GUI.ClientRect.bottom - GUI.ClientRect.top) - h) / 2 + h;
	  clrRect.bottom = GUI.ClientRect.bottom;
	  clrRect.left = GUI.ClientRect.left;
	  clrRect.right = GUI.ClientRect.right;

	  if ((clrRect.bottom - clrRect.top) > 0)
	    {
	      DirectX_FillPrimarySurface (&clrRect);
	      if (!Renderer.Windowed && Renderer.DoubleBuffered)
		DirectX_FillBackbufferSurface (&clrRect);
	    }
	}
    }
}

static RECT dstRect = { 0, 512, 0, 478 };
static RECT srect, drect;
static POINT p_bak;
static RECT GUI_ClientRect_bak;
static BOOL GUI_DoubleRendering_bak;
static int GUI_Stretch_bak;
static int GUI_FilterMethod_bak;

void
GetRenderedRect (LPRECT lpRect)
{
  lpRect->top = dstRect.top;
  lpRect->left = dstRect.left;
  lpRect->bottom = dstRect.bottom;
  lpRect->right = dstRect.right;
}

bool8 CheckRenderRectChanged()
{
  POINT p;

  p.x = 0;
  p.y = 0;
  ClientToScreen (GUI.hWnd, &p);

  if(p_bak.x != p.x ||
     p_bak.y != p.y ||
     GUI_ClientRect_bak.top != GUI.ClientRect.top ||
     GUI_ClientRect_bak.bottom != GUI.ClientRect.bottom ||
     GUI_ClientRect_bak.left != GUI.ClientRect.left ||
     GUI_ClientRect_bak.right != GUI.ClientRect.right ||
     GUI_DoubleRendering_bak != GUI.DoubleRendering ||
     GUI_Stretch_bak != GUI.Stretch ||
     GUI_FilterMethod_bak != GUI.FilterMethod)
    return TRUE;
  return FALSE;
}

static inline void GetDstSize(int Width, int Height, int iWidth, int iHeight, int &dWidth, int &dHeight)
{
  int w = GUI.ClientRect.right - GUI.ClientRect.left;
  int h = GUI.ClientRect.bottom - GUI.ClientRect.top;

  if (GUI.Stretch == 1)
    {
      dWidth = w;
      dHeight = h;
      return;
    }
  else//if (GUI.Stretch == 3 || GUI.Stretch == 2 || GUI.DoubleRendering || !GUI.DoubleRendering)
    {
      int ww = GUI.Stretch == 3 ? 301 : 256;
      int hh = !ExtendHeight ? 224 : 239;

      int xx = min(w / ww, h / hh);

      if(GUI.Stretch != 2 && GUI.Stretch != 3)
	{
	  int mxx = GUI.DoubleRendering ? 2 : 1;

	  if(xx > mxx)
	    xx = mxx;
	}

      dWidth = ww * xx;
      dHeight = hh * xx;

      if(dWidth == 0 && dHeight == 0)
	{
	  if(w >= ww)
	    dWidth = ww;
	  else
	    dWidth = w;
	  if(h >= hh)
	    dHeight = hh;
	  else
	    dHeight = h;
	}

      if(GUI.Stretch == 2 || GUI.Stretch == 3)
	return;
    }

  if((DoubleWidth || DoubleHeight) && Width == iWidth && Height == iHeight)
    return;

  if(w >= iWidth && iWidth > dWidth)
    dWidth = iWidth;

  if(h >= iHeight && iHeight > dHeight)
    dHeight = iHeight;
}

BOOL uowRender(int Width, int Height)
{
  if(!Renderer.DDAvailable)
    return FALSE;

  if (Options.DisplayScreenBufferAll)
    {
      Width = 512;
      Height = 478;
    }

  int iWidth, iHeight;
  SelectRenderMethod (GUI.FilterMethod,
		      Width, Height,
		      GUI.ClientRect.right - GUI.ClientRect.left,
		      GUI.ClientRect.bottom - GUI.ClientRect.top,
		      iWidth, iHeight);

  int dWidth, dHeight;
  GetDstSize(Width, Height, iWidth, iHeight, dWidth, dHeight);
  ClearRenderScreen (dWidth, dHeight);

  POINT p;
  p.x = ((GUI.ClientRect.right - GUI.ClientRect.left) - dWidth) / 2;
  p.y = ((GUI.ClientRect.bottom - GUI.ClientRect.top) - dHeight) / 2;
  ClientToScreen (GUI.hWnd, &p);
  dstRect.left = p.x;
  dstRect.top = p.y;
  dstRect.right = dstRect.left + dWidth;
  dstRect.bottom = dstRect.top + dHeight;

  srect.left = 0;
  srect.right = srect.left + Width;
  srect.top = 0;
  srect.bottom = srect.top + Height;

  drect.left = 0;
  drect.right = drect.left + iWidth;
  drect.top = 0;
  drect.bottom = drect.top + iHeight;

  SSurface Src, Dst;

  Src.Surface = ScreenBuffer;
  Src.Pitch = GFX.Pitch;
  Src.Width = Width;
  Src.Height = Height;

  if(Renderer.lpDDSOffScreen2->IsLost() == DDERR_SURFACELOST)
    Renderer.lpDDSOffScreen2->Restore();

  if (LockSurface2 (Renderer.lpDDSOffScreen2, &Dst) == FALSE)
    return FALSE;

  RenderMethod (&Src, &Dst, &srect, &drect);
  Renderer.lpDDSOffScreen2->Unlock (Dst.Surface);

  //CheckRenderRectChanged()
  p.x = 0;
  p.y = 0;
  ClientToScreen (GUI.hWnd, &p);
  p_bak = p;
  GUI_ClientRect_bak = GUI.ClientRect;
  GUI_DoubleRendering_bak = GUI.DoubleRendering;
  GUI_Stretch_bak = GUI.Stretch;
  GUI_FilterMethod_bak = GUI.FilterMethod;

  return TRUE;
}

static BOOL uowFlip()
{
  BOOL _backbuffer;
  LPDIRECTDRAWSURFACE2 lpDDSurface2;

  //fullscreen DoubleBuffered
  if (!Renderer.Windowed && Renderer.DoubleBuffered)
    {
      _backbuffer = TRUE;
      lpDDSurface2 = Renderer.lpDDSBack2;
    }
  else
    {
      _backbuffer = FALSE;
      lpDDSurface2 = Renderer.lpDDSPrimary2;
    }

  if(lpDDSurface2->IsLost() == DDERR_SURFACELOST)
    lpDDSurface2->Restore();

  vsync_method();

  if (lpDDSurface2->Blt
      (&dstRect, Renderer.lpDDSOffScreen2, &drect, DDBLT_WAIT,
       NULL) != DD_OK)
    {
      return FALSE;
    }
  
  if (_backbuffer)
    {

      if(Renderer.lpDDSPrimary2->IsLost() == DDERR_SURFACELOST)
	Renderer.lpDDSPrimary2->Restore();
      Renderer.lpDDSPrimary2->Flip (NULL, Renderer.wait_for_vsync == 3 ? DDFLIP_NOVSYNC : DDFLIP_WAIT);
    }
  return TRUE;
}

bool8
S9xDeinitUpdate (int Width, int Height)
{
  if(!uowRender(Width, Height) || !uowFlip())
    return FALSE;
  return TRUE;
}

void
uowScreenRedraw (HWND hWnd)
{
  if (!Renderer.Windowed)
    return;

  if(Renderer.wait_for_vsync == 3)
    {
      EnterCriticalSection (&Renderer.cs);
      S9xDeinitUpdate (IPPU.RenderedScreenWidth, IPPU.RenderedScreenHeight);
      LeaveCriticalSection (&Renderer.cs);
    }
  else
    S9xDeinitUpdate (IPPU.RenderedScreenWidth, IPPU.RenderedScreenHeight);
}
  
void VsyncTimerProc(void *tParams)
{
  EnterCriticalSection (&Renderer.cs);
  if(sb_avail)
    uowFlip();
  sb_avail = FALSE;
  LeaveCriticalSection (&Renderer.cs);
}

BOOL FlipProc(void *tParams)
{
  
}

static int
ffs (uint32 mask)
{
  int m = 0;
  if (mask)
    {
      while (!(mask & (1 << m)))
	m++;

      return (m);
    }

  return (0);
}

void
S9xSetWinPixelFormat ()
{
  extern void hq2x_init (unsigned);

  S9xSetRenderPixelFormat (RGB565);
  hq2x_init (16);
  GUI.NeedDepthConvert = FALSE;

  GUI.ScreenDepth = (int)Renderer.DDPixelFormat.dwRGBBitCount;
  if (GUI.ScreenDepth == 15)
    GUI.ScreenDepth = 16;

  GUI.RedShift = ffs (Renderer.DDPixelFormat.dwRBitMask);
  GUI.GreenShift = ffs (Renderer.DDPixelFormat.dwGBitMask);
  GUI.BlueShift = ffs (Renderer.DDPixelFormat.dwBBitMask);

  if ((Renderer.DDPixelFormat.dwFlags & DDPF_RGB) != 0 &&
      GUI.ScreenDepth == 16 &&
      Renderer.DDPixelFormat.dwRBitMask == 0xF800 &&
      Renderer.DDPixelFormat.dwGBitMask == 0x07E0 &&
      Renderer.DDPixelFormat.dwBBitMask == 0x001F)
    {
      GUI.GreenShift += 1;
      S9xSetRenderPixelFormat (RGB565);
      hq2x_init (16);
    }
  else
    if ((Renderer.DDPixelFormat.dwFlags & DDPF_RGB) != 0 &&
	GUI.ScreenDepth == 16 &&
	Renderer.DDPixelFormat.dwRBitMask == 0x7C00 &&
	Renderer.DDPixelFormat.dwGBitMask == 0x03E0 &&
	Renderer.DDPixelFormat.dwBBitMask == 0x001F)
    {
      S9xSetRenderPixelFormat (RGB555);
      hq2x_init (15);
    }
  else
    if ((Renderer.DDPixelFormat.dwFlags & DDPF_RGB) != 0 &&
	GUI.ScreenDepth == 16 &&
	Renderer.DDPixelFormat.dwRBitMask == 0x001F &&
	Renderer.DDPixelFormat.dwGBitMask == 0x07E0 &&
	Renderer.DDPixelFormat.dwBBitMask == 0xF800)
    {
      GUI.GreenShift += 1;
      S9xSetRenderPixelFormat (BGR565);
      hq2x_init (16);
    }
  else
    if ((Renderer.DDPixelFormat.dwFlags & DDPF_RGB) != 0 &&
	GUI.ScreenDepth == 16 &&
	Renderer.DDPixelFormat.dwRBitMask == 0x001F &&
	Renderer.DDPixelFormat.dwGBitMask == 0x03E0 &&
	Renderer.DDPixelFormat.dwBBitMask == 0x7C00)
    {
      S9xSetRenderPixelFormat (BGR555);
      hq2x_init (15);
    }
  else
    if (Renderer.DDPixelFormat.dwRGBBitCount == 8 ||
	Renderer.DDPixelFormat.dwRGBBitCount == 24 ||
	Renderer.DDPixelFormat.dwRGBBitCount == 32)
    {
      S9xSetRenderPixelFormat (RGB565);
      hq2x_init (16);
    }

  if (GUI.ScreenDepth == 8 || GUI.ScreenDepth == 24 || GUI.ScreenDepth == 32)
    GUI.NeedDepthConvert = TRUE;

  if (GUI.ScreenDepth == 24 || GUI.ScreenDepth == 32)
    {
      GUI.RedShift += 3;
      GUI.GreenShift += 3;
      GUI.BlueShift += 3;
    }

  int l = 0;
  int i;
  for (i = 0; i < 6; i++)
    {
      int r = (i * 31) / (6 - 1);
      for (int j = 0; j < 6; j++)
	{
	  int g = (j * 31) / (6 - 1);
	  for (int k = 0; k < 6; k++)
	    {
	      int b = (k * 31) / (6 - 1);
	      FixedColours[l].red = r << 3;
	      FixedColours[l].green = g << 3;
	      FixedColours[l++].blue = b << 3;
	    }
	}
    }

  //int *color_diff = new int [0x10000];
  int diffr, diffg, diffb, maxdiff = 0, won = 0, lost;
  int r, d = 8;
  for (r = 0; r <= (int) MAX_RED; r++)
    {
      int cr, g, q;
      int k = 6 - 1;
      cr = (r * k) / MAX_RED;
      q = (r * k) % MAX_RED;
      if (q > d && cr < k)
	cr++;
      diffr = abs (cr * k - r);
      for (g = 0; g <= (int) MAX_GREEN; g++)
	{
	  int cg, b;
	  k = 6 - 1;
	  cg = (g * k) / MAX_GREEN;
	  q = (g * k) % MAX_GREEN;
	  if (q > d && cg < k)
	    cg++;
	  diffg = abs (cg * k - g);
	  for (b = 0; b <= (int) MAX_BLUE; b++)
	    {
	      int cb;
	      int rgb = BUILD_PIXEL2 (r, g, b);
	      k = 6 - 1;
	      cb = (b * k) / MAX_BLUE;
	      q = (b * k) % MAX_BLUE;
	      if (q > d && cb < k)
		cb++;
	      diffb = abs (cb * k - b);
	      S9xPalette[rgb] = (cr * 6 + cg) * 6 + cb;
	      color_diff[rgb] = diffr + diffg + diffb;
	      if (color_diff[rgb] > maxdiff)
		maxdiff = color_diff[rgb];
	    }
	}
    }

  while (maxdiff > 0 && l < 256)
    {
      int newmaxdiff = 0;
      lost = 0;
      won++;
      for (r = MAX_RED; r >= 0; r--)
	{
	  int g;
	  for (g = MAX_GREEN; g >= 0; g--)
	    {
	      int b;
	      for (b = MAX_BLUE; b >= 0; b--)
		{
		  int rgb = BUILD_PIXEL2 (r, g, b);
		  if (color_diff[rgb] == maxdiff)
		    {
		      if (l >= 256)
			lost++;
		      else
			{
			  FixedColours[l].red = r << 3;
			  FixedColours[l].green = g << 3;
			  FixedColours[l].blue = b << 3;
			  S9xPalette[rgb] = l++;
			}
		      color_diff[rgb] = 0;
		    }
		  else if (color_diff[rgb] > newmaxdiff)
		    newmaxdiff = color_diff[rgb];
		}
	    }
	}
      maxdiff = newmaxdiff;
    }
}

