#include "snes9x.h"
#include "apu_orig.h"
#include "soundux.h"
#include "snesapu.h"
#include "uosnesw.h"
#include "winsound.h"
#include "uow_sound.h"

static bool8 hCriticalSection = FALSE;
CRITICAL_SECTION ssmcs;

static void
SNESSoundCallback (void *buff, int len)
{
  EnterCriticalSection (&ssmcs);
  S9xMixSamplesO ((uint8 *) buff, len >> (so.sixteen_bit ? 1 : 0), 0);
  LeaveCriticalSection (&ssmcs);
}

static BOOL
OpenSoundEvent ()
{
  if (hCriticalSection == FALSE)
    {
      InitializeCriticalSection (&ssmcs);
      hCriticalSection = TRUE;
    }
  return TRUE;
}

static void
CloseSoundEvent ()
{
  if (hCriticalSection)
    {
      DeleteCriticalSection (&ssmcs);
      hCriticalSection = FALSE;
    }
}

bool8
uowSetupSound (long rate, bool8 sixteen_bit, bool8 stereo, HWND hWnd)
{
  WinSoundDeinit ();
  CloseSoundEvent ();

  if (!OpenSoundEvent ())
    {
      CloseSoundEvent ();
      return FALSE;
    }

  Settings.Stereo = stereo ? TRUE : FALSE;
  Settings.SoundPlaybackRate = rate;
  Settings.SixteenBitSound = sixteen_bit ? TRUE : FALSE;

  so.playback_rate = Settings.SoundPlaybackRate;
  so.stereo = Settings.Stereo;
  so.sixteen_bit = Settings.SixteenBitSound;
  //so.buffer_size = (int) (buffersize / Settings.SoundBufferSize);
  so.encoded = FALSE;

  S9xSetPlaybackRate (so.playback_rate);
  SetBufferSize (Settings.SoundMixInterval);

  WAVEFORMATEX wfx;
  DWORD buffersize;

  bool8 r = (bool8) WinSetupSound (rate, (BOOL) sixteen_bit, (BOOL) stereo,
				   Settings.SoundBufferSize,
				   Settings.SoundMixInterval,
				   SNESSoundCallback,
				   &wfx, &buffersize,
				   hWnd, GUI.DSuse,
				   GUI.SoundGlobalFocus,
				   GUI.DirectSoundNotify);

  if (!r && GUI.DSuse && GUI.DirectSoundNotify)
    {
      BOOL ds_notify = GUI.DirectSoundNotify;
      ds_notify = !ds_notify;
      r = (bool8) WinSetupSound (rate, (BOOL) sixteen_bit, (BOOL) stereo,
				 Settings.SoundBufferSize,
				 Settings.SoundMixInterval, SNESSoundCallback,
				 &wfx, &buffersize, hWnd, GUI.DSuse,
				 GUI.SoundGlobalFocus, ds_notify);

      if (r)
	GUI.DirectSoundNotify = ds_notify;
    }

  if (!r)
    {
      BOOL dsuse = GUI.DSuse;
      dsuse = !dsuse;
      r = (bool8) WinSetupSound (rate, (BOOL) sixteen_bit, (BOOL) stereo,
				 Settings.SoundBufferSize,
				 Settings.SoundMixInterval, SNESSoundCallback,
				 &wfx, &buffersize, hWnd, dsuse,
				 GUI.SoundGlobalFocus, GUI.DirectSoundNotify);

      if (!r)
	return FALSE;

      GUI.DSuse = dsuse;
    }

  return TRUE;
}

void
uowReleaseSound ()
{
  WinSoundDeinit ();
  CloseSoundEvent ();
}

void WriteSamples(int32 l, int32 r)
{
  EnterCriticalSection (&ssmcs);
  WriteSamples0(l, r);
  LeaveCriticalSection (&ssmcs);
}

void
S9xGenerateSound ()
{
}

