#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "snapshot.h"
#include "apu_orig.h"
#include "display.h"
#include "controls.h"
#include "controls2.h"
#include "uosnesw.h"
#include "windraw.h"
#include "wintimer.h"
#include "uow_draw.h"
#include "uow_sound.h"
#include "uow_input.h"
#include "render.h"

#include <direct.h>
#include <io.h>

void
S9xParseArg (char **argv, int &i, int argc)
{
#if 0
  if (strcasecmp (argv[i], "-restore") == 0)
    {
      WinDeleteRegistryEntries ();
      WinSetDefaultValues ();
    }
  if (strcasecmp (argv[i], "-hidemenu") == 0)
    {
      GUI.HideMenu = true;
    }
#endif
}

void
S9xExtraUsage ()
{
}

void
S9xTextMode (void)
{
}

void
S9xGraphicsMode ()
{
}

void
S9xExit (void)
{
  MessageBox (NULL, Languages[GUI.Language].S9xExitMessage, "uosnesw",
	      MB_OK | MB_ICONSTOP);
  SendMessage (GUI.hWnd, WM_COMMAND, ID_FILE_EXIT, 0);
}

const char *
GetHomeDirectory ()
{
  static char filename[_MAX_PATH + 1];
  char drive[_MAX_DRIVE + 1];
  char dir[_MAX_DIR + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];

  _splitpath (__argv[0], drive, dir, fname, ext);
  _makepath (filename, drive, dir, NULL, NULL);
  int len = strlen (filename);
  if (len > 0 && filename[--len] == SLASH_CHAR)
    filename[len] = 0;
  return (const char *) filename;
}

extern char InitialROMFileDirectory[];

const char *
S9xGetDirectory (enum s9x_getdirtype dirtype)
{
  static char dirname[PATH_MAX + 1];
  char dir[_MAX_DIR + 1];
  char drive[_MAX_DRIVE + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];
  int len;

  const char *rv = GUI.FreezeFileDir;

  switch (dirtype)
    {
    case DEFAULT_DIR:
    case HOME_DIR:
      rv = Options.HomeDirectory;
      break;

    case SCREENSHOT_DIR:
      rv = Memory.ScreenShotDirectory;
      break;

    case ROM_DIR:
      _splitpath (InitialROMFileDirectory, drive, dir, fname, ext);
      _makepath (dirname, drive, dir, fname, ext);
      len = lstrlen (dirname);
      if (fname[0] == 0 && len > 0 && dirname[len - 1] == '\\')
	dirname[len - 1] = 0;
      rv = dirname;
      break;

    case SRAM_DIR:
      rv = GUI.FreezeFileDir;
      break;

    case BIOS_DIR:
      rv = Options.HomeDirectory;
      break;

    case SPC_DIR:
      rv = Memory.SPCDumpDirectory;
      break;

    case IPS_DIR:
    case CHEAT_DIR:
      rv = GUI.FreezeFileDir;
      break;

    case SNAPSHOT_DIR:
      rv = GUI.FreezeFileDir;
      break;

    case ROMFILENAME_DIR:
      _splitpath (InitialROMFileDirectory, drive, dir, fname, ext);
      _makepath (dirname, drive, dir, fname, ext);
      len = lstrlen (dirname);
      if (fname[0] == 0 && len > 0 && dirname[len - 1] == '\\')
	dirname[len - 1] = 0;
      rv = dirname;
      break;

    case BINDUMP_DIR:		//uosnes added
      rv = Options.BinaryDumpDirectory;
      break;

    default:
      break;
    }
  return rv;
}


const char *
S9xGetFilename (const char *ex, enum s9x_getdirtype dirtype)
{
  static char filename[PATH_MAX + 1];
  char dir[_MAX_DIR + 1];
  char drive[_MAX_DRIVE + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];
  _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
  _snprintf (filename, sizeof (filename), "%s" SLASH_STR "%s%s",
	     S9xGetDirectory (dirtype), fname, ex);
  return (filename);
}

static char numext[256];

const char *
S9xGetFilenameInc (const char *e, enum s9x_getdirtype dirtype)
{
  static char filename[PATH_MAX + 1];
  char dir[_MAX_DIR + 1];
  char drive[_MAX_DRIVE + 1];
  char fname[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];
  unsigned int i = 0;
  const char *d;

  _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
  d = S9xGetDirectory (dirtype);
  do
    {
      _snprintf (filename, sizeof (filename), "%s\\%s_%03d%s", d, fname, i,
		 e);
      sprintf (numext, "_%03d%s", i, e);
      i++;
    }
  while (_access (filename, 0) == 0 && i < 1000);

  return (filename);
}

const char *
S9xGetFilenameIncNumExt ()	//uosnes
{
  return (numext);
}

bool8
S9xOpenSnapshotFile (const char *fname, bool8 read_only, STREAM * file)
{
  char filename[_MAX_PATH + 1];
  char drive[_MAX_DRIVE + 1];
  char dir[_MAX_DIR + 1];
  char fn[_MAX_FNAME + 1];
  char ext[_MAX_EXT + 1];

  _splitpath (fname, drive, dir, fn, ext);
  _makepath (filename, drive, dir, fn, ext[0] == '\0' ? ".000" : ext);

  if (read_only)
    {
      if ((*file = OPEN_STREAM (filename, "rb")))
	return (TRUE);
    }
  else
    {
      if ((*file = OPEN_STREAM (filename, "wb")))
	return (TRUE);
      FILE *fs = fopen (filename, "rb");
      if (fs)
	{
	  sprintf (String, "Freeze file \"%s\" exists but is read only",
		   filename);
	  fclose (fs);
	  S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
	}
      else
	{
	  sprintf (String,
		   "Cannot create freeze file \"%s\". Directory is read-only or does not exist.",
		   filename);

	  S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
	}
    }
  return (FALSE);
}

void
S9xCloseSnapshotFile (STREAM file)
{
  CLOSE_STREAM (file);
}

void
S9xMessage (int type, int, const char *str)
{
#ifdef DEBUGGER
  static FILE *out = NULL;

  if (out == NULL)
    out = fopen ("out.txt", "w");

  fprintf (out, "%s\n", str);
#endif

  S9xSetInfoString (str);

  // if we can't draw on the screen, messagebox it
  // also send to stderr/stdout depending on message type
  switch (type)
    {
    case S9X_INFO:
      if (Settings.StopEmulation)
	fprintf (stdout, "%s\n", str);
      break;
    case S9X_WARNING:
      fprintf (stdout, "%s\n", str);
      if (Settings.StopEmulation)
	MessageBox (GUI.hWnd, str, "Warning", MB_OK | MB_ICONWARNING);
      break;
    case S9X_ERROR:
      fprintf (stderr, "%s\n", str);
      if (Settings.StopEmulation)
	MessageBox (GUI.hWnd, str, "Error", MB_OK | MB_ICONERROR);
      break;
    case S9X_FATAL_ERROR:
      fprintf (stderr, "%s\n", str);
      if (Settings.StopEmulation)
	MessageBox (GUI.hWnd, str, "Fatal Error", MB_OK | MB_ICONERROR);
      break;
    default:
      fprintf (stdout, "%s\n", str);
      break;
    }
}

// MYO
void
S9xHandlePortCommand (s9xcommand_t cmd, int16 data1, int16 data2)
{
  return;
}

//  NYI
const char *
S9xChooseFilename (bool8 read_only)
{
  return NULL;
}

// NYI
const char *
S9xChooseMovieFilename (bool8 read_only)
{
  return NULL;
}

const char *
S9xStringInput (const char *msg)
{
  return NULL;
}

void
S9xToggleSoundChannel (int c)
{
}

bool
S9xPollButton (uint32 id, bool * pressed)
{
  *pressed = false;

  if (id & k_MO)		// mouse
    {
      switch (id & 0xFF)
	{
	case 0:
	  *pressed = (GUI.MouseButtons & 1) /* Left */ ;
	  break;
	case 1:
	  *pressed = (GUI.MouseButtons & 2) /* Right */ ;
	  break;
	}
    }
  else if (id & k_SS)	// superscope
    {
      switch (id & 0xFF)
	{
	case 0:
	  *pressed = GUI.MouseX <= 0 || GUI.MouseY <= 0
	    || GUI.MouseX >= IPPU.RenderedScreenWidth
	    || GUI.MouseY >= IPPU.RenderedScreenHeight;
	  break;
	case 2:
	  *pressed = (GUI.MouseButtons & 2) /* Right */ ;
	  break;
	case 3:
	  *pressed = (GUI.MouseButtons & 4) /* Middle */ 
	    || FunctionCodeStates[MF_SSTURBO];
	  break;
	case 4:
	  *pressed = FunctionCodeStates[MF_SSPAUSE];
	  break;
	case 1:
	  *pressed = (GUI.MouseButtons & 1) /* Left */ ;
	  break;
	}
    }
  else if (id & k_LG)	// justifier
    {
      if (id & k_C1)
	{
	  switch (id & 0xFF)
	    {
	    case 0:
	      *pressed = GUI.MouseX <= 0 || GUI.MouseY <= 0
		|| GUI.MouseX >= IPPU.RenderedScreenWidth
		|| GUI.MouseY >= IPPU.RenderedScreenHeight
		|| (GUI.MouseButtons & 4);
	      break;
	    case 1:
	      *pressed = (GUI.MouseButtons & 1) /* Left */ ;
	      break;
	    case 2:
	      *pressed = (GUI.MouseButtons & 2) /* Right */ ;
	      break;
	    }
	}
      else
	{
	  switch (id & 0xFF)
	    {
	    case 0:
	      *pressed = FunctionCodeStates[MF_LG2_OFFSCREEN];
	      break;
	    case 1:
	      *pressed = FunctionCodeStates[MF_LG2_SHOT];
	      break;
	    case 2:
	      *pressed = FunctionCodeStates[MF_LG2_PAUSE];
	      break;
	    }
	}
    }

  return (true);
}

bool
S9xPollPointer (uint32 id, int16 * x, int16 * y)
{
  if (id & k_PT)
    {
      *x = GUI.MouseX;
      *y = GUI.MouseY;
    }
  else
    *x = *y = 0;
  return (true);
}

bool
S9xPollAxis (uint32 id, int16 * value)
{
  return FALSE;
}

bool8
S9xContinueUpdate (int Width, int Height)
{
  return TRUE;
}

bool8
S9xOpenSoundDevice (int mode, bool8 pStereo, int BufferSize)
{
  return (TRUE);
}

static uint32 IPPU_DisplayedRenderedFrameCount_bak;
static int32 Memory_ROMFramesPerSecond_bak;
static uint8 PPU_BGMode_bak;
static uint8 PPU_0x2133_bak;

void
S9xSyncSpeed (void)
{
  /* Display status bar message */
  if (GUI.hStatusWnd)
    {
      DWORD sb_needrefresh =
	((IPPU.DisplayedRenderedFrameCount !=
	  IPPU_DisplayedRenderedFrameCount_bak
	  || Memory.ROMFramesPerSecond !=
	  Memory_ROMFramesPerSecond_bak) ? SB_NEEDREFRESH_FPS :
	 SB_NEEDREFRESH_NONE) | ((PPU.BGMode != PPU_BGMode_bak
				  ||
				  ((PPU_0x2133_bak ^ Memory.FillRAM[0x2133]) &
				   0x08)) ? SB_NEEDREFRESH_BGMODE :
				 SB_NEEDREFRESH_NONE);
      if (sb_needrefresh)
	{
	  IPPU_DisplayedRenderedFrameCount_bak =
	    IPPU.DisplayedRenderedFrameCount;
	  Memory_ROMFramesPerSecond_bak = Memory.ROMFramesPerSecond;
	  PPU_BGMode_bak = PPU.BGMode;
	  PPU_0x2133_bak = Memory.FillRAM[0x2133];
	  uowDisplayStatusInfo (sb_needrefresh);
	}
    }

  if (!Settings.SPCPlaying && IPPU.RenderThisFrame
      && !(GFX.DoInterlace && GFX.InterlaceFrame == 0))
    {
      if (Renderer.wait_for_vsync == 3)
	{
	  //added This code
	  if(sb_avail && !CheckRenderRectChanged())
	    WaitForSingleObject (Renderer.hVsyncWaitEvent, 100);

	  EnterCriticalSection (&Renderer.cs);
	  sb_avail =
	    uowRender (IPPU.RenderedScreenWidth, IPPU.RenderedScreenHeight);
	  LeaveCriticalSection (&Renderer.cs);
	}
      else
	S9xDeinitUpdate (IPPU.RenderedScreenWidth, IPPU.RenderedScreenHeight);
    }

  bool8 do_frame_adjust = false;

  if ((!Settings.TurboMode && !Settings.FrameAdvance
       && Settings.SkipFrames == AUTO_FRAMERATE)
      || (XxxMovie.Replay_rec == TRUE || XxxMovie.Replay_play == TRUE))
    {
      LONG prev;
      BOOL success;

      if ((success = ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, &prev)) &&
	  prev == 0)
	{
	  WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
	  WaitForSingleObject (GUI.FrameTimerSemaphore, 100);
	}
      else
	{
	  if (success)
	    WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
	  WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
	  do_frame_adjust = true;
	}
    }
  else
    {
      //wait code
      static uint32 TurboSkipFrames;
      if (++TurboSkipFrames >= Settings.TurboSkipFrames)
	{
	  TurboSkipFrames = 0;

	  LONG prev;
	  BOOL success;

	  if ((success = ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, &prev))
	      && prev == 0)
	    {
	      WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
	      WaitForSingleObject (GUI.FrameTimerSemaphore, 100);
	    }
	  else
	    {
	      if (success)
		WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
	      WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
	    }
	}
    }

  if ((!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE) ||
      (XxxMovie.Replay_rec == TRUE || XxxMovie.Replay_play == TRUE))
    {
      if (!do_frame_adjust)
	{
	  IPPU.RenderThisFrame = TRUE;
	  IPPU.SkippedFrames = 0;
	}
      else
	{
	  if (IPPU.SkippedFrames < Settings.AutoMaxSkipFrames)
	    {
	      IPPU.SkippedFrames++;
	      IPPU.RenderThisFrame = FALSE;
	    }
	  else
	    {
	      IPPU.RenderThisFrame = TRUE;
	      IPPU.SkippedFrames = 0;
	    }
	}
    }
  else
    {
      uint32 SkipFrames;
      if (Settings.TurboMode)
	SkipFrames = Settings.TurboSkipFrames;
      else
	SkipFrames =
	  (Settings.SkipFrames == AUTO_FRAMERATE) ? 0 : Settings.SkipFrames;
      if (++IPPU.FrameSkip >= SkipFrames)
	{
	  IPPU.FrameSkip = 0;
	  IPPU.SkippedFrames = 0;
	  IPPU.RenderThisFrame = TRUE;
	}
      else
	{
	  IPPU.SkippedFrames++;
	  IPPU.RenderThisFrame = FALSE;
	}
    }
}

const char *
S9xBasename (const char *f)
{
  const char *p;
  const char *pp = f;
  size_t slen = lstrlen (f);
  //SJIS string
  for (p = f; p < (f + slen); p++)
    {
      if (*(p + 1)
	  && ((*(unsigned char *) p >= 0x81 && *(unsigned char *) p <= 0x9f)
	      || (*(unsigned char *) p >= 0xe0
		  && *(unsigned char *) p <= 0xfc))
	  &&
	  ((*(unsigned char *) (p + 1) >= 0x40
	    && *(unsigned char *) (p + 1) <= 0x7e)
	   || (*(unsigned char *) (p + 1) >= 0x80
	       && *(unsigned char *) (p + 1) <= 0xfc)))
	++p;
      else if (*p == (const char) '\\')
	pp = (p + 1);
    }
  return pp;
}

void
S9xAutoSaveSRAM ()
{
  Memory.SaveSRAM (S9xGetFilename (".srm", SRAM_DIR));
}

void
S9xSetPause (uint32 mask)
{
  Settings.ForcedPause |= mask;
}

void
S9xClearPause (uint32 mask)
{
  Settings.ForcedPause &= ~mask;
  if (!Settings.ForcedPause)
    {
      // Wake up the main loop thread just if its blocked in a GetMessage
      // call.
      PostMessage (GUI.hWnd, WM_NULL, 0, 0);
    }
}

BOOL
uowInit (HWND hWnd)
{
  if (!Memory.Init ())
    {
      MessageBox (hWnd,
		  Languages[GUI.Language].errMemoryInit,
		  "uosnes - Memory Init", MB_OK | MB_ICONSTOP);

      return FALSE;
    }

  if (!DirectX_Init (hWnd, GUI.dwStyleWindowed, GUI.dwStyleFullScreen))
    {
      MessageBox (hWnd,
		  Languages[GUI.Language].errInitDD,
		  "uosnes - DirectX", MB_OK | MB_ICONSTOP);
      return FALSE;
    }

  //check display mode support
  const char *dms = GetDisplayModeString(&Renderer.dm);

  if (strcmp (GUI.DisplayModeString, dms))
    {
      strcpy (GUI.DisplayModeString, dms);
      GUI.FullscreenDisplayModeIndex = GetDefaultDisplayMode ();
    }

  //Invalid GUI.FullscreenDisplayModeIndex value
  if (GUI.FullscreenDisplayModeIndex < 0
      || GUI.FullscreenDisplayModeIndex >= Renderer.dm.size ())
    GUI.FullscreenDisplayModeIndex = GetDefaultDisplayMode ();

  if (!DirectX_SetDisplayMode
      (Renderer.dm[GUI.FullscreenDisplayModeIndex].Width,
       Renderer.dm[GUI.FullscreenDisplayModeIndex].Height,
       GetFilterScale (GUI.FilterMethod),
       Renderer.dm[GUI.FullscreenDisplayModeIndex].Depth,
       Renderer.dm[GUI.FullscreenDisplayModeIndex].Rate,
       TRUE, GUI.DoubleBuffered))
    {
      MessageBox (hWnd,
		  Languages[GUI.Language].errModeDD,
		  "uosnes - DirectDraw", MB_OK | MB_ICONSTOP);
      DirectX_Deinit ();
      return FALSE;
    }

  GUI.ScreenDepth = Renderer.Depth;

  S9xSetWinPixelFormat ();

  if (!RenderInit (GUI.FilterMethod,
		   GUI.ScreenDepth,
		   GUI.RedShift, GUI.GreenShift, GUI.BlueShift))
    return FALSE;

  ScreenBuf1 = new BYTE[EXT_PITCH * EXT_HEIGHT];

  if (!ScreenBuf1)
    return FALSE;

  ScreenBuffer = ScreenBuf1 + EXT_OFFSET;
  GFX.Screen = (uint16 *) ScreenBuffer;
  GFX.Pitch = EXT_PITCH;

  if (!S9xGraphicsInit ())
    {
      MessageBox (hWnd,
		  Languages[GUI.Language].errMemoryInit,
		  "uosnes - Graphics Init", MB_OK | MB_ICONSTOP);
      return FALSE;
    }

  //for uowScreenRedraw ()
  IPPU.RenderedScreenWidth = SNES_WIDTH;
  IPPU.RenderedScreenHeight = SNES_HEIGHT;

  if (!S9xInitAPU ())
    {
      MessageBox (hWnd,
		  Languages[GUI.Language].errMemoryInit,
		  "uosnes - SNES APU Init", MB_OK | MB_ICONSTOP);

      return FALSE;
    }

  //init SNES sound
  int _1msc = 0;
  for (; (_1msc * 1000) < Settings.SoundPlaybackRate; _1msc++);
  int buffer_bytes = _1msc * Settings.SoundMixInterval *
    (Settings.Stereo ? 2 : 1) * (Settings.SixteenBitSound ? 2 : 1);
  S9xInitSound (7, Settings.Stereo, buffer_bytes);

  if (!uowSetupSound (SoundRates[Options.SoundRate],
		      Options.SoundBit == 16, Options.SoundChn == 2, hWnd))
    {
      MessageBox (hWnd,
		  Languages[GUI.Language].errInitSound,
		  "uosnes - Sound", MB_OK | MB_ICONINFORMATION);
    }

  uowJoystickInit (hWnd);

  if (!SetupVsyncTimer (Renderer.wait_for_vsync, VsyncTimerProc, NULL))
    {
      MessageBox (hWnd,
		  "The vsync timer failed to initialize, and set to default(None).",
		  "uosnes - vsync timer", MB_OK | MB_ICONINFORMATION);
      Renderer.wait_for_vsync = Renderer_wait_for_vsync_def;
    }
  return TRUE;
}

void
uowDeinit ()
{
#ifdef DEBUGGER
  extern FILE *trace;
  extern FILE *trace2;
  if (trace)
    fclose (trace);
  if (trace2)
    fclose (trace2);
#endif

  //Release vsync timer resources
  SetupVsyncTimer (0, VsyncTimerProc, NULL);
  //Release input resources
  uowReleaseInput ();
  //Release sound resources
  uowReleaseSound ();
  //Release DirectX resources
  DirectX_Deinit ();

  Memory.Deinit ();
  S9xGraphicsDeinit ();
  RenderDeinit ();

  if (ScreenBuf1)
    {
      delete[]ScreenBuf1;
      ScreenBuf1 = NULL;
    }

  S9xDeinitAPU ();
}

static void
FrameTimerProc (void *tParams)
{
  if (				/*Settings.SkipFrames != AUTO_FRAMERATE || Settings.TurboMode || */
       Settings.Paused || Settings.StopEmulation || Settings.ForcedPause)
    WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
  else
    ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, NULL);
  return;
}

BOOL
uowSetupFrameTimer (DWORD ms)
{
  return SetupFrameTimer (ms, FrameTimerProc, &IPPU.Interlace);
}

