#define STRICT
#include <windows.h>
#include <mmsystem.h>
#include "winthread.h"
#include "windraw.h"

struct SRenderer Renderer;

/*****************
 default settings
*****************/
BOOL Renderer_OffScreenVRAM_def = TRUE;
BOOL Renderer_BackBufferVRAM_def = TRUE;
BOOL Renderer_ForceWindowedDirectDrawClipper_def = TRUE;
int Renderer_wait_for_vsync_def = 0;

static SDisplayMode ddmode = {
  640,
  480,
  16,
  0
};

static HRESULT CALLBACK EnumModesCallback(LPDDSURFACEDESC lpDDSurfaceDesc, LPVOID lpContext);

void
RendererInit ()
{
  FillMemory ((PVOID) &Renderer, sizeof(Renderer), 0);

  //set default value
  Renderer.hWnd = NULL;
  Renderer.lpDD2 = NULL;
  Renderer.lpDDClipper = NULL;
  Renderer.lpDDPalette = NULL;

  Renderer.lpDDSPrimary = NULL;
  Renderer.lpDDSBack = NULL;
  Renderer.lpDDSOffScreen = NULL;

  Renderer.lpDDSPrimary2 = NULL;
  Renderer.lpDDSBack2 = NULL;
  Renderer.lpDDSOffScreen2 = NULL;

  Renderer.Width = Renderer.Height = -1;
  Renderer.Depth = -1;
  Renderer.Windowed = TRUE;
  Renderer.ForceClipper = FALSE;
  Renderer.ForceWindowedDirectDrawClipper =
    Renderer_ForceWindowedDirectDrawClipper_def;
  Renderer.DoubleBuffered = FALSE;
  Renderer.OffScreenVRAM = Renderer_OffScreenVRAM_def;
  Renderer.BackBufferVRAM = Renderer_BackBufferVRAM_def;
  Renderer.wait_for_vsync = Renderer_wait_for_vsync_def;

  Renderer.DDAvailable = FALSE;
}

BOOL
DirectX_Init (HWND hWnd, DWORD dwStyleW, DWORD dwStyleF)
{
  //DirectX_Initialize
  Renderer.hWnd = hWnd;
  Renderer.dwStyleWindowed = dwStyleW;
  Renderer.dwStyleFullScreen = dwStyleF;

  LPDIRECTDRAW lpDD;

  Renderer.dErr = DirectDrawCreate (NULL, &lpDD, NULL);

  if (Renderer.dErr != DD_OK)
    return FALSE;

  if (FAILED
      (lpDD->QueryInterface (IID_IDirectDraw2, (void **)&Renderer.lpDD2)))
    {
      lpDD->Release ();
      lpDD = NULL;
      return FALSE;
    }
  lpDD->Release ();
  lpDD = NULL;
  

  Renderer.lpDD2->EnumDisplayModes(DDEDM_REFRESHRATES,
				  NULL,
				  &Renderer.dm,
				  EnumModesCallback);

  if(Renderer.dm.size() < 1)
    Renderer.dm.push_back(ddmode);

  Renderer.dErr = Renderer.lpDD2->CreateClipper (0, &Renderer.lpDDClipper, NULL);
  if (Renderer.dErr != DD_OK)
    {
      Renderer.lpDD2->Release ();
      Renderer.lpDD2 = NULL;
      return FALSE;
    }

  Renderer.dErr = Renderer.lpDDClipper->SetHWnd (0, Renderer.hWnd);

  if (Renderer.dErr != DD_OK)
    {
      Renderer.lpDDClipper->Release ();
      Renderer.lpDDClipper = NULL;
      Renderer.lpDD2->Release ();
      Renderer.lpDD2 = NULL;
      return FALSE;
    }

  return TRUE;
}

void
DirectX_Deinit ()
{
  if (Renderer.lpDD2 != NULL)
    {
      if (!Renderer.Windowed)
	Renderer.lpDD2->RestoreDisplayMode ();

      if (Renderer.lpDDSPrimary != NULL)
	{
	  Renderer.lpDDSPrimary->Release ();
	  Renderer.lpDDSPrimary = NULL;
	}
      if (Renderer.lpDDSOffScreen != NULL)
	{
	  Renderer.lpDDSOffScreen->Release ();
	  Renderer.lpDDSOffScreen = NULL;
	}
      if (Renderer.lpDDSPrimary2 != NULL)
	{
	  Renderer.lpDDSPrimary2->Release ();
	  Renderer.lpDDSPrimary2 = NULL;
	}
      if (Renderer.lpDDSOffScreen2 != NULL)
	{
	  Renderer.lpDDSOffScreen2->PageUnlock (0);
	  Renderer.lpDDSOffScreen2->Release ();
	  Renderer.lpDDSOffScreen2 = NULL;
	}
      if (Renderer.lpDDClipper != NULL)
	{
	  Renderer.lpDDClipper->Release ();
	  Renderer.lpDDClipper = NULL;
	}
      if (Renderer.lpDDPalette != NULL)
	{
	  Renderer.lpDDPalette->Release ();
	  Renderer.lpDDPalette = NULL;
	}
#if 0
      //It is not necessary
      if (Renderer.lpDDSBack2 != NULL)
	{
	  Renderer.lpDDSBack2->Release ();
	  Renderer.lpDDSBack2 = NULL;
	}
#endif
      Renderer.lpDDSBack = NULL;
      Renderer.lpDDSBack2 = NULL;

      Renderer.lpDD2->Release ();
      Renderer.lpDD2 = NULL;

      Renderer.DDAvailable = FALSE;
    }
  
  Renderer.dm.clear();
}

BOOL
DirectX_SetDisplayMode (int Width, int Height, int Scale,
			int Depth, int Rate,
			BOOL Windowed,
			BOOL DoubleBuffered)
{
  if (!Renderer.lpDD2)
    return FALSE;

  if (Depth < 8)
    Depth = Renderer.Depth;

  if (Renderer.lpDDSPrimary != NULL)
    {
      Renderer.lpDDSPrimary->Release ();
      Renderer.lpDDSPrimary = NULL;
    }
  if (Renderer.lpDDSOffScreen != NULL)
    {
      Renderer.lpDDSOffScreen->Release ();
      Renderer.lpDDSOffScreen = NULL;
    }
  if (Renderer.lpDDSPrimary2 != NULL)
    {
      Renderer.lpDDSPrimary2->Release ();
      Renderer.lpDDSPrimary2 = NULL;
    }
  if (Renderer.lpDDSOffScreen2 != NULL)
    {
      Renderer.lpDDSOffScreen2->PageUnlock (0);
      Renderer.lpDDSOffScreen2->Release ();
      Renderer.lpDDSOffScreen2 = NULL;
    }
  if (Renderer.lpDDPalette != NULL)
    {
      Renderer.lpDDPalette->Release ();
      Renderer.lpDDPalette = NULL;
    }
  Renderer.lpDDSBack = NULL;
  Renderer.lpDDSBack2 = NULL;

  Renderer.DDAvailable = FALSE;

  DDSURFACEDESC ddsd;

  //Window mode
  if (Windowed)
    {
      DoubleBuffered = FALSE;
      if (!Renderer.Windowed)
	{
	  Renderer.lpDD2->RestoreDisplayMode ();
	  SetWindowLong (Renderer.hWnd, GWL_STYLE, Renderer.dwStyleWindowed);
	}
      FillMemory ((PVOID) & ddsd, sizeof (ddsd), 0);
      ddsd.dwSize = sizeof (ddsd);
      ddsd.dwFlags = DDSD_PIXELFORMAT;
      Renderer.dErr = Renderer.lpDD2->GetDisplayMode (&ddsd);
      if (Renderer.dErr != DD_OK)
	{
	  Depth = 8;
	}
      else
	{
	  if (ddsd.ddpfPixelFormat.dwFlags & DDPF_RGB)
	    {
	      Depth = (int) ddsd.ddpfPixelFormat.dwRGBBitCount;
	    }
	  else
	    {
	      Depth = 8;
	    }
	}
      Renderer.dErr =
	Renderer.lpDD2->SetCooperativeLevel (Renderer.hWnd,
					     DDSCL_NORMAL | DDSCL_ALLOWREBOOT);
    }
  //Fullscreen mode
  else
    {
      SetWindowLong (Renderer.hWnd, GWL_STYLE, Renderer.dwStyleFullScreen);
      SetWindowPos (Renderer.hWnd, HWND_TOP, 0, 0, 0, 0,
		    SWP_DRAWFRAME | SWP_FRAMECHANGED | SWP_NOMOVE |
		    SWP_NOSIZE);

      if (Renderer.Windowed ||
	  Width != Renderer.Width ||
	  Height != Renderer.Height ||
	  Depth != Renderer.Depth)
	{
	  Renderer.dErr =
	    Renderer.lpDD2->SetCooperativeLevel (Renderer.hWnd,
						 DDSCL_ALLOWMODEX |
						 DDSCL_EXCLUSIVE |
						 DDSCL_FULLSCREEN |
						 DDSCL_ALLOWREBOOT);

	  if (Renderer.dErr == DD_OK)
	    Renderer.dErr =
	      Renderer.lpDD2->SetDisplayMode (Width, Height, Depth, Rate, 0);
	}
    }

  if (Renderer.dErr != DD_OK)
    return FALSE;

  FillMemory (&ddsd, sizeof (ddsd), 0);

  //Fullscreen mode with backbuffer
  if (!Windowed && DoubleBuffered)
    {
      ddsd.dwSize = sizeof (ddsd);
      ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
      ddsd.ddsCaps.dwCaps =
	DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
      ddsd.dwBackBufferCount = 1;
    }
  //Fullscreen mode without backbuffer
  else if (!Windowed && !DoubleBuffered)
    {
      ddsd.dwSize = sizeof (ddsd);
      ddsd.dwFlags = DDSD_CAPS;
      ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    }
  //Window mode
  else
    {
      ddsd.dwSize = sizeof (ddsd);
      ddsd.dwFlags = DDSD_CAPS;
      ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    }


  //Create PrimarySurface
  Renderer.dErr =
    Renderer.lpDD2->CreateSurface (&ddsd, &Renderer.lpDDSPrimary, NULL);

  //Create PrimarySurface Failed
  if (Renderer.dErr != DD_OK)
    {
      if (DoubleBuffered)
	{
	  FillMemory (&ddsd, sizeof (ddsd), 0);
	  ddsd.dwSize = sizeof (ddsd);
	  ddsd.dwFlags = DDSD_CAPS;
	  ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	  DoubleBuffered = FALSE;
	  Renderer.dErr =
	    Renderer.lpDD2->CreateSurface (&ddsd, &Renderer.lpDDSPrimary, NULL);
	}
      if (Renderer.dErr != DD_OK)
	return FALSE;
    }

  FillMemory (&Renderer.DDPixelFormat, sizeof (Renderer.DDPixelFormat),
	      0);
  Renderer.DDPixelFormat.dwSize = sizeof (Renderer.DDPixelFormat);
  Renderer.lpDDSPrimary->GetPixelFormat (&Renderer.DDPixelFormat);

  if (!Windowed && Renderer.lpDDClipper)
    {
      if (Renderer.lpDDSPrimary)
	{
	  Renderer.lpDDSPrimary->SetClipper (NULL);
	}
      Renderer.lpDDClipper->Release ();
      Renderer.lpDDClipper = NULL;
    }

  if (FAILED
      (Renderer.lpDDSPrimary->
       QueryInterface (IID_IDirectDrawSurface2,
		       (void **)&Renderer.lpDDSPrimary2)))
    {
      return FALSE;
    }
  else
    {
      Renderer.lpDDSPrimary->Release ();
      Renderer.lpDDSPrimary = NULL;
    }

  if (!Windowed && DoubleBuffered)
    {
      DDSCAPS ddscaps;
      RECT src;

      if (!Renderer.BackBufferVRAM)
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER | DDSCAPS_SYSTEMMEMORY;
      else
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
      if (Renderer.lpDDSPrimary2 != NULL)
	{
	  Renderer.dErr =
	    Renderer.lpDDSPrimary2->GetAttachedSurface (&ddscaps,
						       &Renderer.lpDDSBack2);
	}

      if (Renderer.dErr != DD_OK)
	{
	  if (!Renderer.BackBufferVRAM)
	    {
	      ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	      if (Renderer.lpDDSPrimary2 != NULL)
		{
		  Renderer.dErr =
		    Renderer.lpDDSPrimary2->GetAttachedSurface (&ddscaps,
							       &Renderer.
							       lpDDSBack2);
		}
	    }
	  if (Renderer.dErr != DD_OK)
	    return FALSE;
	}

      src.left = 0;
      src.right = Width;
      src.top = 0;
      src.bottom = Height;
      if (Renderer.lpDDSPrimary2 != NULL)
	{
	  Renderer.lpDDSBack2->BltFast (0, 0, Renderer.lpDDSPrimary2, &src,
				       DDBLTFAST_WAIT);
	}
    }
  else
    {
      Renderer.lpDDSBack = NULL;
      Renderer.lpDDSBack2 = NULL;
    }

  ddsd.dwSize = sizeof (ddsd);
  ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
  if (!Renderer.OffScreenVRAM)
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
  else
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
  ddsd.dwWidth = 256 * Scale;
  ddsd.dwHeight = 239 * Scale;

  Renderer.dErr =
    Renderer.lpDD2->CreateSurface (&ddsd, &Renderer.lpDDSOffScreen, NULL);

  if (Renderer.dErr != DD_OK)
    {
      if (!Renderer.OffScreenVRAM)
	{
	  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	  Renderer.dErr =
	    Renderer.lpDD2->CreateSurface (&ddsd, &Renderer.lpDDSOffScreen,
					 NULL);
	}
      if (Renderer.dErr != DD_OK)
	return FALSE;
    }

  if (Renderer.lpDDSPrimary2 != NULL)
    {
      if (FAILED
	  (Renderer.lpDDSOffScreen->
	   QueryInterface (IID_IDirectDrawSurface2,
			   (void **)&Renderer.lpDDSOffScreen2)))
	{
	  return FALSE;
	}
      else
	{
	  Renderer.lpDDSOffScreen2->PageLock (0);
	  Renderer.lpDDSOffScreen->Release ();
	  Renderer.lpDDSOffScreen = NULL;
	}
    }

  if (Depth == 8)
    {
      PALETTEENTRY PaletteEntries[256];

      for (int i = 0; i < 256; i++)
	{
	  PaletteEntries[i].peRed = 0;
	  PaletteEntries[i].peGreen = 0;
	  PaletteEntries[i].peBlue = 0;
	  PaletteEntries[i].peFlags = PC_NOCOLLAPSE;
	}
      Renderer.dErr =
	Renderer.lpDD2->CreatePalette (DDPCAPS_8BIT | DDPCAPS_ALLOW256,
				       PaletteEntries, &Renderer.lpDDPalette,
				       NULL);
      if (Renderer.dErr != DD_OK)
	{
	  Renderer.lpDDPalette = NULL;
	  return FALSE;
	}
    }

  //Get display mode
  FillMemory ((PVOID) & ddsd, sizeof (ddsd), 0);
  ddsd.dwSize = sizeof (ddsd);
  ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH;
  Renderer.lpDD2->GetDisplayMode (&ddsd);

#if 0
  //Sleep
  if (Renderer.Windowed != Windowed ||
      Renderer.Width != (int) ddsd.dwWidth ||
      Renderer.Height != (int) ddsd.dwHeight ||
      Renderer.Depth != Depth)
    {
      Sleep (500);
    }
#endif

  Renderer.Windowed = Windowed;
  Renderer.Depth = Depth;
  Renderer.Height = ddsd.dwHeight;
  Renderer.Width = ddsd.dwWidth;
  Renderer.DoubleBuffered = DoubleBuffered;

  DirectX_ClearOffscreenSurface ();
  DirectX_ClearPrimarySurface ();
  DirectX_ClearBackbufferSurface ();

  Renderer.DDAvailable = TRUE;
  return TRUE;
}

BOOL
LockSurface2 (LPDIRECTDRAWSURFACE2 lpDDSurface, SSurface * lpSurface)
{
  DDSURFACEDESC ddsd;
  HRESULT hResult;
  int retry;

  ZeroMemory (&ddsd, sizeof (DDSURFACEDESC));
  ddsd.dwSize = sizeof (ddsd);

  retry = 0;
  while (TRUE)
    {
      hResult = lpDDSurface->Lock (NULL, &ddsd, DDLOCK_WAIT, NULL);

      if (hResult == DD_OK)
	{
	  lpSurface->Width = ddsd.dwWidth;
	  lpSurface->Height = ddsd.dwHeight;
	  lpSurface->Pitch = ddsd.lPitch;
	  lpSurface->Surface = (unsigned char *) ddsd.lpSurface;
	  return TRUE;
	}

      if (hResult == DDERR_SURFACELOST)
	{
	  retry++;
	  if (retry > 5)
	    {
	      return FALSE;
	    }

	  hResult = lpDDSurface->Restore ();
	  Renderer.ScreenCleared = TRUE;
	  if (hResult != DD_OK)
	    {
	      return FALSE;
	    }

	  continue;
	}

      if (hResult != DDERR_WASSTILLDRAWING)
	return FALSE;
    }
}

//Zero clear Surface2
static inline void
DirectX_ClearSurface2 (LPDIRECTDRAWSURFACE2 lpDDS2)
{
  SSurface clr;

  if (lpDDS2 == NULL || LockSurface2 (lpDDS2, &clr) == false)
    return;

  int pb = Renderer.Depth >> 3;

  for (int y = 0; y < clr.Height; y++)
    FillMemory ((PVOID) (clr.Surface + clr.Pitch * y), clr.Width * pb, 0);

  lpDDS2->Unlock (clr.Surface);
}

//Zero Fill Surface2
static inline void
DirectX_FillSurface2 (LPDIRECTDRAWSURFACE2 lpDDS2, RECT * clrRect)
{
  SSurface clr;

  if (lpDDS2 == NULL || LockSurface2 (lpDDS2, &clr) == false)
    return;

  int pb = Renderer.Depth == 15 ? 2 : (Renderer.Depth >> 3);

  if (clrRect->top >= 0 &&
      clrRect->left >= 0 &&
      clrRect->bottom <= clr.Height && clrRect->right <= clr.Width)
    {

      int w = clrRect->right - clrRect->left;

      for (int y = clrRect->top; y < clrRect->bottom; y++)
	FillMemory ((PVOID)
		    (clr.Surface + clr.Pitch * y + clrRect->left * pb),
		    w * pb, 0);
    }

  lpDDS2->Unlock (clr.Surface);
}

//Zero clear OffscreenSurface
void
DirectX_ClearOffscreenSurface ()
{
  if (Renderer.lpDDSOffScreen2)
    DirectX_ClearSurface2 (Renderer.lpDDSOffScreen2);
}

//Zero clear PrimarySurface
void
DirectX_ClearPrimarySurface ()
{
  if (Renderer.Windowed)
    return;
  if (Renderer.lpDDSPrimary2)
    DirectX_ClearSurface2 (Renderer.lpDDSPrimary2);
}

//Zero clear BackbufferSurface
void
DirectX_ClearBackbufferSurface ()
{
  if (Renderer.Windowed)
    return;
  if (Renderer.lpDDSBack2)
    DirectX_ClearSurface2 (Renderer.lpDDSBack2);
}

//Zero Fill OffscreenSurface
void
DirectX_FillOffscreenSurface (RECT * clrRect)
{
  if (Renderer.lpDDSOffScreen2)
    DirectX_FillSurface2 (Renderer.lpDDSOffScreen2, clrRect);
}

//Zero Fill PrimarySurface
void
DirectX_FillPrimarySurface (RECT * clrRect)
{
  if (Renderer.Windowed)
    return;
  if (Renderer.lpDDSPrimary2)
    DirectX_FillSurface2 (Renderer.lpDDSPrimary2, clrRect);
}

//Zero Fill BackbufferSurface
void
DirectX_FillBackbufferSurface (RECT * clrRect)
{
  if (Renderer.Windowed)
    return;
  if (Renderer.lpDDSBack2)
    DirectX_FillSurface2 (Renderer.lpDDSBack2, clrRect);
}

void
DirectX_SetClipper (BOOL clip)
{
  if (!Renderer.DDAvailable || !Renderer.hWnd || !Renderer.Windowed)
    return;

  if (clip || Renderer.ForceWindowedDirectDrawClipper || Renderer.ForceClipper)
    {
      if (Renderer.lpDD2 && Renderer.lpDDClipper == NULL)
	{
	  Renderer.dErr =
	    Renderer.lpDD2->CreateClipper (0, &Renderer.lpDDClipper, NULL);
	  if (Renderer.dErr == DD_OK)
	    {
	      Renderer.dErr = Renderer.lpDDClipper->SetHWnd (0, Renderer.hWnd);
	      if (Renderer.dErr != DD_OK)
		return;
	    }
	  else
	    {
	      Renderer.lpDDClipper = NULL;
	      return;
	    }
	}
      if (Renderer.lpDDSPrimary2)
	{
	  Renderer.lpDDSPrimary2->SetClipper (Renderer.lpDDClipper);
	}
    }
  else
    {
      if (Renderer.lpDDSPrimary2)
	{
	  Renderer.lpDDSPrimary2->SetClipper (NULL);
	}
      if (Renderer.lpDDClipper != NULL)
	{
	  Renderer.lpDDClipper->SetHWnd (0, NULL);
	  Renderer.lpDDClipper->Release ();
	  Renderer.lpDDClipper = NULL;
	}
    }
}

static HRESULT CALLBACK EnumModesCallback(LPDDSURFACEDESC lpDDSurfaceDesc, LPVOID lpContext)
{
  if((lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 8 &&
      lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 15 &&
      lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 16 &&
      lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 24 &&
      lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 32) ||
     (lpDDSurfaceDesc->dwWidth < 256 ||
      lpDDSurfaceDesc->dwHeight < 224))
    {
      return DDENUMRET_OK; // keep going without adding mode to list
    }

  SDisplayMode curmode;

  curmode.Width = lpDDSurfaceDesc->dwWidth;
  curmode.Height = lpDDSurfaceDesc->dwHeight;
  curmode.Depth = lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount;
  curmode.Rate = lpDDSurfaceDesc->dwRefreshRate;

  std::vector<SDisplayMode> *modeVector=(std::vector<SDisplayMode> *)lpContext;
  modeVector->push_back(curmode);
  
  return DDENUMRET_OK;
}

int GetDefaultDisplayMode()
{
  if(Renderer.dm.size() < 2)
    return 0;

  int i;

  for(i = 0; i < Renderer.dm.size(); i++)
    if(Renderer.dm[i].Width == ddmode.Width &&
       Renderer.dm[i].Height == ddmode.Height &&
       Renderer.dm[i].Depth == ddmode.Depth)
      return i;

  for(i = 0; i < Renderer.dm.size(); i++)
    if(Renderer.dm[i].Width >= ddmode.Width &&
       Renderer.dm[i].Height >= ddmode.Height &&
       Renderer.dm[i].Depth >= 15)
      return i;

  for(i = 0; i < Renderer.dm.size(); i++)
    if(Renderer.dm[i].Width >= ddmode.Width &&
       Renderer.dm[i].Height >= ddmode.Height)
      return i;

  return 0;
}

const char * GetDisplayModeString(std::vector<SDisplayMode> *dm)
{
  static char dms[1024 + 1];
  char dmss[16];
  for (int i = 0; i < Renderer.dm.size (); i++)
    {
      if (lstrlen (dms) >= 1024 - 16)
	break;
      wsprintf (dmss, "%d", Renderer.dm[i].Width);
      lstrcat (dms, dmss);
      if (lstrlen (dms) >= 1024 - 16)
	break;
      wsprintf (dmss, "%d", Renderer.dm[i].Height);
      lstrcat (dms, dmss);
      if (lstrlen (dms) >= 1024 - 16)
	break;
      wsprintf (dmss, "%d", Renderer.dm[i].Depth);
      lstrcat (dms, dmss);
      if (lstrlen (dms) >= 1024 - 16)
	break;
      wsprintf (dmss, "%d", Renderer.dm[i].Rate);
      lstrcat (dms, dmss);
    }
  return dms;
}

void vsync_method()
{
  if(Renderer.Windowed || !Renderer.DoubleBuffered)
    {
      switch (Renderer.wait_for_vsync)
	{
	case 1:
	  Renderer.lpDD2->WaitForVerticalBlank (DDWAITVB_BLOCKBEGIN, 0);
	  break;

	case 2:
	  WaitForSingleObject (Renderer.hVsyncWaitEvent, 100);
	  break;

	default:
	  break;
	}
    }
}

static int vth;
static HANDLE hVsyncEvent0;
static HANDLE hVsyncEvent;

struct SVsyncTimerThreadParams
{
  int method;
  void (*VsyncTimerProc) (void *tParams);
  void *tParams;
};

static void
VsyncTimerThreadProc (void *tParams, HANDLE hThread, HANDLE hThreadInitEvent, BOOL *init_success, BOOL *exit)
{
  SVsyncTimerThreadParams vp = *((SVsyncTimerThreadParams *) tParams);

  DWORD dwScanLine;
  HRESULT r = Renderer.lpDD2->GetScanLine(&dwScanLine);
  if(r != DD_OK && r != DDERR_VERTICALBLANKINPROGRESS)
    {
      SetEvent (hThreadInitEvent);
      return;
    }

  char EventName[64];
  DWORD ThreadID = GetCurrentThreadId();
  wsprintf(EventName, "%08X", ThreadID);
  lstrcat(EventName, "windraw-hVsyncTimerEvent");

  HANDLE hVsyncTimerEvent0 = CreateEvent (NULL, FALSE, FALSE, (LPCTSTR)EventName);

  if(hVsyncTimerEvent0 == FALSE)
    {
      SetEvent (hThreadInitEvent);
      return;
    }

  TIMECAPS tc;
  if (timeGetDevCaps (&tc, sizeof (TIMECAPS)) == TIMERR_NOERROR)
    timeBeginPeriod (tc.wPeriodMin);
  else
    tc.wPeriodMin = 0;

  HANDLE hVsyncTimerEvent =
    OpenEvent (EVENT_MODIFY_STATE, FALSE, (LPCTSTR)EventName);
  HANDLE hVsyncTimerWaitEvent =
    OpenEvent (SYNCHRONIZE, FALSE, (LPCTSTR)EventName);
  UINT uDelay = (UINT) tc.wPeriodMin;
  UINT TimerID;

  if (hVsyncTimerEvent == FALSE ||
      hVsyncTimerWaitEvent == FALSE ||
      (TimerID = timeSetEvent (uDelay,
			       0,
			       (LPTIMECALLBACK) hVsyncTimerEvent,
			       0,
			       TIME_PERIODIC | TIME_CALLBACK_EVENT_SET)) ==
      (UINT) NULL)
    {

      if (hVsyncTimerWaitEvent != FALSE)
	CloseHandle (hVsyncTimerWaitEvent);
      if (hVsyncTimerEvent != FALSE)
	CloseHandle (hVsyncTimerEvent);
      CloseHandle (hVsyncTimerEvent0);

      if (tc.wPeriodMin)
	timeEndPeriod (tc.wPeriodMin);

      SetEvent (hThreadInitEvent);
      return;
    }

  *init_success = TRUE;
  SetEvent (hThreadInitEvent);

  SetThreadPriority (hThread, THREAD_PRIORITY_TIME_CRITICAL);

  BOOL pvbs = FALSE;
  DWORD ps = 0;

  while (*exit == FALSE)
    {
      WaitForSingleObject (hVsyncTimerWaitEvent, INFINITE);

      DWORD dwScanLine;
      HRESULT r = Renderer.lpDD2->GetScanLine(&dwScanLine);

      if(r == DDERR_VERTICALBLANKINPROGRESS || dwScanLine < ps)
	{
	  if(!pvbs)
	    {
	      if(Renderer.wait_for_vsync == 3)
		vp.VsyncTimerProc (vp.tParams);
	      SetEvent (hVsyncEvent);
	    }
	  pvbs = TRUE;
	  ps = 0;
	}
      else
	{
	  if(pvbs)
	    ResetEvent (hVsyncEvent);
	  pvbs = FALSE;
	  ps = dwScanLine;
	}
    }

  timeKillEvent (TimerID);

  CloseHandle (hVsyncTimerWaitEvent);
  CloseHandle (hVsyncTimerEvent);
  CloseHandle (hVsyncTimerEvent0);

  if (tc.wPeriodMin)
    timeEndPeriod (tc.wPeriodMin);

  SetThreadPriority (hThread, THREAD_PRIORITY_NORMAL);

}

static void CloseVsyncEvent()
{
  if(Renderer.hVsyncWaitEvent != FALSE)
    {
      CloseHandle (Renderer.hVsyncWaitEvent);
      Renderer.hVsyncWaitEvent = FALSE;
    }
  if(hVsyncEvent != FALSE)
    {
      CloseHandle (hVsyncEvent);
      hVsyncEvent = FALSE;
    }
  if(hVsyncEvent0 != FALSE)
    {
      CloseHandle (hVsyncEvent0);
      hVsyncEvent0 = FALSE;
    }
  if(Renderer.hCriticalSection != FALSE)
    {
      DeleteCriticalSection (&Renderer.cs);
      Renderer.hCriticalSection = FALSE;
    }
}

BOOL
SetupVsyncTimer (int method,
		 void (*VsyncTimerProc) (void *tParams),
		 void *tParams)
{
  if (VsyncTimerProc == NULL || (method > 1 && Renderer.lpDD2 == NULL))
    return FALSE;

  if(vth)
    {
      RemoveThread (vth);
      vth = 0;
    }

  CloseVsyncEvent();

  if(method < 2)
    return TRUE;

  hVsyncEvent0 = CreateEvent (NULL, FALSE, FALSE, "windraw-hVsyncEvent0");

  if(hVsyncEvent0 == FALSE)
    return FALSE;

  Renderer.hVsyncWaitEvent = OpenEvent (SYNCHRONIZE, FALSE, "windraw-hVsyncEvent0");
  hVsyncEvent = OpenEvent (EVENT_MODIFY_STATE, FALSE, "windraw-hVsyncEvent0");

  if(Renderer.hVsyncWaitEvent == FALSE || hVsyncEvent == FALSE)
    {
      CloseVsyncEvent();
      return FALSE;
    }

  if (Renderer.hCriticalSection == FALSE)
    {
      InitializeCriticalSection (&Renderer.cs);
      Renderer.hCriticalSection = TRUE;
    }

  SVsyncTimerThreadParams vp;

  vp.method = method;
  vp.VsyncTimerProc = VsyncTimerProc;
  vp.tParams = tParams;

  vth = AddThread(VsyncTimerThreadProc, &vp);
  if(vth)
    return TRUE;

  CloseVsyncEvent();
  return FALSE;
}

