#ifndef _WINDRAW_H_
#define _WINDRAW_H_

#include <ddraw.h>
#include <vector>

struct SDisplayMode
{
  int Width;
  int Height;
  int Depth;
  int Rate;
};

struct SRenderer
{
  HWND hWnd;
  DWORD dwStyleWindowed;
  DWORD dwStyleFullScreen;

  BOOL DDAvailable;
  HRESULT dErr;

  LPDIRECTDRAW2 lpDD2;
  LPDIRECTDRAWCLIPPER lpDDClipper;
  LPDIRECTDRAWPALETTE lpDDPalette;

  LPDIRECTDRAWSURFACE lpDDSPrimary;
  LPDIRECTDRAWSURFACE lpDDSBack;
  LPDIRECTDRAWSURFACE lpDDSOffScreen;

  LPDIRECTDRAWSURFACE2 lpDDSPrimary2;
  LPDIRECTDRAWSURFACE2 lpDDSBack2;
  LPDIRECTDRAWSURFACE2 lpDDSOffScreen2;

  DDPIXELFORMAT DDPixelFormat;

  int Width;
  int Height;
  int Depth;

  BOOL Windowed;

  BOOL ForceClipper;
  BOOL ForceWindowedDirectDrawClipper;

  BOOL DoubleBuffered;
  BOOL OffScreenVRAM;
  BOOL BackBufferVRAM;
  int wait_for_vsync;

  BOOL ScreenCleared;

  std::vector<SDisplayMode> dm;

  HANDLE hVsyncWaitEvent;
  BOOL hCriticalSection;
  CRITICAL_SECTION cs;
};

struct SSurface
{
  PBYTE Surface;
  DWORD Pitch;
  DWORD Width;
  DWORD Height;
};

//Renderer proc
void RendererInit ();
BOOL DirectX_SetDisplayMode (int Width, int Height, int Scale,
			     int Depth, int Rate,
			     BOOL Windowed, BOOL DoubleBuffered);
BOOL DirectX_Init (HWND hWnd, DWORD dwStyleWindowed,
		   DWORD dwStyleFullScreen);
void DirectX_Deinit ();
void DirectX_ClearPrimarySurface ();
void DirectX_ClearOffscreenSurface ();
void DirectX_ClearBackbufferSurface ();
void DirectX_FillPrimarySurface (RECT * clrRect);
void DirectX_FillOffscreenSurface (RECT * clrRect);
void DirectX_FillBackbufferSurface (RECT * clrRect);
void DirectX_SetClipper (BOOL clip);
BOOL LockSurface2 (LPDIRECTDRAWSURFACE2 lpDDSurface, SSurface * lpSurface);
int GetDefaultDisplayMode();
const char * GetDisplayModeString(std::vector<SDisplayMode> *dm);
void vsync_method();
BOOL SetupVsyncTimer (int method,
		      void (*VsyncTimerProc) (void *tParams),
		      void *tParams);

/*****************
 default settings
*****************/
extern BOOL Renderer_OffScreenVRAM_def;
extern BOOL Renderer_BackBufferVRAM_def;
extern BOOL Renderer_ForceWindowedDirectDrawClipper_def;
extern int  Renderer_wait_for_vsync_def;

//Renderer struct
extern struct SRenderer Renderer;

#endif /* _WINDRAW_H_ */

