/*
  Copyright (C) 2010 sanmaiwashi
*/

#define STRICT
#include <windows.h>
#include <winbase.h>
#include <process.h>
#include "winthread.h"

#define MAX_THREADS 16

struct SThreadParams
{
  int h;
  HANDLE hThread;
  unsigned ThreadID;
  void (*ThreadProc) (void *tParams, HANDLE hThread, HANDLE hThreadInitEvent,
		      BOOL * init_success, BOOL * exit);
  void *tParams;
  BOOL init_success;
  BOOL exit;
};

static SThreadParams tp[MAX_THREADS];
static HANDLE hThreadInitEvent0 = FALSE;
static HANDLE hThreadInitEvent1 = FALSE;
static HANDLE hThreadInitEvent2 = FALSE;

static int
GetValidThreadHandle ()
{
  int h = 1;
  for (;;)
    {
      int i;
      for (i = 0; i < MAX_THREADS; i++)
	if (tp[i].h == h)
	  {
	    h++;
	    break;
	  }
      if (i >= MAX_THREADS)
	break;
    }
  return h > MAX_THREADS ? 0 : h;
}

static int
GetValidThreadParamsIndex (int h)
{
  for (int i = 0; i < MAX_THREADS; i++)
    if (tp[i].h == h)
      return i;
  return -1;
}

static void
CloseThreadInitEvent ()
{
  if (hThreadInitEvent2 != FALSE)
    {
      CloseHandle (hThreadInitEvent2);
      hThreadInitEvent2 = FALSE;
    }
  if (hThreadInitEvent1 != FALSE)
    {
      CloseHandle (hThreadInitEvent1);
      hThreadInitEvent1 = FALSE;
    }
  if (hThreadInitEvent0 != FALSE)
    {
      CloseHandle (hThreadInitEvent0);
      hThreadInitEvent0 = FALSE;
    }
}

static unsigned __stdcall
ThreadProc0 (void *tParams)
{
  SThreadParams *tp = (SThreadParams *) tParams;
  tp->ThreadProc (tp->tParams, tp->hThread, hThreadInitEvent2,
		  &tp->init_success, &tp->exit);
  _endthreadex (0);
  return 0;
}

int
AddThread (void (*ThreadProc)
	   (void *tParams, HANDLE hThread, HANDLE hThreadInitEvent,
	    BOOL * init_success, BOOL * exit), void *tParams)
{
  if (ThreadProc == NULL)
    return 0;

  int i = GetValidThreadParamsIndex (0);
  if (i < 0)
    return 0;

  int h = GetValidThreadHandle ();
  if (h == 0)
    return 0;

  tp[i].ThreadProc = ThreadProc;
  tp[i].tParams = tParams;
  tp[i].init_success = FALSE;
  tp[i].exit = FALSE;

  hThreadInitEvent0 =
    CreateEvent (NULL, FALSE, FALSE, "winthread-hThreadInitEvent0");

  if (hThreadInitEvent0 == FALSE)
    return 0;

  hThreadInitEvent1 =
    OpenEvent (SYNCHRONIZE, FALSE, "winthread-hThreadInitEvent0");
  hThreadInitEvent2 =
    OpenEvent (EVENT_MODIFY_STATE, FALSE, "winthread-hThreadInitEvent0");

  if (hThreadInitEvent1 == FALSE || hThreadInitEvent2 == FALSE)
    {
      CloseThreadInitEvent ();
      return 0;
    }

  tp[i].h = h;

  tp[i].hThread = (HANDLE) _beginthreadex (NULL,
					   0,
					   &ThreadProc0,
					   &tp[i], 0, &tp[i].ThreadID);

  if ((int) tp[i].hThread == 0)
    {
      tp[i].h = 0;
      CloseThreadInitEvent ();
      return 0;
    }

  // init_success?
  WaitForSingleObject (hThreadInitEvent1, INFINITE);
  CloseThreadInitEvent ();
  if (tp[i].init_success == FALSE)
    {
      //Wait thread end.
      WaitForSingleObject (tp[i].hThread, INFINITE);
      CloseHandle (tp[i].hThread);
      tp[i].h = 0;
      return 0;
    }
  return h;
}

void
RemoveThread (int h)
{
  if (h == 0 || h > MAX_THREADS)
    return;

  int i = GetValidThreadParamsIndex (h);
  if (i < 0)
    return;

  tp[i].exit = TRUE;
  WaitForSingleObject (tp[i].hThread, INFINITE);
  CloseHandle (tp[i].hThread);
  tp[i].h = 0;
}
