/*
  Copyright (C) 2010 sanmaiwashi
*/

#define STRICT
#include <windows.h>
#include <mmsystem.h>
#include "winthread.h"
#include "wintimer.h"
#include <math.h>

static int fth;

struct SFrameTimerThreadParams
{
  DWORD ms;
  void (*FrameTimerProc) (void *tParams);
  void *tParams;
};

BOOL PerformanceCounterExist;
LARGE_INTEGER PerfTicksPerSecond;

static inline double get_ntsc_step(unsigned char interlace)
{
  double step;
  if(!interlace)
    step = (double)1000.0/(21477272.0/(341*4*262-2));
  else
    step = (double)1000.0/(21477272.0/(341*4*262+682));
  return step;
}

static inline double get_pal_step(unsigned char interlace)
{
  double step;
  if(!interlace)
    step = (double)1000.0/(21281370.0/(341*4*312-2));
  else
    step = (double)1000.0/(21281370.0/(341*4*312+682));
  return step;
}

static void
FrameTimerThreadProc (void *tParams, HANDLE hThread, HANDLE hThreadInitEvent, BOOL *init_success, BOOL *exit)
{
  SFrameTimerThreadParams tp = *((SFrameTimerThreadParams *) tParams);
  unsigned char *interlace = (unsigned char *)tp.tParams;

  char EventName[64];
  DWORD ThreadID = GetCurrentThreadId();
  wsprintf(EventName, "%08X", ThreadID);
  lstrcat(EventName, "wintimer-hFrameTimerEvent");

  HANDLE hFrameTimerEvent0 = CreateEvent (NULL, FALSE, FALSE, (LPCTSTR)EventName);

  if(hFrameTimerEvent0 == FALSE)
    {
      SetEvent (hThreadInitEvent);
      return;
    }

  TIMECAPS tc;
  if (timeGetDevCaps (&tc, sizeof (TIMECAPS)) == TIMERR_NOERROR)
    timeBeginPeriod (tc.wPeriodMin);
  else
    tc.wPeriodMin = 0;

  HANDLE hFrameTimerEvent =
    OpenEvent (EVENT_MODIFY_STATE, FALSE, (LPCTSTR)EventName);
  HANDLE hFrameTimerWaitEvent =
    OpenEvent (SYNCHRONIZE, FALSE, (LPCTSTR)EventName);
  UINT uDelay = (UINT) tc.wPeriodMin;
  UINT TimerID;

  if (hFrameTimerEvent == FALSE ||
      hFrameTimerWaitEvent == FALSE ||
      (TimerID = timeSetEvent (uDelay,
			       0,
			       (LPTIMECALLBACK) hFrameTimerEvent,
			       0,
			       TIME_PERIODIC | TIME_CALLBACK_EVENT_SET)) ==
      (UINT) NULL)
    {

      if (hFrameTimerWaitEvent != FALSE)
	CloseHandle (hFrameTimerWaitEvent);
      if (hFrameTimerEvent != FALSE)
	CloseHandle (hFrameTimerEvent);
      CloseHandle (hFrameTimerEvent0);

      if (tc.wPeriodMin)
	timeEndPeriod (tc.wPeriodMin);

      SetEvent (hThreadInitEvent);
      return;
    }

  *init_success = TRUE;
  SetEvent (hThreadInitEvent);

  SetThreadPriority (hThread, THREAD_PRIORITY_TIME_CRITICAL);

  // NTSC
  if (tp.ms == 16)
    {
      double step = get_ntsc_step(*interlace);
      double pass = step;
      DWORD last_ms = timeGetTime ();
      while (*exit == FALSE)
	{
	  WaitForSingleObject (hFrameTimerWaitEvent, INFINITE);
	  DWORD cur_ms = timeGetTime ();
	  while ((cur_ms - last_ms) >= (int)pass)
	    {
	      last_ms = last_ms + (int)pass;
	      tp.FrameTimerProc (tp.tParams);
	      pass = fmod(pass, 1.0);
	      step = get_ntsc_step(*interlace);
	      pass += step;
	    }
	}
    }
  // PAL
  else if (tp.ms == 20)
    {
      double step = get_pal_step(*interlace);
      double pass = step;
      DWORD last_ms = timeGetTime ();
      while (*exit == FALSE)
	{
	  WaitForSingleObject (hFrameTimerWaitEvent, INFINITE);
	  DWORD cur_ms = timeGetTime ();
	  while ((cur_ms - last_ms) >= (int)pass)
	    {
	      last_ms = last_ms + (int)pass;
	      tp.FrameTimerProc (tp.tParams);
	      pass = fmod(pass, 1.0);
	      step = get_pal_step(*interlace);
	      pass += step;
	    }
	}
    }
  // ms
  else
    {
      DWORD last_ms = timeGetTime ();
      while (*exit == FALSE)
	{
	  WaitForSingleObject (hFrameTimerWaitEvent, INFINITE);
	  DWORD cur_ms = timeGetTime ();
	  while ((cur_ms - last_ms) >= tp.ms)
	    {
	      last_ms = last_ms + tp.ms;
	      tp.FrameTimerProc (tp.tParams);
	    }
	}
    }

  timeKillEvent (TimerID);

  CloseHandle (hFrameTimerWaitEvent);
  CloseHandle (hFrameTimerEvent);
  CloseHandle (hFrameTimerEvent0);

  if (tc.wPeriodMin)
    timeEndPeriod (tc.wPeriodMin);

  SetThreadPriority (hThread, THREAD_PRIORITY_NORMAL);
}

BOOL
SetupFrameTimer (DWORD ms,
		 void (*FrameTimerProc) (void *tParams),
		 void *tParams)
{
  if (FrameTimerProc == NULL)
    return FALSE;

  if(fth)
    {
      RemoveThread (fth);
      fth = 0;
    }

  if(ms == 0)
    return TRUE;

  SFrameTimerThreadParams tp;

  tp.ms = ms;
  tp.FrameTimerProc = FrameTimerProc;
  tp.tParams = tParams;

  fth = AddThread(FrameTimerThreadProc, &tp);
  if(fth)
    return TRUE;
  return FALSE;
}

